#ifndef POOL_IDATASVC_H
#define POOL_IDATASVC_H

#include "StorageSvc/poolDb.h"
#include "StorageSvc/DbStatus.h"
#include "StorageSvc/IStorageSvc.h"
#include "POOLCore/Token.h"
#include "Reflex/Type.h"
#include <typeinfo>


namespace pool  {
  class IStorageSvc;
  class DataHandler;
  class Transaction;
  class Placement;
  class CacheRow;
  struct DbaseDescriptors;
  namespace {
    template<class T> struct Delete  {
      static void call(void* ptr)  { T* p = (T*)ptr; delete p;    }
    };
  }


  /** class IDataSvc
  */
  class IDataSvc  {
  public:
    typedef ROOT::Reflex::Type ClassH;
    typedef std::vector<CacheRow*> Cache;
    /// Reference to default data handler
    DataHandler*      m_dataHandler;
    /// Cache of objects
    Cache             m_cache;
    /// Reference to storage service
    IStorageSvc*      m_storageSvc;
    /// Database descriptors
    DbaseDescriptors* m_files;
    /// Session handle
    SessionH          m_session;
    /// Flag to not delete any object
    bool              m_doDeleteObjects;
    /// Printout flag
    long              m_debug;
  private:
    /// Register object for write (Internal call)
    DbStatus i_registerForWrite(Transaction* trContext, 
                                const Placement& placeHint,
                                const std::type_info& typ,
                                void* ptr_obj);
    DbStatus i_readObject(const Token* pTok, 
                          void*&       object,
                          ClassH&      clH);
    /// Update object
    DbStatus i_update(Transaction* trContext, 
                      const std::type_info& typ,
                      void* pObj);
    /// Destroy object
    DbStatus i_destroy( Transaction* trContext, 
                        const std::type_info& typ,
                        void* pObj);

  public:
    /// Standard Constructor
    IDataSvc();
    /// Standard Destructor
    virtual ~IDataSvc();
    /// Clear the data cache
    DbStatus clear();
    /// Enable printouts
    void setPrintout(long lvl);
    /// Number of objects in the cache
    long size() const   {   return m_cache.size();    }
    /// Flag to not delete any object
    void setDeleteObjects(bool value)  { m_doDeleteObjects = value; }
    /// Initialize and return storage service object
    IStorageSvc* storageSvc();
    /// Database session handle
    SessionH session();
    /// Load class
    ClassH loadClass(const std::string& nam);
    /// Register object for write
    template < class T >
    DbStatus registerForWrite(Transaction* trContext, 
                              const Placement& placeHint, 
                              T* pObj)
    {
      return i_registerForWrite(trContext, placeHint, typeid(T), pObj);
    }
    /// Register object for write
    template < class T >
    DbStatus update(Transaction* trContext, T* pObj) 
    {
      return i_update(trContext, typeid(T), pObj);
    }
    /// Register object for write
    template < class T >
    DbStatus destroy(Transaction* trContext, T* pObj) 
    {
      return i_destroy(trContext, typeid(T), pObj);
    }
    /// Read single object according to token
    DbStatus find_or_load_object(Token* tok, void*& object);
    /// Find an existing object in the cache
    DbStatus findToken(const void* object, Token*& tok);
    /// Register token with the cache
    DbStatus regToken(Token* pTok, Token*& tok);

    /// Start Transaction
    DbStatus startTransaction(ConnectionH dbcConn,
                              Transaction*& context);
    /// Commit cache content  
    DbStatus commit(          Transaction* trContext);

    /// Rollback cache content  
    DbStatus rollback(        Transaction* trContext);

    /// Connect to a logical database unit.
    DbStatus connect(         const std::string& refDbName,
                              long               accessmode,
                              ConnectionH&       dbcConn);

    /// Reconnect to a logical database unit.
    DbStatus reconnect(       ConnectionH&       dbcConn,
                              long               accessmode);

    /// Disconnect from a logical database unit.
    DbStatus disconnect(      ConnectionH        dbcConn);
  };
}
#endif // POOL_IDATASVC_H
