// $Id: Ref.h,v 1.2 2005/07/11 16:49:54 frankb Exp $
//====================================================================
//	Reference class definitions
//--------------------------------------------------------------------
//
//	Package    : Pool (The POOL project)
//	Author     : M.Frank
//====================================================================
#ifndef POOL_REF_H
#define POOL_REF_H

// Pool include files
#include "POOLCore/Reference.h"
#include "StorageSvc/DbStatus.h"

/**
    POOL namespace declaration
*/
namespace pool  {

  // Forward declarations
  class Token;
  class IDataSvc;

  /** template class RefBase.

      Description:
      Small class to easy following references.
      Simple and stupid:
      The ref has 3 states:
        HAS_NONE:    useless, uninitialized
        HAS_OBJECT:  together with oid from load callback
                     reference should be loadable via the data 
                     service
        HAS_TOKEN:   Token of the target object.
                     reference should be loadable via the data 
                     service

      Two ingredients are global and must be set by the data service
      at load time:
      - Reference to the data service

      @author   M.Frank
      @version  1.0
      @date     10/11/2002
  */
  class RefBase : public Reference {
  public:
  protected:
    /// Reference to object token 
    mutable Token* m_token;
    IDataSvc*      m_cache;
    mutable void*  m_data;
  public:
    /// Load referenced object in memory; optionally throw exception on failure
    DbStatus load(void*& ptr, bool throw_if) const;
    /// Constructor
    RefBase();
    /// Constructor
    RefBase(IDataSvc* c, Token* p);
    /// Copy constructor
    RefBase(const RefBase& c);
    /// Standard Destructor
    virtual ~RefBase();
    /// Attach data service to the reference
    void setDataSvc(IDataSvc* pSvc)             { m_cache = pSvc; }
    /// Access to raw object pointer
    virtual void* object() const                { return m_data;  }
    /// Set token object for this reference
    /** Any existing token object will be released.
        @param pToken   [IN] reference to new Token object.
      */
    void setToken(Token* pToken)  const;
    /// Access to the token object
    Token* token()  const                        {   return m_token;   }
    /// Operator needed by STL (dummy)
    bool operator <(const RefBase& /*r*/) const { return true;  }
    /// Operator needed by STL (dummy)
    bool operator >(const RefBase& /*r*/) const { return false; }
    /// Set flag to debug tokens
    static void debugTokens(bool val);
  };

  /** template class Ref<T>.

      Reference template class.

      Description:
      The specialized reference class allows to mimic the behaviour
      or a normal pointer and at the same time supports load-on-demand.
      This class must be small to avoid code bloat and should only 
      implement the minimal functionality.
      Major code execution should be implemented in a generic way
      by the base class.

      @author   M.Frank
      @version  1.0
      @date     10/11/2002
  */
  template <class T> class Ref : public RefBase  {
  protected:
    /// Reference to target object (only valid after loading)
    mutable T* m_p;
  public:
    /// Standard constructor with no arguments
    explicit Ref()     : m_p(0)           {                               }
    /// Extern reference: points to valid token
    explicit Ref(IDataSvc* c, Token* p) : RefBase(c, p), m_p(0)  {        }
    /// Embedded reference: Constructor with initial assignment of pointer
    explicit Ref(T* p) : m_p(p)           {                               }
    /// Copy Constructor
    Ref(const Ref<T>& r) : RefBase(r), m_p(r.m_p) { m_data = m_p;         }
    /// Destructor (NON-virtual!)
    virtual ~Ref()                        {                               }
    /// Assignment operator to make reference behave like a pointer
    Ref& operator=(T* ptr)           {
      m_data = m_p = ptr;
      return *this;
    }
    /// Assignment operator to make reference behave like a pointer
    Ref& operator=(const Ref<T>& copy) {
      m_data  = m_p = copy.m_p;
      setDataSvc(copy.m_cache);
      setToken(copy.m_token);
      return *this;
    }
    /// Equality operator
    bool operator==(const Ref& cmp) const  {
      if ( 0 != m_p && 0 != cmp.m_p )
        return m_p == cmp.m_p;
      return cmp.ptr(false) == ptr(false);
    }
    /// De-reference operator: loads object on demand (CONST)
    operator const T* () const            {  return ptr(false);               }
    /// De-reference operator: loads object on demand 
    operator T* ()                        {  return ptr(false);               }
    /// De-reference operator: allow member access (CONST)
    const T* operator->() const           {  return ptr(true);                }
    /// De-reference operator: allow member access
    T* operator->()                       {  return ptr(true);                }
    /// De-reference operator (CONST)
    const T& operator*() const            {  return *ptr(true);               }
    /// De-reference operator
    T& operator*()                        {  return *ptr(true);               }
    /// NON-Equality operator
    bool operator!=(const Ref& cmp) const { return !(this->operator==(cmp));  }
    /// 'not' operator for consistency with pointers common use
    bool operator!()  const               { return ptr(false) == 0;           }
    /// Check if the current ref is referencing some persistent object
    bool is_null()  const                 { return ptr(false) == 0;           }
    /// Access to raw object
    virtual void* object() const          { return m_p;                       } 

    /// Access true pointer behind smart reference.
    /** This member call possibly throws exception on failure.
        @param throw_if   [IN]  Flag to indicate that an exception should 
                                be thrown on failure.
        @return On success valid pointer to object
    */
    T* ptr(bool throw_if=true)  const {
      if ( !m_p )  {
        load(m_data, throw_if);
        m_p = (T*)m_data;
      }
      return m_p;
    }
  };

  /// Friend helper to check for object existence (will load object)
  template <class T> inline 
  bool operator != (const Ref<T>& ref, int)   
  {  return ref.ptr(false) != 0;                                        }

  /// Friend helper to check for object existence (will load object)
  template <class T> inline 
  bool operator != (int, const Ref<T>& ref)
  {  return ref.ptr(false) != 0;                                        }

}
#endif
