#include "pool/Placement.h"
#include "StorageSvc/DbPrint.h"
#include "StorageSvc/DbObjectCallBack.h"
#include "Reflex/Type.h"

#include <vector>
#include <iostream>
#include <typeinfo>
#include <stdexcept>
#include <limits>
#ifndef WIN32
#include <unistd.h>
#endif

#include "pool/IDataSvc.h"

#ifdef __use_refs
#include "TNamed.h"
#endif

#define _CHECK(x)  checkError(x,#x,true)

namespace pool  {

  /** @class TestExecuter
    *
    * @author  M.Frank
    * @version 1.0
    */
  struct TestExecuter  {
    std::string m_label;
    TestExecuter(const std::string& l);
    virtual ~TestExecuter() {}
    virtual const std::string label() const { return m_label; }
    virtual DbStatus runme(int argc, char** argv) const = 0;
  };

  /** @class TestRunner
    *
    * @author  M.Frank
    * @version 1.0
    */
  template <class T> struct TestRunner : public TestExecuter {
    virtual DbStatus runme(int argc, char** argv) const  {
      T test;
      return test.run(argc, argv);
    }
    TestRunner(const char* l) : TestExecuter(l) {}
  };

  typedef ROOT::Reflex::Type ClassH;

  /** @class TestBase
    *
    * @author M.Frank
    */
  class TestBase  {
  private:
    IDataSvc*        m_cache;  // Primitive m_cache...
    IStorageSvc*     m_storageSvc;
  public: 
    enum { READ, WRITE, UPDATE, DESTROY };
    typedef std::vector<std::string> StringVec;
    typedef ROOT::Reflex::Type       ClassH;

    int              m_action;
    bool             m_doInit;
    long             m_type;
    std::string      m_dbName1;
    std::string      m_dbName2;
    std::string      m_criteria;
    StringVec        m_libraries;
    StringVec        m_options[2];
    size_t           m_innerLoops;
    size_t           m_outerLoops;
    int              m_objSize;

    bool getOpt(const std::string& w, const std::string& o, std::string& r);
    bool hasOpt(const std::string& w) const;
  public: 

    /// Standard Constructor
    TestBase(long storage_type, 
             const std::string& f1,
             const std::string& f2,
             bool  init=true);

    /// Standard Destructor
    virtual ~TestBase();

    /// Initialize the plugin manager
    static DbStatus initPlugins();

    /// Print test usage    
    static void usage(const char* pgm);

    /// Load dynamic library
    static void* loadLibrary(const std::string& name);

    static void accessRootClass(const std::string& name);
    /// Type name
    static std::string typeName(const std::type_info& info);

    /// Initialize and return cache object
    IDataSvc* cache()  const;

    /// Initialize and return storage service object
    IStorageSvc* storageSvc()  const;

    /// Create POOL token from basics
    Token* makeToken(const std::string& db, const std::string& cnt, int tech, const std::string& cl, int i1, int i2) const;

    /// Parse arguments
    void updateArgs(int argc, char** argv);

    /// Utility function to check for errors during execution
    static void checkError(DbStatus sc, const std::string& msg, bool throw_exc=false);

    /// Load class information
    DbStatus loadClass(const std::string& nam, 
                       ClassH& cl,
                       Guid& guid);

    /// Retrieve GUID for a single class
    Guid getGuid(const std::string& cl_name) const;

    /// Initialize the test
    virtual DbStatus initialize();

    /// Write objects to the persistent store
    virtual DbStatus write();

    /// read objects from the persistent store
    virtual DbStatus read();

    /// Update objects from the persistent store
    virtual DbStatus update();

    /// Update objects from the persistent store
    virtual DbStatus destroy();

    /// Run the test
    virtual DbStatus run(int argc, char** argv);

    /// Dump root streamer info
    void dumpClass(const std::string& nam);

    /// Dump root streamer info
    void dumpStreamerInfo(const std::string& nam);

    /// Dump root real data content
    void dumpRealData(const std::string& nam);

    /// Dump root data member content
    void dumpDataMembers(const std::string& nam);

    /// Printout for new loop cycle
    void printRound(size_t i, size_t m);
  };
}

