#ifndef TEST_H_
#define TEST_H_

#include <algorithm>
#include <vector>
#include <map>
#include <functional>

struct DataObject  {
public:
  int i;    //!
  int j;    //!
  int k;    //!
  int l;    //!
  int m;    //!
  DataObject() {}
  virtual ~DataObject() {}
  virtual void release() {}
};

struct IInterface  {
public:
  virtual void release() = 0;
};

namespace Relations   {
  template < class FROM, class TO> struct RelationTypeTraits {
    /// type traits version  
    enum { version = 1 };
    /// actual type of "FROM" object 
    typedef FROM      From;
    /// actual type of "TO" object 
    typedef TO        To;    
    /// "equality" function object for "To" objects 
    typedef typename std::equal_to<To>         EqualT;

    struct Entry: public std::pair<From,To> {
      /// shortcut to the own base
      typedef std::pair<From,To>  Pair;

      /// constructor
      Entry( const From& f = From () , 
             const To&   t = To   () ): Pair( f , t ) {};
    };
    typedef typename std::vector<Entry> Entries;
  };
}

template<class FROM, class TO> struct IRelation 
: public virtual IInterface  {
public:
  /// the type traits structure
  typedef Relations::RelationTypeTraits<FROM,TO> TypeTraits;
  /// actual "FROM" type
  typedef typename TypeTraits::From                       From;
  /// actual "TO" type
  typedef typename TypeTraits::To                         To;
  /// shortcut to own type
  typedef IRelation<FROM,TO>                     OwnType;
  /// shortcut to "direct" type
  typedef IRelation<FROM,TO>                     DirectType;
  /// shortcut to "inverse" type
  typedef IRelation<TO,FROM>                     InverseType;

protected:

  virtual ~IRelation() {}
};

namespace Relations {
  template<class FROM,class TO> struct RelationBase  {  
  public:    
    /// short cut to own type 
    typedef RelationBase<FROM,TO>                  OwnType;
    /// short cut to type of inverse relations 
    typedef RelationBase<TO,FROM>                  InvType;
    /// shortcut to type traits structure 
    typedef Relations::RelationTypeTraits<FROM,TO> TypeTraits;
    /// actual "FROM" type 
    typedef typename TypeTraits::From                       From;
    /// actual "TO"   type 
    typedef typename TypeTraits::To                         To;  
    /// Entry type  
    typedef typename TypeTraits::Entry                      Entry;
    /// container  type  
    typedef typename TypeTraits::Entries                    Entries;
    /// size_type 
    typedef typename Entries::size_type                     size_type;
  public:
    RelationBase() : m_entries ()    {}
    /// destructor (virtual)
    virtual ~RelationBase() {};
    // the actual storage of references 
    mutable Entries m_entries; ///< the actual storage of references 
  };

  template<class FROM,class TO> struct Relation 
    : virtual public IRelation<FROM,TO> {  
  public:
    /// short cut for own     type
    typedef          Relation<FROM,TO>                OwnType;
    /// short cut for inverse type
    typedef          Relation<TO,FROM>                InvType;
    /// short cut for interface 
    typedef          IRelation<FROM,TO>               IBase;
    /// short cut for the actual implementation type 
    typedef typename Relations::RelationBase<FROM,TO> Base;
    /// shortcut for inverse base
    typedef typename Relations::RelationBase<TO,FROM> InvBase;
    /// shortcut for direct base type 
    typedef          Base                             Direct;
    /// shortcut for inverse base type 
    typedef          InvBase                          Inverse;
    /// shortcut for direct  interface 
    typedef          IBase                            IDirect;
    /// shortcut for inverse base type 
    typedef typename IDirect::InverseType             IInverse;
    /// basic types from Interface 
    typedef typename IBase::From  From;
    /// basic types from Interface 
    typedef typename IBase::To    To;
  public:
    Direct   m_direct;   ///< the holder of all direct relations
    Inverse* m_inverse;  ///! the pointer to inverse table
    
    /// the default constructor
    Relation () : m_direct() , m_inverse(0)     {}
    virtual ~Relation() {}
    virtual void release() {}
  };
}

template<class FROM,class TO> struct Relation1D :
  public         DataObject,
  virtual public IRelation<FROM,TO>,
  virtual public Relations::Relation<FROM,TO>
{
public:
  /// short cut for own     type
  typedef Relation1D<FROM,TO>              OwnType;
  /// short cut for inverse type
  typedef Relation1D<TO,FROM>              InvType;
  /// short cut for interface 
  typedef IRelation<FROM,TO>               IBase;
  /// short cut for the actual implementation type 
  typedef typename Relations::Relation<FROM,TO>     Base;
public:
  /// the default constructor
  Relation1D ( const size_t reserve = 0 ) : IBase(), Base(), DataObject()   {  };
  virtual ~Relation1D()  {  }
};
#endif
