#include "pool/TestBase.h"
#include "StorageSvc/DbPrint.h"
#include "POOLCore/strcasecmp.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/DbReflex.h"
#include <utility>
#include <stdexcept>

#include "TROOT.h"
#include "TList.h"
#include "Api.h"
#include "TClass.h"
#include "TRealData.h"
#include "TDataMember.h"
#include "TStreamerInfo.h"
#include "TMemberInspector.h"
#include "TStreamerElement.h"
#ifdef WIN32
  extern "C" __declspec(dllimport) void* __stdcall LoadLibraryA(const char*);
  #define LOAD_LIB(x)  (void*)::LoadLibraryA( #x )
#else
  #include <dlfcn.h>
  #include <unistd.h>
  #define LOAD_LIB(x)  ::dlopen( "lib" #x , RTLD_NOW)
#endif

unsigned long getLastError();
const std::string getErrorString(unsigned long error);
using namespace pool;

namespace pool  {
  typedef pool::DbPrint DbLog;
  void setDumpLinkTables(DbPrintLvl::MsgLevel value);
  int  numTokenInstances();
  const std::string typeName(const std::type_info&);
}

namespace pool {
  void genMD5(const std::string& s, void* code);
}

namespace {
  static std::vector< TestExecuter* >& tests()  {
    static std::vector< TestExecuter* > s_tests;
    return s_tests;
  }
  std::pair<std::string,std::string> splitOpt(const std::string& optString)    {
    const char* c = optString.c_str();
    std::string first = optString;
    std::string second = optString;
    if( *c == '-' || *c == '/' )         {
      ++c;
      first = second = c;
      const char* ind = ::strchr(c, '=');
      if ( ind )  {
        first = optString.substr(1,ind-c);
        second = ++ind;
      }
      else {
        ind = ::strchr(c,' ');
        if ( ind )  {
          first = optString.substr(1,ind-c);
          second = ++ind;
        }
      }
    }
    return std::make_pair(first, second);
  }
  bool getpackageOpt(const std::string& which, const std::string& optString, std::string& result)    {
    std::pair<std::string,std::string> r=splitOpt(optString);
    if ( ::strncasecmp(r.first.c_str(), which.c_str(), which.length())==0 )   {
      result = r.second;
      return true;
    }
    return false;
  }
}

extern "C" long runTest(int argc, char** argv)    {
  std::vector< TestExecuter* >::const_iterator j;
  if ( argc < 2 || argv[1][0] == '-' )  {
    std::cout << "Insufficient number of arguments." << std::endl;
    std::cout << "test <test-name> -opt <-opt>"      << std::endl;
    std::cout << "type test <test-name> -? for help" << std::endl;
    return Error;
  }
  int i;
  std::string o;
  DbPrint::setMsgVerbosity(DbPrintLvl::Info);
  for (i = 0; i < argc; ++i) {
    if ( getpackageOpt("p", argv[i], o) )   {
      if ( o.length() > 0 )   {
        switch( toupper(o[0]))  {
          case 'V':  DbPrint::setMsgVerbosity(DbPrintLvl::Verbose);    break;
          case 'D':  DbPrint::setMsgVerbosity(DbPrintLvl::Debug);      break;
          case 'I':  DbPrint::setMsgVerbosity(DbPrintLvl::Info);       break;
          case 'W':  DbPrint::setMsgVerbosity(DbPrintLvl::Warning);    break;
          case 'E':  DbPrint::setMsgVerbosity(DbPrintLvl::Error);      break;
          case 'F':  DbPrint::setMsgVerbosity(DbPrintLvl::Fatal);      break;
        }
      }
    }
    else if ( getpackageOpt("ld", argv[i], o) ) {
      TestBase::loadLibrary(o);
    }
  }
  std::cout << "Command line:" << std::endl;
  for (i = 0; i < argc; ++i )  {
    std::cout << argv[i] << " ";
  }
  std::cout << std::endl;
  const char* wanted = argv[1];
  for ( j=tests().begin(); j != tests().end(); ++j )  {
    if ( strncmp(wanted, (*j)->label().c_str(), strlen(argv[1])) == 0 )  {
      return (*j)->runme(argc-2, &argv[2]);
    }
  }
  std::cout << "Unknown test -- nothing to be executed.";
  return Error;
}

TestExecuter::TestExecuter(const std::string& l) : m_label(l) {
  tests().push_back(this);
}

/// Standard Constructor
TestBase::TestBase(long type,const std::string& f1,const std::string& f2,bool init) {
  m_type    = type;
  m_cache   = 0;
  m_dbName1 = f1;
  m_dbName2 = f2;
  m_doInit  = init;
  m_storageSvc = 0;
  m_innerLoops = 400;
  m_outerLoops = 1;
  m_objSize = 100;
  //m_libraries.push_back("RootStorageSvc");
  setDumpLinkTables(DbPrintLvl::Always);
}

/// Standard Destructor
TestBase::~TestBase()   {
  deletePtr( m_cache );
  releasePtr( m_storageSvc );
}

/// Type name
std::string TestBase::typeName(const std::type_info& info)  {
  return pool::typeName(info);
}

/// Create POOL token from basics
Token* TestBase::makeToken(const std::string& db, const std::string& cnt, int tech, const std::string& cl, int i1, int i2) const {
  Token* t1 = new Token();
  t1->setDb(db);
  t1->setCont(cnt);
  t1->setTechnology(tech);
  t1->setClassID(getGuid(cl));
  t1->oid().first  = i1;
  t1->oid().second = i2;
  return t1;
}

/// Printout for new loop cycle
void TestBase::printRound(size_t i, size_t m)  {
  if ( (i % m) == 0 ) {
    std::cout << "======================= New round [" << i << "]=====================" << std::endl;
    cache()->setPrintout(1);
  }
  else {
    cache()->setPrintout(0);
  }
}

/// Retrieve GUID for a single class
Guid TestBase::getGuid(const std::string& cl_name)  const {
  ClassH cl = cache()->loadClass(cl_name);
  if ( cl ) {
    ROOT::Reflex::PropertyList pl = cl.Properties();
    if ( pl.HasProperty("ClassID") )  {
      return Guid(pl.PropertyAsString("ClassID"));
    }
    Guid guid(Guid::null());
    genMD5(DbReflex::fullTypeName(cl),&guid);
    return guid;
  }
  std::cout << "WARNING: no CLASSID for class " << cl_name << std::endl;
  Guid guid(Guid::null());
  genMD5(cl_name,&guid);
  return guid;
}

/// Initialize the test
DbStatus TestBase::initialize()   {
  initPlugins();
  for(StringVec::const_iterator j=m_libraries.begin(); j != m_libraries.end(); ++j) {
    if ( 0 == loadLibrary(*j) )  {
      std::cout << "Failed to load library:" << *j << std::endl;
      return Error;
    }
    else  {
      std::cout << "Successfully loaded library:" << *j << std::endl;
    }
  }
  return Success;
}

/// Initialize and return cache object
IDataSvc* TestBase::cache()  const  {
  if ( 0 == m_cache )  {
    (const_cast<TestBase*>(this))->m_cache = new IDataSvc();
  }
  return m_cache;
}

/// Initialize and return storage service object
IStorageSvc* TestBase::storageSvc()  const  {
  return cache()->storageSvc();
}

/// Utility function to check for errors during execution
void TestBase::checkError(DbStatus sc, const std::string& msg, bool throw_exc)  {
  if ( !sc.isSuccess() )  {
    DbPrint log ("Logger");
    log << DbPrintLvl::Error;
    if ( throw_exc )  {
      log << "[Fatal-Error]: ";
    }
    log << msg << DbPrint::endmsg;
    if ( throw_exc )  {
      throw std::runtime_error(msg);
    }
  }
}

/// Print test usage    
void TestBase::usage(const char* pgm)   {
  std::cout << pgm << " -opt <-opt>                               \n"
            << "  -r         Read produced files                  \n"
            << "  -w         Write mode                           \n"
            << "  -u         Update mode                          \n"
            << "  -d         Destroy mode                         \n"
            << "  -c=<val>   Set selection criteria               \n"
            << "  -t         Use ROOT tree mode to store data.    \n"
            << "  -k         Use ROOT keyed objects to store data.\n"
            << "  -l=<lib>   Add dynamic library to be loaded.    \n"
            << "  -p=<val>   Set printout level (V,D,I,W,E,F)     \n"
            << "  -f1=<name> Name of data file #1.                \n"
            << "  -f2=<name> Name of data file #2.                \n"
            << "  -n=<val>   Total number of inner loops.         \n"
            << "  -m=<val>   Total number of outer loops.         \n"
            << "  -osize=<val> Object size                        \n"
            << "  -h         Show this help                       \n"
            << "  -?         Show this help                       \n"
            << std::endl;
  ::exit(0);
}

DbStatus TestBase::initPlugins() {
  return DbStatus::Success;
}

DbStatus TestBase::loadClass(const std::string& nam,ClassH& cl,Guid& guid) {
  cl = cache()->loadClass(nam);
  if ( 0 == cl ) {
    std::cout << "Cannot find class \"" << nam << "\"" << std::endl;
    return DbStatus::Error;
  }
  guid = getGuid(nam);
  return guid==Guid::null() ? DbStatus::Error : DbStatus::Success;
}

/// Load dynamic library
void* TestBase::loadLibrary(const std::string& name)
{
#ifdef WIN32
  std::string lib_name = name;
  void* handle = ::LoadLibraryA( name.c_str() );
  if ( 0 == handle )  {
    std::cout << getErrorString(getLastError()) << std::endl;
  }
#else
  std::string lib_name = "lib"+name+".so";
  void* handle = ::dlopen( lib_name.c_str() , RTLD_NOW|RTLD_GLOBAL);
  if ( 0 != handle && name == "test_dict" ) {
    handle = ::dlopen( "libtest_dictDict.so" , RTLD_NOW|RTLD_GLOBAL);
  }
#endif
  if ( 0 == handle )    {
    std::cout << "Failed to load dynamic library:" << name 
              << " <" << lib_name << ">"
              << std::endl;
  }
  return handle;
}

/// Dump root streamer info
void TestBase::dumpClass(const std::string& nam)   {
  std::cout << "==================================================================" << std::endl;
  std::cout << "===            CLASS: " << nam << std::endl;
  std::cout << "==================================================================" << std::endl;
  dumpStreamerInfo(nam);
  dumpRealData(nam);
  dumpDataMembers(nam);
}

/// Dump root streamer info
void TestBase::dumpStreamerInfo(const std::string& nam)   {
  TClass* cl = gROOT->GetClass(nam.c_str());
  if ( cl )  {
    std::cout << "=== STREAMER Info of: " << nam << std::endl;
    cl->GetStreamerInfo()->ls();
  }
  else {
    std::cout << "dumpStreamerInfo> No TClass structure available for:" << nam << std::endl;
  }
}

/// Dump root real data info
void TestBase::dumpRealData(const std::string& nam)   {
  TClass* cl = gROOT->GetClass(nam.c_str());
  if ( cl )  {
    TRealData *dm;
    TIter   next(cl->GetListOfRealData());
    std::cout << "=== REAL DATA Info of:" << nam << std::endl;
    while ((dm = (TRealData *) next()))  {
      std::cout << std::setw(16) << std::left << "showRealdata>"
          << cl->GetName() << "::" << dm->GetName() 
          << " Member:"  << (void*)dm 
          << ","  << (void*)dm->GetStreamer()
          << "," << (void*)dm->GetDataMember() << std::endl
          << std::setw(16) << std::left << " " << dm->GetDataMember()->GetFullTypeName()
          << " Offset:" << dm->GetThisOffset()
          << std::endl;
    }
  }
  else {
    std::cout << "dumpRealData> No TClass structure available for:" << nam << std::endl;
  }
}

/// Dump root data member info
void TestBase::dumpDataMembers(const std::string& nam)   {
  TClass* cl = gROOT->GetClass(nam.c_str());
  if ( cl )  {
    TDataMember *dm;
    DbPrint log(  "CINT");
    TIter   next(cl->GetListOfDataMembers());
    std::cout << "=== DATA MEMBERS Info of:" << nam << std::endl;
    while ((dm = (TDataMember*) next()))  {
      std::cout << std::setw(16) << std::left << "showDataMembers>"
          << cl->GetName() << "::" << dm->GetName() 
          << " Member:"  << (void*)dm 
          << ","  << dm->GetTypeName()
          << std::setw(16) << std::left << " " << (void*)dm->GetClass()
          << ", " << dm->GetFullTypeName()
          << std::endl;
    }
  }
  else {
    std::cout << "dumpDataMembers> No TClass structure available for:" << nam << std::endl;
  }
}

/// Execute write
DbStatus TestBase::write()    {
  std::cout << "This test has no WRITE mode." << std::endl;
  return DbStatus::Error;
}

/// Execute read
DbStatus TestBase::read()   {
  std::cout << "This test has no READ mode." << std::endl;
  return DbStatus::Error;
}

/// Execute update
DbStatus TestBase::update()   {
  std::cout << "This test has no UPDATE mode." << std::endl;
  return DbStatus::Error;
}

/// Execute update
DbStatus TestBase::destroy()  {
  std::cout << "This test has no DESTROY mode." << std::endl;
  return DbStatus::Error;
}

void TestBase::accessRootClass(const std::string& name) {
  TClass* cl = gROOT->GetClass(name.c_str());
  if ( cl ) return;
  std::cout << "Class " << name << " is not accessible to ROOT." << std::endl;
}

/// Run the test
DbStatus TestBase::run(int argc, char** argv)   {
  try {
    updateArgs(argc, argv);
    DbStatus sc = initialize();
    if ( sc.isSuccess() )   {
      std::cout << "Test setup for :" << typeName(typeid(*this)) << std::endl 
        << "===========" << std::endl
        << "  Technology:" << DbType(m_type).storageName()   
        << " [" << (void*)m_type << "]" << std::endl;
      if ( m_dbName1.length() > 0 )  {
        std::cout << "  Data file #1: " << m_dbName1 << std::endl;
      }
      else {
        std::cout << "  Data file #1: is VOID" << std::endl;
      }
      if ( m_dbName2.length() > 0 )  {
        std::cout << "  Data file #2: " << m_dbName2 << std::endl;
      }
      std::cout << "  Number of OUTER loops:" << m_outerLoops << std::endl;
      std::cout << "  Number of INNER loops:" << m_innerLoops << std::endl;
      std::cout << "  Object size:          " << m_objSize    << std::endl;
      std::cout << "  Selection criteria:   " << m_criteria   << std::endl;
      std::cout << std::endl;
      switch(m_action) {
        case READ:
          std::cout << "===========  Running READ test ===" << std::endl;
          sc = read();
          break;
        case WRITE:
          std::cout << "===========  Running WRITE test ===" << std::endl;
          sc = write();
          break;
        case UPDATE:
          std::cout << "===========  Running UPDATE test ===" << std::endl;
          sc = update();
          break;
        case DESTROY:
          std::cout << "===========  Running DESTROY test ===" << std::endl;
          sc = destroy();
          break;
        default:
          break;
      }
      std::cout << "Number of token instances: " << numTokenInstances() << std::endl;
    }
    return sc.code();
  }
  catch(std::exception& e)   {
    std::cout << "The test has thrown an exception:" << e.what() << std::endl;
  }
  catch(...)    {
    std::cout << "The test has thrown an unknown exception." << std::endl;
  }
  return Error;
}

bool TestBase::getOpt(const std::string& which, 
                            const std::string& optString, 
                            std::string& result)
{
   if ( getpackageOpt(which, optString, result) )  {
     m_options[0].push_back(which);
     m_options[1].push_back(result);
     return true;
   }
   return false;
}

bool TestBase::hasOpt(const std::string& w)  const  {
  StringVec::const_iterator i;
  for(i=m_options[0].begin(); i != m_options[0].end(); ++i) {
    if ( strncmp((*i).c_str(), w.c_str(), w.length()) == 0 ) {
      return true;
    }
  }
  return false;
}

/// Parse arguments
void TestBase::updateArgs(int argc, char** argv)  {
  std::string o;
  for ( int i = 0; i < argc; ++i) {
    if      ( getOpt("r", argv[i], o) )  {
      m_action  = READ;
    }
    else if ( getOpt("w", argv[i], o) )  {
      m_action  = WRITE;
    }
    else if ( getOpt("upd", argv[i], o) )  {
      m_action  = UPDATE;
    }
    else if ( getOpt("d", argv[i], o) )  {
      m_action  = DESTROY;
    }
    else if ( getOpt("c", argv[i], o) )  {
      m_criteria  = o;
    }
    else if ( getOpt("k", argv[i], o) )  {
      m_type    = ROOTKEY_StorageType.type();
    }
    else if ( getOpt("t", argv[i], o) )  {
      m_type    = ROOTTREE_StorageType.type();
    }
    else if ( getOpt("f1",argv[i], o) )  {
      m_dbName1 = o;
    }
    else if ( getOpt("f2",argv[i], o) )  {
      m_dbName2 = o;
    }
    else if ( getOpt("h", argv[i], o) )  {
      usage(argv[0]);
    }
    else if ( getOpt("?", argv[i], o) )  {
      usage(argv[0]);
    }
    else if ( getOpt("p", argv[i], o) )   {
    }
    else if ( getOpt("l", argv[i], o) ) {
      m_libraries.push_back(o);
    }
    else if ( getOpt("n", argv[i], o) )   {
      int n = m_innerLoops;
      if ( 1 == ::sscanf(o.c_str(),"%d",&n) )  {
        m_innerLoops = n;
      }
    }
    else if ( getOpt("m", argv[i], o) )   {
      int n = m_outerLoops;
      if ( 1 == ::sscanf(o.c_str(),"%d",&n) )  {
        m_outerLoops = n;
      }
    }
    else if ( getOpt("osize", argv[i], o) )   {
      int n = m_objSize;
      if ( 1 == ::sscanf(o.c_str(),"%d",&n) )  {
        m_objSize = n;
      }
    }
    else  {
      std::pair<std::string,std::string> r=splitOpt(argv[i]);
      m_options[0].push_back(r.first);
      m_options[1].push_back(r.second);
    }
  }
}
