#include "pool/TestBase.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/DbReflex.h"
#include "StorageSvc/DbSelect.h"
#include "StorageSvc/DbOption.h"
#include "StorageSvc/DbColumn.h"
#include "StorageSvc/DbTypeInfo.h"
#include "StorageSvc/IStorageSvc.h"
#include "StorageSvc/IStorageExplorer.h"
#include "StorageSvc/FileDescriptor.h"
#include <stdexcept>

namespace pool  {

  void checkError(DbStatus sc, const std::string& msg, bool throw_exc)  {
    if ( !sc.isSuccess() )    {
      DbPrint log ("Logger");
      log << DbPrintLvl::Error;
      if ( throw_exc )  {
        log << "[FATAL-ERROR]: ";
      }
      log << msg << DbPrint::endmsg;
      if ( throw_exc )  {
        throw std::runtime_error(msg);
      }
    }
  }
  template<class T>
  DbStatus printDomOpt(const std::string& nam, IStorageExplorer* exp, SessionH sesH, const char* o="") {
    T ival;
    DbOption opt(nam, o);
    DbStatus sc = exp->getDomainOption(sesH, opt);
    if ( !sc.isSuccess() )   {
      checkError(sc,"FAILED: getDomainOption(" + nam + ")", true);
      return sc;
    }
    std::string desc = "Domain option ["+nam+"]";
    if ( opt.option().length() > 0 ) desc += "["+opt.option()+"]";
    opt._getValue(ival);
    std::cout << "Getting " << std::setw(40) << std::left << desc << ival << std::endl;
    return sc;
  }

  template<class T>
  DbStatus setDomOpt(const std::string& nam, T val, IStorageExplorer* exp, SessionH sesH) {
    DbOption opt(nam,"",val);
    DbStatus sc = exp->setDomainOption(sesH, opt);
    if ( !sc.isSuccess() )   {
      checkError(sc,"FAILED: setDomOption(" + nam + ")", true);
    }
    std::string desc = "Domain option ["+nam+"]";
    std::cout << "Setting " << std::setw(40) << std::left << desc << val << std::endl;
    return sc;
  }

  template<class T>
  DbStatus printDbOpt(const std::string& nam, IStorageExplorer* exp, FileDescriptor& dsc) {
    T ival;
    DbOption opt(nam);
    DbStatus sc = exp->getDatabaseOption(dsc, opt);
    if ( !sc.isSuccess() )   {
      checkError(sc,"FAILED: getDatabaseOption(" + nam + ")", true);
      return sc;
    }
    opt._getValue(ival);
    std::cout << "Getting " << std::setw(40) << std::left << "Database option ["+nam+"]:" << ival << std::endl;
    return sc;
  }

  template<class T>
  DbStatus printContOpt(const std::string& nam, 
                      IStorageExplorer* exp,
                      FileDescriptor& dsc,
                      const std::string& cntName,
                      const std::string& brName="") {
    T val;
    DbOption opt(nam, brName);
    DbStatus sc = exp->getContainerOption(dsc, cntName, opt);
    if ( !sc.isSuccess() )   {
      checkError(sc,"FAILED: getContainerOption(" + nam + ")", true);
      return sc;
    }
    opt._getValue(val);
    std::string desc = "Container option ["+cntName;
    if ( !brName.empty() )  {
      desc += "["+brName+"]";
    }
    desc += "."+nam+"]:  ";
    std::cout << "Getting " << std::setw(50) << std::left << desc << val << std::endl;
    return sc;
  }

  template<class T>
  DbStatus setDbOpt(const std::string& nam, T val, IStorageExplorer* exp, FileDescriptor& dsc) {
    DbOption opt(nam,"",val);
    DbStatus sc = exp->setDatabaseOption(dsc, opt);
    if ( !sc.isSuccess() )   {
      checkError(sc,"FAILED: setDbOption(" + nam + ")", true);
    }
    std::cout << "Setting " << std::setw(40) << std::left << "Database option ["+nam+"]:" << val << std::endl;
    return sc;
  }

  template<class T>
  DbStatus setContOpt(const std::string& nam, T val,
                    IStorageExplorer* exp,
                    FileDescriptor& dsc,
                    const std::string& cntName) 
  {
    DbOption opt(nam, "test", val);
    DbStatus sc = exp->setContainerOption(dsc, cntName, opt);
    if ( !sc.isSuccess() )   {
      checkError(sc,"FAILED: setContainerOption(" + nam + ")", true);
    }
    std::string desc = "Container option ["+cntName+"."+nam+"]:  ";
    std::cout << "Setting " << std::setw(50) << std::left << desc << val << std::endl;
    return sc;
  }

  /** @class DbOptions
    *
    */
  class DbOptions : public TestBase {
  public: 

    /// Standard Constructor
    DbOptions() : TestBase(ROOTKEY_StorageType.type(), "test_1_1.root", "test_1_2.root") {
      m_criteria = "m_char<=3";
    }

    /// Standard Destructor
    virtual ~DbOptions()               {    }

    void printDom(IStorageExplorer* exp, SessionH sesH)    {
      printDomOpt<int>("DEFAULT_COMPRESSION",exp, sesH);
      setDomOpt<int>  ("DEFAULT_COMPRESSION",3, exp, sesH);
      printDomOpt<int>("DEFAULT_COMPRESSION",exp, sesH);
      setDomOpt<int>  ("DEFAULT_COMPRESSION",1, exp, sesH);
      printDomOpt<int>("DEFAULT_SPLITLEVEL", exp, sesH);
      setDomOpt<int>  ("DEFAULT_SPLITLEVEL", 15, exp, sesH);
      printDomOpt<int>("DEFAULT_SPLITLEVEL", exp, sesH);
      setDomOpt<int>  ("DEFAULT_SPLITLEVEL", 99, exp, sesH);
      printDomOpt<int>("DEFAULT_AUTOSAVE",   exp, sesH);
      setDomOpt<int>  ("DEFAULT_AUTOSAVE",   12345600, exp, sesH);
      printDomOpt<int>("DEFAULT_AUTOSAVE",   exp, sesH);
      setDomOpt<int>  ("DEFAULT_AUTOSAVE",   10000000, exp, sesH);
      printDomOpt<int>("DEFAULT_BUFFERSIZE", exp, sesH);
      setDomOpt<int>  ("DEFAULT_BUFFERSIZE", 1024, exp, sesH);
      printDomOpt<int>("DEFAULT_BUFFERSIZE", exp, sesH);
      setDomOpt<int>  ("DEFAULT_BUFFERSIZE", 16000, exp, sesH);
      printDomOpt<int>("ABORT_LEVEL", exp, sesH);
      setDomOpt<int>  ("ABORT_LEVEL", 3, exp, sesH);
      printDomOpt<int>("ABORT_LEVEL", exp, sesH);
      setDomOpt<int>  ("ABORT_LEVEL", 4001, exp, sesH);
      printDomOpt<int>("ERROR_LEVEL", exp, sesH);
      setDomOpt<int>  ("ERROR_LEVEL", 3, exp, sesH);
      printDomOpt<int>("ERROR_LEVEL", exp, sesH);
      setDomOpt<int>  ("ERROR_LEVEL", -1, exp, sesH);
      printDomOpt<int>("NUM_CLASSES",        exp, sesH);
      printDomOpt<int>("NUM_TYPES",          exp, sesH);
      printDomOpt<int>("ERRNO",              exp, sesH);
      printDomOpt<void*>("CLASS", exp, sesH, "TClass");
      printDomOpt<void*>("CLASS", exp, sesH, "TTree");
      printDomOpt<void*>("CLASS", exp, sesH, "TBranch");
      printDomOpt<void*>("CLASS", exp, sesH, "test");
    }

    /// Print DB options
    DbStatus printDbOptions(IStorageExplorer* exp, FileDescriptor& dsc)  {
      printDbOpt<int>   ("DEFAULT_COMPRESSION", exp, dsc);
      setDbOpt<int>     ("DEFAULT_COMPRESSION", 2, exp, dsc);
      printDbOpt<int>   ("DEFAULT_COMPRESSION", exp, dsc);
      setDbOpt<int>     ("DEFAULT_COMPRESSION", 1, exp, dsc);
      printDbOpt<int>   ("DEFAULT_SPLITLEVEL",  exp, dsc);
      setDbOpt<int>     ("DEFAULT_SPLITLEVEL",  15, exp, dsc);
      printDbOpt<int>   ("DEFAULT_SPLITLEVEL",  exp, dsc);
      setDbOpt<int>     ("DEFAULT_SPLITLEVEL",  99, exp, dsc);
      printDbOpt<int>   ("DEFAULT_AUTOSAVE",    exp, dsc);
      setDbOpt<int>     ("DEFAULT_AUTOSAVE",    123450, exp, dsc);
      printDbOpt<int>   ("DEFAULT_AUTOSAVE",    exp, dsc);
      setDbOpt<int>     ("DEFAULT_AUTOSAVE",    1000000, exp, dsc);
      printDbOpt<int>   ("DEFAULT_BUFFERSIZE",  exp, dsc);
      setDbOpt<int>     ("DEFAULT_BUFFERSIZE",  12340, exp, dsc);
      printDbOpt<int>   ("DEFAULT_BUFFERSIZE",  exp, dsc);
      setDbOpt<int>     ("DEFAULT_BUFFERSIZE",  10000, exp, dsc);
      printDbOpt<int>   ("COMPRESSION_LEVEL",   exp, dsc);
      setDbOpt<int>     ("COMPRESSION_LEVEL",   2, exp, dsc);
      printDbOpt<int>   ("COMPRESSION_LEVEL",   exp, dsc);
      setDbOpt<int>     ("COMPRESSION_LEVEL",   1, exp, dsc);
      printDbOpt<double>("COMPRESSION_FACTOR",  exp, dsc);
      printDbOpt<int>   ("BEST_BUFFER",         exp, dsc);
      printDbOpt<double>("BYTES_WRITTEN",       exp, dsc);
      printDbOpt<double>("BYTES_READ",          exp, dsc);
      printDbOpt<int>   ("BYTES_FREE",          exp, dsc);
      printDbOpt<int>   ("BYTES_INFO",          exp, dsc);
      printDbOpt<int>   ("BYTES_KEYS",          exp, dsc);
      // printDbOpt<int>   ("NFREE",               exp, dsc);
      printDbOpt<int>   ("NKEYS",               exp, dsc);
      printDbOpt<double>("FILEBYTES_WRITTEN",   exp, dsc);
      printDbOpt<double>("FILEBYTES_READ",      exp, dsc);
      printDbOpt<int>   ("FILE_DESCRIPTOR",     exp, dsc);
      printDbOpt<int>   ("FILE_VERSION",        exp, dsc);
      printDbOpt<int>   ("FILE_SIZE",           exp, dsc);
      printDbOpt<int>   ("FILE_ERROR",          exp, dsc);
      printDbOpt<int>   ("FILE_END",            exp, dsc);
      printDbOpt<void*> ("FILE_KEYS",           exp, dsc);
      printDbOpt<void*> ("TFILE",               exp, dsc);
      printDbOpt<long long> ("TREE_MAX_SIZE",    exp, dsc);
      setDbOpt<long long>   ("TREE_MAX_SIZE",  8*1900000000LL, exp, dsc);
      printDbOpt<long long> ("TREE_MAX_SIZE",    exp, dsc);
      setDbOpt<long long>   ("TREE_MAX_SIZE",  1900000000LL, exp, dsc);
      printDbOpt<long long> ("TREE_MAX_SIZE",    exp, dsc);
      setDbOpt<double>  ("FILEBYTES_WRITTEN",  100.0, exp, dsc);
      setDbOpt<double>  ("FILEBYTES_READ",     100.0, exp, dsc);
      printDbOpt<double>("FILEBYTES_WRITTEN",   exp, dsc);
      printDbOpt<double>("FILEBYTES_READ",      exp, dsc);
      setDbOpt<int>     ("PRINT",               0, exp, dsc);

      return DbStatus::Success;
    }

    /// Print DB options
    DbStatus printKeyContOptions(IStorageExplorer* exp, 
                                  FileDescriptor& dsc,
                                  const std::string& cntName)  
    {
      printContOpt<void*> ("DIRECTORY",           exp, dsc, cntName);
      printContOpt<void*> ("DIR_MOTHER",          exp, dsc, cntName);
      printContOpt<void*> ("DIR_FILE",            exp, dsc, cntName);
      printContOpt<void*> ("DIR_LIST_KEYS",       exp, dsc, cntName);
      printContOpt<int>   ("DIR_NKEYS",           exp, dsc, cntName);
      printContOpt<int>   ("DIR_PRINT",           exp, dsc, cntName);
      printContOpt<int>   ("DIR_WRITABLE",        exp, dsc, cntName);
      printContOpt<int>   ("DIR_MODIFIED",        exp, dsc, cntName);
      //printContOpt<int>   ("DIR_LS",              exp, dsc, cntName);
      return poolDb::Success;
    }

    /// Print DB options
    DbStatus printTreeContOptions(IStorageExplorer* exp, 
                                  FileDescriptor& dsc,
                                  const std::string& cntName)  {
      printContOpt<void*>   ("TREE",                exp, dsc, cntName);
      printContOpt<int>     ("TREE_ENTRIES",        exp, dsc, cntName);
      printContOpt<int>     ("TREE_NBRANCHES",      exp, dsc, cntName);
      printContOpt<double>  ("TREE_TOTAL_BYTES",    exp, dsc, cntName);
      printContOpt<double>  ("TREE_ZIP_BYTES",      exp, dsc, cntName);
      printContOpt<int>     ("TREE_FILE_NUMBER",    exp, dsc, cntName);
      printContOpt<long long>("TREE_MAX_SIZE",       exp, dsc, cntName);
      setContOpt<long long>  ("TREE_MAX_SIZE",       1900000000LL*4,exp, dsc, cntName);
      printContOpt<long long>("TREE_MAX_SIZE",       exp, dsc, cntName);
      setContOpt<long long>  ("TREE_MAX_SIZE",       1900000000LL,exp, dsc, cntName);
      printContOpt<long long>("TREE_MAX_SIZE",       exp, dsc, cntName);
      printContOpt<int>     ("TREE_MAX_VIRTUAL_SIZE", exp, dsc, cntName);
      //setContOpt<int>     ("TREE_PRINT",               0,exp, dsc, cntName);

      DbOption o("TREE_NBRANCHES", "");
      o._setValue(0);
      DbStatus sc = exp->getContainerOption(dsc, cntName, o);
      int nBranches = 0;
      o._getValue(nBranches);
      for (int ibr = 0; ibr<nBranches; ++ibr)  {
        DbOption opt("TREE_BRANCH_IDX_NAME", "");
        opt._setValue(ibr);
        sc = exp->getContainerOption(dsc, cntName, opt);
        const char* brName = 0;
        opt._getValue(brName);
        printContOpt<void*> ("BRANCH",              exp, dsc, cntName, brName);
        printContOpt<int>   ("BRANCH_ENTRIES",      exp, dsc, cntName, brName);
        printContOpt<int>   ("BRANCH_MAX_BASKETS",  exp, dsc, cntName, brName);
        printContOpt<int>   ("BRANCH_BASKET_SIZE",  exp, dsc, cntName, brName);
        printContOpt<double>("BRANCH_TOTAL_SIZE",   exp, dsc, cntName, brName);
        printContOpt<double>("BRANCH_TOTAL_BYTES",  exp, dsc, cntName, brName);
        printContOpt<double>("BRANCH_ZIP_BYTES",    exp, dsc, cntName, brName);
        printContOpt<int>   ("BRANCH_COMPRESSION_LEVEL",   exp, dsc, cntName, brName);
        printContOpt<int>   ("BRANCH_NLEAVES",      exp, dsc, cntName, brName);
        printContOpt<int>   ("BRANCH_SPLIT_LEVEL",  exp, dsc, cntName, brName);
      }
      return DbStatus::Success;
    }

    void printDb(IStorageExplorer* exp, FileDescriptor& dsc)
    {
      DbStatus sc = printDbOptions(exp, dsc);
      checkError(sc,"Access database parameters");

      std::vector<ShapeH> shapes;
      IStorageExplorer::Parameters pars;

      sc = exp->dbParams(dsc, pars);
      checkError(sc,"Cannot access database parameters", true);
      std::cout << " ============== Parameters ==============" << std::endl;
      for(IStorageExplorer::Parameters::iterator l=pars.begin(); l != pars.end(); ++l)
      {
        std::cout << (*l).first << "=" << (*l).second << std::endl;
      }

      sc = exp->shapes(dsc, shapes);
      checkError(sc,"Cannot access database shapes", true);
      std::cout << " ============== Shapes ==================" << std::endl;
      for(std::vector<ShapeH>::const_iterator q=shapes.begin(); q != shapes.end(); ++q)
      {
        const DbTypeInfo* pShape = dynamic_cast<const DbTypeInfo*>(*q);
        const DbTypeInfo::Columns& cols = pShape->columns();
        std::cout << "---> Shape[" << pShape->shapeID().toString() << "]: ";
        std::cout << "[" << cols.size() << " Column(s)]" << std::endl;
        if ( pShape->clazz() )
          std::cout << "---->Class:" << DbReflex::fullTypeName(pShape->clazz()) << std::endl;
        else
          std::cout << "---->Class:" << "<not availible>" << std::endl;
        for (size_t ic=0; ic < cols.size();++ic)  {
          const DbColumn* c = cols[ic];
          std::cout << "---->[" << ic << "]:" << c->name()
                    << " Typ:" << c->typeName() << " ["<< c->typeID() << "]"
                    << " Size:" << c->size()
                    << " Offset:" << c->offset()
                    << " #Elements:" << c->nElement()
                    << std::endl;
        }
      }

      std::vector<const Token*> conts;
      sc = exp->associations(dsc, conts);
      checkError(sc,"Cannot access database associations", true);
      std::cout << " ============== Associations ============" << std::endl;
      for(std::vector<const Token*>::iterator m=conts.begin(); m != conts.end(); ++m)
      {
        std::cout << (*m)->toString() << std::endl;
      }

      sc = exp->containers(dsc, conts);
      checkError(sc,"Cannot access database containers", true);
      std::cout << " ============== Containers ==============" << std::endl;
      std::vector<const Token*>::iterator p;
      for(p=conts.begin(); p != conts.end(); ++p)
      {
        std::cout << (*p)->contID() << ":" << DbType((*p)->technology()).storageName() << std::endl;
      }
      for(p=conts.begin(); p != conts.end(); ++p)  {
        std::cout << " ============== Contain Options of:" 
                  << (*p)->contID() << ":" << DbType((*p)->technology()).storageName() 
                  << std::endl;
        if ( (*p)->technology() == ROOTTREE_StorageType.type() )  {
          DbSelect sel("");
          sc = exp->select(dsc, (*p)->contID(), sel);
          sc = printTreeContOptions(exp, dsc, (*p)->contID());
          checkError(sc,"Access database parameters", true);
        }
        else if ( (*p)->technology() == ROOTKEY_StorageType.type() )  {
          DbSelect sel("");
          sc = exp->select(dsc, (*p)->contID(), sel);
          sc = printKeyContOptions(exp, dsc, (*p)->contID());
          checkError(sc,"Access database parameters", true);
        }
      }
    }

    /// Read objects from the persistent store
    DbStatus read()  {
      SessionH sesH = 0;
      DbStatus sc = DbStatus::Error;
      std::string criteria = m_criteria;
      IStorageSvc* storSvc = storageSvc();
      IStorageExplorer* exp = 0;
      sc = storSvc->queryInterface(IStorageExplorer::interfaceID(),(void**)&exp);
      checkError(sc,"queryInterface(IStorageExplorer::interfaceID(),(void**)&exp)");
      if ( !exp ) {
        std::cout << "Cannot access storage explorer object." << std::endl;
        return DbStatus::Error;
      }
      sc = exp->startSession(poolDb::READ, m_type, sesH);
      checkError(sc,"startSession(poolDb::READ, m_type, sesH)", true);

      printDom(exp, sesH);

      FileDescriptor dsc1(m_dbName1, m_dbName1);
      sc = exp->connect(sesH, poolDb::READ, dsc1);
      checkError(sc,"connect(sesH, poolDb::READ, dsc1)", true);
      printDb(exp, dsc1);
      sc = exp->disconnect(dsc1);
      checkError(sc,"disconnect(dsc1)", true);

      FileDescriptor dsc2(m_dbName2, m_dbName2);
      sc = exp->connect(sesH, poolDb::READ, dsc2);
      checkError(sc,"connect(sesH, poolDb::READ, dsc2)", true);
      printDb(exp, dsc2);
      sc = exp->disconnect(dsc2);
      checkError(sc,"disconnect(dsc2)");

      std::cout << "DbStorageSvc::release()=" << exp->release() << std::endl;
      return sc;
    }
  };
}

pool::TestRunner<pool::DbOptions> s_testDbOptions("DbOptions");
