#ifndef ATTRIBUTETYPES_H
#define ATTRIBUTETYPES_H

/// @file 
/// @author Jakub.Moscicki@cern.ch (kuba)

#include <string>
#include <map>

namespace pool {

  /// String type which holds the symbolic names of attribute types.
  /// These names may be stored in the persistent storage as they always remain the same on all platforms.
  typedef std::string typename_string_t;
  
  /// String type holding sql symbolic names of attribute types. \implementation.
  typedef std::string sql_typename_string_t;
  
  /**
   * Trait class for mapping from static C++ type to runtime type identifiers.
   * T is a value type stored in the Attribute.
   *
   * If needed, for performance reasons, typeid_t may be introduced. Such type identifier would be a non-persistent-capable (not compatible between
   * compilers/platforms) number identyfying a type for internal handling which otherwise done with a string type containing a type name would
   * not be efficient.
   */
  
  template<class T>
  class AttributeStaticTypeInfo
  {
  public:
    
    /// Symbolic, portable type names. Examples: "int","string"
    static typename_string_t type_name();
    
    /// SQL type names (mainly for internal implementation SQL backends).
    static sql_typename_string_t sql_type_name();

    /// Root type name (defines mapping to root storage types for Root backends)
    /// As defined in: http://root.cern.ch/root/htmldoc/TTree.html
    static char root_type_char();
  };
  
  /// Map runtime typename string into root type character.
  char get_root_type_char(const typename_string_t& tn);


  /// Fast string converter: convert from string to some type and store 
  /// the result in the memory location. This is used in conjuncion with
  /// attribute value objects to implement very fast destringification
  /// for mysql collections.
  typedef void (*str_to_mem_function)(const char *c, void *mem);

  /// Get pointer to the converter for a given type name.
  /// Converters are defined for each supported type in 
  /// AttributePredefinedTypes.cpp
  str_to_mem_function get_str_mem_converter(const typename_string_t& tn);

  class BaseAttributeFactory;
  class AttributeBaseValue;
  
  /**
   * Singleton factory for attribute implementations.
   *
   * Allows to create an attribute object (more precisely, the value object which is contained in the attribute object) from a type name string.
   * So effectively is allows you to map from "string" to std::string at runtime.
   * 
   */
  class AttributeImplFactory
  {
  public:
    /// Create typed attribute value object.
    AttributeBaseValue* createAttributeImpl(typename_string_t type_name);
    
    /// Size of the typed attribute value object.
    size_t sizeOfAttributeObject(typename_string_t type_name);

    /// Returns the type of the attribute
    const std::type_info* typeOfAttributeObject(typename_string_t type_name) const;

    /// Returns the type_name of the attribute
    std::string typeNameOfAttributeObject( const std::type_info& type ) const;

    /// Singleton access.
    static AttributeImplFactory& theFactory();
    
    /// for friendly use only not for general public.
    void _addref(BaseAttributeFactory&, typename_string_t type_name);
    
  private: 
    AttributeImplFactory();
    
  };
}

#endif


