#include "Collection/CollectionProxy.h"

#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeSpecification.h"

#include "CollectionBase/CollectionFactory.h"
#include "CollectionBase/CollectionDescription.h"
#include "CollectionBase/ICollectionColumn.h"
#include "CollectionBase/ICollectionDataEditor.h"
#include "CollectionBase/ICollectionSchemaEditor.h"
#include "CollectionBase/CollectionRowBuffer.h"

#include "DataSvc/IDataSvc.h"

#include "boost/tokenizer.hpp"

#include "POOLCore/Exception.h"

#include <iostream>
using namespace std;

pool::CollectionProxy::CollectionProxy( pool::IDataSvc* dataSvc,
                                 std::string type,
                                 std::string connection,
                                 std::string name,
                                 pool::ICollection::OpenMode openMode ):
  m_collection( 0 ),
  m_dataSvc( dataSvc ),
  m_isNewCollection( true ),
  m_attrListSpec( 0 )
{
   pool::CollectionFactory* factory = CollectionFactory::get();
   if( factory ) {
      CollectionDescription description( name, type, connection );
      pool::ISession* session = 0;
      if ( dataSvc ) session = &( dataSvc->session() );
      if( type != "PHYSICAL_NAME" &&
          type != "LOGICAL_NAME" &&
          type != "GUID" ) {      
         m_collection = factory->create( type, description, openMode, session );
         
      } else {
         const std::string& logicalName = connection;
         if( openMode == ICollection::CREATE )
            m_collection = factory->createAndRegister( description, 0, false, logicalName, 0, session);
         else if( openMode == ICollection::CREATE_AND_OVERWRITE )
            m_collection = factory->createAndRegister( description, 0, true, logicalName, 0, session);
         else {
            bool readOnly = (openMode == ICollection::READ);
  
            if( type == "PHYSICAL_NAME" )
               m_collection = factory->openWithPhysicalName( name, 0, readOnly, session);
            else if( type == "LOGICAL_NAME" )
               m_collection = factory->openWithLogicalName( name, 0, readOnly, session);
            else if( type == "GUID" )
               m_collection = factory->openWithGuid( name, 0 , readOnly, session);
         }
      }
  
      if( openMode == ICollection::READ || openMode == ICollection::UPDATE )
         m_isNewCollection = false;
  }
}


pool::CollectionProxy::CollectionProxy( pool::IDataSvc* dataSvc,
                                 pool::ICollection* icoll ):
      m_collection( icoll ),
      m_dataSvc( dataSvc ),
      m_isNewCollection( true ),
      m_attrListSpec( 0 )
{}


pool::CollectionProxy::CollectionProxy( const CollectionProxy& rhs ) :
      m_collection( rhs.m_collection ),
      m_dataSvc( rhs.m_dataSvc ),
      m_isNewCollection( rhs.m_isNewCollection ),
      m_attrListSpec( rhs.m_attrListSpec )
{
   rhs.m_collection = 0;
   rhs.m_attrListSpec = 0;
}


pool::CollectionProxy::~CollectionProxy()
{
  delete m_collection;
  if( m_attrListSpec ) m_attrListSpec->release();
}



void
pool::CollectionProxy::makeAttrListSpec() const
{
   if( m_attrListSpec ) m_attrListSpec->release();
   m_attrListSpec = new coral::AttributeListSpecification;
   if( m_collection ) {
      const ICollectionDescription &description = m_collection->description();
      for( int f=0; f < description.numberOfCollectionFragments(); f++ ) {
         for( int c=0; c < description.numberOfAttributeColumns( f ); c++ ) {
            const pool::ICollectionColumn&        column = description.attributeColumn(c, f);
            if( column.name() != description.eventReferenceColumnName() ) 
               m_attrListSpec->extend( column.name(), column.type() );
         }
      }
   }
}
               

coral::AttributeListSpecification&
pool::CollectionProxy::attributeListSpecification()
{
   makeAttrListSpec();
   return *m_attrListSpec;
}


const coral::AttributeListSpecification&
pool::CollectionProxy::attributeListSpecification() const
{
   makeAttrListSpec();
   return *m_attrListSpec;
}



bool
pool::CollectionProxy::
add( const RefBase& pref, const coral::AttributeList& attributeList )
{
   if( !m_collection )
      return false;

  // if the collection was just created without specification, use the attributeList to define the collection columns
  if( m_isNewCollection ) {
     if( !m_attrListSpec ) {
        m_attrListSpec = new coral::AttributeListSpecification;
     }
     if( !m_attrListSpec->size() ) {
        for( coral::AttributeList::const_iterator attrIter = attributeList.begin();
             attrIter != attributeList.end(); ++attrIter ) {
           m_attrListSpec->extend( attrIter->specification().name(), attrIter->specification().type() );
        }
     } else {
        for( coral::AttributeList::const_iterator attIter = attributeList.begin();
             attIter != attributeList.end(); ++attIter ) {
           if( m_attrListSpec->index( attIter->specification().name() ) < 0 )
              throw pool::Exception( "Layout of attribute lists not compatible",
                                     "CollectionProxy::add", 
                                     "CollectionProxy");
        }
     }
        
     for( coral::AttributeListSpecification::const_iterator specIter = m_attrListSpec->begin();
          specIter != m_attrListSpec->end(); ++specIter ) {
        m_collection->schemaEditor().insertColumn( specIter->name(), specIter->typeName() );
     }
     m_isNewCollection = false;
  }
        
  CollectionRowBuffer &rowBuffer = m_collection->dataEditor().rowBuffer();

  // set the 'main' token
  Token &token = rowBuffer.tokenList()[ m_collection->description().eventReferenceColumnName() ];
  pref.token()->setData( &token );

  // set all attributes
  for( coral::AttributeList::const_iterator attrIter = attributeList.begin();
       attrIter != attributeList.end();
       ++attrIter )
  {
     const std::string attr_name = attrIter->specification().name();
     const std::string attr_type = attrIter->specification().typeName();
     //attrIter->getValue( rowBuffer.attributeList()[ attr_name ].data<std::string>() );
     if( attr_type == "Token" ) {
        attrIter->data<Token>().setData( &rowBuffer.tokenList()[ attr_name ] );
     } else {
        rowBuffer.attributeList()[ attr_name ] = *attrIter;
     }
  }
  m_collection->dataEditor().insertRow( rowBuffer ); 
  return true;
}


bool
pool::CollectionProxy::commit()
{
   return ( m_collection && (m_collection->commit(),true) );
}


bool
pool::CollectionProxy::rollback() {
   return ( m_collection && (m_collection->rollback(),true) );
}


bool
pool::CollectionProxy::close() {
   if( m_collection ) m_collection->close();
   return m_collection;
}

bool
pool::CollectionProxy::open() {
   return ( m_collection && (m_collection->open(),true) );
}

bool
pool::CollectionProxy::isOpen() const {
   return m_collection && m_collection->isOpen();
}


std::string
pool::CollectionProxy::name() const
{
   return m_collection ? 
      m_collection->description().name() :
      std::string("undefined");
}


std::string 
pool::CollectionProxy::connection() const{
  return m_collection ? 
    m_collection->description().connection() :
    std::string("undefined");
}


std::string 
pool::CollectionProxy::type() const{
  return m_collection ? 
    m_collection->description().type() :
    std::string("undefined");
}


pool::ICollection::OpenMode 
pool::CollectionProxy::openMode() const
{
   return m_collection->openMode();
}




pool::CollectionProxy::Iterator
pool::CollectionProxy::select( std::string primaryQuery,
                               std::string /*secondaryQuery*/,
                               std::string queryOptions )
{
   if( !( m_collection && m_dataSvc ) ) return Iterator();

//   cout << ">> Creating query for the collectionProxy" << endl;
   pool::ICollectionQuery *collQuery = m_collection->newQuery(); 
   collQuery->setCondition( primaryQuery );
   if( queryOptions.size() ) {
      typedef boost::tokenizer<boost::char_separator<char> > Tizer;
      boost::char_separator<char> sep(" ,");
      Tizer tizer( queryOptions, sep );
      Tizer::iterator token = tizer.begin();
      bool        selectAll = true;
      while( token != tizer.end() ) {
         if( *token == "FetchOne" ) {
             collQuery->setRowCacheSize(1);
         } else if( *token == "FetchAll" ) {
            collQuery->setRowCacheSize(1000);  // MN:  random number..
         } else if( *token == "SELECT" ) {
            selectAll = false;
         } else  if( *token == "*" ) {
            selectAll = true;            
            break;
         } else {
            collQuery->addToOutputList( *token );
         }
         ++token;
      }
      if( selectAll )  collQuery->selectAll();
      // should we add the main token to the output list too?
   } else {
      //MN FIX - is that a good default?
      collQuery->selectAll();
   } 
   return Iterator( collQuery, 0, m_dataSvc );
}



pool::IDataSvc*
pool::CollectionProxy::dataSvc() {
   return m_dataSvc;
}


const pool::IDataSvc*
pool::CollectionProxy::dataSvc() const {
  return m_dataSvc;
}


pool::ICollection*
pool::CollectionProxy::collection() {
  return m_collection;
}


const pool::ICollection*
pool::CollectionProxy::collection() const {
   return m_collection;
}




//----------------------------- Iterator

pool::CollectionProxy::Iterator::Iterator():
      m_query( 0 ),
      m_cursor( 0 ),
      m_dataSvc( 0 ),
      m_token( 0 ),
      m_advancedCursor( true ),
      m_attrListSpec( 0 ),
      m_attributeList( 0 ),
      m_valid( false )
{}


pool::CollectionProxy::Iterator::Iterator( pool::ICollectionQuery* query,
                                           pool::ICollectionCursor* cursor,
                                           pool::IDataSvc* dataSvc ):
  m_query( query ),
  m_cursor( cursor ), 
  m_dataSvc( dataSvc ),
  m_token( 0 ),
  m_advancedCursor( true ),
  m_attrListSpec( 0 ),
  m_attributeList( 0 ),
  m_valid( false )
{
   if( !cursor ) {
      // cout << "CollectionProxy::Iterator: Executing query " << endl;
      m_cursor = &m_query->execute();
   }
}


pool::CollectionProxy::Iterator::Iterator( const Iterator& rhs ):
  m_query( rhs.m_query),
  m_cursor( rhs.m_cursor),
  m_dataSvc( rhs.m_dataSvc ),
  m_token( 0 ),
  m_advancedCursor( true ),
  m_attrListSpec( rhs.m_attrListSpec ),
  m_attributeList( rhs.m_attributeList ),
  m_valid( rhs.m_valid )
{
   rhs.m_cursor = 0;
   rhs.m_query = 0;
   rhs.m_attrListSpec = 0;
   rhs.m_attributeList = 0;
}


pool::CollectionProxy::Iterator::~Iterator()
{
   if( m_query ) {
      // cursor is deleted by the query
      delete m_query;
   } else {
      delete m_cursor;
   }
   delete m_attributeList;
   if( m_attrListSpec ) m_attrListSpec->release();
   if( m_token ) m_token->release();
}



void
pool::CollectionProxy::Iterator::makeAttributeList() const
{
   if( m_attrListSpec ) m_attrListSpec->release();
   m_attrListSpec = new coral::AttributeListSpecification;
   //MN FIX unsafe hack to skip the 'main' token
   for( TokenList::const_iterator tokIter = ++m_cursor->currentRow().tokenList().begin();
        tokIter !=  m_cursor->currentRow().tokenList().end(); ++tokIter ) {
      // cout << "CollectionProxy::Iterator::makeAttributeList() - adding token " << tokIter.tokenName() << endl;
      m_attrListSpec->extend( tokIter.tokenName(), "Token" );
   }
   for( coral::AttributeList::const_iterator attrIter = m_cursor->currentRow().attributeList().begin();
        attrIter != m_cursor->currentRow().attributeList().end(); ++attrIter ) {
      // cout << "CollectionProxy::Iterator::makeAttributeList() - adding attribute " << attrIter->specification().name() << endl;
      m_attrListSpec->extend( attrIter->specification().name(), attrIter->specification().type() );
   }

   delete m_attributeList;
   m_attributeList = new coral::AttributeList( *m_attrListSpec );
}


const coral::AttributeList&
pool::CollectionProxy::Iterator::attributeList() const
{
   if( !m_attributeList )  {
      makeAttributeList();
      m_advancedCursor  = true;
   }
   if( m_advancedCursor ) {
      //MN FIX unsafe hack to skip the 'main' token
      for( TokenList::const_iterator tokIter = ++m_cursor->currentRow().tokenList().begin();
           tokIter !=  m_cursor->currentRow().tokenList().end(); ++tokIter ) {
         (*m_attributeList)[ tokIter.tokenName() ].setValue( *tokIter );
      }
      for( coral::AttributeList::const_iterator attrIter = m_cursor->currentRow().attributeList().begin();
           attrIter != m_cursor->currentRow().attributeList().end(); ++attrIter ) {
         (*m_attributeList)[ attrIter->specification().name() ] = *attrIter;
      }
      m_advancedCursor = false;
   }
   return *m_attributeList;
}


void
pool::CollectionProxy::Iterator::getTokenFromCursor() const
{
   if ( ! m_token ) {
      //m_token = const_cast<Token*>( &m_cursor->eventRef() );
      // m_token->addRef();
      m_token = new Token( &m_cursor->eventRef() );
   }
}
 

bool
pool::CollectionProxy::Iterator::next()
{
  if( !( m_cursor && m_dataSvc ) )
     return false;
  if( m_token ) {
     m_token->release();
     m_token = 0;
  }
  m_valid = m_cursor->next();
  m_advancedCursor = true;
  return m_valid;
}
