#define POOL_DATASVC_CPP 1
#ifdef WIN32
  #pragma warning ( disable : 4786 )
#endif

#include "DataSvc/DataSvcException.h"
#include "DataSvc/DataSvcContext.h"
#include "DataSvc.h"
#include "CacheSvc.h"
#include "CacheTransactionObserver.h"

#include "PersistencySvc/IPersistencySvc.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/IPersistencySvcFactory.h"
#include "PersistencySvc/PersistencySvcException.h"

/// factory for the Persistency service
void pool::DataSvc::_setPersistencySvc(IFileCatalog* aFileCatalog,
                                       ITokenValidator* aTokenValidator,
                                       IDataTransform* aDataTransform)
{

  pool::IPersistencySvcFactory* psfactory = pool::IPersistencySvcFactory::get();
  if ( ! psfactory ) {
    throw DataSvcException("Could not retrieve an IPersistencySvc factory");
  }
  m_persSvc = psfactory->create( "PersistencySvc", *aFileCatalog );
  if ( ! m_persSvc ) {
    throw DataSvcException( "Could not create a PersistencySvc");
  }
  if (aTokenValidator) m_persSvc->setTokenValidator( *aTokenValidator );
  if (aDataTransform) m_persSvc->setDataTransform(*aDataTransform);
}

/// constructor with an already created PersistencySvc
pool::DataSvc::DataSvc(IPersistencySvc* aPersSvc): m_persSvc(aPersSvc), 
                                                   m_cacheSvc(0),
                                                   m_transactionObserver(0),
                                                   m_owned(false){ 
  if(!m_persSvc){
    throw DataSvcException(
          "Cannot construct DataSvc: null PersistencySvc pointer");
  }
  CacheSvc* cacheSvc = 
            new CacheSvc(aPersSvc,ObjectDeletePolicy::defaultPolicy());
  m_cacheSvc = cacheSvc;
  m_transactionObserver = new CacheTransactionObserver(*cacheSvc);
  m_persSvc->registerObserver( *m_transactionObserver );
} 

pool::DataSvc::DataSvc(IFileCatalog* aFileCatalog,
                       ITokenValidator* aTokenValidator,
                       IDataTransform* aDataTransform,
                       const ObjectDeletePolicy& aPolicy) : m_persSvc(0), 
                                                            m_cacheSvc(0),
                                                            m_transactionObserver(0),
                                                            m_owned(false){
  _setPersistencySvc(aFileCatalog,aTokenValidator,aDataTransform);
  m_owned = true;
  CacheSvc* cacheSvc = new CacheSvc(m_persSvc,aPolicy);
  m_cacheSvc = cacheSvc;

  m_transactionObserver = new CacheTransactionObserver(*cacheSvc);
  m_persSvc->registerObserver( *m_transactionObserver );
}

pool::DataSvc::DataSvc(const DataSvcContext& aCtx): m_persSvc(0), 
                                                    m_cacheSvc(0),
                                                    m_transactionObserver(0),
                                                    m_owned(false){  
  _setPersistencySvc(aCtx.fileCatalog(),aCtx.tokenValidator(),aCtx.dataTransform());
  m_owned = true;
  CacheSvc* cacheSvc = new CacheSvc(m_persSvc,aCtx.objectDeletePolicy());
  m_cacheSvc = cacheSvc;
  m_transactionObserver = new CacheTransactionObserver(*cacheSvc);
  m_persSvc->registerObserver( *m_transactionObserver );
}

/// Standard destructor
pool::DataSvc::~DataSvc() { 
  m_persSvc->deregisterObserver( *m_transactionObserver );
  if(m_owned){
    delete m_persSvc;
  }
  delete m_cacheSvc;
  delete m_transactionObserver;
}

/// Returns a reference to the IDataTransform object.
pool::IDataTransform& pool::DataSvc::dataTransform(){
  return m_persSvc->dataTransform();
}

/// Returns a referenc to the IFileCatalog instance in use.
pool::IFileCatalog& pool::DataSvc::fileCatalog(){
  return m_persSvc->fileCatalog();
}

/// Returns the underlying global session
pool::ISession& pool::DataSvc::session(){
  return m_persSvc->session();
}

/// Returns the current transaction
pool::ITransaction& pool::DataSvc::transaction(){
  return m_persSvc->session().transaction();
}

// Clear Cache
void pool::DataSvc::clearCache(){
  m_cacheSvc->clearCache();
}

// Reset Cache
void pool::DataSvc::resetCache(){
  m_cacheSvc->resetCache();
}

pool::ICacheSvc& pool::DataSvc::cacheSvc(){
  return *m_cacheSvc;
}

pool::IConfiguration& pool::DataSvc::configuration(){
  return m_persSvc->configuration();
}

