#ifndef POOL_HMAP_K_TABLE_H
#define POOL_HMAP_K_TABLE_H 

#ifdef __GNUC__
#include <ext/hash_map>
//#include <ext/stl_hash_fun.h>
#define EXTSTD __gnu_cxx

#elif _WIN32
#include "POOLCore/win32/stl_hash.h"
#define EXTSTD stdext

#elif __ICC
#include <hash_map>
#define EXTSTD std

#endif

namespace pool  {

  /** @class Hmap_k_table Hmap_k_table.h
      Templated container class. Designed to be a look up table to map a set of pointers
      to an other set. The implementation uses an hash_map<K,T> where the K is
      the type used as akey.

      @author Giacomo Govi
      @date 25/07/2002
   **/
  template <class K,class T> class Hmap_k_table {    

    public:

      // convenience types
      typedef std::pair<const K, T> Value;

      typedef std::pair<K, T> Entry;

      typedef std::pair<bool,T> Result;

      typedef std::vector<Entry> Entries;

      // container type
      typedef EXTSTD::hash_map<K, T> KTable;  	 
      // iterator type
      typedef typename KTable::const_iterator const_iterator;

      // default constructor
      Hmap_k_table() : theTable() {}

      // constructor with initial capacity
#ifdef __GNUC__
      Hmap_k_table(unsigned int init_size) : theTable(init_size){}
#else
      Hmap_k_table(unsigned int) : theTable(){}
#endif

      /** put - method to store a pointer value for a specified pointer Key.
          @param aK the primary key
          @param anObj the object to be stored in the row
	  @return a Result value holding: 
	  bool: true if the key was present already in the table; false otherwise
	  void*: the pointer mapped before in the first case; NULL otherwise
      */
      const Result put(const K& aK,T& anObj);

      /** get - method to get the Value instance stored for a specified key.
          @param aK the primary key
	  @return a Result value holding: 
	  bool: true if the key was present in the table; false otherwise
	  void*: the pointer mapped in the first case; NULL otherwise
      */
      const Result get(const K& aK);

      /** del - method to delete from the table  an Entry corresponding to a specified pointer key
          @param aK the key of the entry to be deleted
	  @return a Result value holding:
	  bool: true if the key was present in the table; false otherwise
	  void* : the pointer in the entry deleted in the first case; NULL otherwise
      */
      const Result del(const K& aK); 

      /** erase - method to erase from the Table an Entry for a given key,
	  without to retrieve the present item before.
          @param aK the key of the entry
	  @return a size_t = 1 if the key was in the table; 0 otherwise

      */
      unsigned int erase(const K& aK);

      /** clear - method to erase all the entry of the table
      */
      void clear();

      /** begin iterator
      */
      const_iterator begin();

      /** end iterator
      */
      const_iterator end();

      /** size - method to retrieve the number of entries of the table
	  @return the number of entries
      */
      unsigned int size();

    private:
	
      // private instance of the hash_map container
      KTable theTable;

  };  
}

// - method to store a pointer value for a specified pointer Key.
template <class K,class T> 
const typename pool::Hmap_k_table<K,T>::Result 
               pool::Hmap_k_table<K,T>::put(const K& aK,T& anObj) {
  Value aVal(aK,anObj);
  typedef typename KTable::iterator iterator_type;
  std::pair<iterator_type,bool> p = theTable.insert(aVal);
  Result res(false,(T)0);
  bool inserted = p.second;
  if(!inserted){
    T oldObj = (*(p.first)).second;
    theTable.erase(aK);
    theTable.insert(aVal);
    res.second = oldObj;
    res.first = true;
  }
  return res;
}

// method to get the Value instance stored for a specified key.
template <class K,class T> const typename pool::Hmap_k_table<K,T>::Result 
                                          pool::Hmap_k_table<K,T>::get(const K& aK) {
  Result res(false,(T)0);
  typename KTable::iterator item = theTable.find(aK);
  if(item != theTable.end()){
    res.first = true;
    res.second = (*item).second;
  }
  return res;
}

// method to delete from the table  an Entry corresponding to a specified pointer key
template <class K,class T> const typename pool::Hmap_k_table<K,T>::Result 
                                          pool::Hmap_k_table<K,T>::del(const K& aK) {
  Result res(false,(T)0);
  typename KTable::iterator item = theTable.find(aK);
  if(item != theTable.end()){
    res.first = true;
    res.second = (*item).second;
    theTable.erase(item);
  }
  return res;
}

// method to erase from the Table an Entry for a given key,
template <class K,class T> unsigned int pool::Hmap_k_table<K,T>::erase(const K& aK) {
  return theTable.erase(aK);
}

// method to erase all the entry of the table
template <class K,class T> void pool::Hmap_k_table<K,T>::clear(){
  theTable.clear();
}

// begin iterator
template <class K,class T> typename pool::Hmap_k_table<K,T>::const_iterator 
                                    pool::Hmap_k_table<K,T>::begin() {
  return theTable.begin();
}

// end iterator
template <class K, class T> typename pool::Hmap_k_table<K,T>::const_iterator 
                                     pool::Hmap_k_table<K,T>::end() {
  return theTable.end();
}

// method to retrieve the number of entries of the table
template <class K, class T> unsigned int pool::Hmap_k_table<K,T>::size(){
  return theTable.size();
}

#endif  // POOL_HMAP_K_TABLE_H




