#define POOL_REF_CPP 1

#include "DataSvc/RefBase.h"
#include "DataSvc/RefException.h"
#include "DataSvc/ObjectDeletePolicy.h"

#include "POOLCore/Token.h"
#include "PersistencySvc/Placement.h"
#include "StorageSvc/DbType.h"

// default delete policy
pool::RefBase::DeletePolicy pool::RefBase::_defaultDeletePolicy(){
  DeletePolicy ret = DO_DELETE;
  if(!ObjectDeletePolicy::defaultPolicy().onRef()){
    ret = DONOT_DELETE;
  }
  return ret;
}

// default constructor
pool::RefBase::RefBase() : 
               Reference(),
               m_cacheLink(),
               m_ptr(_defaultDeletePolicy()==DO_DELETE),
               m_refType(),
               m_refTypePtr(0),
               m_fullTypePtr(0){
}

pool::RefBase::RefBase(const std::type_info& refType, 
                       DeletePolicy policy): 
               Reference(),
               m_cacheLink(),
               m_ptr(policy==DO_DELETE),
               m_refType(refType),
               m_refTypePtr(0),
               m_fullTypePtr(0){
} 

// cache service only constructor
pool::RefBase::RefBase(IDataSvc* aDataSvc, const std::type_info& refType) : 
               Reference(),
               m_cacheLink(aDataSvc),
               m_ptr(_defaultDeletePolicy()==DO_DELETE),
               m_refType(refType),
               m_refTypePtr(0),
               m_fullTypePtr(0){
  if(m_cacheLink){
    SharedAnyPtr tmp(m_cacheLink.deleteObjectFlag());
    m_ptr = tmp;
  }
}

// cache service only constructor
pool::RefBase::RefBase(ICacheSvc* aCacheSvc, const std::type_info& refType) : 
               Reference(),
               m_cacheLink(aCacheSvc),
               m_ptr(_defaultDeletePolicy()==DO_DELETE),
               m_refType(refType),
               m_refTypePtr(0),
               m_fullTypePtr(0){
  if(m_cacheLink){
    SharedAnyPtr tmp(m_cacheLink.deleteObjectFlag());
    m_ptr = tmp;
  }
}

pool::RefBase::RefBase(IDataSvc* aDataSvc, 
                       const AnyPtr& object,
                       const std::type_info& refType) : 
               Reference(),
               m_cacheLink(aDataSvc),
               m_ptr(_defaultDeletePolicy()==DO_DELETE),
               m_refType(refType),
               m_refTypePtr(0),
               m_fullTypePtr(0){
  if(m_cacheLink){
    SharedAnyPtr tmp(m_cacheLink.deleteObjectFlag());
    m_ptr = tmp;  // here only pass the deletepolicy
    tmp = object; // now is filled
    m_cacheLink.bindCacheEntry(tmp); 
  } else {
    m_ptr = object;
  }
}

pool::RefBase::RefBase(ICacheSvc* aCacheSvc, 
                       const AnyPtr& object,
                       const std::type_info& refType) : 
               Reference(),
               m_cacheLink(aCacheSvc),
               m_ptr(_defaultDeletePolicy()==DO_DELETE),
               m_refType(refType),
               m_refTypePtr(0),
               m_fullTypePtr(0){
  if(m_cacheLink){
    SharedAnyPtr tmp(m_cacheLink.deleteObjectFlag());
    m_ptr = tmp;  // here only pass the deletepolicy
    tmp = object; // now is filled
    m_cacheLink.bindCacheEntry(tmp); 
  } else {
    m_ptr = object;
  }
}

pool::RefBase::RefBase(IDataSvc* aDataSvc, 
                       const Token& aToken, const std::type_info& refType) : 
               Reference(),
               m_cacheLink(aDataSvc),
               m_ptr(_defaultDeletePolicy()==DO_DELETE),
               m_refType(refType),
               m_refTypePtr(0),
               m_fullTypePtr(0){
  if(m_cacheLink){
    SharedAnyPtr tmp(m_cacheLink.deleteObjectFlag());
    m_ptr = tmp;
    m_cacheLink.bindCacheEntry(refType,aToken); 
  }
}

pool::RefBase::RefBase(ICacheSvc* aCacheSvc, 
                       const Token& aToken, const std::type_info& refType) : 
               Reference(),
               m_cacheLink(aCacheSvc),
               m_ptr(_defaultDeletePolicy()==DO_DELETE),
               m_refType(refType),
               m_refTypePtr(0),
               m_fullTypePtr(0){
  if(m_cacheLink){
    SharedAnyPtr tmp(m_cacheLink.deleteObjectFlag());
    m_ptr = tmp;
    m_cacheLink.bindCacheEntry(refType,aToken); 
  }
}

pool::RefBase::RefBase(IDataSvc* aDataSvc, 
                       const std::string& stringToken, 
                       const std::type_info& refType) : 
               Reference(),
               m_cacheLink(aDataSvc),
               m_ptr(_defaultDeletePolicy()==DO_DELETE),
               m_refType(refType),
               m_refTypePtr(0),
               m_fullTypePtr(0){
  if(m_cacheLink){
    SharedAnyPtr tmp(m_cacheLink.deleteObjectFlag());
    m_ptr = tmp;
    Token* tok = new Token();
    tok->fromString(stringToken);
    m_cacheLink.bindCacheEntry(refType,*tok); 
    tok->release();
  }
}

pool::RefBase::RefBase(ICacheSvc* aCacheSvc, 
                       const std::string& stringToken, 
                       const std::type_info& refType) : 
               Reference(),
               m_cacheLink(aCacheSvc),
               m_ptr(_defaultDeletePolicy()==DO_DELETE),
               m_refType(refType),
               m_refTypePtr(0),
               m_fullTypePtr(0){
  if(m_cacheLink){
    SharedAnyPtr tmp(m_cacheLink.deleteObjectFlag());
    m_ptr = tmp;
    Token* tok = new Token();
    tok->fromString(stringToken);
    m_cacheLink.bindCacheEntry(refType,*tok); 
    tok->release();
  }
}

pool::RefBase::RefBase(const RefBase& aCopy) :
               Reference(aCopy),
               m_cacheLink(aCopy.m_cacheLink),
               m_ptr(aCopy.m_ptr),
               m_refType(aCopy.m_refType),
               m_refTypePtr(0),
               m_fullTypePtr(0) {
}

pool::RefBase::~RefBase(){
}

pool::RefBase& pool::RefBase::operator=(const RefBase& aCopy) {
  if(this == &aCopy){
    return *this;
  }
  m_ptr = aCopy.m_ptr;
  m_cacheLink = aCopy.m_cacheLink;
  m_refType = aCopy.m_refType;
  m_refTypePtr = 0;
  m_fullTypePtr = 0;
  return *this;  
}

void pool::RefBase::copy(const RefBase& aRefBase) {
  if(this != &aRefBase){
    if(aRefBase.m_cacheLink){
      copyDeep(aRefBase);
    } else {
      copyShallow(aRefBase);
    }
  }
}

void pool::RefBase::copyDeep(const RefBase& aRefBase) {
  if(this != &aRefBase){
    m_ptr = aRefBase.m_ptr;
    m_cacheLink.copy(m_refType,aRefBase.m_cacheLink);
    m_refTypePtr = 0;
    m_fullTypePtr = 0;
  }
}

void pool::RefBase::copyShallow(const RefBase& aRefBase) {
  if(this != &aRefBase){
    SharedAnyPtr tmp;
    const SharedAnyPtr& theObject = aRefBase.object(); 
    if(theObject.isOfType(m_refType)){
      tmp = theObject;
    }
    if(m_cacheLink){
      m_cacheLink.bindCacheEntry(tmp);
    } else {
      m_ptr = tmp;
    }
    m_refTypePtr = 0;
    m_fullTypePtr = 0;
  }
}

// method to check if the current ref is referencing an already loaded 
// transient instance
bool pool::RefBase::isOpen() const {
  bool ret = true;
  if(m_cacheLink){
    ret = m_cacheLink.isObjectLoaded();
  }
  return ret;
}

bool pool::RefBase::isNull() const {
  return _pointee(false)==0;
}

bool pool::RefBase::isConnected() const {
  return m_cacheLink;
}

void pool::RefBase::setCacheSvc(ICacheSvc* aCacheSvc) const {  
  m_cacheLink.connect(aCacheSvc);
  m_refTypePtr = 0;
  m_fullTypePtr = 0;
  if(m_ptr  && m_cacheLink){
    m_cacheLink.bindCacheEntry(m_ptr);
    m_ptr.reset();
  }
}

pool::ICacheSvc* pool::RefBase::cacheSvc() const {
  return m_cacheLink.cacheSvc();
}


// retrieve the related cache line
const pool::CacheEntry* pool::RefBase::cacheEntry() const{
  return m_cacheLink.cacheEntry();
}

bool pool::RefBase::markWrite(const Placement& place) const {
  return m_cacheLink.sendWriteRequest(place);
}

bool pool::RefBase::markMultiwrite(const Placement& place) const {
  return m_cacheLink.sendMultiWriteRequest(place);
}

bool pool::RefBase::markUpdate() const {
  return m_cacheLink.sendUpdateRequest();
}

bool pool::RefBase::markDelete() const {
  return m_cacheLink.sendDeleteRequest();
}
     
const pool::Token* pool::RefBase::token() const {
  return m_cacheLink.getToken();
}

pool::Placement pool::RefBase::placement() const {
  const Token* tok = token();
  Placement place;
  if(tok){
    place.setDatabase(tok->dbID(),pool::DatabaseSpecification::FID );
    place.setContainerName(tok->contID());
    place.setTechnology(tok->technology());
    //place.setShapeID(tok->classID());
  }
  return place;
}

void pool::RefBase::setToken(const Token& aToken){
  m_ptr.reset();
  if(!m_refType){
    throw RefException("Class not found in the dictionary");
  }
  m_cacheLink.bindCacheEntry(m_refType,aToken);
  m_refTypePtr = 0;
  m_fullTypePtr = 0;
}

const std::string pool::RefBase::toString() const {
  return token() ? token()->toString() : "";
}

const pool::SharedAnyPtr& pool::RefBase::object() const {
  const SharedAnyPtr* ret = &m_ptr;
  if(m_cacheLink){
    ret = &m_cacheLink.getObject();
  } 
  return *ret;
}

Reflex::Type pool::RefBase::objectType() const {
  return object().get().getClass();
}

Reflex::Type pool::RefBase::refType() const {
  return m_refType.getClass();
}

bool pool::RefBase::equals(const RefBase &rR) const {
  if(isConnected() && rR.isConnected()){
    return _compareCacheLink(rR);
  } else {
    return object()==rR.object();
  }
}

// reset the stored reference to 0 -> needed?
void pool::RefBase::clear() {
  m_cacheLink.unbind();
  m_ptr.reset();
  m_refTypePtr = 0;
  m_fullTypePtr = 0;
}

void pool::RefBase::reset() {
  m_refTypePtr = 0;
  m_fullTypePtr = 0;
  if(m_cacheLink){
    m_cacheLink.resetObject();
  } else {
    m_ptr.forceDelete();
  }
}

void pool::RefBase::_copyNoTypeCheck(const RefBase& aRefBase) {
  if(this != &aRefBase){
    if(aRefBase.m_cacheLink){
      _copyDeepNoTypeCheck(aRefBase);
    } else {
      _copyShallowNoTypeCheck(aRefBase);
    }
    m_refTypePtr = 0;
    m_fullTypePtr = 0;
  }
}

void pool::RefBase::_copyDeepNoTypeCheck(const RefBase& aRefBase) {
  if(this != &aRefBase){
    m_ptr = aRefBase.m_ptr;
    m_cacheLink = aRefBase.m_cacheLink;
    m_refTypePtr = 0;
    m_fullTypePtr = 0;
  }
}

void pool::RefBase::_copyShallowNoTypeCheck(const RefBase& aRefBase) {
  if(this != &aRefBase){
    if(m_cacheLink){
      m_cacheLink.bindCacheEntry(aRefBase.object());
    } else {
      m_ptr = aRefBase.object();
    }
    m_refTypePtr = 0;
    m_fullTypePtr = 0;
  }
}

// method to set (or reset) the real pointer from a non-cache-bound Ref 
void pool::RefBase::_setPtr(const AnyPtr& aPtr) {
  m_refTypePtr = 0;
  m_fullTypePtr = 0;
  if(m_cacheLink){
    SharedAnyPtr tmp(aPtr);
    m_cacheLink.bindCacheEntry(tmp);
  } else {
    m_ptr = aPtr;
  }
}

// return the object pointer
void* pool::RefBase::_pointee( bool throw_flag)  const {
  // real pointer as it is in the cache location
  const void* cachePtr = object().constPtr();
  // if pointer different from its copy in the RefBase 
  if((cachePtr!=m_fullTypePtr)){
    // only reset local pointer values
    m_fullTypePtr = cachePtr;
    m_refTypePtr = 0;
  }
  // if the offset-corrected pointer is not in the RefBase 
  if(!m_refTypePtr){
    // re-calculate all the local RefBase parameters
    m_refTypePtr = object().ptr(m_refType);
  }
  void* objPtr = m_refTypePtr; 
  if(!objPtr){
      if(throw_flag)  
        throw RefException("Null pointer retrieved.");
  }
  return objPtr;
} 

bool pool::RefBase::_compareCacheLink(const RefBase& aRefBase) const {
  bool ret = false;
  CacheEntry* anEntry = aRefBase.m_cacheLink.cacheEntry();
  CacheEntry* m_cacheEntry = m_cacheLink.cacheEntry();
  if(anEntry && m_cacheEntry){
    if(anEntry == m_cacheEntry){
      return true;
    }
    const Token* tok = token();
    const Token* aRefTok =  aRefBase.token();
    if(tok && aRefTok){
      ret = (*tok == *aRefTok);
    } else {
      ret = (object()==aRefBase.object());
    }
  }
  return ret;
}
