#ifndef LFCCATALOG_LFCCATALOG_H
#define LFCCATALOG_LFCCATALOG_H

#include "FileCatalog/FCImpl.h"
#include "FileCatalog/FCCommon.h"

#include "FileCatalog/FCException.h"

#include "CoralBase/MessageStream.h"

extern "C" {
#include "lfc_api.h"
#include "serrno.h"
}

#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>

#include <pthread.h>
#include <sys/time.h>
#include <stdlib.h>

#include <string>
#include <list>
#include <map>

#include "LFCCatalog/gfal_api.h"

#define MAX_DELAYED_PFNS 10000
#define LFC_ERRBUFSZ 1023

#define LFC_LIST_FUNC_TYPE(x)  FCBuf<pool::LFCCatalog::udata>    &buf,            \
                               void                              *&priv_data,     \
                               std::string                       value,           \
                               pool::LFCCatalog::match_types_t   match_type,      \
                               FCBuf<pool::LFCCatalog::udata>    *universal_set,  \
                               pool::LFCCatalog::distinct_t      &set_distinct,   \
                               pool::LFCCatalog::constraint_list &cl,             \
                               int                               flags

#define MRET(x,y) ((pool::LFCQueryEvaluator<x>::retrieveData_t *)y)

#define LFC_NO_NEED_SET ((void *)0L)
#define LFC_NEED_SET ((void *)1L)

namespace pool {
  /**
   * @class LFCCatalog LFCCatalog.h
   *
   * Implementation of the FCImpl interface for a LFC Catalog
   */
  
  class LFCCatalog : public pool::FCImpl {
  public:

    struct udata {
      enum udata_primary_t { NOTYPE, FID, LFN, PFN };
      enum udata_complete_t { EMPTY, PARTIAL, FULL };

      udata() {
        m_lfn_state = EMPTY;
        m_pfn_state = EMPTY;
        m_type = NOTYPE;
      }
      udata(const FileCatalog::FileID &rfid) {
        init(rfid);
      }
      udata(const LFNEntry &rlfn) {
        init(rlfn);
      }
      udata(const PFNEntry &rpfn) {
        init(rpfn);
      }

      void init(const FileCatalog::FileID &rfid) {
        m_lfn_state = EMPTY;
        m_pfn_state = EMPTY;
        m_fid = rfid;
        m_type = FID;
      }
      void init(const LFNEntry &rlfn) {
        m_fid = rlfn.guid();
        m_primary_lfn.clear();
        m_primary_lfn.push_back(rlfn);
        m_type = LFN;
        m_lfn_state = PARTIAL;
        m_lfns.clear();
        m_lfns.push_back(rlfn);
        m_pfn_state = EMPTY;
      }
      void init(const PFNEntry &rpfn) {
        m_fid = rpfn.guid();
        m_primary_pfn.clear();
        m_primary_pfn.push_back(rpfn);
        m_type = PFN;
        m_lfn_state = EMPTY;
        m_pfn_state = PARTIAL;
        m_pfns.clear();
        m_pfns.push_back(rpfn);
      }

      udata_primary_t get_type() const {
        return m_type;
      }

      void get_item(FileCatalog::FileID &fid) const {
        fid = this->guid();
        return;
      }

      void get_item(LFNEntry &lfn) const {
        lfn = this->lfn();
        return;
      }

      void get_item(PFNEntry &pfn) const {
        pfn = this->pfn();
        return;
      }

      const FileCatalog::FileID &guid() const {
        return m_fid;
      }
      const LFNEntry &lfn() const {
        if (m_type != LFN) {
          throw FCbackendException("LFCCatalog::udata::lfn","Object not of type LFN");
        }
        return m_primary_lfn[0];
      }
      const PFNEntry &pfn() const {
        if (m_type != PFN) {
          throw FCbackendException("LFCCatalog::udata::pfn","Object not of type PFN");
        }
        return m_primary_pfn[0];
      }
      const std::vector<LFNEntry> &lfns() const {
        if (m_type == LFN) {
          return m_primary_lfn;
        }
        return m_lfns;
      }
      const std::vector<PFNEntry> &pfns() const {
        if (m_type == PFN) {
          return m_primary_pfn;
        }
        return m_pfns;
      }
      bool need_lfns() const {
        if (m_type == LFN) return false;
        return (m_lfn_state != FULL);
      }
      bool need_pfns() const {
        if (m_type == PFN) return false;
        return (m_pfn_state != FULL);
      }
      void set_full(const std::vector<LFNEntry> &lfns) {
        m_lfns = lfns;
        m_lfn_state = FULL;
      }
      void set_full(const std::vector<PFNEntry> &pfns) {
        m_pfns = pfns;
        m_pfn_state = FULL;
      }
      udata pfns(size_t num) {
        udata d = *this;
        d.m_type = PFN;
        d.m_primary_lfn.clear();
        d.m_primary_pfn.clear();
        d.m_primary_pfn.push_back(m_pfns[num]);
        return d;
      }
      udata lfns(size_t num) {
        udata d = *this;
        d.m_type = LFN;
        d.m_primary_lfn.clear();
        d.m_primary_pfn.clear();
        d.m_primary_lfn.push_back(m_lfns[num]);
        return d;
      }
      udata guid(size_t __attribute__((unused)) num) {
        udata d = *this;
        d.m_type = FID;
        d.m_primary_lfn.clear();
        d.m_primary_pfn.clear();
        return d;
      }

      bool operator== (const udata &b) const {
        if (this->m_type == NOTYPE || b.m_type == NOTYPE) {
          throw FCbackendException("LFCCatalog::udata::operator==","One of the operands has invalid type");
        }
        if (this->m_type != b.m_type) {
          throw FCbackendException("LFCCatalog::udata::operator==","The operands are of different type");
        }
        switch(this->m_type) {
          case FID:
            return (LFCCatalog::compare(this->guid(), b.guid()));
            break;
          case PFN:
            return (LFCCatalog::compare(this->pfn(), b.pfn()));
            break;
          case LFN:
            return (LFCCatalog::compare(this->lfn(), b.lfn()));
            break;
          default:
            throw FCbackendException("LFCCatalog::udata::operator==","The operands are of unknown type");
            break;
        }
        return false;
      }

      bool operator!= (const udata &b) const {
        return !(*this == b);
      }
        
      private:

      FileCatalog::FileID m_fid;
      std::vector<LFNEntry> m_primary_lfn;
      std::vector<PFNEntry> m_primary_pfn;
      udata_primary_t m_type;
      std::vector<LFNEntry> m_lfns;
      std::vector<PFNEntry> m_pfns;
      udata_complete_t m_lfn_state;
      udata_complete_t m_pfn_state;

    };

    enum match_types_t { LITERAL, WILDCARD };

    enum list_flags_t { LIST_BEGIN, LIST_CONTINUE, LIST_END, LIST_CONSTRAINTS };

    enum distinct_t { YES, NO, MAYBE, DIST_UNSET };


    typedef std::vector<std::pair<match_types_t, std::string> > clist_t;

    struct constraint_list {
      size_t m_n,m_n_literal;

      clist_t fids, lfns, pfns;

      void clear() { fids.clear(); lfns.clear(); pfns.clear(); m_n = 0; m_n_literal = 0; }

      constraint_list() { clear(); }

      size_t n() const { return m_n; }

      size_t n_literal() const { return m_n_literal; }

      void add(constraint_list &l) {
        for(size_t i=0;i<l.fids.size();++i) {
          fids.push_back(l.fids[i]);
          ++m_n;
          if (l.fids[i].first == LITERAL) ++m_n_literal;
        }
        for(size_t i=0;i<l.lfns.size();++i) {
          lfns.push_back(l.lfns[i]);
          ++m_n;
          if (l.lfns[i].first == LITERAL) ++m_n_literal;
        }
        for(size_t i=0;i<l.pfns.size();++i) {
          pfns.push_back(l.pfns[i]);
          ++m_n;
          if (l.pfns[i].first == LITERAL) ++m_n_literal;
        }
      }

      void lfns_add(const match_types_t &mt, const std::string &s) {
        lfns.push_back(std::make_pair(mt,s));
        if (mt == LITERAL) ++m_n_literal;
        ++m_n;
      }

      void pfns_add(const match_types_t &mt, const std::string &s) {
        pfns.push_back(std::make_pair(mt,s));
        if (mt == LITERAL) ++m_n_literal;
        ++m_n;
      }

      void fids_add(const match_types_t &mt, const std::string &s) {
        fids.push_back(std::make_pair(mt,s));
        if (mt == LITERAL) ++m_n_literal;
        ++m_n;
      }
      
    };

    /// Constructor
    LFCCatalog();
    
    /// Destructor
    virtual ~LFCCatalog();
    
    /// Methods inherited from FCImpl
    
    /// Connects to the database
    void connect();
    
    /// Disconnects from the database
    void disconnect() const;
    
    /// Starts a new transaction
    void start() const;
    
    /// Commits a transaction
    void commit( pool::FileCatalog::CommitMode cm = pool::FileCatalog::REFRESH ) const;
    
    /// Rolls back a transaction
    void rollback() const;
    
    /// Creats a new (invented) primary LFN, registers a PFN to it and
    /// returns the file id of the new LFN
    void registerPFN( const std::string& pfn,
		      const std::string& filetype,
		      pool::FileCatalog::FileID& fid ) const;
    
    /// Registers an LFN:
    void registerLFN( const std::string& pfn,
		      const std::string& lfn ) const;
    
    /// Adds a new replica (PFN) for a file identified by an existing PFN.
    void addReplicaPFN( const std::string& pf,
			const std::string& rpf ) const;

    /// Adds a PFN to an existing file id. If GUID doesn't already exist
    /// a new primary LFN is invented with this GUID, and the PFN registered to it
    void addPFNtoGuid( const pool::FileCatalog::FileID& guid,
		       const std::string& pf,
		       const std::string& filetype ) const;
    
    /// Renames a PFN
    void renamePFN( const std::string& pf,
		    const std::string& newpf ) const;
    
    /// Looks up a file given its PFN
    void lookupFileByPFN( const std::string& pf,
			  pool::FileCatalog::FileID& fid,
			  std::string& ftype ) const;
    
    /// Looks up a file given its LFN
    void lookupFileByLFN( const std::string& lf,
			  pool::FileCatalog::FileID& fid) const;
    
    /// Looks up the best PFN
    void lookupBestPFN( const pool::FileCatalog::FileID& fid,
			const FileCatalog::FileOpenMode& omode, 
			const FileCatalog::FileAccessPattern& amode,
			std::string& pf,
			std::string& filetype ) const;
    
    /// Inserts a PFN entry to an existing guid
    /// ie. adds a replica to an existing primary LFN
    void insertPFN( pool::PFNEntry& pentry) const;
    
    /// Inserts another LFN entry for an existing guid
    /// ie. creates a symlink to an existing primary LFN
    void insertLFN( pool::LFNEntry& lentry) const;
    
    /// Deletes a PFN entry
    void deletePFN( const std::string& pf ) const;
    
    /// Deletes an LFN entry
    void deleteLFN( const std::string& lf ) const;
    
    /// Deletes an entry from the file catalog
    void deleteEntry(const pool::FileCatalog::FileID& guid) const;
    
    /// Checks whether the catalog is read-only
    bool isReadOnly() const;
    
    /// Queries the PFNs
    bool retrievePFN( const std::string& query,
		      pool::FCBuf<pool::PFNEntry>& buf,
		      const size_t& start );
    
    /// Queries the LFNs
    bool retrieveLFN( const std::string& query,
		      pool::FCBuf<pool::LFNEntry>& buf,
		      const size_t& start );
    
    /// Queries the GUID
    bool retrieveGuid( const std::string& query,
		       pool::FCBuf<pool::FileCatalog::FileID>& buf,
		       const size_t& start );

    /// does this catalogue instance expect to return distinct results for qulk queries?
    bool returns_distinct() const {
      return m_want_distinct;
    }

    /// used as keepalive for transaction connection
    void keepalive() const {
      if ( !m_transactionsta.test(0) || !m_transactionsta.test(1) ) return;

      time_t newtime;
      (void)time(&newtime);
      if (newtime>=m_timek && newtime-m_timek<30) return;

      m_timek = newtime;

      struct lfc_filestat fs;
      (void)lfc_stat("/",&fs);
    }

    static bool compare(const FileCatalog::FileID &, const FileCatalog::FileID &);
    static bool compare(const LFNEntry &, const LFNEntry &);
    static bool compare(const PFNEntry &, const PFNEntry &);

    static bool compare(const udata &, const udata &);

    static void  L_regularise_path(std::string &path);

    void L_qualify(const std::string &unqual, std::string &qual) const;

    void L_unqualify(LFNEntry &lfn) const;
    void L_unqualify(PFNEntry &pfn) const;
    void L_unqualify(FileCatalog::FileID &guid) const;
    void L_unqualify(const std::string &in, std::string &out) const;

    /* for thread handling in Cglobals */
    static pthread_mutex_t thread_data_mutex;
    static int threads_installed;
    static std::map<pthread_t, const LFCCatalog *> thread_2_cat;
    static std::map<const LFCCatalog *, pthread_t> cat_2_thread;
    static std::map<int, bool> tid_map;
    static std::map<int *, pthread_key_t> gkey_2_pkey;
    static pthread_key_t tids_key;
    static bool tids_key_created;
    static int next_tid;
    static bool gave_interfere_warning;
    typedef std::map<pthread_t, const LFCCatalog *>::iterator thread_2_cat_itr;
    typedef std::map<const LFCCatalog *, pthread_t>::iterator cat_2_thread_itr;
    typedef std::map<int *, pthread_key_t>::iterator gkey_2_pkey_itr;

    /* per object thread variables */
    mutable int m_tid;
    mutable unsigned int m_thread_enter_count;
    mutable unsigned int m_thread_acquire_count;
    mutable pthread_cond_t m_thread_cond;
    mutable std::map<int *, void *> m_gkey_to_vp;
    typedef std::map<int *, void *>::iterator gkey_to_vp_itr;

    /* thread handling methods */
    void my_init_object_tid();
    void my_threadspecific_install();
    void my_init_object_keys();
    void my_free_object_keys();
    struct lfc_enter_routine {
      lfc_enter_routine(const LFCCatalog *obj) {
        lfc_take_mutex lk(&thread_data_mutex);
        m_cat = obj;
        pthread_t pself = pthread_self();
        cat_2_thread_itr itr;
        itr = cat_2_thread.find(m_cat);
        if (itr != cat_2_thread.end()) {
          if (m_cat->m_thread_acquire_count == 0 && (m_cat->m_thread_enter_count==0 || pthread_equal(pself, itr->second))) {
            ++m_cat->m_thread_enter_count;
          } else {
            ++m_cat->m_thread_acquire_count;
            int ret;
            do {
              ret = pthread_cond_wait(&m_cat->m_thread_cond, &thread_data_mutex);
            } while(ret == EINTR);
            if (ret) {
              throw FCbackendException("LFCCatalog::lfc_enter_routine::lfc_enter_routine","Error while waiting for condition variable");
            }
            --m_cat->m_thread_acquire_count;
            if (++m_cat->m_thread_enter_count != 1) {
              throw FCbackendException("LFCCatalog::lfc_enter_routine::lfc_enter_routine","Unexpected thread enter count");
            }
          }
        } else {
          throw FCbackendException("LFCCatalog::lfc_enter_routine::lfc_enter_routine","The current catalog object is unkown");
        }
        cat_2_thread[m_cat] = pself;
        if (thread_2_cat.find(pself) != thread_2_cat.end()) {
          saved_cat =  thread_2_cat[pself];
        } else {
          saved_cat = 0;
        }
        thread_2_cat[pself] = m_cat;

        if (threads_installed<0 && !gave_interfere_warning) {
          int oc=0;
          for(itr=cat_2_thread.begin();itr!=cat_2_thread.end();++itr) {
            if (pthread_equal(pself,itr->second)) oc++;
          }
          if (oc>1) {
            coral::MessageStream fclog( "LFCCatalog::lfc_enter_routine");
            fclog<<coral::Warning<< "*** There are multiple LFCCatalog interface instances active in this thread" <<coral::MessageStream::endmsg;
            fclog<<coral::Warning<< "*** but the interface was unable to install a custom Cglobals routine:" <<coral::MessageStream::endmsg;
            fclog<<coral::Warning<< "*** ACTIVE CONNECTIONS WITHIN A THREAD WILL INTERFERE WITH EACH OTHER" <<coral::MessageStream::endmsg;
            gave_interfere_warning = true;
          }
        }
      }
      ~lfc_enter_routine() {
        lfc_take_mutex lk(&thread_data_mutex);
        pthread_t pself = pthread_self();
        cat_2_thread_itr itr;
        itr = cat_2_thread.find(m_cat);
        if (itr != cat_2_thread.end()) {
          if (pthread_equal(pself, itr->second)) {
            if (saved_cat == 0) {
              thread_2_cat.erase(pself);
            } else {
              thread_2_cat[pself] = saved_cat;
            }
            if (--m_cat->m_thread_enter_count == 0) {
              if (m_cat->m_thread_acquire_count != 0) {
                if (pthread_cond_signal(&m_cat->m_thread_cond)) {
                  throw FCbackendException("LFCCatalog::lfc_enter_routine::~lfc_enter_routine","Failure to signal on condition variable");
                }
              }
            }
          } else {
            throw FCbackendException("LFCCatalog::lfc_enter_routine::~lfc_enter_routine","The current thread is not in control of this object");
          }
        } else {
          throw FCbackendException("LFCCatalog::lfc_enter_routine::~lfc_enter_routine","The current catalog object is unkown");
        }
      }
      const LFCCatalog *m_cat,*saved_cat;
    };
    struct lfc_take_mutex {
      lfc_take_mutex(pthread_mutex_t *mutex) {
        m_mutex = mutex;
        if (pthread_mutex_lock(m_mutex)) {
          throw FCbackendException("LFCCatalog::lfc_take_mutex::lfc_take_mutex","Failure to lock mutex");
        }
      }
      ~lfc_take_mutex() {
        if (pthread_mutex_unlock(m_mutex)) {
          throw FCbackendException("LFCCatalog::lfc_take_mutex::~lfc_take_mutex","Failure to unlock mutex");
        }
      }
      pthread_mutex_t *m_mutex;
    };
    struct my_key_val_pair {
      int *key;
      void *val;
    };
    static int my_get_thread_specific(int *, void **);
    static int my_set_thread_specific(int *, void *);
    static int my_get_tid();
    static void my_key_destructor(void *);
    static void my_tid_destructor(void *);
    static void my_release_tid(int);
    static int my_calc_new_tid();

    /* for mutex implimenting in Cmutex */
    struct my_cmutex_s {
      bool locked;
      pthread_t owner_id;
      pthread_cond_t wakeup_cond;
      unsigned int number_waiting;
    };
    static pthread_mutex_t my_cmutex_data_mutex;
    static std::map<void *, my_cmutex_s> my_cmutex_data;
    static int my_cmutex_installed;
    typedef std::map<void *, my_cmutex_s>::iterator my_cmutex_data_itr;

    /* cmutex handling methods */
    void my_cmutex_install();
    static int my_cmutex_lockroutine(void *, int);
    static int my_cmutex_unlockroutine(void *);

  private:

    /// InsertPFN with indictaion of whether the GUID/PFN already exists
    void insertPFN(const pool::PFNEntry& pentry, const int &, const int &) const;

    void commitAllDelayedPFN() const;

    LFNEntry generateLFN(const FileCatalog::FileID &) const;
    void generateLFN(LFNEntry &my_lfn) const;

    void fid_to_entry(const FileCatalog::FileID &fid, std::vector<PFNEntry> &pfn_vec, const bool ignore_enoent) const;
    void fid_to_entry(const FileCatalog::FileID &fid, std::vector<LFNEntry> &lfn_vec, const bool ignore_enoent) const;

    void translate_entry(const std::string value, FileCatalog::FileID &fid, std::vector<PFNEntry> &pfn_vec, const bool ignore_enoent) const;
    void translate_entry(const std::string value, FileCatalog::FileID &fid, std::vector<LFNEntry> &lfn_vec, const bool ignore_enoent) const;

    static bool has_wildcard(const std::string &s);

    static const char *wild_underscore_strstr(const char *haystack, const char *needle);
    static bool string_wildmatch(const std::string &s, const std::string &match);
    static void breakup_path(const std::string path, std::vector<std::string> &components);

    static bool lfn_wildmatch(const LFNEntry &lfn, const std::string &match);
    static bool lfn_wildmatch(const LFNEntry &lfn, const std::vector<std::string> &match_elements);

    static bool pfn_wildmatch(const PFNEntry &pfn, const std::string &match);
    static bool guid_wildmatch(const FileCatalog::FileID &fid, const std::string &match);

    struct path_list_t {

      struct dir_element {
        std::string dir_path;
        lfc_DIR *dirp;
        std::string curr_name;
        FileCatalog::FileID fid;
        mode_t filemode;
        std::vector<PFNEntry> pfns;
        bool advance;
        bool ext_query;
        bool first_readdir;
      };  

      std::vector<dir_element> de;
      std::vector<std::string> names;
      bool end;
      LFCCatalog *m_cat;

      size_t number_read_dir;
      size_t dir_needed_stat;

      void close_all_dirs() {
        std::vector<dir_element>::reverse_iterator de_itr;
        for(de_itr=de.rbegin();de_itr!=de.rend();++de_itr) {
          if (de_itr->dirp != NULL) {
            (void) lfc_closedir(de_itr->dirp);
            // used to catch exceptions on the pop_cd, but must not
            // ignore as chdir failure will have aborted transaction
            m_cat->pop_cd();
          }
        }
        de.clear();
      };

      path_list_t() { number_read_dir=0; dir_needed_stat=0; };
      ~path_list_t() { close_all_dirs(); };

      void init(LFCCatalog *cat, std::string path) {
        m_cat = cat;

        names.clear();
        close_all_dirs();
      
        breakup_path(path, names);

        end = true;
        if (names.size() > 0) end=false;
      };

      udata next() {
        coral::MessageStream fclog( "LFCCatalog::path_list_t::next");

        size_t nlevel = names.size();
        size_t curr = de.size();

        while (!end) {

          if (curr > 0 && (curr == nlevel || de[curr-1].advance)) {
            dir_element &e = de[curr-1];
            if (e.advance) {
              if (e.dirp == NULL) {
                de.pop_back();
                --curr;
                if (curr>0) {
                  de[curr-1].advance = true;
                } else {
                  end = true;
                }
                continue;
              }

              void *ds;

              // Can only choose which type of readdir to use on a newly opened dirp
              if (e.first_readdir) {
                e.first_readdir = false;

                // not clear if this is a reasonable way to determine which to use:
                // for now always use readdirg
                //if (number_read_dir < 1000 || dir_needed_stat > 0.05*number_read_dir) {

                if (1) {
                  e.ext_query = true;
                } else {
                  e.ext_query = false;
                }
              }

              serrno = 0;
              if (e.ext_query) {
                //ds = lfc_readdirg(e.dirp);
                ds = lfc_readdirxr(e.dirp,NULL);
              } else {
                ds = lfc_readdir(e.dirp);
              }
   
              // check for null ds //   
              if (ds == NULL) {
                int save_errno = serrno;
                (void) lfc_closedir(e.dirp);
                e.dirp = NULL;
                m_cat->pop_cd();
                if (save_errno) {
                  serrno = save_errno;
                  std::string s = m_cat->L_build_serr("Error listing directory");
                  fclog<<coral::Warning<< s <<coral::MessageStream::endmsg;
                  throw FCbackendException("LFCCatalog::path_list_t::next",s);
                }
                continue;
              }

              // record that we read a directory entry
              ++number_read_dir;

              if (e.ext_query) {
                //struct lfc_direnstatg *p = (struct lfc_direnstatg *)ds;
                //e.curr_name.assign(p->d_name);
                //e.fid = p->guid;
                //e.filemode = p->filemode;

                struct lfc_direnrep *p = (struct lfc_direnrep *)ds;
                e.curr_name.assign(p->d_name);
                e.fid = p->guid;
                e.filemode = p->filemode;
                e.pfns.clear();
                struct lfc_rep_info *rep = p->rep;
                for(int i=0;i<p->nbreplicas;++i) {
                  std::string filetype;
                  PFNEntry pfn(rep->sfn, e.fid, filetype);
                  e.pfns.push_back(pfn);
                  rep++;
                }
              } else {
                struct dirent *p = (struct dirent *)ds;
                e.curr_name.assign(p->d_name);
                e.fid = "";
              }

              if (string_wildmatch(e.curr_name, names[curr-1])) {
                // path element matches: if the filemode info is available, also check:
                // if we're at the last element it has to be regular or link, otherwise it has to be directory
                if (!e.ext_query || (curr==nlevel && (e.filemode & (S_IFLNK | S_IFREG))) || (curr<nlevel && (e.filemode & S_IFDIR))) {
                  e.advance = false;
                }
              }

              continue;
            }

            // at this point advance is false, and curr==nlevel

            e.advance = true;
            std::string lfnstr = e.dir_path;
            if (lfnstr != "/") lfnstr += "/";
            lfnstr += e.curr_name;

            // we're going to need to use the stat info, record this fact
            if (e.dirp != NULL) ++dir_needed_stat;

            // if we don't have the fid get it
            if (e.fid.empty()) {
              lfc_filestatg statbuf;
              if (lfc_statg(lfnstr.c_str(), NULL, &statbuf)<0) {
                // skip if statg fails
                continue;
              }

              e.fid = statbuf.guid;
              e.filemode = statbuf.filemode;

              // if we were expect to also have the list of PFNs get them
              if (e.ext_query && e.pfns.empty()) {
                m_cat->fid_to_entry(e.fid, e.pfns, true);
              }
            }

            if (!e.fid.empty() && (e.filemode & (S_IFLNK | S_IFREG))) {
              // only return it if we have the guid and it is of desired type
              LFNEntry lfn2(lfnstr, e.fid);
              udata ret(lfn2);
              if (e.ext_query) {
                ret.set_full(e.pfns);
              }
              return ret;
            }
            continue;
          }
           
          if (curr >= nlevel) {
            fclog<<coral::Warning<< "unexpected error" <<coral::MessageStream::endmsg;
            throw FCbackendException("LFCCatalog::path_list_t::next","unexpected error");
          }

          dir_element e;

          if (curr> 0) {
            e.dir_path = de[curr-1].dir_path;
            if (curr > 1) e.dir_path += "/";
            e.dir_path += de[curr-1].curr_name;
          } else {
            e.dir_path = "/";
          }

          if (!has_wildcard(names[curr])) {
            e.dirp = NULL;
            e.advance = false;
            e.curr_name = names[curr];
          } else {
            std::string hostport = m_cat->m_host;
            if (!m_cat->m_port.empty()) {
              hostport += std::string(":") + m_cat->m_port;
            }
            if (curr == 0) {
              if (!m_cat->push_cd("/")) {
                std::string s = m_cat->L_build_serr("Could not chdir to /");
                fclog<<coral::Warning<< s <<coral::MessageStream::endmsg;
                throw FCbackendException("LFCCatalog::path_list_t::next",s);
              }
              e.dirp = lfc_opendirxg((char *)hostport.c_str(),"/",NULL);
            } else {
              e.dirp = NULL;
              // if we're trying to change to/open something that is not a directory these will fail
              // nb in fact a chdir failure would abort transaction and push_cd will throw.
              // Therefore ensure above that we only consider directories
              if (m_cat->push_cd(de[curr-1].dir_path)) {
                e.dirp = lfc_opendirxg((char *)hostport.c_str(),de[curr-1].curr_name.c_str(),NULL);
                if (e.dirp == NULL) m_cat->pop_cd();
              }
            }
            e.first_readdir = true;
            e.advance = true;
          }
          de.push_back(e);
          ++curr;
        }
        LFNEntry lfn("","");
        udata ret(lfn);
        return ret;
      };
    };

    struct constraint_checker {

      constraint_checker() { m_init = false; };
                               
      ~constraint_checker() {};

      void lfns_add(const match_types_t &mt, const std::string &s) {
        coral::MessageStream fclog( "LFCCatalog::constraint_checker::lfns_add");
        if (!m_init) {
          fclog<<coral::Warning<< "Not initalised" <<coral::MessageStream::endmsg;
          throw  FCbackendException("LFCCatalog::constraint_checker::lfns_add","Not initalised");
        }
        m_cl.lfns_add(mt,s);
      }

      void pfns_add(const match_types_t &mt, const std::string &s) {
        coral::MessageStream fclog( "LFCCatalog::constraint_checker::lfns_add");
        if (!m_init) {
          fclog<<coral::Warning<< "Not initalised" <<coral::MessageStream::endmsg;
          throw  FCbackendException("LFCCatalog::constraint_checker::pfns_add","Not initalised");
        }
        m_cl.pfns_add(mt,s);
      }

      void fids_add(const match_types_t &mt, const std::string &s) {
        coral::MessageStream fclog( "LFCCatalog::constraint_checker::lfns_add");
        if (!m_init) {
          fclog<<coral::Warning<< "Not initalised" <<coral::MessageStream::endmsg;
          throw  FCbackendException("LFCCatalog::constraint_checker::fids_add","Not initalised");
        }
        m_cl.fids_add(mt,s);
      }

      bool check(const udata &citem) const {
        udata the_item = citem;
        return check(the_item);
      }

      bool check(udata &the_item) const {
        coral::MessageStream fclog( "LFCCatalog::constraint_checker::check");

        if (!m_init) {
          fclog<<coral::Warning<< "Not initalised" <<coral::MessageStream::endmsg;
          throw FCbackendException("LFCCatalog::check (on item)","Not initalised");
        }

        bool matched = true;
        bool repeat;
        clist_t::const_iterator fids_itr, lfns_itr, pfns_itr;
        std::vector<LFNEntry>::const_iterator this_lfns_itr;
        std::vector<PFNEntry>::const_iterator this_pfns_itr;

        FileCatalog::FileID my_fid = the_item.guid();

        if (my_fid.empty()) {
          fclog<<coral::Warning<< "Item to be checked is empty" << coral::MessageStream::endmsg;
          throw FCbackendException("LFCCatalog::check (on item)","Item to be checked is empty");
        }

        // check on constraints on the fid itself
        for(fids_itr=m_cl.fids.begin(); fids_itr != m_cl.fids.end(); ++fids_itr) {
          if (fids_itr->first == LITERAL) {
            FileCatalog::FileID fid(fids_itr->second);
            if (!compare(my_fid, fid)) {
              return false;
            }
          } else {
            if (!guid_wildmatch(my_fid, fids_itr->second)) {
              return false;
            }
          }
        }

        do {
          repeat = false;

          // check on constraints on the lfn
          for(lfns_itr=m_cl.lfns.begin(); lfns_itr != m_cl.lfns.end(); ++lfns_itr) {
            bool an_lfn_matches = false;
            if (lfns_itr->first == LITERAL) {
              LFNEntry lfn(lfns_itr->second, my_fid);
              for(this_lfns_itr=the_item.lfns().begin(); this_lfns_itr != the_item.lfns().end(); ++this_lfns_itr) {
                if (compare(*this_lfns_itr, lfn)) {
                  an_lfn_matches = true;
                  break;
                }
              }
            } else {
              std::vector<std::string> match_elements;
              breakup_path(lfns_itr->second, match_elements);
              for(this_lfns_itr=the_item.lfns().begin(); this_lfns_itr != the_item.lfns().end(); ++this_lfns_itr) {
                if (lfn_wildmatch(*this_lfns_itr, match_elements)) {
                  an_lfn_matches = true;
                  break;
                }
              }
            }
            if (!an_lfn_matches) {
              matched = false;
              break;
            }
          }

          // if we didn't match this time but dont have a full set of lfns..
          if (!matched && m_cl.lfns.size() != 0 && the_item.need_lfns()) {
            // get LFNs for my_fid, don't ignore ENOENT as there should be at least one LFN
            std::vector<LFNEntry> lfns;
            m_cat->fid_to_entry( my_fid, lfns, false);
            the_item.set_full(lfns);
            repeat = true;
          }
        } while(repeat);

        if (!matched)
          return matched;


        do {
          repeat = false;
        
          for(pfns_itr=m_cl.pfns.begin(); pfns_itr != m_cl.pfns.end(); ++pfns_itr) {
            bool a_pfn_matches = false;
            if (pfns_itr->first == LITERAL) {
              std::string filetype;
              PFNEntry pfn(pfns_itr->second, my_fid, filetype);
              for(this_pfns_itr=the_item.pfns().begin(); this_pfns_itr != the_item.pfns().end(); ++this_pfns_itr) {
                if (compare(*this_pfns_itr,pfn)) {
                  a_pfn_matches = true;
                  break;
                }
              }
            } else {
              for(this_pfns_itr=the_item.pfns().begin(); this_pfns_itr != the_item.pfns().end(); ++this_pfns_itr) {
                if (pfn_wildmatch(*this_pfns_itr, pfns_itr->second)) {
                  a_pfn_matches = true;
                  break;
                }
              }
            }
            if (!a_pfn_matches) {
              matched = false;
              break;
            }
          }

          // if we didn't match this time but dont have a full set of pfns..
          if (!matched && m_cl.pfns.size() != 0 && the_item.need_pfns()) {
            // get PFNs for my_fid, ignore ENOENT
            std::vector<PFNEntry> pfns;
            m_cat->fid_to_entry(my_fid, pfns, true);
            the_item.set_full(pfns);
            repeat = true;
          }  
        } while(repeat);
        return matched;
      };
        
      void init(pool::LFCCatalog *cat, constraint_list &c) {
        m_cl = c;
        m_cat = cat;
        m_init = true;
      };

      bool m_init;
      constraint_list m_cl;
      pool::LFCCatalog *m_cat;
    };

    void L_get_lfnentry_from_lfn(LFC_LIST_FUNC_TYPE(LFNEntry));
    void L_get_lfnentry_from_pfn(LFC_LIST_FUNC_TYPE(LFNEntry));
    void L_get_lfnentry_from_guid(LFC_LIST_FUNC_TYPE(LFNEntry));

    void L_get_pfnentry_from_lfn(LFC_LIST_FUNC_TYPE(PFNEntry));
    void L_get_pfnentry_from_pfn(LFC_LIST_FUNC_TYPE(PFNEntry));
    void L_get_pfnentry_from_guid(LFC_LIST_FUNC_TYPE(PFNEntry));

    void L_get_fid_from_lfn(LFC_LIST_FUNC_TYPE(FileCatalog::FileID));
    void L_get_fid_from_pfn(LFC_LIST_FUNC_TYPE(FileCatalog::FileID));
    void L_get_fid_from_guid(LFC_LIST_FUNC_TYPE(FileCatalog::FileID));

    char *m_errbuf;

    void destroyRetData() const;

    // stuff for the CWD
    mutable std::vector<std::string> m_cwds;

    bool push_cd(const std::string &new_dir) const {
      coral::MessageStream fclog( "LFCCatalog::push_cd");

      if (!m_transactionsta.test(1)) {
        fclog<<coral::Warning<< "not connected" <<coral::MessageStream::endmsg;
        throw FCbackendException("LFCCatalog::push_cd","not connected");
      }

      // attempt to get the first directory, but don't stop if that's not possible
      if (m_cwds.empty()) {
        char *p = lfc_getcwd(NULL, CA_MAXPATHLEN+1);
        if (p != NULL) {
          std::string initial_dir(p);
          free(p);
          m_cwds.push_back(initial_dir);
        }
      }

      if (lfc_chdir(new_dir.c_str()) < 0) {
        // chdir aborts transaction if it fails!
        std::string s = L_build_serr("could not chdir to " + new_dir);
        this->L_transactionDone();
        fclog<<coral::Warning<< s <<coral::MessageStream::endmsg;
        throw FCbackendException("LFCCatalog::push_cd",s);
        // failed to change directory
        // return false;
      }
      m_cwds.push_back(new_dir);
      return true;
    };

    void pop_cd() const {
      coral::MessageStream fclog( "LFCCatalog::pop_cd");

      if (!m_transactionsta.test(1)) {
        fclog<<coral::Warning<< "not connected" <<coral::MessageStream::endmsg;
        throw FCbackendException("LFCCatalog::pop_cd","not connected");
      }

      if (m_cwds.size() > 1) {
        std::string target = m_cwds[m_cwds.size()-2];
        if (lfc_chdir(target.c_str()) < 0) {
          std::string s = L_build_serr("could not chdir to " + target);
          this->L_transactionDone();
          fclog<<coral::Warning<< s <<coral::MessageStream::endmsg;
          throw FCbackendException("LFCCatalog::pop_cd",s);
        }       
        m_cwds.pop_back();
      }
    };

    void L_transactionDone() const;

    std::string L_build_err(const std::string &) const;
    std::string L_build_serr(const std::string &) const;

    // mutable, since they should be cleared after a transation finishes
    // and commit() & rollback() are const
    mutable void *m_retPFN;
    mutable void *m_retLFN;
    mutable void *m_retGUID;

    // delayed PFN registration
    struct delayed_t {
     public:
      typedef std::vector<PFNEntry> pfns_t;
      typedef std::map<FileCatalog::FileID, pfns_t> guid_2_pfns_t;
      typedef std::map<std::string, FileCatalog::FileID> pfn_2_guid_t;
      delayed_t() {
        clear();
      }
      void clear() {
        lfc_take_mutex lk(&numb_mutex);
        numb -= pfn_2_guid.size();
        guid_2_pfns.clear();
        pfn_2_guid.clear();
      }
      bool empty() const {
        return guid_2_pfns.empty();
      }
      size_t total_size() const {
        lfc_take_mutex lk(&numb_mutex);
        return numb;
      }
      pfns_t get_any_pfns() const {
        pfns_t set;
        if (this->empty()) return set;
        return guid_2_pfns.begin()->second;
      }
      bool has_pfname(const std::string &pfname) const {
        if (pfname.empty()) return false;
        pfn_2_guid_t::const_iterator it = pfn_2_guid.find(pfname);
        return (it != pfn_2_guid.end());
      }
      bool has_guid(const FileCatalog::FileID &guid) const {
        if (guid.empty()) return false;
        guid_2_pfns_t::const_iterator it = guid_2_pfns.find(guid);
        return (it != guid_2_pfns.end());
      }
      void insert_pfn(const PFNEntry &pfn) {
        if (pfn.pfname().empty() || pfn.guid().empty()) {
          throw FCbackendException("LFCCatalog::delayed_t::insert_pfn","Badly formed PFN");
        }
        if (pfn_2_guid.find(pfn.pfname()) != pfn_2_guid.end()) return;
        guid_2_pfns_t::const_iterator it = guid_2_pfns.find(pfn.guid());
        if (it != guid_2_pfns.end()) {
          guid_2_pfns[pfn.guid()].push_back(pfn);
        } else {
          pfns_t pfns;
          pfns.push_back(pfn);
          guid_2_pfns[pfn.guid()] = pfns;
        }
        pfn_2_guid[pfn.pfname()] = pfn.guid();
        lfc_take_mutex lk(&numb_mutex);
        ++numb;
      }
      void remove_guid(const FileCatalog::FileID &guid) {
        if (guid.empty()) return;
        guid_2_pfns_t::const_iterator it = guid_2_pfns.find(guid);
        if (it == guid_2_pfns.end()) return;
        size_t n=0;
        for(pfns_t::const_iterator it2 = it->second.begin(); it2 != it->second.end(); ++it2) {
          pfn_2_guid.erase(it2->pfname());
          ++n;
        }
        guid_2_pfns.erase(guid);
        if (n) {
          lfc_take_mutex lk(&numb_mutex);
          numb -= n;
        }
      }
      void remove_pfname(const std::string &pfname) {
        if (pfname.empty()) return;
        pfn_2_guid_t::const_iterator it = pfn_2_guid.find(pfname);
        if (it == pfn_2_guid.end()) return;
        FileCatalog::FileID fid = it->second;
        guid_2_pfns_t::iterator it2 = guid_2_pfns.find(fid);
        pfns_t &plist = it2->second;
        for(pfns_t::iterator it3 = plist.begin(); it3 != plist.end(); ++it3) {
          if (it3->pfname() == pfname) {
            plist.erase(it3);
            break;
          }
        }
        if (plist.empty()) {
          guid_2_pfns.erase(fid);
        }
        pfn_2_guid.erase(pfname);
        lfc_take_mutex lk(&numb_mutex);
        --numb;
      }
      pfns_t find(const FileCatalog::FileID &guid, const std::string &pfname) {
        FileCatalog::FileID fid = guid;
        pfns_t ret;
        if (!pfname.empty()) {
          pfn_2_guid_t::const_iterator it = pfn_2_guid.find(pfname);
          if (it != pfn_2_guid.end()) {
            fid = it->second;
          }
        }
        if (!fid.empty()) {
          guid_2_pfns_t::const_iterator it = guid_2_pfns.find(fid);
          if (it != guid_2_pfns.end()) {
            const pfns_t &plist = it->second;
            for(pfns_t::const_iterator it2=plist.begin(); it2!=plist.end(); ++it2) {
              if (pfname.empty() || it2->pfname() == pfname) {
                ret.push_back(*it2);
              }
            }
          }
        }
        return ret;
      }

     private:
      static pthread_mutex_t numb_mutex;
      static size_t numb;
      guid_2_pfns_t guid_2_pfns;
      pfn_2_guid_t pfn_2_guid;
    };

    mutable delayed_t m_delayed_pfn;

    // use session or transaction
    bool m_want_session;

    // the path prefix to use for generated LFNs
    std::string m_gen_path;

    // the path prefix to use for relative LFNs
    std::string m_lfc_home;

    // does the user want distinct results for bulk queries?
    bool m_want_distinct;

    // should we avoid dangling symlinks on delete or not?
    bool m_allow_dangling_lfns;

    // host for connection
    mutable std::string m_host;

    // port for connection
    mutable std::string m_port;

    // time at which the last 'keepalive' was sent
    mutable time_t m_timek;
  };
  
}

#endif
