/*
 * Some ancillary parsing routines for LFCCatalog
 * Author : James Casey and David Smith
 */
#include "LFCCatalog/LFCParser.h"
#include "FileCatalog/FCException.h"
#include <string>
#include <ctype.h>

#include "CoralBase/MessageStream.h"

namespace pool {
  LFCQueryException::LFCQueryException(const std::string& from,
				       const std::string& reason) : 
    FCbackendException(std::string("Invalid LFC Query in ")+from, reason){
  }
  /*
  LFCQueryPhrase::LFCQueryPhrase() {

  }
  */
  LFCParser::LFCParser() {
  }
  
  LFCParser::~LFCParser() {
  }

  /** parse the query :
   *  * split into words
   *  * run through state machine to validate query
   *  * contruct set of QueryPhrases (LFCCatalog specific)
   */

  std::vector<LFCQueryPhrase> LFCParser::parse(const std::string& query) const {
    coral::MessageStream fclog( "LFCParser::parse");

    std::vector<std::string> words =  split_query(query);
    // and now validate
    int state = START;
    std::vector<LFCQueryPhrase> phrases;

    for(std::vector<std::string>::const_iterator it = words.begin();
	it != words.end(); ++it) {
      state = do_transition(state, *it, phrases);
    }
    // check we end properly
    state = do_transition(state, "", phrases);
    if(state != LFCParser::END) {
      fclog<<coral::Warning<< "Incorrectly constructed query" <<coral::MessageStream::endmsg;
      throw LFCQueryException("LFCParser::parse",
			      "Incorrectly constructed query");
    }
    
    return phrases;
  }

  // helper methods
  std::vector<std::string> LFCParser::split_query(const std::string& query) const {
    coral::MessageStream fclog( "LFCParser::split_query");

    std::vector<std::string> words;
    std::string this_word;
    const char *p;
    char c;
    bool single_quote=false,
         double_quote=false,
         in_word     =false;

    enum { ONE=1, TWO, THREE } state=ONE;
    p = query.c_str();

    while((c = *p++) != '\0') {

      // break processing
      if (!(single_quote || double_quote)) {
        bool word_break=false;

        switch(state) {
          case ONE:
            if (c=='=') {   
              state = TWO;  
              if (in_word) word_break=true;   
            } else if (isblank(c)) {
              word_break = true;
            }
            break;

         case TWO:
            if (c != '=') {
              if (in_word) {
                state = THREE;
                word_break = true;
              } else if (isblank(c)) {
                state = ONE;
                word_break = true;
              }
            }
            break;
               
         case THREE:
            if (c == '=' && !in_word) {
              state = TWO;
            } else if (isblank(c)) {
              state = ONE;
              word_break = true;
            }
            break;
        }
              
        // maybe do the word break
        if (word_break) {
          if (in_word) {

            words.push_back(this_word);
            in_word=false;

            // if it is not a blank, reprocess this char again
            if (!isblank(c)) --p;
          }
          continue;
        }
      }
      // not (unquoted) whitespace

      if (c == '\'' && !double_quote) {
        single_quote = !single_quote;
        continue;
      }
      if (c == '"' && !single_quote) {
        double_quote = !double_quote;
        continue;
      }
      // not (unquoted) quote mark

      // so this is part of a word, add it to the current or start a new one
      if (in_word) {
        this_word += c;
      } else {
        in_word = true;
        this_word = c;
      }
    }
    // add on the last word, if any

    if (in_word) words.push_back(this_word);
    // check for unclosed quotes
    if (double_quote || single_quote) {
      fclog<<coral::Warning<< "Missing quote mark in query" <<coral::MessageStream::endmsg;
      throw LFCQueryException("LFCParser::split_query",
			      "Missing quote mark in query");
    }

    return words;
  }

  int LFCParser::do_transition(const int& state, const std::string& word,
			       std::vector<LFCQueryPhrase>& phrases) const {

    coral::MessageStream fclog( "LFCParser::do_transition");

    if(word == "") {
      if(state == VALUE) {
	// this is the only good end transition
	return END;
      } else 
	// return cur state as last state
	return state;
    }
    
    if(state == START || state == JOIN) {
      if(compare(word, "PFNAME") == 0) {
	LFCQueryPhrase p(LFCQueryPhrase::PFN);
	phrases.push_back(p);
      } else if (compare(word, "LFNAME") == 0) {
	phrases.push_back( LFCQueryPhrase(LFCQueryPhrase::LFN));
      } else if (compare(word,"GUID") == 0) {
      	phrases.push_back( LFCQueryPhrase(LFCQueryPhrase::GUID));
      }  else {
        std::string s = std::string("Invalid name in query: ") + word;
        fclog<<coral::Warning<< s <<coral::MessageStream::endmsg;
	throw LFCQueryException("LFCParser::do_transition",s);
      }
      return NAME;
    } else if (state == NAME) {
      if(word == "=") { 
	phrases.back().setOp("=");
	return COMPARATOR;
      }  else if (compare(word, "LIKE") == 0) {
	phrases.back().setOp("LIKE");
	return COMPARATOR;
      }      
      std::string s = std::string("Invalid comparator in query: ") + word;
      fclog<<coral::Warning<< s <<coral::MessageStream::endmsg;
      throw LFCQueryException("LFCParser::do_transition",s);

    } else if(state == COMPARATOR) {
      phrases.back().setValue(word);
      return VALUE;
    } else if(state == VALUE) {
      if(compare(word, "AND") == 0) {
	LFCQueryPhrase p(LFCQueryPhrase::JOIN);
	p.setOp("AND");
	phrases.push_back(p);
	return JOIN;
      } else if( compare(word, "OR") == 0) {
	LFCQueryPhrase p(LFCQueryPhrase::JOIN);
	p.setOp("OR");
	phrases.push_back(p);
	return JOIN;
      }
      std::string s = std::string("Invalid operator in query: ") + word;
      fclog<<coral::Warning<< s <<coral::MessageStream::endmsg;
      throw LFCQueryException("LFCParser::do_transition",s);

    }
    fclog<<coral::Warning<< "Invalid State" <<coral::MessageStream::endmsg;
    throw LFCQueryException("LFCParser::do_transition",
			    "Invalid State");
  }
  
  int LFCParser::compare(const std::string& str1, const std::string& str2) const {
    std::string::const_iterator it1 = str1.begin(), it2 = str2.begin();

    while( (it1 != str1.end()) && (it2 != str2.end())) {
      if( ::toupper(*it1) != ::toupper(*it2))
	return (::toupper(*it1) < ::toupper(*it2)) ? -1 : 1;
      ++it1; ++it2;
    }
    size_t size1 = str1.size(), size2 = str2.size();
    if(size1 == size2)
      return 0;
    return (size1 < size2) ? -1 : 1;
  }
};
  
