/*
 * Copyright (C) 2004 by CERN
 */

/*
 * @(#)$RCSfile: lfc_ifce.cpp,v $ $Revision: 1.10 $ $Date: 2008/02/20 09:38:29 $ CERN James Casey
 */
#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <uuid/uuid.h>

extern "C" {
#include "lfc_api.h"
#include "serrno.h"
}
#include "LFCCatalog/gfal_api.h"

#include "LFCCatalog/LFCCatalog.h"

#define ALLOC_BLOCK_SIZE 16 /* the block size to allocate new pointers in */

/** extract a hostname from a SURL.  We search for "://" to get the start of
    the hostname.  Then we keep going to the next slash, colon or end of the
    SURL. */
char *
get_hostname(const char *path, char *errbuf, int errbufsz) {
  char *start;
  char *cp;
  char *result;
  char c;
  char sav_path[CA_MAXSFNLEN+1];

  strcpy(sav_path, path);

  start = strchr(sav_path, ':');
  if ( start == NULL || *(start+1) != '/' || *(start+2) != '/') {
    result = strdup("unknown");
  } else {
    /* point start to beginning of hostname */
    start += 3;
    for(cp = start; *cp != '\0' && *cp != ':' && *cp != '/'; cp++)
      ;
    c = *cp;
    *cp = '\0';
    result = strdup(start);
    *cp = c;
  }
  
  if (result != NULL && (strlen(result) == 0 || strlen(result) >  CA_MAXHOSTNAMELEN)) {
    free(result);
    result = strdup("unknown");
  }

  if(result == NULL) {
    gfal_errmsg(errbuf, errbufsz, "Could not allocate memory for a result");
    errno = ENOMEM;
    return (NULL);
  }

  return result;
}

static int 
lfc_init (char *errbuf, int errbufsz) {
  lfc_seterrbuf(errbuf, errbufsz);
  return (0);
}

/* transaction never aborted */
/* ENOENT returned if GUID does not exist */
int lfc_replica_exists (const char *guid, char *errbuf, int errbufsz) {
  struct lfc_filereplica *rp_vec;
  int ret,inum;

  if(lfc_init(errbuf, errbufsz) < 0)
    return (-1);

  rp_vec = NULL;
  ret = lfc_getreplica(NULL, guid, NULL, &inum, &rp_vec);

  if (ret != 0) {
    int serr = serrno;
    if (serr < 1000) {
      errno = serr;
    } else {
      gfal_errmsg(errbuf, errbufsz, sstrerror(serr));
      errno = ECOMM;
    }

    return(-1);
  }

  if (rp_vec == NULL) {
    inum=0;
  } else {
    free(rp_vec);
  }

  if (inum)
    return(1);

  return(0);
}

/* never aborts transaction */
int
lfc_getfilesizeg(const char *guid, GFAL_LONG64 *sizep, char *errbuf, int errbufsz)
{
  struct lfc_filestatg statg;

  if(lfc_init(errbuf, errbufsz) < 0)
    return (-1);

  if(lfc_statg(NULL, guid, &statg) < 0) {
    if (serrno < 1000) 
      errno = serrno;
    else {
      gfal_errmsg(errbuf, errbufsz, sstrerror(serrno));
      errno = ECOMM;
    }
    return (-1);
  }
  
  *sizep = (u_signed64) statg.filesize;
  return (0);
}

/** lfc_guidforpfn : Get the guid for a replica.  If the replica does not
    exist, fail with ENOENT
    any existing transaction is never rolled back
*/
char *
lfc_guidforpfn (const char *pfn, char *errbuf, int errbufsz)
{
  char *p;
  struct lfc_filestatg statg;

  if(lfc_init(errbuf, errbufsz) < 0)
    return (NULL);

  if(lfc_statr(pfn, &statg) < 0) {
    if(serrno < 1000)
      errno = serrno;
    else {
      gfal_errmsg(errbuf, errbufsz, sstrerror(serrno));
      errno = ECOMM;
    }
    return (NULL);
  }
  if((p = strdup(statg.guid)) == NULL) {
    return (NULL);
  }
  return (p);
}

/* never causes transaction to be rolled back */
int
lfc_guid_exists (const char *guid, char *errbuf, int errbufsz)
{
  struct lfc_filestatg statg;

  if(lfc_init(errbuf, errbufsz) < 0)
    return (-1);

  if(lfc_statg(NULL, guid, &statg) < 0) {
    if(serrno == ENOENT) 
	return (0);
    if (serrno < 1000) 
      errno = serrno;
    else {
      gfal_errmsg(errbuf, errbufsz, sstrerror(serrno));
      errno = ECOMM;
    }
    return (-1);
  }
  return (1);
}

/* on error transaction is aborted */
int
lfc_register_pfn (const char *guid, const char *pfn, char *errbuf, int errbufsz)
{
  char *hostname;

  if(lfc_init(errbuf, errbufsz) < 0)
    return (-1);
  
  if((hostname = get_hostname(pfn, errbuf, errbufsz)) == NULL) {
    return (-1);
  }
  /* We always have available permanent files at the minute */
  if(lfc_addreplica(guid, NULL, hostname, pfn, '-', 'P', NULL, NULL) < 0) {
    if (serrno < 1000)
      errno = serrno;
    else {
      gfal_errmsg(errbuf, errbufsz, sstrerror(serrno));
      errno = ECOMM;
    }
    free(hostname);
    return (-1);
  }
  free(hostname);
  return (0);
}

/* transaction never rolledback
   ENOENT returned if GUID does not exist, empty list for no replicas
*/
char **
lfc_surlsfromguid (const char *guid, char *errbuf, int errbufsz)
{
  struct lfc_filereplica *rp_vec;
  char **p;
  int ret, inum;
  
  if(lfc_init(errbuf, errbufsz) < 0)
    return (NULL);

  rp_vec = NULL;
  ret = lfc_getreplica(NULL, guid, NULL, &inum, &rp_vec);

  if (ret != 0) {
    int serr = serrno;
    if (serr < 1000) {
      errno = serr;
    } else {
      gfal_errmsg(errbuf, errbufsz, sstrerror(serr));
      errno = ECOMM;
    }
    return(NULL);
  }

  if (rp_vec == NULL)
    inum=0;

  /* allocate some memory for the pointers */
  if((p = (char**)malloc(sizeof(char*) * (inum+1))) == NULL) {
    return (NULL);
  }

  p[inum] = NULL;

  for(int i=0;i<inum;++i) {
    struct lfc_filereplica *rp = &rp_vec[i];
    p[i] = strdup(rp->sfn);
  }

  if (rp_vec != NULL)
    free(rp_vec);

  return(p);
}

/* transaction never rolled back
   ENOENT is returned if either GUID does not exist or if there are no replicas
*/
char *
lfc_surlfromguid (const char *guid, char *errbuf, int errbufsz)
{
  char **surls;
  char **cp;
  char *result;
  int inum;
  
  if(lfc_init(errbuf, errbufsz) < 0)
    return (NULL);

  if((surls = lfc_surlsfromguid(guid, errbuf, errbufsz)) == NULL) {
    /* pass through error number */
    return (NULL);
  }

  inum=0;
  while(surls[inum] != NULL) inum++;

  if (inum == 0) {
    free(surls);
    errno = ENOENT;
    return(NULL);
  }

  result = getbestfile(surls, inum, errbuf, errbufsz);

  for(cp = surls; *cp != NULL; ++cp) {
    if(*cp != result) {
      free (*cp);
    }
  }
  free (surls);
  return (result);
}

/** lfc_unregister_pfn : We unregister a pfn from a guid, but only if it is a
    replica for that guid. It is an error for both the guid or pfn not to exist
    or for the guid to not correspond to the pfn
    Transaction is rolledback on error
*/
int
lfc_unregister_pfn (const char *guid, const char *pfn, char *errbuf, int errbufsz)
{
  if(lfc_init(errbuf, errbufsz) < 0)
    return (-1);
  
  if(lfc_delreplica(guid, NULL, pfn) < 0) {
    if(serrno < 1000) 
      errno = serrno;
    else {
      gfal_errmsg(errbuf, errbufsz, sstrerror(serrno));
      errno = ECOMM;
    }
    return (-1);
  }
  return (0);
}

/* transaction never rolledback */
char *
lfc_guidfromlfn (const char *lfn, char *errbuf, int errbufsz)
{
  struct lfc_filestatg statg;
  char *p;

  if(lfc_init(errbuf, errbufsz) < 0)
    return (NULL);

  if(lfc_statg(lfn, NULL, &statg) < 0) {
    if(serrno < 1000)
      errno = serrno;
    else {
      gfal_errmsg(errbuf, errbufsz, sstrerror(serrno));
      errno = ECOMM;
    }
    return (NULL);
  }
  if (strlen(statg.guid)==0) {
    errno = ENOENT;
    if (statg.filemode & S_IFDIR)
      errno = EISDIR;
    return(NULL);
  }
  if((p = strdup(statg.guid)) == NULL)
    return (NULL);
  return (p);
}

/* ENOENT returned if guid does not exist
   Transaction never rolledback
*/
char **
lfc_lfnsforguid (const char *guid, char *errbuf, int errbufsz)
{
  struct lfc_linkinfo *li_vec;
  struct lfc_filestatg statg;
  char **p;
  int ret, inum;

  if(lfc_init(errbuf, errbufsz) < 0)
    return (NULL);

  if(lfc_statg(NULL, guid, &statg) < 0) {
    if(serrno == ENOENT) {
      errno = ENOENT;
      return(NULL);
    }
  }

  li_vec = NULL;
  ret = lfc_getlinks(NULL, guid, &inum, &li_vec);

  if (ret != 0) {
    int serr = serrno;

    if (serr == ENOENT)
      serr = ECOMM;

    if (serr < 1000) {
      errno = serr;
    } else {
      gfal_errmsg(errbuf, errbufsz, sstrerror(serr));
      errno = ECOMM;
    }
    return(NULL);
  }

  if (li_vec == NULL)
    inum=0;

  if (inum == 0) {
    if (li_vec != NULL)
      free(li_vec);

    errno = ENOENT;
    return(NULL);
  }

  /* allocate some memory for the pointers */
  if((p = (char**)malloc(sizeof(char*) * (inum+1))) == NULL) {
    return (NULL);
  }

  p[inum] = NULL;

  for(int i=0;i<inum;++i) {
    struct lfc_linkinfo *li = &li_vec[i];
    p[i] = strdup(li->path);
  }

  if (li_vec != NULL)
    free(li_vec);

  return(p);
}

/* transaction aborted on error */
int
lfc_create_alias (const char *guid, const char *lfn, GFAL_LONG64 size, char *errbuf, int errbufsz)
{
  if(lfc_init(errbuf, errbufsz) < 0)
    return (-1);

  //lfc_starttrans();
  if(lfc_creatg(lfn, guid, 0777) < 0) {
    if(serrno < 1000) 
      errno = serrno;
    else {
      gfal_errmsg(errbuf, errbufsz, sstrerror(serrno));
      errno = ECOMM;
    }
    return (-1);
  }
  if(size!=0 && lfc_setfsizeg(guid, size, NULL, NULL) < 0) {
    if(serrno < 1000)
      errno = serrno;
    else {
      gfal_errmsg(errbuf, errbufsz, sstrerror(serrno));
      errno = ECOMM;
    }
    return (-1);
  }
  //lfc_endtrans();
  return (0);
}

/* transaction abort on error except for ENOENT */
int
lfc_register_alias (const char *guid, const char *lfn, char *errbuf, int errbufsz)
{
  struct lfc_filestatg statg;
  char master_lfn[CA_MAXPATHLEN+1];
  
  if(lfc_init(errbuf, errbufsz) < 0)
    return (-1);

  //lfc_starttrans();
  if(lfc_statg(NULL, guid, &statg) < 0) {
    /* have to return ENOENT */
    errno = ENOENT;
    return(-1);
  }

  /* now we do a getpath() to get the master lfn */
  if (lfc_getpath(NULL, statg.fileid, master_lfn) <0 ) {
    /* have to return ENOENT */
    errno = ENOENT;
    return(-1);
  }

  /* and finally register */
  if(lfc_symlink(master_lfn, lfn) < 0) {
    if (serrno == ENOENT)
      serrno = ECOMM;

    if (serrno < 1000)
      errno = serrno;
    else {
      gfal_errmsg(errbuf, errbufsz, sstrerror(serrno));
      errno = ECOMM;
    }
    return (-1);
  }
  //lfc_endtrans();
  return (0);
}

/* abort transaction on any error
   The GUID is unused; it does not have to correspond to the lfn
*/
int 
lfc_unregister_alias (const char __attribute__((unused)) *guid, const char *lfn, char *errbuf, int errbufsz)
{
  if(lfc_init(errbuf, errbufsz) < 0)
    return (-1);

  //lfc_starttrans();

  if(lfc_unlink(lfn) < 0) {
    if(serrno < 1000) 
      errno = serrno;
    else {
      gfal_errmsg(errbuf, errbufsz, sstrerror(serrno));
      errno = ECOMM;
    }
    return (-1);
  }
  //lfc_endtrans();
  return (0);
}

int 
lfc_mkdirp(const char *path, mode_t mode, char *errbuf, int errbufsz)
{
  int c;
  char *lastslash = NULL;
  char *p;
  char *p1;
  char *q;
  char sav_path[CA_MAXPATHLEN+1];
  struct lfc_filestatg statbuf;
  uuid_t uuid;
  char uuid_buf[CA_MAXGUIDLEN+1];

  if(lfc_init(errbuf, errbufsz) < 0)
    return (-1);

  if (strlen (path) >= sizeof(sav_path)) {
    gfal_errmsg(errbuf, errbufsz, "Path too long");
    errno = ENAMETOOLONG;
    return (-1);
  }
  strcpy (sav_path, path);
  p1 = strchr (sav_path, '/');
  p = strrchr (sav_path, '/');
  while (p > p1) {
    if (lastslash == NULL) lastslash = p;
    *p = '\0';
    c = lfc_statg (sav_path, NULL, &statbuf);
    if (c == 0) {
      *p = '/';
      break;
    }
    if (serrno != ENOENT) {
      if(serrno < 1000) 
	errno = serrno;
      else {
	gfal_errmsg(errbuf, errbufsz, sstrerror(serrno));
	errno = ECOMM;
      }
      return (c);
    }
    q = strrchr (sav_path, '/');
    *p = '/';
    p = q;
  }
  c = 0;
  while (c == 0 && (p = strchr (p + 1, '/')) && p <= lastslash) {
    *p = '\0';
    uuid_generate(uuid);
    uuid_unparse(uuid, uuid_buf);
    c = lfc_mkdirg (sav_path, uuid_buf, mode);
    if(c != 0) {
      if (serrno < 1000)
	errno = serrno;
      else {
	gfal_errmsg(errbuf, errbufsz, sstrerror(serrno));
	errno = ECOMM;
      }
    }
    *p = '/';
  }
  return (c);
}

