#include "MappingBuilder.h"
#include "XMLDatabaseManagerDriver.h"
#include "ObjectRelationalAccess/ObjectRelationalNames.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingSchema.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingPersistency.h"
#include "ObjectRelationalAccess/ObjectRelationalException.h"
#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingRules.h"
#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingBuilder.h"
#include "ObjectRelationalAccess/ObjectRelationalTableNameRationalizer.h"
#include "ObjectRelationalAccess/ObjectRelationalClassUtils.h"
#include "ObjectRelationalAccess/ObjectRelationalDatabaseDescriptor.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingDescriptor.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include <sstream>
#include <fstream>
#include <memory>
#include "Reflex/Type.h"
#include "CoralBase/Exception.h"
#include "CoralBase/MessageStream.h"

#include "RVersion.h"
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
using namespace ROOT;
#endif
  
pool::MappingBuilder::MappingBuilder( coral::ISchema& schema ):
  m_initialized(false),
  m_builtMappings(),
  m_containerMappings(),
  m_mappingVersions(),
  m_schema(schema),
  m_mappingPersistency(schema),
  m_containerTableMap(){
}

pool::MappingBuilder::~MappingBuilder(){
  deleteMappings();
}

void pool::MappingBuilder::init(bool forWriting){
  if(!m_initialized){
    if(forWriting) m_mappingPersistency.initializeForWriting();
    m_mappingVersions.clear();
    std::map<std::string,std::string> vers = m_mappingPersistency.retrieveMappingVersions();
    for(std::map<std::string,std::string>::const_iterator iV=vers.begin();
        iV!=vers.end();++iV){
      m_mappingVersions.insert(std::make_pair(iV->first,true));
    }
    m_containerTableMap = m_mappingPersistency.containerTableMap();
    m_initialized = true;
  }
}


void pool::MappingBuilder::deleteMappings(){
  for(std::map<std::string,std::pair<ObjectRelationalMapping*,ClassMetadata> >::iterator iM=m_builtMappings.begin();
      iM!=m_builtMappings.end();++iM) {
    delete iM->second.first;
  }
  m_builtMappings.clear();
}

void pool::MappingBuilder::buildMapping(const ObjectRelationalMappingDescriptor* mappingDescriptor){
  this->init(true);
  std::auto_ptr<ObjectRelationalTableNameRationalizer> rationalizer(new ObjectRelationalTableNameRationalizer( m_schema ));
  
  const std::vector<ObjectRelationalMappingDescriptorElement*>& classDescriptors =  mappingDescriptor->classDescriptors();
  std::vector<ObjectRelationalMappingDescriptorElement*>::const_iterator iClassDescriptor;
  ObjectRelationalDefaultMappingBuilder defaultBuilder;
  
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  for(iClassDescriptor=classDescriptors.begin();iClassDescriptor!=classDescriptors.end();iClassDescriptor++){

    std::string className = (*iClassDescriptor)->name();

    Reflex::Type dict = Reflex::Type::ByName( className );

    if ( ! dict ) {
      throw ObjectRelationalException( "Dictionary information for class \"" + className + "\" not found.",
                                       "MappingBuilder::buildMapping");
    }
    std::string classVersion = pool::ObjectRelationalDefaultMappingRules::defaultClassVersion( className );

    Reflex::PropertyList classProps = dict.Properties();

    if(classProps.HasProperty(pool::ObjectRelationalDefaultMappingRules::classVersionPropertyNameInDictionary())){
      classVersion = classProps.PropertyAsString(pool::ObjectRelationalDefaultMappingRules::classVersionPropertyNameInDictionary());
    }

    bool buildMapping = true;
    std::string containerName("");
    try{
      containerName = (*iClassDescriptor)->attribute(pool::ObjectRelationalMappingDescriptorElement::poolContainerNameAttr());
    } catch (const ObjectRelationalException&){
    }
    if(!containerName.empty()){
      std::string useMapping("");
      try{
        useMapping = (*iClassDescriptor)->attribute(ObjectRelationalMappingDescriptorElement::useMappingAttr());
      } catch (const ObjectRelationalException&){
      }
      if(!useMapping.empty()){
        buildMapping = false;
        std::map<std::string,bool>::const_iterator iM=m_mappingVersions.find(useMapping);
        if(iM==m_mappingVersions.end()){
          throw ObjectRelationalException( "Mapping \""+ useMapping +"\" specified for container \""+containerName+"\" does not exist in the database.",
                                           "MappingBuilder::buildMapping");
        } else {
          m_containerMappings.insert(std::make_pair(containerName,std::make_pair(classVersion,useMapping)));
        }
      }
      
    }

    if(buildMapping){

      std::string mappingVersion("");
      try{
        mappingVersion = (*iClassDescriptor)->attribute(ObjectRelationalMappingDescriptorElement::mappingVersionAttr());
      } catch (const ObjectRelationalException&){
      }

      if(mappingVersion.empty()){
        for ( int iteration = 0;; ++iteration ) {
          mappingVersion = pool::ObjectRelationalDefaultMappingRules::newMappingVersionForClass( className, iteration );
          bool found = false;
          for ( std::map<std::string,bool>::reverse_iterator iVersion = m_mappingVersions.rbegin();
                iVersion != m_mappingVersions.rend(); ++iVersion ) {
            if ( iVersion->first == mappingVersion ) {
              found = true;
              break;
            }
          }
          if ( ! found ) break;
        }
      } else {
        std::map<std::string,bool>::const_iterator iVer = m_mappingVersions.find(mappingVersion);
        if(iVer!=m_mappingVersions.end()){
          throw ObjectRelationalException( "Mapping \""+ mappingVersion +"\" already exists in the database.",
                                           "MappingBuilder::buildMapping");          
        }
        
      }
      
      m_mappingVersions.insert(std::make_pair(mappingVersion,false));

      std::string baseMappingVersion("");
      bool useBase = false;
      try{
        baseMappingVersion = (*iClassDescriptor)->attribute(ObjectRelationalMappingDescriptorElement::baseMappingAttr());
      } catch (const ObjectRelationalException&){
      }
      if(!baseMappingVersion.empty()){
        std::map<std::string,bool>::const_iterator iBm = m_mappingVersions.find( baseMappingVersion );
        if(iBm == m_mappingVersions.end() || !iBm->second){
          throw ObjectRelationalException( "Base Mapping \""+ baseMappingVersion +"\" does not exist in the database.",
                                           "MappingBuilder::buildMapping");
        } else {
          useBase = true;
        }
      }
      
      ObjectRelationalMapping* defaultMapping = 0;
      std::map<std::string,std::pair<ObjectRelationalMapping*,ClassMetadata> >::iterator iM = m_builtMappings.find(mappingVersion);
      if(iM==m_builtMappings.end()){
        defaultMapping = new ObjectRelationalMapping(mappingVersion);
        ClassMetadata mdata = std::make_pair(classVersion,containerName);
        m_builtMappings.insert(std::make_pair(mappingVersion,std::make_pair(defaultMapping,mdata)));
      } else {
        defaultMapping = iM->second.first;
      }

      Reflex::Type dictionaryEntry = dict;
      className = dictionaryEntry.Name(Reflex::SCOPED);

      log << coral::Debug << "Processing class " << className << coral::MessageStream::endmsg;
      
      if(defaultMapping->className().empty()){
        if ( !defaultBuilder.processClass(*defaultMapping, dictionaryEntry) ){
          throw ObjectRelationalException( "Processing of default mapping for class \"" + className + "\" failed.",
                                           "MappingBuilder::buildMapping" );
        }
      }
      if(defaultMapping->className()!=className){
        throw ObjectRelationalException( "Mapping for class \"" + className + "\" not found.",
                                         "MappingBuilder::buildMapping");
      }

      // rationalize mapping element
      rationalizer->rationalizeMappingElement( defaultMapping->element() );
      if(useBase){
        ObjectRelationalMapping* baseMapping = m_mappingPersistency.retrieveMapping( baseMappingVersion );
        if(baseMapping) {
          std::auto_ptr<ObjectRelationalMapping> baseM(baseMapping);
          if(baseM->className()==className){
            defaultMapping->override(*baseMapping);
          }
        } else {
          throw ObjectRelationalException( "Could not retrieve the version \"" + baseMappingVersion + "\" as base mapping.",
                                           "MappingBuilder::buildMapping" );          
        }
      }
      (*iClassDescriptor)->alter(defaultMapping->element(),defaultBuilder.tableRegister());
      //check table usage
      std::set<std::string> involvedTables = defaultMapping->element().involvedTables();
      for(std::set<std::string>::const_iterator iT=involvedTables.begin();
          iT!=involvedTables.end();++iT){
        std::map<std::string,std::string>::const_iterator iCt = m_containerTableMap.find(*iT);
        if(iCt==m_containerTableMap.end()){
          log << coral::Debug << "Table \"" << *iT << "\" will be reserved for container \""<< containerName << "\""<< coral::MessageStream::endmsg; 
          m_containerTableMap.insert(std::make_pair(*iT,containerName));
        } else {
          log << coral::Debug << "Table \"" << *iT << "\" is reserved for container \""<< iCt->second << "\""<< coral::MessageStream::endmsg;
          if(iCt->second!=containerName){
            throw ObjectRelationalException( "Table \""+*iT+"\" involved in mapping of class \"" + className +
                                             "\" for container \"" + containerName+"\" is already used by container \""+iCt->second+"\".",
                                             "MappingBuilder::buildMapping" );            
          }
        }
      }
      
    }
    
  }
}

bool pool::MappingBuilder::exportMapping(const std::string& fileName){
  if(m_builtMappings.size()){
    std::auto_ptr<ObjectRelationalDatabaseDescriptor> databaseDescriptor(new ObjectRelationalDatabaseDescriptor);
    for(std::map<std::string,std::pair<ObjectRelationalMapping*,ClassMetadata> >::iterator iM=m_builtMappings.begin();
        iM!=m_builtMappings.end();++iM){
      pool::ObjectRelationalContainerDescriptor& contDescriptor = databaseDescriptor->addContainer(iM->second.second.second,"");
      if(!iM->second.first->className().empty()){
        pool::ObjectRelationalClassDescriptor* classDescriptor = contDescriptor.addClassStore(iM->second.first->element().variableType(),iM->first);
        ObjectRelationalMappingDescriptor* mappDescr =  classDescriptor->newMapping();
        mappDescr->import(*(iM->second.first));
      }
    }
    XMLDatabaseManagerDriver driver;
    std::ofstream outFile;
    outFile.open (fileName.c_str(), std::fstream::out);
    driver.write(*databaseDescriptor,outFile);
    outFile.close();
    return true;
  } else {
    return false;
  }
}

bool pool::MappingBuilder::retrieveMapping(const std::string& version,const std::string& fileName){
  ObjectRelationalMappingSchema mappingSchema(m_schema);
  if(!mappingSchema.existTables()){
    throw ObjectRelationalException( "Mapping database not found.",
                                     "MappingBuilder::retrieveMapping");
  }
  ObjectRelationalMapping* mapping = m_mappingPersistency.retrieveMapping( version );
  if(!mapping) return false;
  std::string containerName = m_mappingPersistency.containerForMappingVersion( version );
  
  std::auto_ptr<ObjectRelationalDatabaseDescriptor> databaseDescriptor(new ObjectRelationalDatabaseDescriptor);
  pool::ObjectRelationalContainerDescriptor& contDescriptor = databaseDescriptor->addContainer(containerName,"");
  if(!mapping->className().empty()){
    pool::ObjectRelationalClassDescriptor* classDescriptor = contDescriptor.addClassStore(mapping->element().variableType(),version);
    ObjectRelationalMappingDescriptor* mappDescr =  classDescriptor->newMapping();
    mappDescr->import(*mapping);
  }
  delete mapping;
  XMLDatabaseManagerDriver driver;
  std::ofstream outFile;
  outFile.open (fileName.c_str(), std::fstream::out);
  driver.write(*databaseDescriptor,outFile);
  outFile.close();
  return true;
}

bool pool::MappingBuilder::retrieveMapping(const std::string& fileName, bool oldVersionSchema){
  //ObjectRelationalMappingSchema mappingSchema(m_schema);
  //if(!mappingSchema.existTables()){
  //  throw ObjectRelationalException( "Mapping database not found.",
  //                                   "MappingBuilder::retrieveMapping");
  //}
  std::auto_ptr<ObjectRelationalDatabaseDescriptor> databaseDescriptor(new ObjectRelationalDatabaseDescriptor);
  XMLDatabaseManagerDriver driver;
  std::ofstream outFile;
  outFile.open (fileName.c_str(), std::fstream::out);
  std::map<std::string,std::string > versions = m_mappingPersistency.retrieveMappingVersions();
  
  for(std::map<std::string,std::string >::const_iterator iMv = versions.begin();
      iMv != versions.end(); iMv++){
    const ObjectRelationalMapping* mapping = 0;
    if(oldVersionSchema) {
      mapping = m_mappingPersistency.retrieveMappingFromOldSchema( iMv->first );
    } else {
      mapping = m_mappingPersistency.retrieveMapping( iMv->first );
    }
    if(mapping){
      pool::ObjectRelationalContainerDescriptor& contDescriptor = databaseDescriptor->addContainer(iMv->second,"");
      if(!mapping->className().empty()){
        pool::ObjectRelationalClassDescriptor* classDescriptor = contDescriptor.addClassStore(mapping->element().variableType(),iMv->first);
        ObjectRelationalMappingDescriptor* mappDescr =  classDescriptor->newMapping();
        mappDescr->import(*mapping);
      }
      delete mapping;
      mapping = 0;
    }
  }
  driver.write(*databaseDescriptor,outFile);
  outFile.close();
  return true;
}

std::map<std::string,std::string >
pool::MappingBuilder::mappingVersions(){
  ObjectRelationalMappingSchema mappingSchema(m_schema);
  if(!mappingSchema.existTables()){
    throw ObjectRelationalException( "Mapping database not found.",
                                     "MappingBuilder::retrieveMappingVersions");
  }
  return m_mappingPersistency.retrieveMappingVersions();
}

std::map<std::pair<std::string,std::string>, std::string >
pool::MappingBuilder::mappingForClassVersions(){
  ObjectRelationalMappingSchema mappingSchema(m_schema);
  if(!mappingSchema.existTables()){
    throw ObjectRelationalException( "Mapping database not found.",
                                     "MappingBuilder::retrieveMappingVersions");
  }
  return m_mappingPersistency.retrieveMappingForClassVersions();  
}

std::map<std::string, std::string>
pool::MappingBuilder::containerTableMap(){
  ObjectRelationalMappingSchema mappingSchema(m_schema);
  if(!mappingSchema.existTables()){
    throw ObjectRelationalException( "Mapping database not found.",
                                     "MappingBuilder::retrieveMappingVersions");
  }
  return m_mappingPersistency.containerTableMap();  
}

void
pool::MappingBuilder::selectMapping(const std::string& version, std::map<std::pair<std::string,std::string>,std::pair<std::string,std::string> >& mappingStruct){
  m_mappingPersistency.retrieveMapping(version,mappingStruct);
}

bool 
pool::MappingBuilder::removeMapping( const std::string& version, bool removeTables, bool dryRun ){
  ObjectRelationalMappingSchema mappingSchema(m_schema);
  if(!mappingSchema.existTables()){
    throw ObjectRelationalException( "Mapping database not found.",
                                     "MappingBuilder::removeMapping");
  }
  bool ret = true;
  if( removeTables ) ret = m_mappingPersistency.removeMaterializedTables( version, dryRun );
  if( ret && !dryRun ) ret = m_mappingPersistency.removeMapping(version);
  return ret;
}

void pool::MappingBuilder::finalizeMapping(bool evolve, bool dryRun){
  if(m_containerMappings.size() || m_builtMappings.size()){
    for(std::map<std::string,std::pair<ObjectRelationalMapping*,ClassMetadata> >::const_iterator iM=m_builtMappings.begin();
        iM!=m_builtMappings.end();++iM){
      if(!dryRun) {
        if(!m_mappingPersistency.storeMapping( *(iM->second.first),iM->second.second.first,iM->second.second.second)){
          throw ObjectRelationalException( "Could not store the mapping.","MappingBuilder::finalizeMapping");
        }
      }
      if(!m_mappingPersistency.materializeMapping( *(iM->second.first), evolve, dryRun )){
        throw ObjectRelationalException( "Could not materialize the mapping.","MappingBuilder::finalizeMapping");
      }
    }
    for(std::map<std::string,std::pair<std::string,std::string> >::const_iterator iC=m_containerMappings.begin();
        iC!=m_containerMappings.end();++iC){
      if(!m_mappingPersistency.assignMapping( iC->second.first, iC->second.second, iC->first)){
        throw ObjectRelationalException( "Could not assign the mapping.","MappingBuilder::finalizeMapping");
      }
    }
  } else {
    coral::MessageStream log( pool::ObjectRelationalAccess::serviceName());
    log << coral::Debug << "No mapping to finalize..." << coral::MessageStream::endmsg;
  }
}

bool pool::MappingBuilder::upgradeMappingSchema(){
  ObjectRelationalMappingSchema mappingSchema(m_schema);
  if(!mappingSchema.existsOldSchema()) return false;
  mappingSchema.createTmpTables();
  // retrieve the versions and copy them in the tmp tables
  std::map<std::string,std::string> versionData = m_mappingPersistency.retrieveMappingVersions();
  m_mappingPersistency.storeMappingVersions( versionData, true );
  
  // retrieve the class versions and copy them in the tmp tables
  {
    std::map<std::pair<std::string,std::string>,std::string> classVersionData = m_mappingPersistency.retrieveMappingForClassVersions();
    m_mappingPersistency.storeMappingForClassVersions( classVersionData, true );
  }

  // retrieve the mapping data, export in xml file and copy them in the tmp tables
  std::auto_ptr<ObjectRelationalDatabaseDescriptor> databaseDescriptor(new ObjectRelationalDatabaseDescriptor);
  XMLDatabaseManagerDriver driver;
  std::ofstream outFile;
  outFile.open ("mapping.xml", std::fstream::out);
  for( std::map<std::string,std::string>::const_iterator iV = versionData.begin();
       iV!=versionData.end();++iV){
    ObjectRelationalMapping* mapping = m_mappingPersistency.retrieveMappingFromOldSchema( iV->first );
    pool::ObjectRelationalContainerDescriptor& contDescriptor = databaseDescriptor->addContainer(iV->second,"");
    if(!mapping->className().empty()){
      pool::ObjectRelationalClassDescriptor* classDescriptor = contDescriptor.addClassStore( mapping->element().variableType(),iV->first );
      ObjectRelationalMappingDescriptor* mappDescr =  classDescriptor->newMapping();
      mappDescr->import(*mapping);
    }

    m_mappingPersistency.storeMapping( *mapping, true );
    delete mapping;
  }
  driver.write(*databaseDescriptor,outFile);
  outFile.close();
  
  // drop old mapping tables
  mappingSchema.dropTables(true);
  // create new schema
  mappingSchema.createTables();
  // store version data in new schema
  m_mappingPersistency.storeMappingVersions( versionData );
  // store class version data in new schema
  {
    std::map<std::pair<std::string,std::string>,std::string> classVersionData = m_mappingPersistency.retrieveMappingForClassVersions( true );
    m_mappingPersistency.storeMappingForClassVersions( classVersionData );
  }
  // store mapping data in new schema
  for( std::map<std::string,std::string>::const_iterator iV = versionData.begin();
       iV!=versionData.end();++iV){
    ObjectRelationalMapping* mapping = m_mappingPersistency.retrieveMapping( iV->first, true );
    m_mappingPersistency.storeMapping( *mapping );
    delete mapping;
  }
  // drop tmp tables
  mappingSchema.dropTmpTables();
  return true;
}




