#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingBuilder.h"
#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingRules.h"
#include "ObjectRelationalAccess/ObjectRelationalNames.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include "ObjectRelationalAccess/ObjectRelationalClassUtils.h"
#include "ObjectRelationalAccess/ObjectRelationalException.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/MessageStream.h"
#include "POOLCore/Reference.h"
#include "Reflex/Reflex.h"
#include <sstream>

const size_t pool::ObjectRelationalDefaultMappingBuilder::MaxColumnsPerTable = 100;

pool::ObjectRelationalDefaultMappingBuilder::ObjectRelationalDefaultMappingBuilder():
  m_tableRegister(),
  m_log( *(new coral::MessageStream(pool::ObjectRelationalAccess::serviceName())) ),
  m_arraysInBlobs( false )
{}

pool::ObjectRelationalDefaultMappingBuilder::ObjectRelationalDefaultMappingBuilder(bool arraysInBlobs):
  m_tableRegister(),
  m_log( *(new coral::MessageStream(pool::ObjectRelationalAccess::serviceName())) ),
  m_arraysInBlobs( arraysInBlobs )
{}

pool::ObjectRelationalDefaultMappingBuilder::~ObjectRelationalDefaultMappingBuilder()
{
  delete &m_log;
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processClass( pool::ObjectRelationalMapping& mapping,
                                                           const Reflex::Type& classDictionary )
{
  std::string className = classDictionary.Name(Reflex::SCOPED);
  m_log << coral::Debug << "Processing mapping for class \""<< className << "\"." << coral::MessageStream::endmsg;

  size_t sz = sizeInColumns( classDictionary );
  m_log << coral::Info << "Class  \""<< className << "\" requires "<<sz<<" columns."<< coral::MessageStream::endmsg;
  if(sz > MaxColumnsPerTable){
    m_log << coral::Error << "Cannot process default mapping for class \""<<className<<"\" to the relational schema: number of columns ("<<sz<<") required exceedes maximum="<< MaxColumnsPerTable << coral::MessageStream::endmsg;
    return false;    
  }

  std::string tableName = pool::ObjectRelationalDefaultMappingRules::tableNameForClass( classDictionary.Name() );
  if(!m_tableRegister.checkTable(tableName)){
    m_tableRegister.insertTable(tableName);
  } else {
    m_log << coral::Error << "Table \"" <<tableName << "\" cannot be allocated to map class \""<<className <<"\" (it's already allocated.)"<< coral::MessageStream::endmsg;
    return false;
  }

  // Define the top level element
  pool::ObjectRelationalMappingElement& classElement = mapping.setElement( className, tableName );
  // Set the id of the class
  classElement.setColumnNames( std::vector< std::string >( 1, pool::ObjectRelationalDefaultMappingRules::columnNameForId() ) );
  return this->processObject( classElement, classDictionary );
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processDependentClass( ObjectRelationalMapping& parentClassMapping,
                                                                    const Reflex::Type& classDictionary )
{
  std::string className = classDictionary.Name(Reflex::SCOPED);
  m_log << coral::Debug << "Processing mapping for class \""<< className << "\"." << coral::MessageStream::endmsg;

  size_t sz = sizeInColumns( classDictionary );
  m_log << coral::Info << "Class  \""<< className << "\" requires "<<sz<<" columns."<< coral::MessageStream::endmsg;
  if(sz > MaxColumnsPerTable){
    m_log << coral::Error << "Cannot process default mapping for class \""<<className<<"\" to the relational schema: number of columns ("<<sz<<") required exceedes maximum="<< MaxColumnsPerTable << coral::MessageStream::endmsg;
    return false;    
  }

  std::string tableName = pool::ObjectRelationalDefaultMappingRules::tableNameForVariable( classDictionary.Name(), parentClassMapping.element().tableName() );
  if(!m_tableRegister.checkTable(tableName)){
    m_tableRegister.insertTable(tableName);
  } else {
    m_log << coral::Error << "Table \"" <<tableName << "\" cannot be allocated to map class \""<<className <<"\" (it's already allocated.)"<< coral::MessageStream::endmsg;
    return false;
  }

  pool::ObjectRelationalMappingElement& classElement = parentClassMapping.appendDependentElement( className, tableName );
  classElement.setDependency( parentClassMapping.element() );

  // Set the id of the class
  const std::vector<std::string>& parentClassColumns = parentClassMapping.element().columnNames();
  std::vector<std::string> columns;
  columns.reserve( parentClassColumns.size() + 1 );
  columns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForRefColumn() );
  for ( std::vector<std::string>::const_iterator iColumn = parentClassColumns.begin();
        iColumn != parentClassColumns.end(); ++iColumn ) {
    columns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForId() + "_" + *iColumn );
  }
  classElement.setColumnNames( columns );
  return this->processObject( classElement, classDictionary );
}

bool
pool::ObjectRelationalDefaultMappingBuilder::buildCArrayElementTree( ObjectRelationalMappingElement& topElement ){
  if( topElement.elementType()!=ObjectRelationalMappingElement::CArray ){
    return false;
  }
  Reflex::Type elementType = Reflex::Type::ByName(topElement.variableType());
  elementType = ObjectRelationalClassUtils::resolvedType( elementType );
  Reflex::Type arrayElementType = elementType.ToType();
  std::string contentTypeName = arrayElementType.Name();
  std::string variableNameForSchema = pool::ObjectRelationalDefaultMappingRules::variableNameForArrayColumn( arrayElementType );
  return processItem( topElement, contentTypeName, variableNameForSchema, arrayElementType);
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processPrimitive( pool::ObjectRelationalMappingElement& parentElement,
                                                               const std::string& attributeName,
                                                               const std::string& attributeNameForSchema,
                                                               const Reflex::Type& attributeType )
{
  const std::type_info* attrType = &attributeType.TypeInfo();
  if(attributeType.IsEnum()) attrType = &typeid(int);
  std::string typeName = coral::AttributeSpecification::typeNameForId(*attrType);

  return processLeafElement(pool::ObjectRelationalMappingElement::primitiveMappingElementType(),
                            parentElement,
                            attributeName,
                            attributeNameForSchema,
                            typeName);
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processBlob( pool::ObjectRelationalMappingElement& parentElement,
                                                          const std::string& attributeName,
                                                          const std::string& attributeNameForSchema,
                                                          const Reflex::Type& attributeType )
{
  std::string className = attributeType.Name(Reflex::SCOPED);
  return processLeafElement(pool::ObjectRelationalMappingElement::blobMappingElementType(),
                            parentElement,
                            attributeName,
                            attributeNameForSchema,
                            className);
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processLeafElement( const std::string& elementType,
                                                                 ObjectRelationalMappingElement& parentelement,
                                                                 const std::string& attributeName,
                                                                 const std::string& attributeNameForSchema,
                                                                 const std::string& typeName ){
  if(parentelement.find( attributeName )!=parentelement.end()){
    m_log << coral::Error << "Attribute name \"" << attributeName << "\" is already defined in the mapping element of variable \""<<parentelement.variableName()<<"\"."<< coral::MessageStream::endmsg;
    return false;
  }
  
  if(!m_tableRegister.checkTable( parentelement.tableName())){
    m_log << coral::Error << "Table \"" << parentelement.tableName() << "\" has not been allocated."<< coral::MessageStream::endmsg;
    return false;
  } else {
    pool::ObjectRelationalMappingElement& me = parentelement.appendSubElement( elementType, attributeName, typeName, parentelement.tableName() );

    std::string inputCol = pool::ObjectRelationalDefaultMappingRules::columnNameForVariable( attributeNameForSchema, parentelement.variableNameForSchema() );
    std::string columnName(inputCol);
    unsigned int i=0;
    while(m_tableRegister.checkColumn(parentelement.tableName(),columnName)){
      columnName = pool::ObjectRelationalDefaultMappingRules::newNameForSchemaObject( inputCol, i, pool::ObjectRelationalDefaultMappingRules::MaxColumnNameLength );
      i++;
    }
    m_tableRegister.insertColumn(parentelement.tableName(),columnName);
    me.setColumnNames( std::vector< std::string >( 1, columnName ) );
  }

  return true;  
}

bool 
pool::ObjectRelationalDefaultMappingBuilder::processEmbeddedClass( pool::ObjectRelationalMappingElement& parentelement,
                                                                   const std::string& attributeName,
                                                                   const std::string& attributeNameForSchema,
                                                                   const Reflex::Type& attributeType,
                                                                   bool isRef )
{
  std::string className = attributeType.Name(Reflex::SCOPED);
  std::string elementType = ( isRef ) ? pool::ObjectRelationalMappingElement::poolReferenceMappingElementType() : pool::ObjectRelationalMappingElement::objectMappingElementType();
  pool::ObjectRelationalMappingElement& me = parentelement.appendSubElement( elementType, attributeName, className, parentelement.tableName() );
  me.setVariableNameForSchema(parentelement.variableNameForSchema(),attributeNameForSchema);

  std::vector< std::string > idColumns = parentelement.columnNames();
  if ( parentelement.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
    idColumns.pop_back();
    idColumns.pop_back();
  }
  if ( isRef ) {
    idColumns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForOID_1( attributeName, me.scopeName() ) );
    idColumns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForOID_2( attributeName, me.scopeName() ) );
  }

  me.setColumnNames( idColumns );
  return (isRef) ? true : this->processObject( me, attributeType );
}


bool
pool::ObjectRelationalDefaultMappingBuilder::processObject( pool::ObjectRelationalMappingElement& mappingElement,
                                                            const Reflex::Type& objType )
{
  std::string className = objType.Name(Reflex::SCOPED);
  std::string tableName = mappingElement.tableName();
  
  // resolve possible typedef chains
  Reflex::Type objectType = ObjectRelationalClassUtils::resolvedType(objType);
  // loop over the data members (including the ones from the base classes)
  objectType.UpdateMembers();
  std::vector<std::pair<std::pair<std::string,std::string>,Reflex::Type> > carraysMembers;
  for ( size_t i=0; i< objectType.DataMemberSize(); i++){

    Reflex::Member objectMember = objectType.DataMemberAt(i);
    // Skip the transient ones
    if ( objectMember.IsTransient() ) continue;

    // Retrieve the field type
    Reflex::Type type = ObjectRelationalClassUtils::resolvedType( objectMember.TypeOf() );
    
    // check if the member is from a class in the inheritance tree
    std::string scope("");
    Reflex::Type declaringType = ObjectRelationalClassUtils::resolvedType( objectMember.DeclaringType());
    if( declaringType != objectType ){
      scope = declaringType.Name(Reflex::SCOPED);
    }
    // Retrieve the field name
    std::string objectMemberName = pool::ObjectRelationalDefaultMappingRules::scopedVariableName( objectMember.Name(), scope );
    std::string objectNameForSchema = pool::ObjectRelationalDefaultMappingRules::scopedVariableForSchemaObjects( objectMember.Name(), scope );
    
    // Check for the existence of the dictionary information
    if ( !type ){
      m_log << coral::Error << "Field \"" <<objectMemberName << "\" in class \""<<className <<"\" is unknown type"<< coral::MessageStream::endmsg;
      return false;
    }
    
    if ( type.IsArray() ){
      carraysMembers.push_back( std::make_pair( std::make_pair(objectMemberName,objectNameForSchema), type ));
    } else {
      if ( ! this->processItem( mappingElement, objectMemberName, objectNameForSchema, type ) ) return false;      
    }  
  }
  
  // C-arrays are processed separately and at the end.
  for( size_t i=0; i< carraysMembers.size(); i++){
    if ( ! this->processItem( mappingElement, carraysMembers[i].first.first, carraysMembers[i].first.second, carraysMembers[i].second ) ) return false; 
  }
  
  return true;
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processItem( pool::ObjectRelationalMappingElement& parentElement,
                                                          const std::string& attributeName,
                                                          const std::string& attributeNameForSchema,
                                                          const Reflex::Type& attributeType )
{

  Reflex::Type resType = ObjectRelationalClassUtils::resolvedType( attributeType );
  // primitive and string
  if ( resType.IsFundamental() || resType.IsEnum() || resType.TypeInfo() == typeid(std::string) ) {
    // Correct the string type name
    return this->processPrimitive( parentElement,
                                   attributeName,
                                   attributeNameForSchema,
                                   resType );
  }
  else if ( resType.IsArray() ){
    
    if ( m_arraysInBlobs ){
      return this->processBlob( parentElement,
                                attributeName,
                                attributeNameForSchema,
                                resType );
    }
    if( parentElement.isDependentTree() ) {
      m_log << coral::Error << "Containers or C-Arrays on dependent trees are not supported" << coral::MessageStream::endmsg;      
      return false;
    }
    return this->processCArray( parentElement,
                                attributeName,
                                attributeNameForSchema,
                                resType);
  }
  else if ( pool::ObjectRelationalClassUtils::isTypeContainer( resType ) ) {

    if ( m_arraysInBlobs ){
      return this->processBlob( parentElement,
                                attributeName,
                                attributeNameForSchema,
                                resType );
    }
    if( parentElement.isDependentTree() ) {
      m_log << coral::Error << "Containers or C-Arrays on dependent trees are not supported" << coral::MessageStream::endmsg;      
      return false;
    }
    return this->processArray( parentElement,
                               attributeName,
                               attributeNameForSchema,
                               resType);
  }
  else if ( resType.IsPointer() || resType.IsReference() ){
    m_log << coral::Error << "Pointers or references are not supported" << coral::MessageStream::endmsg;
    return false;
  }
  else if ( pool::ObjectRelationalClassUtils::isTypePoolPointer( resType )){
    if( parentElement.isDependentTree() ) {
      m_log << coral::Error << "Pool pointers on dependent trees are not supported" << coral::MessageStream::endmsg;      
      return false;
    }
    return this->processPoolPtr( parentElement,
                                 attributeName,
                                 attributeNameForSchema,
                                 resType );
  }
  else if ( pool::ObjectRelationalClassUtils::isTypePoolPolymorphicPointer( resType )){
    if( parentElement.isDependentTree() ) {
      m_log << coral::Error << "Pool pointers on dependent trees are not supported" << coral::MessageStream::endmsg;      
      return false;
    }
    return this->processPoolPolyPtr( parentElement,
                                     attributeName,
                                     attributeNameForSchema,
                                     resType );
  }
  else { // embeddedobject
    bool isRef = ( resType.TypeInfo() == typeid(pool::Reference) ||
                   resType.HasBase( Reflex::Type::ByTypeInfo( typeid(pool::Reference) ) ) );
    return this->processEmbeddedClass( parentElement,
                                       attributeName,
                                       attributeNameForSchema,
                                       attributeType,
                                       isRef );
  }
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processArray( pool::ObjectRelationalMappingElement& parentelement,
                                                           const std::string& attributeName,
                                                           const std::string& attributeNameForSchema,
                                                           const Reflex::Type& attributeType ){
  std::ostringstream fullNameForSchema;
  fullNameForSchema << parentelement.variableNameForSchema() << "_" << attributeNameForSchema;
  std::string initialTable = pool::ObjectRelationalDefaultMappingRules::tableNameForVariable( fullNameForSchema.str(),parentelement.mainTableName());
  std::string arrayTable(initialTable);
  unsigned int i=0;
  while(m_tableRegister.checkTable(arrayTable)){
    arrayTable = pool::ObjectRelationalDefaultMappingRules::newNameForSchemaObject( initialTable, i, pool::ObjectRelationalDefaultMappingRules::MaxTableNameLength );
    i++;
  }
  m_tableRegister.insertTable(arrayTable);

  std::string className = attributeType.Name(Reflex::SCOPED);

  std::string elementType = pool::ObjectRelationalMappingElement::arrayMappingElementType();
  if(pool::ObjectRelationalClassUtils::isTypePVector(attributeType)){
    elementType = pool::ObjectRelationalMappingElement::poolArrayMappingElementType();
  }
  pool::ObjectRelationalMappingElement& me = parentelement.appendSubElement( elementType,
                                                                             attributeName,
                                                                             className,
                                                                             arrayTable );
  std::vector<std::string> parentColumns = parentelement.columnNames();
  if ( parentelement.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
    parentColumns.pop_back();
    parentColumns.pop_back();
  }
  std::vector<std::string> columns;
  columns.reserve( parentColumns.size() + 1 );
  // first comes the position
  columns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForPosition() );
  for ( std::vector<std::string>::const_iterator iColumn = parentColumns.begin();
        iColumn != parentColumns.end(); ++iColumn ) {
    columns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForId() + "_" + *iColumn );
  }
  me.setColumnNames( columns );
  me.setVariableNameForSchema(parentelement.variableNameForSchema(),attributeNameForSchema);
  
  bool singleItemContainer =  pool::ObjectRelationalClassUtils::isTypeNonAssociativeContainer(attributeType);
  bool associativeContainer =  pool::ObjectRelationalClassUtils::isTypeAssociativeContainer(attributeType);

  Reflex::Type contentType;
  Reflex::Type keyType;
  
  if( singleItemContainer ){
    contentType = ObjectRelationalClassUtils::containerValueType(attributeType);
  }
  else if ( associativeContainer ) { // This is an associative container type
    contentType = ObjectRelationalClassUtils::containerDataType( attributeType );
    keyType = ObjectRelationalClassUtils::containerKeyType( attributeType );
    if( !keyType || !ObjectRelationalClassUtils::resolvedType(keyType) ){
      m_log << coral::Error << "cannot decide the type of the key item of the container \"" << attributeType.Name(Reflex::SCOPED) << "\"." << coral::MessageStream::endmsg;
      return false;
    }
  }
  else { // Not supported container
    m_log << coral::Error << "Unsupported container type=\""<<attributeType.Name(Reflex::SCOPED) <<"\"."<< coral::MessageStream::endmsg;
    return false;
  }

  if( !contentType || !ObjectRelationalClassUtils::resolvedType(contentType) ){
    m_log << coral::Error << "cannot decide the type of the content item of the container \"" << attributeType.Name(Reflex::SCOPED) << "\"." << coral::MessageStream::endmsg;
    return false;
  }
  if ( keyType ) {
    std::string keyTypeName = keyType.Name();
    std::string keyTypeNameForSchema = ObjectRelationalDefaultMappingRules::variableNameForContainerKey();
    if ( !( this->processItem( me, keyTypeName, keyTypeNameForSchema, keyType ) ) )
      return false;
  }
  std::string contentTypeName = contentType.Name();
  std::string contentTypeNameForSchema = ObjectRelationalDefaultMappingRules::variableNameForContainerValue();
  return this->processItem( me, contentTypeName, contentTypeNameForSchema, contentType );
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processCArray( ObjectRelationalMappingElement& parentelement,
                                                            const std::string& attributeName,
                                                            const std::string& attributeNameForSchema,
                                                            const Reflex::Type& attributeType ){
  
  Reflex::Type arrayElementType = attributeType.ToType();
  if( !arrayElementType || !ObjectRelationalClassUtils::resolvedType( arrayElementType ) ){
    m_log << coral::Error << "cannot decide the type of the content of the array \"" << attributeType.Name(Reflex::SCOPED) << "\"." << coral::MessageStream::endmsg;
    return false;
  }

  if(!m_tableRegister.checkTable(parentelement.tableName())){
    m_log << coral::Error << "Table \"" <<parentelement.tableName() << "\" has not been allocated."<< coral::MessageStream::endmsg;
    return false;
  }
  size_t arraySizeInColumns = sizeInColumns( attributeType, true );
  m_log << coral::Debug << "Attribute \""<< parentelement.variableType()<<"::"<<attributeName<<"\" of type \""<<attributeType.Name(Reflex::SCOPED)<<"\" requires "<<arraySizeInColumns<<" columns."<<coral::MessageStream::endmsg;
  //if(! pool::ObjectRelationalClassUtils::isTypeContainer( arrayElementType ) ) {
  if( arraySizeInColumns < MaxColumnsPerTable ) {
      //size_t arraySizeInColumns = sizeInColumns( attributeType, true );
    size_t columnsInTable = m_tableRegister.numberOfColumns(parentelement.tableName()) + arraySizeInColumns;
    if( columnsInTable < MaxColumnsPerTable ){
        m_log << coral::Debug << "Attribute \""<< parentelement.variableType()<<"::"<<attributeName<<"\" is mapped with Inline columns."<<coral::MessageStream::endmsg;
        return processInlineCArrayItem(parentelement,attributeName,attributeNameForSchema,attributeType, arrayElementType);    
    }
  }
  m_log << coral::Debug << "Attribute \""<< parentelement.variableType()<<"::"<<attributeName<<"\" is mapped in C-Arrays tables."<<coral::MessageStream::endmsg;
  return processCArrayItem(parentelement,attributeName,attributeNameForSchema,attributeType,arrayElementType );
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processCArrayItem( ObjectRelationalMappingElement& parentelement,
                                                                const std::string& attributeName,
                                                                const std::string& attributeNameForSchema,
                                                                const Reflex::Type& attributeType,
                                                                const Reflex::Type& arrayElementType ){
  std::ostringstream fullNameForSchema;
  fullNameForSchema << parentelement.variableNameForSchema() << "_" << attributeNameForSchema;
  std::string initialTable = pool::ObjectRelationalDefaultMappingRules::tableNameForVariable( fullNameForSchema.str(),parentelement.mainTableName());
  std::string arrayTable(initialTable);
  unsigned int i=0;
  while(m_tableRegister.checkTable(arrayTable)){
    arrayTable = pool::ObjectRelationalDefaultMappingRules::newNameForSchemaObject( initialTable, i, pool::ObjectRelationalDefaultMappingRules::MaxTableNameLength );
    i++;
  }
  m_tableRegister.insertTable(arrayTable);
  
  std::string className = attributeType.Name(Reflex::SCOPED);
  pool::ObjectRelationalMappingElement& me = parentelement.appendSubElement( pool::ObjectRelationalMappingElement::CArrayMappingElementType(),
                                                                             attributeName, className, arrayTable );
  std::vector<std::string> parentColumns = parentelement.columnNames();
  if ( parentelement.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
    parentColumns.pop_back();
    parentColumns.pop_back();
  }
  std::vector<std::string> columns;
  columns.reserve( parentColumns.size() + 1 );
  columns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForPosition() );
  for ( std::vector<std::string>::const_iterator iColumn = parentColumns.begin();
        iColumn != parentColumns.end(); ++iColumn ) {
    columns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForId() + "_" + *iColumn );
  }
  me.setColumnNames( columns );
  me.setVariableNameForSchema(parentelement.variableNameForSchema(),attributeNameForSchema);

  std::string contentTypeName = arrayElementType.Name();
  std::string variableNameForSchema = pool::ObjectRelationalDefaultMappingRules::variableNameForArrayColumn( attributeType );
  return this->processItem( me, contentTypeName, variableNameForSchema, arrayElementType );  
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processInlineCArrayItem( ObjectRelationalMappingElement& parentelement,
                                                                      const std::string& attributeName,
                                                                      const std::string& attributeNameForSchema,
                                                                      const Reflex::Type& attributeType,
                                                                      const Reflex::Type& arrayElementType ){
  std::string className = attributeType.Name(Reflex::SCOPED);
  std::string mappingElementType = pool::ObjectRelationalMappingElement::inlineCArrayMappingElementType();
  pool::ObjectRelationalMappingElement& me = parentelement.appendSubElement( mappingElementType, attributeName, className, parentelement.tableName() );
  std::vector< std::string > idColumns = parentelement.columnNames();
  if ( parentelement.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
    idColumns.pop_back();
    idColumns.pop_back();
  }
  me.setColumnNames( idColumns );
  me.setVariableNameForSchema(parentelement.variableNameForSchema(),attributeNameForSchema);
  
  for(size_t i=0;i<attributeType.ArrayLength();i++){
    if ( ! this->processItem( me,
                              ObjectRelationalDefaultMappingRules::variableNameForArrayIndex(attributeName,i),
                              ObjectRelationalDefaultMappingRules::variableNameForArrayColumn(i),
                              arrayElementType ) ) return false;
  }
  return true;  
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processPoolPtr( ObjectRelationalMappingElement& parentelement,
                                                             const std::string& attributeName,
                                                             const std::string& attributeNameForSchema,
                                                             const Reflex::Type& attributeType ){
  std::ostringstream fullNameForSchema;
  fullNameForSchema << parentelement.variableNameForSchema() << "_" << attributeNameForSchema;
  std::string initialTable = pool::ObjectRelationalDefaultMappingRules::tableNameForVariable( fullNameForSchema.str(),parentelement.mainTableName());
  std::string objPtrTable(initialTable);
  unsigned int i=0;
  while(m_tableRegister.checkTable(objPtrTable)){
    objPtrTable = pool::ObjectRelationalDefaultMappingRules::newNameForSchemaObject( initialTable, i, pool::ObjectRelationalDefaultMappingRules::MaxTableNameLength );
    i++;
  }
  m_tableRegister.insertTable(objPtrTable);
  
  std::string className = attributeType.Name(Reflex::SCOPED);

  pool::ObjectRelationalMappingElement& me = parentelement.appendSubElement(
    pool::ObjectRelationalMappingElement::poolPointerMappingElementType(),attributeName, className, objPtrTable );
  std::vector<std::string> parentColumns = parentelement.columnNames();
  if ( parentelement.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
    parentColumns.pop_back();
    parentColumns.pop_back();
  }
  std::vector<std::string> columns;
  columns.reserve( parentColumns.size() );
  for ( std::vector<std::string>::const_iterator iColumn = parentColumns.begin();
        iColumn != parentColumns.end(); ++iColumn ) {
    columns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForId() + "_" + *iColumn );
  }
  me.setColumnNames( columns );
  me.setVariableNameForSchema(parentelement.variableNameForSchema(),attributeNameForSchema);
  
  Reflex::Type ptrType = attributeType.TemplateArgumentAt(0);
  if( !ptrType ){
      m_log << coral::Error << "cannot resolve the type of the pool pointer \"" << attributeType.Name(Reflex::SCOPED) << "\"." << coral::MessageStream::endmsg;;
      return false;    
  }
  
  std::string ptrTypeName = ptrType.Name();
  return this->processItem( me, ptrTypeName, ptrTypeName, ptrType );  
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processPoolPolyPtr( ObjectRelationalMappingElement& parentelement,
                                                                 const std::string& attributeName,
                                                                 const std::string& attributeNameForSchema,
                                                                 const Reflex::Type& attributeType ){
  
  if(parentelement.find( attributeName )!=parentelement.end()){
    m_log << coral::Error << "Attribute name \"" << attributeName << "\" is already defined in the mapping element of variable \""<<parentelement.variableName()<<"\"."<< coral::MessageStream::endmsg;
    return false;
  }
  
  if(!m_tableRegister.checkTable( parentelement.tableName())){
    m_log << coral::Error << "Table \"" << parentelement.tableName() << "\" has not been allocated."<< coral::MessageStream::endmsg;
    return false;
  } else {
    std::string typeName = attributeType.Name(Reflex::SCOPED);
    pool::ObjectRelationalMappingElement& me = parentelement.appendSubElement( pool::ObjectRelationalMappingElement::poolPolymorphicPointerMappingElementType(), attributeName, typeName, parentelement.tableName() );

    std::string inputCol = pool::ObjectRelationalDefaultMappingRules::columnNameForVariable( attributeNameForSchema, parentelement.variableNameForSchema(), false );
    std::string columnName(inputCol);
    unsigned int i=0;
    while(m_tableRegister.checkColumn(parentelement.tableName(),columnName)){
      columnName = pool::ObjectRelationalDefaultMappingRules::newNameForSchemaObject( inputCol, i, pool::ObjectRelationalDefaultMappingRules::MaxColumnNameLength );
      i++;
    }
    m_tableRegister.insertColumn(parentelement.tableName(),columnName);
    me.setColumnNames( std::vector< std::string >( 1, columnName ) );
  }
  return true;
}

pool::ObjectRelationalTableRegister&
pool::ObjectRelationalDefaultMappingBuilder::tableRegister(){
  return m_tableRegister;
}

size_t
pool::ObjectRelationalDefaultMappingBuilder::sizeInColumns(const Reflex::Type& topLevelClassType,
                                                           bool inlinedCArraysOnly){
  size_t sz=0;
  _sizeInColumns(topLevelClassType, sz, inlinedCArraysOnly);
  return sz;
}

void
pool::ObjectRelationalDefaultMappingBuilder::_sizeInColumns(const Reflex::Type& topLevelClassType,
                                                            size_t& sz, bool inlinedCArraysOnly){
  // resolve possible typedef chains
  Reflex::Type typ = ObjectRelationalClassUtils::resolvedType( topLevelClassType );
  bool isPoolPolyPointer = pool::ObjectRelationalClassUtils::isTypePoolPolymorphicPointer(typ);

  // primitive and string
  if ( typ.IsFundamental() || typ.TypeInfo() == typeid(std::string) || typ.IsEnum() || isPoolPolyPointer ) {
    ++sz;
  } else if (typ.IsArray()){
    size_t arraySize = typ.ArrayLength();
    Reflex::Type arrayType = typ.ToType();
    size_t arrayElementSize = 0;
    _sizeInColumns(arrayType, arrayElementSize, true );
    size_t totSize = arraySize*arrayElementSize;
    size_t currSize = sz + totSize;
    if(inlinedCArraysOnly || (currSize < MaxColumnsPerTable)) sz += totSize;
  } else if (typ.TypeInfo() == typeid(pool::Reference) ||
             typ.HasBase( Reflex::Type::ByTypeInfo( typeid(pool::Reference) ) )){
    sz += 2;
  } else {
  
    bool isContainer =  pool::ObjectRelationalClassUtils::isTypeNonAssociativeContainer(typ) ||
      pool::ObjectRelationalClassUtils::isTypeAssociativeContainer(typ);
    bool isPoolPointer = pool::ObjectRelationalClassUtils::isTypePoolPointer(typ);
    if( !isContainer && !isPoolPointer ){
      
      // loop over the data members
      typ.UpdateMembers();
      std::vector<Reflex::Type> carrays;
      for ( size_t i=0; i< typ.DataMemberSize(); i++){
        Reflex::Member objMember = typ.DataMemberAt(i);

        // Skip the transient ones
        if ( objMember.IsTransient() ) continue;

        // Retrieve the field type
        Reflex::Type objMemberType = objMember.TypeOf();

        if(objMemberType.IsArray()){
          carrays.push_back( objMemberType );
        } else {
          _sizeInColumns(objMemberType,sz, inlinedCArraysOnly );
        }
      }
      for( size_t i=0; i< carrays.size(); i++){
        _sizeInColumns( carrays[i], sz, inlinedCArraysOnly );
      }
      
    }
  }
}
