//$Id: IPersistencySvc.h,v 1.25 2008/05/13 14:39:26 witoldp Exp $

#ifndef PERSISTENCYSVC_IPERSISTENCYSVC
#define PERSISTENCYSVC_IPERSISTENCYSVC

// includes
#include <utility>
#include "Reflex/Type.h"

#include "RVersion.h"

/**
    pool namespace
*/
namespace pool {

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
typedef ROOT::Reflex::Type TypeH;
#else
typedef Reflex::Type TypeH;
#endif


  // Forward declarations
  class IFileCatalog;
  class ITokenValidator;
  class IDataTransform;
  class IConfiguration;
  class Placement;
  class Token;
  class DataHandler;
  class ISession;
  class Guid;
  class ITransactionObserver;

  /** @class IPersistencySvc IPersistencySvc.h PersistencySvc/IPersistencySvc.h
   * 
   *  IPersistencySvc is the abstract interface for all services which execute
   *  the conversion between transient and persistent objects.
   *
   */
  class IPersistencySvc { 
  public:
    /// Constructor
    explicit IPersistencySvc( IFileCatalog& ) {}

    /// Empty destructor
    virtual ~IPersistencySvc() {}

    /// Returns the file catalog in use
    virtual IFileCatalog& fileCatalog() = 0;

    /// Set the attached file catalog
    virtual void setFileCatalog( IFileCatalog& catalog ) = 0;

    /// Returns the token validator in use
    virtual ITokenValidator& tokenValidator() = 0;

    /// Set the attached object token validator
    virtual void setTokenValidator( ITokenValidator& validator ) = 0;

    /// Returns a reference to the IDataTransform object.
    virtual IDataTransform& dataTransform() = 0;

    /// Sets the attached object data transform
    virtual void setDataTransform( IDataTransform& dataTransform ) = 0;

    /// Returns the persistency configuration
    virtual IConfiguration& configuration() = 0;

    /** Retrieves an object from persistent store and return with type information
     *  The handle to the reflection class is necessary to later delete the object.
     *
     * @param  transientType [IN]  reference to the type of the corresponding transient class
     * @param  token         [IN]  reference to the token for the object
     * @param  handler       [IN]  optional data handler e.g. to interprete references
     *
     * @return std::pair< void*, TypeH > The data and the class description.
     *
     * In case of failure zeros are returned.
     *
     */
    virtual std::pair< void*, TypeH > readObject( const pool::Guid& transientType,
                                                               const Token& token,
                                                               DataHandler* handler = 0 ) = 0;

    virtual std::pair< void*, TypeH > readObject( const TypeH& transientType,
                                                               const Token& token,
                                                               DataHandler* handler = 0 ) = 0;

    /** Retrieves an object from persistent store and return with type information
     *  The handle to the reflection class is necessary to later delete the object.
     *  The Guid of the transient class is assumed to be the classID of the token
     *
     * @param  token   [IN]  reference to the token for the object
     * @param  handler [IN]  optional data handler e.g. to interprete references
     *
     * @return std::pair< void*, const TypeH > The data and the class description.
     *
     * In case of failure zeros are returned.
     *
     */
    virtual std::pair< void*, TypeH > readObject( const Token& token,
                                                                DataHandler* handler = 0 ) = 0;

    /**  registerForWrite registers an object for writing to the persistent medium
     *   The optional handler is only necessary if during the load phase 
     *   higher level interactions with the framework are necessary.
     *
     * @param  place        [IN]  the placement hint
     * @param  object       [IN]  pointer to transient object which will be written
     * @param  type         [IN]  reflection class description with the layout of transient object
     * @param  handler      [IN]  optional data handler e.g. to interprete references
     *
     * @return Token*   the token address of the persistent object. I case of failure 0 is returned.
     *
     */
    virtual Token* registerForWrite( const Placement& place,
                                     const void* object,
                                     const TypeH& type,
                                     DataHandler* handler = 0 ) = 0;
    
    /**  updateObject updates an object which is already written in the persistent medium.
     *   higher level interactions with the framework are necessary.
     *
     * @param  object       [IN]  pointer to transient object which will be updates
     * @param  type         [IN]  reflection class description with the layout of transient object
     * @param  token        [IN]  reference to the token for the object
     * @param  handler      [IN]  optional data handler e.g. to interprete references
     *
     * @return bool   Indicates whether operation was successfull.
     *
     */
    virtual bool updateObject( const void* object,
                               const TypeH& type,
                               const Token& token,
                               DataHandler* handler = 0 ) = 0;

    /** Deletes an object from the persistent store
     *
     * @param  token        [IN]  reference to the token for the object
     *
     * @return bool   Indicates whether operation was successfull.
     *
     */
    virtual bool destroyObject( const Token& token ) = 0;

    /// Returns the underlying global session
    virtual ISession& session() = 0;

    /// Registers a transaction observer
    virtual bool registerObserver( ITransactionObserver& observer ) = 0;

    /// Deregisters a transaction observer
    virtual bool deregisterObserver( ITransactionObserver& observer ) = 0;

  protected:
    /// No copy constructor, and no assignment operator
    IPersistencySvc() {}
    IPersistencySvc( const IPersistencySvc& );
    IPersistencySvc& operator=( const IPersistencySvc& );

  }; // class IPersistencySvc

} // namespace pool 

#endif // PERSISTENCYSVC_IPERSISTENCYSVC
