#ifndef RELATIONALFILECATALOG_RELATIONALFILECATALOGIMPLEMENTATION_H
#define RELATIONALFILECATALOG_RELATIONALFILECATALOGIMPLEMENTATION_H 1

#include "FileCatalog/FCImpl.h"
#include <map>

namespace coral {
  class ISessionProxy;
  class AttributeList;
}

namespace pool {

  namespace RelationalFileCatalog {

    class QueryHolder;

    /**
     * @class RelationalFileCatalogImplementation RelationalFileCatalogImplementation.h
     *
     * Implementation of the FCImpl interface for a relational database
     */

    class RelationalFileCatalogImplementation : public pool::FCImpl
    {
    public:
      /// Constructor
      RelationalFileCatalogImplementation();

      /// Destructor
      virtual ~RelationalFileCatalogImplementation();

      /// Methods inherited from FCImpl

      /// Registers a new PFN, returns the fild id.
      virtual void registerPFN( const std::string& pfn,
                                const std::string& filetype,
                                pool::FileCatalog::FileID& fid ) const;

      /// Registers an LFN for a given PFN
      virtual void registerLFN( const std::string& pfn,
                                const std::string& lfn ) const;

      /// Adds a new replica PFN for a given file identified by its PFN.
      virtual void addReplicaPFN( const std::string& pf,
                                  const std::string& rpf ) const;

      /** Adds a PFN to a fileID. If fileID exists, adds a replica PFN; if fileID doesnot exist, adds the pair.
       * For performance reasons, the implementation does NOT check the existance of the fileID explicitly.
       * An exception from the backend library is used whenever possible.
       */
      virtual void addPFNtoGuid( const pool::FileCatalog::FileID& guid,
                                 const std::string& pf,
                                 const std::string& filetype ) const;

      /// Renames a PFN
      virtual void renamePFN( const std::string& pf,
                              const std::string& newpf ) const;

      /// Looks up a file given its PFN
      virtual void lookupFileByPFN( const std::string& pf,
                                    pool::FileCatalog::FileID& fid,
                                    std::string& ftype ) const;

      /// Looks up a file fiven its LFN
      virtual void lookupFileByLFN( const std::string& lf,
                                    pool::FileCatalog::FileID& fid) const;

      /// Looks up the best PFN
      virtual void lookupBestPFN( const pool::FileCatalog::FileID& fid,
                                  const FileCatalog::FileOpenMode& omode, 
                                  const FileCatalog::FileAccessPattern& amode,
                                  std::string& pf,
                                  std::string& filetype ) const;

      /// Inserts a PFN entry
      virtual void insertPFN( pool::PFNEntry& pentry) const;

      /// Inserts an LFN entry
      virtual void insertLFN( pool::LFNEntry& lentry) const;

      /// Deletes a PFN entry
      virtual void deletePFN( const std::string& pf ) const;

      /// Deletes an LFN entry
      virtual void deleteLFN( const std::string& lf ) const;

      /// Deletes an entry from the file catalog
      virtual void deleteEntry(const pool::FileCatalog::FileID& guid) const;

      /// Checks whether the catalog is read-only
      virtual bool isReadOnly() const;

      /// Queries the PFNs
      virtual bool retrievePFN( const std::string& query,
                                pool::FCBuf<pool::PFNEntry>& buf,
                                const size_t& start );

      /// Queries the LFNs
      virtual bool retrieveLFN( const std::string& query,
                                pool::FCBuf<pool::LFNEntry>& buf,
                                const size_t& start );

      /// Queries the GUID
      virtual bool retrieveGuid( const std::string& query,
                                 pool::FCBuf<pool::FileCatalog::FileID>& buf,
                                 const size_t& start );

      /// Connects to the database
      virtual void connect();

      /// Disconnects from the database
      virtual void disconnect() const;

      /// Starts a new transaction
      virtual void start() const;

      /// Commits a transaction
      virtual void commit( pool::FileCatalog::CommitMode cm = pool::FileCatalog::REFRESH ) const;

      /// Rolls back a transaction
      virtual void rollback() const;

      /// THE METADATA PART TO BE USED BY THE SPLIT FILE CATALOG VERSION
        
      /// Inserts a meta data entry
      void insertMetaData( pool::MetaDataEntry& mentry) const;

      /// Registers meta data information for a given FID
      void registerMetaData( const pool::FileCatalog::FileID& fid,
                             pool::MetaDataEntry& meta ) const;

      /// Creates a meta data specification
      void createMetaDataSpec( pool::MetaDataEntry& schemadef) const;

      /// Updates the meta data specification
      void updateMetaDataSpec( pool::MetaDataEntry& schemadef,
                               const pool::FileCatalog::MetaUpdateMode metamode=pool::FileCatalog::NO_DELETE ) const;

      /// Retrieves the meta data specification
      void getMetaDataSpec( pool::MetaDataEntry& schemadef) const;

      /// Drops the meta data specification
      void dropMetaDataSpec() const;

      /// Deletes the meta data information for a given guid
      void deleteMetaData(const pool::FileCatalog::FileID& guid) const;

      /// Queries the meda data
      bool retrieveMetaData( const std::string& query,
                             pool::FCBuf<pool::MetaDataEntry>& buf,
                             const size_t& start );
      
    private:
      /// Loads the meta data specification from the database
      void loadMetaDataSpecification() const;

    private:
      mutable coral::ISessionProxy*              m_session;
      mutable std::map< std::string, int >       m_fieldNameToIndex;
      QueryHolder*                               m_queryHolder;
      mutable coral::AttributeList*              m_metaData;
      mutable coral::AttributeList*              m_rowData;
    };

  }

}

#endif
