#include "ExistingDataScanner.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ICursor.h"
#include "RelationalAccess/ITable.h"
#include "RelationalStorageServiceNames.h"
#include <sstream>   

pool::RelationalStorageService::ExistingDataScanner::ExistingDataScanner( coral::ISchema& schema ):
                                                                          m_schema(schema),
                                                                          m_query(0),
									                                                        m_resultSet(0),
                                                                          m_mappingVersion(""),
                                                                          m_className(""),
                                                                          m_tableName(""),
                                                                          m_maxChunkSize(10000),
                                                                          m_chunkSize(),
                                                                          m_entries(),
                                                                          m_writtenObjects(0),
                                                                          m_deletedObjects(0),
                                                                          m_totalRows(0),
                                                                          m_tableMap(){}

pool::RelationalStorageService::ExistingDataScanner::~ExistingDataScanner(){
  if(m_query){
    delete m_query;
    m_query = 0;
    m_resultSet = 0;
  }
}

void pool::RelationalStorageService::ExistingDataScanner::setChunkSize(unsigned int size){
  m_chunkSize=size;
}

// loads in the instance memory a chunk of rows retrieved by a partial table scan
void pool::RelationalStorageService::ExistingDataScanner::prepareScan( const ObjectRelationalMappingElement& mappingElement,std::string version ){

  m_mappingVersion = version;
  m_className = mappingElement.variableType();
  m_tableName = mappingElement.tableName();
  m_totalRows = 0;
  
  std::string idColumn = mappingElement.columnNames()[0];
  
  m_query = m_schema.tableHandle(m_tableName).newQuery();
  m_query->addToOutputList(idColumn,"ID"); // for the moment only 'single' primary key is supported
  m_query->defineOutputType("ID",coral::AttributeSpecification::typeNameForType<int>());
  m_query->addToOrderList(idColumn); 
  m_query->setRowCacheSize( m_maxChunkSize );
  m_resultSet = &m_query->execute();
}

bool pool::RelationalStorageService::ExistingDataScanner::scan() {
  if(!m_query) return false;
  unsigned long numRow = 0;
  unsigned long maxId = 0;
  bool result = false;
  bool incomplete = false;
  while( m_resultSet->next() ) {
    result = true;
    incomplete = false;
    const coral::AttributeList& row = m_resultSet->currentRow();
    int id = row["ID"].data< int >();
    Entry entry;
    entry.className = m_className;
    entry.mappingVersion = m_mappingVersion;
    entry.targetOid = id;
    m_entries.push_back(entry);
    numRow++;
    m_chunkSize++;
    maxId = (unsigned long)id;
    if(m_chunkSize==m_maxChunkSize){
      incomplete = true;
      break;
    } 
  }

  if(numRow){
    unsigned long prevMaxId = 0;
    std::map<std::string,std::pair<unsigned long,int> >::const_iterator iTable = m_tableMap.find(m_tableName);
    if(iTable != m_tableMap.end()){
      prevMaxId = iTable->second.second;
      m_tableMap.erase(m_tableName);
    } else {
      // the lastOidMap does not contain the table
      m_writtenObjects += 1;
      m_deletedObjects += 1;
    }
    m_totalRows += numRow;
    m_tableMap.insert(std::make_pair(m_tableName,std::make_pair(m_totalRows,maxId)));
    m_writtenObjects += (maxId-prevMaxId);
    m_deletedObjects += (maxId-numRow-prevMaxId);
    m_totalRows += numRow;
  }

  if(!incomplete){
    delete m_query;
    m_query = 0;
    m_resultSet = 0;
  }

  return result;
  
}

bool pool::RelationalStorageService::ExistingDataScanner::scanAll() {
  if(!m_query) return false;
  bool result = false;
  unsigned long numRow = 0;
  unsigned long maxId = 0;
  while( m_resultSet->next() ) {
    result = true;
    const coral::AttributeList& row = m_resultSet->currentRow();
    int id = row["ID"].data< int >();
    maxId = (unsigned long)id;
    numRow++;
    m_chunkSize++;
  }
  if(numRow){
    m_writtenObjects += (maxId+1);
    m_deletedObjects += (maxId+1-numRow);
    m_totalRows = numRow;
    std::map<std::string,std::pair<unsigned long,int> >::const_iterator iTable = m_tableMap.find(m_tableName);
    if(iTable != m_tableMap.end()){
      m_tableMap.erase(m_tableName);
    } 
    m_tableMap.insert(std::make_pair(m_tableName,std::make_pair(m_totalRows,maxId)));
  }
  delete m_query;
  m_query = 0;
  m_resultSet = 0;

  return result;  
}

// loads in the instance memory a chunk of rows retrieved by a partial table scan
void pool::RelationalStorageService::ExistingDataScanner::prepareScanForNewRows( const std::string& containerTable, const ObjectRelationalMappingElement& mappingElement,std::string version ){

  m_mappingVersion = version;
  m_className = mappingElement.variableType();
  m_tableName = mappingElement.tableName();
  m_totalRows = 0;
  
  std::string idColumn = mappingElement.columnNames()[0];
  m_query = m_schema.tableHandle(m_tableName).newQuery();
  m_query->addToOutputList(idColumn,"ID"); // for the moment only 'single' primary key is supported
  m_query->defineOutputType("ID",coral::AttributeSpecification::typeNameForType<int>());
  m_query->addToOrderList(idColumn); 

  std::ostringstream cond;
  cond << mappingElement.columnNames()[0]
         << " NOT IN (SELECT " << pool::RelationalStorageService::RelationalStorageServiceNames::targetOIDVariableNameInPolymorphicContainerTable() 
         << " FROM " << containerTable
         << " WHERE " << pool::RelationalStorageService::RelationalStorageServiceNames::classNameVariableNameInPolymorphicContainerTable()
     << "=:" << pool::RelationalStorageService::RelationalStorageServiceNames::classNameVariableNameInPolymorphicContainerTable()
     << ")";
  coral::AttributeList classData;
  classData.extend<std::string>(pool::RelationalStorageService::RelationalStorageServiceNames::classNameVariableNameInPolymorphicContainerTable());
  classData.begin()->data<std::string>() = m_className ;
  m_query->setCondition(cond.str(),classData);   

  m_query->setRowCacheSize( m_maxChunkSize );
  m_resultSet = &m_query->execute();
}

const std::vector<pool::RelationalStorageService::ExistingDataScanner::Entry>& pool::RelationalStorageService::ExistingDataScanner::entries() const{
  return m_entries;
}

unsigned long pool::RelationalStorageService::ExistingDataScanner::writtenObjects() const {
  return m_writtenObjects;
}
      
unsigned long pool::RelationalStorageService::ExistingDataScanner::deletedObjects() const {
  return m_deletedObjects;
}

unsigned long pool::RelationalStorageService::ExistingDataScanner::storedObjects() const {
  unsigned long storedObj = 0;
  for(std::map<std::string,std::pair<unsigned long,int> >::const_iterator iTable = m_tableMap.begin();
       iTable!=m_tableMap.end();iTable++){
    storedObj += iTable->second.first;
  }
  return storedObj;
}

unsigned long pool::RelationalStorageService::ExistingDataScanner::storedObjects(const std::string& tableName) const {
  unsigned long storedObj = 0;
  std::map<std::string,std::pair<unsigned long,int> >::const_iterator iTable = m_tableMap.find(tableName);
  if(iTable != m_tableMap.end()){
    storedObj = iTable->second.first;
  }
  return storedObj;
}

const std::map<std::string,std::pair<unsigned long,int> >& pool::RelationalStorageService::ExistingDataScanner::tableMap() const {
  return m_tableMap;
}

bool pool::RelationalStorageService::ExistingDataScanner::isChunkFull() const {
  return m_chunkSize == m_maxChunkSize;
}

void pool::RelationalStorageService::ExistingDataScanner::resetChunk() {
  m_chunkSize = 0;
  m_entries.clear();
}

void pool::RelationalStorageService::ExistingDataScanner::reset() {
  m_entries.clear();
  m_tableMap.clear();
  m_writtenObjects = 0;
  m_deletedObjects = 0;
  m_totalRows = 0;
  m_chunkSize = 0;
  m_mappingVersion = "";
  m_className = "";
  m_tableName = "";
  if(m_query){
    delete m_query;
    m_query = 0;
    m_resultSet = 0;
  }
}

unsigned int pool::RelationalStorageService::ExistingDataScanner::chunkSize() const {
  return m_chunkSize;
}
