#include "MappingManager.h"
#include "RelationalStorageServiceException.h"
#include "RelationalStorageServiceNames.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingPersistency.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingBuilder.h"
#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingRules.h"
#include "ObjectRelationalAccess/ObjectRelationalTableNameRationalizer.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "CoralBase/MessageStream.h"
#include "Reflex/Type.h"
#include <memory>

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
using namespace ROOT;
#endif

pool::RelationalStorageService::MappingManager::MappingManager( const std::string& containerName, coral::ISchema& schema ):
  m_schema( schema ),
  m_containerName( containerName ),
  m_persistency( new pool::ObjectRelationalMappingPersistency( schema ) ),
  m_mappingVersionsForClass(),
  m_mappings()
{}


pool::RelationalStorageService::MappingManager::~MappingManager()
{
  this->clear();
  delete m_persistency;
}


void
pool::RelationalStorageService::MappingManager::clear()
{
  m_mappingVersionsForClass.clear();
  for ( std::map< std::string, std::map< std::string, pool::ObjectRelationalMapping* > >::iterator iMappingsForClass = m_mappings.begin();
        iMappingsForClass != m_mappings.end(); ++iMappingsForClass ) {
    for ( std::map< std::string, pool::ObjectRelationalMapping* >::iterator iMappingForVersion = iMappingsForClass->second.begin();
          iMappingForVersion != iMappingsForClass->second.end(); ++iMappingForVersion ) {
      delete iMappingForVersion->second;
    }
  }
  m_mappings.clear();
}

const pool::ObjectRelationalMapping&
pool::RelationalStorageService::MappingManager::mappingForClass( const Reflex::Type& dictionaryEntry, bool writingEnabled )
{
  // The class parameters
  std::string className = dictionaryEntry.Name(Reflex::SCOPED);
  std::string classVersion = versionOfClass( dictionaryEntry );

  if(writingEnabled) m_persistency->initializeForWriting();

  std::map< std::string, std::map< std::string, ObjectRelationalMapping* > >::iterator iMappingsForClass = m_mappings.find( className );
  if(iMappingsForClass != m_mappings.end()){
    std::map< std::string, pool::ObjectRelationalMapping* >::const_iterator iM = iMappingsForClass->second.find( classVersion );
    if(iM != iMappingsForClass->second.end()) return *iM->second;
  } else {
    iMappingsForClass = m_mappings.insert( std::make_pair( className, std::map< std::string, ObjectRelationalMapping* >() ) ).first;
  }
  std::map< std::string, ObjectRelationalMapping* >& mappingsForClass = iMappingsForClass->second;
  
  std::string mappingVersion("");
  std::map<std::string,std::string>::const_iterator iMv =  m_mappingVersionsForClass.find( classVersion );
  if( iMv != m_mappingVersionsForClass.end() ){
    mappingVersion = iMv->second;
  } else {
    mappingVersion = m_persistency->selectMappingVersion( classVersion, m_containerName );
    m_mappingVersionsForClass.insert(std::make_pair( classVersion, mappingVersion));
  }
  
  if(!mappingVersion.empty()){
    pool::ObjectRelationalMapping* mappingStructure = m_persistency->retrieveMapping( mappingVersion );
    if ( mappingStructure && (mappingStructure->className() == className) ) {
      // OK, the mapping is valid
      mappingsForClass.insert( std::make_pair( classVersion, mappingStructure ) );
      return *mappingStructure;
    } else {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "No entry for class \"" + className +
                                                                               "\" in mapping version \"" + mappingVersion + "\"",
                                                                               "MappingManager::mappingForClass" );
    }
  } 

  if(!writingEnabled){
    throw pool::RelationalStorageService::RelationalStorageServiceException( "No mapping version available for container \"" + m_containerName +
                                                                             "\" and class version \""+classVersion+"\".",
                                                                             "MappingManager::mappingForClass" );
  }
  
  // Create a new mapping using the default rules
  std::string newMappingVersion = m_persistency->newMappingVersionForClass( className );  
  pool::ObjectRelationalMapping* newMapping = new pool::ObjectRelationalMapping( newMappingVersion );
  pool::ObjectRelationalDefaultMappingBuilder builder;
  if ( ! builder.processClass( *newMapping, dictionaryEntry ) ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not generate a mapping for class \"" + className + "\"",
                                                                             "MappingManager::mappingForClass" );
  }

  // Rationalize the table and variable names.
  pool::ObjectRelationalTableNameRationalizer rationalizer( m_schema );
  rationalizer.rationalizeMappingElement( newMapping->element() );
  
  // Store the new mapping.
  if ( !m_persistency->storeMapping( *newMapping, classVersion, m_containerName ) ){
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not store the new mapping for class \"" + className + "\"",
                                                                             "MappingManager::mappingForClass" );
  }
  // Materialize the new mapping.
  if ( !m_persistency->materializeMapping( *newMapping, false ) ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not materialize the new mapping for class \"" + className + "\"",
                                                                             "MappingManager::mappingForClass" );
  }
  mappingsForClass.insert( std::make_pair( classVersion ,newMapping ));
  return *newMapping;
}

const pool::ObjectRelationalMappingElement& 
pool::RelationalStorageService::MappingManager::mappingForDependingClass( const Reflex::Type& classDict,const Reflex::Type& parentDict, bool writingEnabled ) 
{
  // The class parameters
  std::string className = classDict.Name(Reflex::SCOPED);
  std::string parentClassName = parentDict.Name(Reflex::SCOPED);
  std::string parentClassVersion = versionOfClass( parentDict );

  pool::ObjectRelationalMapping& mapping = const_cast<pool::ObjectRelationalMapping&>(mappingForClass( parentDict, writingEnabled ));
  std::map< std::string, pool::ObjectRelationalMappingElement >::iterator iDep = mapping.dependentElements().find( className );
  if(iDep != mapping.dependentElements().end()){
    return iDep->second;
  }
  
  if(!writingEnabled){
    throw pool::RelationalStorageService::RelationalStorageServiceException( "No mapping version available for container \"" + m_containerName +
                                                                             "\" and dependent class \""+className+"\".",
                                                                             "MappingManager::mappingForDependingClass" );
  }

  pool::ObjectRelationalDefaultMappingBuilder builder;
  if ( ! builder.processDependentClass( mapping, classDict ) ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not generate a mapping for class \"" + className + "\"",
                                                                             "MappingManager::mappingForDependingClass" );
  }

  iDep = mapping.dependentElements().find( className );
  if(iDep == mapping.dependentElements().end()){
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not obtain a mapping for class \"" + className + "\"",
                                                                             "MappingManager::mappingForDependingClass" );
  }
  // Rationalize the table and variable names.
  pool::ObjectRelationalTableNameRationalizer rationalizer( m_schema );
  rationalizer.rationalizeMappingElement( iDep->second );
  
  // Store the new mapping.
  size_t newId = mapping.numberOfDependentElements()+1;
  std::stringstream newElementId;
  newElementId << newId;
  if ( !m_persistency->storeMappingElement( mapping.version(), newElementId.str(), iDep->second ) ){
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not store the new mapping for class \"" + className + "\"",
                                                                             "MappingManager::mappingForDependingClass" );
  }
  // Materialize the new mapping.
  if ( !m_persistency->materializeMappingElement( iDep->second, false ) ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not materialize the new mapping for class \"" + className + "\"",
                                                                             "MappingManager::mappingForDependingClass" );
  }
  return iDep->second;
}

std::string
pool::RelationalStorageService::MappingManager::versionOfClass( const Reflex::Type& dictionary ){
  std::string className = dictionary.Name(Reflex::SCOPED);
  Reflex::PropertyList classProps = dictionary.Properties();
  std::string classVersion = pool::ObjectRelationalDefaultMappingRules::defaultClassVersion(className);
  if(classProps.HasProperty(pool::ObjectRelationalDefaultMappingRules::classVersionPropertyNameInDictionary())){
    classVersion = classProps.PropertyAsString(pool::ObjectRelationalDefaultMappingRules::classVersionPropertyNameInDictionary());
  }
  return classVersion;
}

