#include "ObjectRelationalDataBuffer.h"
#include "ContainerHandlerFactory.h"
#include "IContainerHandler.h"
#include "RelationalStorageServiceException.h"
#include "POOLCore/IBlobStreamingService.h"
#include "POOLCore/Context.h"

#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "ObjectRelationalAccess/ObjectRelationalClassUtils.h"
#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingRules.h"

#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"

#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/Blob.h"

#include "Reflex/Member.h"
#include "RVersion.h"

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
using namespace ROOT;
#endif

bool pool::RelationalStorageService::ObjectRelationalDataBuffer::existAttribute(const std::string& attributeName, const coral::AttributeList& data){
  bool found = false;
  for(coral::AttributeList::const_iterator iAttr = data.begin();
      iAttr!=data.end() && !found; ++iAttr){
    if( iAttr->specification().name() == attributeName ) found = true;
  }
  return found;
}

pool::RelationalStorageService::ObjectRelationalDataBuffer::ObjectRelationalDataBuffer( pool::Context* context):
  m_context(context),
  m_mapping( 0 ),
  m_offsets(),
  m_metadata(),
  m_data( 0 ),
  m_table( 0 ),
  m_dictionary(),
  m_offsetToParent( 0 ),
  m_keyItems( 0 ),
  m_keyData( false ),
  m_containerHandler( 0 ),
  m_arrayBuffers(),
  m_blobWriters(),
  m_poolPointerBuffers(),
  m_poolPolymorphicPointerBuffers(),
  m_updateClause(""),
  m_idClause("")
{}



pool::RelationalStorageService::ObjectRelationalDataBuffer::~ObjectRelationalDataBuffer()
{
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalDataBuffer* >::iterator iBuffer = m_arrayBuffers.begin();
        iBuffer != m_arrayBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_arrayBuffers.clear();
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalDataBuffer* >::iterator iBuffer = m_poolPointerBuffers.begin();
        iBuffer != m_poolPointerBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_poolPointerBuffers.clear();
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalDataBuffer* >::iterator iBuffer = m_poolPolymorphicPointerBuffers.begin();
        iBuffer != m_poolPolymorphicPointerBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_poolPolymorphicPointerBuffers.clear();
  if ( m_containerHandler ) delete m_containerHandler;
  if ( m_data ) delete m_data;
  for( std::map< int,pool::IBlobWriter* >::iterator iBlobWriter=m_blobWriters.begin();
       iBlobWriter!=m_blobWriters.end();
       iBlobWriter++)
  {
    delete iBlobWriter->second;
  } 
}

bool
pool::RelationalStorageService::ObjectRelationalDataBuffer::build( coral::ISchema& schema,
                                                                   const pool::ObjectRelationalMappingElement& mapping,
                                                                   const TypeH& dict )
{
  m_table = &(schema.tableHandle( mapping.tableName() ) );
  m_mapping = &mapping;
  // resolve possible typedef chains
  TypeH dictionary = ObjectRelationalClassUtils::resolvedType(dict);
  m_dictionary = dictionary;
  m_data = new coral::AttributeList;
  this->buildBuffer( mapping, dictionary, schema, 0, false );
  // Build the actual buffer
  return true;
}



int
pool::RelationalStorageService::ObjectRelationalDataBuffer::buildBuffer( const pool::ObjectRelationalMappingElement& mapping,
                                                                         const TypeH& dictionary,
                                                                         coral::ISchema& schema,
                                                                         int offset,
                                                                         bool keyData )
{
  int retVal = 0;
  if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Primitive ) { // Primitives
    if ( !( dictionary.IsFundamental() || dictionary.TypeInfo() == typeid(std::string) || dictionary.IsEnum()) ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               "\" is not declared as primitive in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildBuffer" );
    }
    if(!existAttribute( mapping.columnNames()[0], *m_data )){
      const std::type_info* attrType = &dictionary.TypeInfo();
      if(dictionary.IsEnum()) attrType = &typeid(int);
      m_data->extend( mapping.columnNames()[0], *attrType );
    }
    m_offsets.push_back( std::make_pair(offset,-1) );
    ++retVal;
    appendToUpdateClause( mapping.columnNames()[0] );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Blob ) { // Blobs
    if(!existAttribute( mapping.columnNames()[0], *m_data )){
      m_data->extend<coral::Blob>( mapping.columnNames()[0] );
    }
    m_offsets.push_back( std::make_pair(offset,-1) );
    ++retVal;
    appendToUpdateClause( mapping.columnNames()[0] );
    // book also the corresponding Blob Writer
    pool::IBlobWriter* blobWriter = 0;
    std::map< int,pool::IBlobWriter* >::const_iterator iBlobWriter=m_blobWriters.find(offset);
    if( iBlobWriter!=m_blobWriters.end() )
    {
      blobWriter = iBlobWriter->second;
    }
    if(!blobWriter)
    {
      pool::IBlobStreamingService* blobStreamingService = m_context->blobStreamer();
      if ( !blobStreamingService ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Cannot stream to blob variable \""+mapping.variableName()+"\": No blob streaming Service registered.","ObjectRelationalDataBuffer::buildBuffer" );
      }
      // fix me : version is currently not used...
      blobWriter = blobStreamingService->newWriter(TypeH::ByName(mapping.variableType()),"");
      m_blobWriters.insert(std::make_pair(offset,blobWriter));
    }
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Object ||
            mapping.elementType() == pool::ObjectRelationalMappingElement::DependentObject ||
            mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) { // Objects
    // Get the id columns first
    std::vector< std::string > columnNames = mapping.columnNames();
    if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
      columnNames.pop_back();
      columnNames.pop_back();
    }
    bool setIdClause = false;
    if(m_idClause.empty()) setIdClause = true;
    unsigned int iCol=0;
    for ( std::vector< std::string >::const_iterator iColumn = columnNames.begin();
          iColumn != columnNames.end(); ++iColumn, ++iCol ) {
      if(!existAttribute( *iColumn, *m_data )){
        const std::type_info* colType = &typeid(int);
        if(mapping.elementType() == pool::ObjectRelationalMappingElement::DependentObject && iCol==0) colType = &typeid(std::string);
        m_data->extend( *iColumn, *colType );
      }
      if(setIdClause) {
        if(!m_idClause.empty()) m_idClause.append(" AND ");
        m_idClause.append( *iColumn+" = :"+*iColumn );
      }
    }

    if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
      
      // Book the OID columns
      size_t index = columnNames.size();
      int stubOffset = -( 1 + offset );
      m_data->extend< int >( mapping.columnNames()[index] );
      m_offsets.push_back( std::make_pair(stubOffset,-1) );
      appendToUpdateClause(mapping.columnNames()[index]);
      ++index;
      m_data->extend< int >( mapping.columnNames()[index] );
      m_offsets.push_back( std::make_pair(stubOffset,-1) );
      appendToUpdateClause(mapping.columnNames()[index]);
      
    } else {

      // update dictionary to include base classes members
      dictionary.UpdateMembers();
      // Loop over the fields of the class.
      for ( unsigned int i=0;i<dictionary.DataMemberSize();i++){

        Reflex::Member field = dictionary.DataMemberAt(i);
      
        // Ignore the transients
        if ( field.IsTransient() ) continue;

        // Get the member type and resolve possible typedef chains
        TypeH field_type = ObjectRelationalClassUtils::resolvedType( field.TypeOf() );
        if ( ! field_type ) {
          throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for field \"" +
                                                                                   field.Name() + "\" of class \"" +
                                                                                   dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                                   "ObjectRelationalDataBuffer::buildBuffer" );
        }
      
        // check if the member is from a class in the inheritance tree
        std::string scope("");
        TypeH declaringType = ObjectRelationalClassUtils::resolvedType( field.DeclaringType());
        if( declaringType != dictionary ){
          scope = declaringType.Name(Reflex::SCOPED|Reflex::FINAL);
        }
        // Get the field name
        std::string fieldName = pool::ObjectRelationalDefaultMappingRules::scopedVariableName( field.Name(), scope );

        // Retrieve the relevant mapping element
        pool::ObjectRelationalMappingElement::const_iterator iFieldElement = mapping.find( fieldName );
        if ( iFieldElement == mapping.end() ) {
          throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + fieldName + "\" not found in the mapping element",
                                                                                   "ObjectRelationalDataBuffer::buildBuffer" );
        }
        const pool::ObjectRelationalMappingElement& fieldElement = iFieldElement->second;

        retVal += this->buildBuffer( fieldElement, field_type, schema, offset + field.Offset(), keyData );
      }
    }
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::InlineCArray ) { // InlineCArrays
    // Get the id columns first
    std::vector< std::string > columnNames = mapping.columnNames();
    for ( std::vector< std::string >::const_iterator iColumn = columnNames.begin();
          iColumn != columnNames.end(); ++iColumn ) {
      if(!existAttribute( *iColumn, *m_data )){
        m_data->extend< int >( *iColumn );
      }
      appendToUpdateClause(*iColumn);
    }

      // Get the member type and resolve possible typedef chains
    TypeH elementType = ObjectRelationalClassUtils::resolvedType( dictionary.ToType() );
    if ( ! elementType ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException(
        "Missing dictionary information for the element of array \"" +dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"","ObjectRelationalDataBuffer::buildBuffer" );
    }
    // Loop over the elements of the array.
    for ( unsigned int i=0;i<dictionary.ArrayLength();i++){

      // Form the element name
      std::string arrayElementLabel = ObjectRelationalDefaultMappingRules::variableNameForArrayIndex( mapping.variableName(),i);

      // Retrieve the relevant mapping element
      pool::ObjectRelationalMappingElement::const_iterator iElement = mapping.find( arrayElementLabel );
      if ( iElement == mapping.end() ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Array Element \"" + arrayElementLabel + "\" not found in the mapping element",
                                                                                 "ObjectRelationalDataBuffer::buildBuffer" );
      }
      const pool::ObjectRelationalMappingElement& element = iElement->second;

      retVal += this->buildBuffer( element, elementType, schema, offset + i*elementType.SizeOf(), keyData );
    }
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Array ) { // Containers
    if ( ! ObjectRelationalClassUtils::isTypeContainer( dictionary ) ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as Array in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildBuffer" );
    }
    this->buildArrayBuffer( mapping, dictionary, schema, offset, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolArray) { // PoolArrays
    if ( ! ObjectRelationalClassUtils::isTypePVector( dictionary ) ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as PoolArray in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildBuffer" );
    }
    this->buildArrayBuffer( mapping, dictionary, schema, offset, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::CArray ) { // C-Arrays
    if ( ! dictionary.IsArray() ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as a C-Array in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildBuffer" );
    }
    this->buildCArrayBuffer( mapping, dictionary, schema, offset, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolPointer ) { // Pool Pointers
    if ( ! ObjectRelationalClassUtils::isTypePoolPointer( dictionary )) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as a Pool Pointer in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildBuffer" );
    }
    this->buildPoolPointerBuffer( mapping, dictionary, schema, offset, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolPolymorphicPointer ) { // Pool Poly Pointers
    if ( ! ObjectRelationalClassUtils::isTypePoolPolymorphicPointer( dictionary )) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as a Pool PolymorphicPointer in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildBuffer" );
    }
    if(!existAttribute( mapping.columnNames()[0], *m_data )){
      m_data->extend( mapping.columnNames()[0], typeid(std::string) );
    }
    m_offsets.push_back( std::make_pair(offset,m_metadata.size()) );
    m_metadata.push_back( mapping.variableType() );
    ++retVal;
    appendToUpdateClause( mapping.columnNames()[0] );
    this->buildPoolPolymorphicPointerBuffer( *m_mapping, mapping, dictionary, offset );
  }
  else { // Do not treat other types yet...
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Cannot treat anything else than objects, primitives, C-arrays and containers for the moment...",
                                                                             "ObjectRelationalDataBuffer::buildBuffer" );
  }

  return retVal;
}




bool
pool::RelationalStorageService::ObjectRelationalDataBuffer::buildArrayBuffer( const pool::ObjectRelationalMappingElement& mapping,
                                                                              const TypeH& dictionary,
                                                                              coral::ISchema& schema,
                                                                              int offset,
                                                                              bool keyData )
{
  pool::RelationalStorageService::ObjectRelationalDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalDataBuffer(m_context);
  m_arrayBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_offsetToParent = offset;
  newBuffer->m_table = &(schema.tableHandle( mapping.tableName() ) );
  newBuffer->m_containerHandler = pool::RelationalStorageService::ContainerHandlerFactory::newContainerHandler( dictionary );
  if ( ! newBuffer->m_containerHandler ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not create a container handler",
                                                                             "ObjectRelationalDataBuffer::buildArrayBuffer" );
  }

  newBuffer->m_keyData = keyData;
  // Construct the buffer
  newBuffer->m_data = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  for ( std::vector< std::string >::const_iterator iColumn = columnNames.begin();
        iColumn != columnNames.end(); ++iColumn ) {
    newBuffer->m_data->extend< int >( *iColumn );
  }

  TypeH field_type;
  // Get the type of the container...
  if ( ObjectRelationalClassUtils::isTypeAssociativeContainer(dictionary)){
    
    TypeH key_type = ObjectRelationalClassUtils::containerKeyType(dictionary);

    TypeH keyResolvedType = ObjectRelationalClassUtils::resolvedType(key_type);
    if ( ! key_type || !keyResolvedType ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the key type of the container \"" +
                                                                               dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                               "ObjectRelationalDataBuffer::buildArrayBuffer" );
    }
    std::string keyName = key_type.Name();
    // Retrieve the relevant mapping element
    pool::ObjectRelationalMappingElement::const_iterator iFieldElement = mapping.find( keyName );
    if ( iFieldElement == mapping.end() ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + keyName + "\" not found in the mapping element",
                                                                               "ObjectRelationalDataBuffer::buildArrayBuffer" );
    }
    newBuffer->m_keyItems = newBuffer->buildBuffer( iFieldElement->second, keyResolvedType, schema, 0, true );
    field_type = ObjectRelationalClassUtils::containerDataType(dictionary);
  } else {
    field_type = ObjectRelationalClassUtils::containerValueType(dictionary);
  }

  TypeH fieldResolvedType = ObjectRelationalClassUtils::resolvedType(field_type);
  // Check the component type
  if ( ! field_type || !fieldResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the content type of the container \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalDataBuffer::buildArrayBuffer" );
  }
  
  std::string fieldName = field_type.Name();
  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iFieldElement = mapping.find( fieldName );
  if ( iFieldElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + fieldName + "\" not found in the mapping element",
                                                                             "ObjectRelationalDataBuffer::buildArrayBuffer" );
  }
  newBuffer->buildBuffer( iFieldElement->second, fieldResolvedType, schema, 0, false );
  
  return true;
}

bool
pool::RelationalStorageService::ObjectRelationalDataBuffer::buildCArrayBuffer( const pool::ObjectRelationalMappingElement& mapping,
                                                                              const TypeH& dictionary,
                                                                              coral::ISchema& schema,
                                                                              int offset,
                                                                              bool keyData )
{
  pool::RelationalStorageService::ObjectRelationalDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalDataBuffer(m_context);
  m_arrayBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_offsetToParent = offset;
  newBuffer->m_table = &(schema.tableHandle( mapping.tableName() ) );
  newBuffer->m_containerHandler = pool::RelationalStorageService::ContainerHandlerFactory::newContainerHandler( dictionary );
  if ( ! newBuffer->m_containerHandler ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not create a container handler",
                                                                             "ObjectRelationalDataBuffer::buildCArrayBuffer" );
  }

  newBuffer->m_keyData = keyData;
  // Construct the buffer
  newBuffer->m_data = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  for ( std::vector< std::string >::const_iterator iColumn = columnNames.begin();
        iColumn != columnNames.end(); ++iColumn ) {
    newBuffer->m_data->extend< int >( *iColumn );
  }
  
  // Check the array type
  TypeH arrayType = dictionary.ToType();
  TypeH arrayResolvedType = ObjectRelationalClassUtils::resolvedType(arrayType);
  // Check the component type
  if ( ! arrayType || !arrayResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the element type of the array \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalDataBuffer::buildCArrayBuffer" );
  }
  
  std::string arrayTypeName = arrayType.Name();
  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iElement = mapping.find( arrayTypeName );
  if ( iElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + arrayTypeName + "\" not found in the mapping element",
                                                                             "ObjectRelationalDataBuffer::buildCArrayBuffer" );
  }
  newBuffer->buildBuffer( iElement->second, arrayResolvedType, schema, 0, false );
  
  return true;
}

bool
pool::RelationalStorageService::ObjectRelationalDataBuffer::buildPoolPointerBuffer( const pool::ObjectRelationalMappingElement& mapping,
                                                                                    const ROOT::Reflex::Type& dictionary,
                                                                                    coral::ISchema& schema,
                                                                                    int offset,
                                                                                    bool keyData )
{
  pool::RelationalStorageService::ObjectRelationalDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalDataBuffer(m_context);
  m_poolPointerBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_offsetToParent = offset;
  newBuffer->m_table = &(schema.tableHandle( mapping.tableName() ) );
  newBuffer->m_keyData = keyData;
  // Construct the buffer
  newBuffer->m_data = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  for ( std::vector< std::string >::const_iterator iColumn = columnNames.begin();
        iColumn != columnNames.end(); ++iColumn ) {
    newBuffer->m_data->extend< int >( *iColumn );
  }

  // Check the  type
  ROOT::Reflex::Type ptrType = dictionary.TemplateArgumentAt(0);
  ROOT::Reflex::Type ptrResolvedType = ObjectRelationalClassUtils::resolvedType(ptrType);
  // Check the component type
  if ( ! ptrType || !ptrResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the type of the pointer \"" +
                                                                             dictionary.Name(ROOT::Reflex::SCOPED|ROOT::Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalDataBuffer::buildPoolPointerBuffer" );
  }
  
  std::string ptrTypeName = ptrType.Name();
  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iElement = mapping.find( ptrTypeName );
  if ( iElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Item \"" + ptrTypeName + "\" not found in the mapping element",
                                                                             "ObjectRelationalDataBuffer::buildPoolPointerBuffer" );
  }
  newBuffer->buildBuffer( iElement->second, ptrResolvedType, schema, 0, false );
  
  return true;
}

bool
pool::RelationalStorageService::ObjectRelationalDataBuffer::buildPoolPolymorphicPointerBuffer( const pool::ObjectRelationalMappingElement& topLevelMapping,
                                                                                               const pool::ObjectRelationalMappingElement& polyPtrMapping,
                                                                                               const ROOT::Reflex::Type& dictionary,
                                                                                               int offset )
{
  pool::RelationalStorageService::ObjectRelationalDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalDataBuffer(m_context);
  m_poolPolymorphicPointerBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &topLevelMapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_offsetToParent = offset;
  //newBuffer->m_table = &(schema.tableHandle( mapping.tableName() ) );
  //newBuffer->m_keyData = keyData;
  std::string columnKey = ObjectRelationalDefaultMappingRules::referenceColumnKey(polyPtrMapping.tableName(),polyPtrMapping.columnNames()[0]);
  newBuffer->m_metadata.push_back( columnKey );

  // Construct the buffer
  //newBuffer->m_data = new coral::AttributeList;

  // Check the  type
  ROOT::Reflex::Type ptrType = dictionary.TemplateArgumentAt(0);
  ROOT::Reflex::Type ptrResolvedType = ObjectRelationalClassUtils::resolvedType(ptrType);
  // Check the component type
  if ( ! ptrType || !ptrResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the type of the pointer \"" +
                                                                             dictionary.Name(ROOT::Reflex::SCOPED|ROOT::Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalDataBuffer::buildPoolPointerBuffer" );
  }
  return true;
}

void
pool::RelationalStorageService::ObjectRelationalDataBuffer::appendToUpdateClause(const std::string& columnName){
  std::string columnSetClause(columnName+" = :"+columnName);
  if(!m_updateClause.empty()) m_updateClause.append(", ");
  m_updateClause.append(columnSetClause);
}

coral::AttributeList&
pool::RelationalStorageService::ObjectRelationalDataBuffer::data()
{
  return *m_data;
}

const std::vector< std::pair<int,int> >&
pool::RelationalStorageService::ObjectRelationalDataBuffer::offsets() const
{
  return m_offsets;
}

const std::vector< std::string >&
pool::RelationalStorageService::ObjectRelationalDataBuffer::metadata() const {
  return m_metadata;
}

std::vector< std::string >&
pool::RelationalStorageService::ObjectRelationalDataBuffer::metadata() {
  return m_metadata;
}

int
pool::RelationalStorageService::ObjectRelationalDataBuffer::keyItems() const
{
  return m_keyItems;
}

int
pool::RelationalStorageService::ObjectRelationalDataBuffer::offsetToParent() const
{
  return m_offsetToParent;
}

coral::ITable&
pool::RelationalStorageService::ObjectRelationalDataBuffer::table()
{
  return *m_table;
}

const TypeH&
pool::RelationalStorageService::ObjectRelationalDataBuffer::dictionary() const
{
  return m_dictionary;
}

pool::RelationalStorageService::IContainerHandler&
pool::RelationalStorageService::ObjectRelationalDataBuffer::containerHandler()
{
  return *m_containerHandler;
}


bool
pool::RelationalStorageService::ObjectRelationalDataBuffer::isKeyData() const
{
  return m_keyData;
}


pool::RelationalStorageService::ObjectRelationalDataBuffer::iterator
pool::RelationalStorageService::ObjectRelationalDataBuffer::begin()
{
  return m_arrayBuffers.begin();
}

pool::RelationalStorageService::ObjectRelationalDataBuffer::iterator
pool::RelationalStorageService::ObjectRelationalDataBuffer::end()
{
  return m_arrayBuffers.end();
}

size_t
pool::RelationalStorageService::ObjectRelationalDataBuffer::size() const
{
  return m_arrayBuffers.size();
}

const std::vector< pool::RelationalStorageService::ObjectRelationalDataBuffer* >&
pool::RelationalStorageService::ObjectRelationalDataBuffer::poolPointerBuffers() const {
  return m_poolPointerBuffers;
}

const std::vector< pool::RelationalStorageService::ObjectRelationalDataBuffer* >&
pool::RelationalStorageService::ObjectRelationalDataBuffer::poolPolymorphicPointerBuffers() const {
  return m_poolPolymorphicPointerBuffers;
}

const pool::ObjectRelationalMappingElement& pool::RelationalStorageService::ObjectRelationalDataBuffer::mapping() const {
  return *m_mapping;
}

/// Returns a blob version of the specified data, using the blob streamer registered in the pool context
/// The offset is used to lookup the writer in the map
const coral::Blob&
pool::RelationalStorageService::ObjectRelationalDataBuffer::writeBlob(int offset, const void* data)
{
  pool::IBlobWriter* blobWriter = 0;
  
  std::map< int,pool::IBlobWriter* >::const_iterator iBlobWriter=m_blobWriters.find(offset);
  if( iBlobWriter!=m_blobWriters.end() )
  {
    blobWriter = iBlobWriter->second;
  } else 
  {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Cannot stream to blob: The streamer has not been booked.","ObjectRelationalDataBuffer::writeBlob" );
  }
  return blobWriter->write(data);
}

const std::string& 
pool::RelationalStorageService::ObjectRelationalDataBuffer::updateClause() const{
  return m_updateClause;
}

const std::string& 
pool::RelationalStorageService::ObjectRelationalDataBuffer::idClause() const{
  return m_idClause;
}
