#include "ObjectUpdater.h"
#include "ObjectRelationalDataBuffer.h"
#include "IContainerHandler.h"
#include "IContainerIteratorHandler.h"
#include "RelationalStorageServiceException.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/IBulkOperation.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Blob.h"
#include "StorageSvc/DataCallBack.h"
#include "StorageSvc/DbDataHandlerGuard.h"
#include "POOLCore/Reference.h"
#include "POOLCore/Token.h"
#include "Reflex/Member.h"
#include <sstream>

pool::RelationalStorageService::ObjectUpdater::ObjectUpdater( coral::ISchema& schema,
                                                              const Reflex::Type& dictionary,
                                                              MappingManager& mappingManager,
                                                              pool::DbDatabase* databaseHandle ):
  DbIOHandler(databaseHandle),
  ObjectWriter(schema, dictionary, mappingManager, databaseHandle),
  m_schema(schema)
{
}


pool::RelationalStorageService::ObjectUpdater::~ObjectUpdater()
{
}

bool
pool::RelationalStorageService::ObjectUpdater::update( int oid,
                                                       const void* data,
                                                       pool::DataCallBack& callback )
{
  return update(oid,dataBuffer(),data,callback);
}

bool
pool::RelationalStorageService::ObjectUpdater::update( int oid,
                                                       ObjectRelationalDataBuffer& dataBuff,
                                                       const void* data,
                                                       pool::DataCallBack& callback )
{
  coral::AttributeList& buffer = dataBuff.data();

  // Update first the row of the top level table
  coral::AttributeList::iterator iAttribute = buffer.begin();

  // Skip the identity columns
  for ( size_t iColumn = 0; iColumn < dataBuff.mapping().columnNames().size(); ++iColumn ) ++iAttribute;

  // Copy the data
  std::vector<std::pair<int,int> >::const_iterator iOffset = dataBuff.offsets().begin();
  while ( iAttribute != buffer.end() ) {
    int offset = iOffset->first;
    if ( offset < 0 ) {
      pool::Token::OID_t oid;
      const void* refAddress = static_cast< const char* >( data ) - ( 1 + offset );
      // Call here the ref handler to get the oid...
      pool::DbDataHandlerGuard handlerGuard( &callback, m_databaseContext );
      callback.notify( pool::DataCallBack::PUT, pool::DataCallBack::REFERENCE, refDictionary(), const_cast< void* >( refAddress ), &oid );
      iAttribute->data<int>() =  oid.first;
      ++iAttribute;
      ++iOffset;
      iAttribute->data<int>() =  oid.second;
    }
    else {
      if( iAttribute->specification().type() == typeid(coral::Blob) )
      {
        iAttribute->bind<coral::Blob>( dataBuff.writeBlob( offset, static_cast< const char* >( data ) + offset ) );
      } else 
      {
        iAttribute->setValueFromAddress( static_cast< const char* >( data ) + offset );
      }
    }
    ++iAttribute;
    ++iOffset;
  }

  // Set the oid. For the moment set only the first identity column. In principle there should be a plugin taking
  // care of the OID -> identity values mapping.
  buffer.begin()->data< int >() =  oid;

  // Update the rows into the table
  dataBuff.table().dataEditor().updateRows( dataBuff.updateClause(), dataBuff.idClause(), buffer );

  // Update the arrays...
  // now remove the object dependencies...
  deleteObjectDependencies(dataBuff.mapping(),oid, dataBuff.mapping().tableName());
  //  re-writing the arrays now...
  std::vector<int> identity( 1, oid );
  for ( pool::RelationalStorageService::ObjectRelationalDataBuffer::iterator iBuffer = dataBuff.begin();
        iBuffer != dataBuff.end(); ++iBuffer ) {
    
    if ( ! this->updateArray( oid,
                              **iBuffer,
                              static_cast< const char* >( data ) + (*iBuffer)->offsetToParent() )) return false;
    if ( ! this->writeArray( identity, **iBuffer,
                             static_cast< const char* >( data ) + (*iBuffer)->offsetToParent(),
                             callback ) ) return false;
  }
  // Update the pool pointers...
  for ( std::vector<ObjectRelationalDataBuffer*>::const_iterator iBuffer = dataBuff.poolPointerBuffers().begin();
        iBuffer != dataBuff.poolPointerBuffers().end(); ++iBuffer ) {
    if ( ! this->updatePoolPointer( oid, *(*iBuffer),
                                   static_cast< const char* >( data ) + (*iBuffer)->offsetToParent(),
                                   callback ) ) return false;
  }
  return true;
}

bool
pool::RelationalStorageService::ObjectUpdater::updatePoolPointer( int oid,
                                                                  ObjectRelationalDataBuffer& dataBuff,
                                                                  const void* data,
                                                                  pool::DataCallBack& callback )
{
  ROOT::Reflex::Object ptrObj(dataBuff.dictionary(),const_cast<void*>(data));

#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
  bool isNull = *static_cast<bool*>(ptrObj.Invoke("operator!").Address());
#else
  bool isNull;
  ptrObj.Invoke("operator!",isNull);
#endif

  if(!isNull){
#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
    void* ptrData = ptrObj.Invoke("operator*").Address();
#else
    void* ptrData;
    ptrObj.Invoke("operator*",ptrData);
#endif
    return update(oid,dataBuff,ptrData,callback);  
  }
  return true;
}

bool pool::RelationalStorageService::ObjectUpdater::updateArray( int oid,
                                                                 pool::RelationalStorageService::ObjectRelationalDataBuffer& buffer,
                                                                 const void* address ){
  bool ret = true;
  if(buffer.mapping().elementType()==pool::ObjectRelationalMappingElement::PoolArray) {
    pool::RelationalStorageService::IContainerHandler& containerHandler = buffer.containerHandler();
    size_t arraySize = containerHandler.size(address);
    size_t persistentSize = containerHandler.persistentSize(address);
    if(persistentSize>arraySize){
      ret = deleteArrayData( buffer.mapping(),oid, (int)arraySize, buffer.mapping().tableName());
    } 
  } 
  return ret;
}

bool
pool::RelationalStorageService::ObjectUpdater::deleteObjectDependencies( const pool::ObjectRelationalMappingElement& mapping,
                                                                         int oid, const std::string& topLevelTable, bool cascade )
{
  bool execQuery = true;
  if ( mapping.tableName()==topLevelTable || mapping.elementType() == pool::ObjectRelationalMappingElement::Object) execQuery = false;
  // the first delete execute on the top of the hierarchy makes all the depending tables to be also cleaned up
  if( execQuery ) cascade = true;
  
  for ( pool::ObjectRelationalMappingElement::const_iterator iMapping = mapping.begin();
        iMapping != mapping.end(); ++iMapping ) {
    pool::ObjectRelationalMappingElement::ElementType elementType = iMapping->second.elementType();
    
    if ( elementType == pool::ObjectRelationalMappingElement::Object ||
         elementType == pool::ObjectRelationalMappingElement::Array ||
         elementType == pool::ObjectRelationalMappingElement::CArray ||
         ((elementType == pool::ObjectRelationalMappingElement::PoolArray) && cascade) ) {
      this->deleteObjectDependencies( iMapping->second, oid, topLevelTable, cascade );
    }
  }

  if( !execQuery ) return true;
  if ( mapping.tableName()==topLevelTable || mapping.elementType() == pool::ObjectRelationalMappingElement::Object) return true;
  size_t columnId = mapping.columnNames().size() - 1;
  std::string oidColumn = mapping.columnNames()[ columnId ];
  coral::AttributeList condition;
  condition.extend< int >( oidColumn );
  condition.begin()->data< int >() = oid;
  this->m_schema.tableHandle( mapping.tableName() ).dataEditor().deleteRows( oidColumn + " = :" + oidColumn,
                                                                             condition );
  return true;
}

bool
pool::RelationalStorageService::ObjectUpdater::deleteArrayData( const pool::ObjectRelationalMappingElement& mapping,
                                                                int oid, int fromIndex, const std::string& topLevelTable )
{
  for ( pool::ObjectRelationalMappingElement::const_iterator iMapping = mapping.begin();
        iMapping != mapping.end(); ++iMapping ) {
    pool::ObjectRelationalMappingElement::ElementType elementType = iMapping->second.elementType();
    if ( elementType == pool::ObjectRelationalMappingElement::Object ||
         elementType == pool::ObjectRelationalMappingElement::Array ||
         elementType == pool::ObjectRelationalMappingElement::PoolArray ||
         elementType == pool::ObjectRelationalMappingElement::CArray ) {
      this->deleteArrayData( iMapping->second, oid, fromIndex, topLevelTable );
    }
  }

  if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Object) return true;
  size_t columnId = mapping.columnNames().size() - 1;
  size_t indexId = mapping.columnNames().size() - 2;
  std::string oidColumn = mapping.columnNames()[ columnId ];
  std::string indexColumn = mapping.columnNames()[ indexId ];
  std::stringstream condition;
  condition << oidColumn;
  condition << " = :" << oidColumn;
  condition << " AND " << indexColumn;
  condition << " >= :" << indexColumn;
  coral::AttributeList conditionData;
  conditionData.extend< int >( oidColumn );
  conditionData.extend< int >( indexColumn );
  coral::AttributeList::iterator iter = conditionData.begin();
  iter->data< int >() = oid;
  ++iter;
  iter->data< int >() = fromIndex;
  this->m_schema.tableHandle( mapping.tableName() ).dataEditor().deleteRows( condition.str(), conditionData );
  return true;
}


