#include "CoralBase/Exception.h"
#include "CoralBase/MessageStream.h"
#include "ObjectRelationalAccess/ObjectRelationalDatabaseDescriptor.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingDescriptor.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingUtilities.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingSchema.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingPersistency.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalStorageService/RelationalPoolDatabaseUtilities.h"
#include "RelationalStorageServiceException.h"
#include "RelationalStorageServiceNames.h"
#include "POOLCore/Context.h"
#include "POOLCore/Guid.h"
#include "POOLCore/Token.h"
#include "StorageSvc/poolDb.h"
#include "StorageSvc/IStorageSvc.h"
#include "StorageSvc/IStorageExplorer.h"
#include "StorageSvc/FileDescriptor.h"
#include "StorageSvc/DbStatus.h"
#include "StorageSvc/DbReflex.h"
#include "StorageSvc/DbSelect.h"
#include "StorageSvc/DbDomain.h"
#include "StorageSvc/DbSession.h"
#include "StorageSvc/DbDatabase.h"
#include "StorageSvc/DbContainer.h"
#include "StorageSvc/DbTransaction.h"
#include "StorageSvc/DbTypeInfo.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/Transaction.h"
#include "RelationalPoolContainer.h"
#include "RelationalPoolDatabase.h"
#include "PoolDatabaseManager.h"
#include <typeinfo>
#include <vector>
#include <memory>

#include "RVersion.h"
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
 typedef ROOT::Reflex::Type TypeH;
#else
 typedef Reflex::Type TypeH;
#endif

pool::RelationalPoolDatabaseUtilities::RelationalPoolDatabaseUtilities() :  ObjectRelationalUtilities(), 
                                                                            m_connectionString(""){
}

pool::RelationalPoolDatabaseUtilities::~RelationalPoolDatabaseUtilities(){
}

void pool::RelationalPoolDatabaseUtilities::setUpDatabase(const std::string& fileName,
                                                          const std::string& outFile,
                                                          bool evolve,
                                                          bool dry ){
  init();

  // retrieve the setup directives from the xml file
  ObjectRelationalMappingUtilities mappingUtilities;
  pool::ObjectRelationalDatabaseDescriptor* ptr = mappingUtilities.loadDescriptor(fileName);
  if(!ptr){
    throw pool::RelationalStorageService::RelationalStorageServiceException("Database descriptor not build.","RelationalPoolDatabaseUtilities::setUpDatabase");
  }
  std::auto_ptr<pool::ObjectRelationalDatabaseDescriptor> dbstruct(ptr);

  mappingUtilities.setMessageVerbosityLevel(messageVerbosityLevel());
  try{
    mappingUtilities.connect(m_connectionString,coral::Update);
    mappingUtilities.session()->transaction().start();
    mappingUtilities.buildAndMaterializeMapping( *dbstruct, outFile, evolve, dry);
    mappingUtilities.session()->transaction().commit();
  } catch (const coral::Exception& e){
    try {
      mappingUtilities.session()->transaction().rollback();
    } catch (const coral::Exception& /* e */ ){
    }
    throw RelationalStorageService::RelationalStorageServiceException(e.what(),"RelationalPoolDatabaseUtilities::setUpDatabase");    
  }

  if(!dry && !evolve){
    setUpDatabase( *dbstruct );
  }  
  
}

pool::RelationalPoolDatabaseUtilities::DatabaseInfo pool::RelationalPoolDatabaseUtilities::queryDatabase(){
  init();
  DatabaseInfo dbInfo;
  pool::IStorageSvc* storageSvc = pool::createStorageSvc( 0, "StorageSvc");
  if ( ! storageSvc ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not create a StorageSvc object",
                                         "RelationalStorageService::RelationalPoolDatabaseUtilities::queryDatabase" );
  }
  storageSvc->addRef();
  void* pVoid = 0;
  pool::DbStatus sc = storageSvc->queryInterface( pool::IStorageExplorer::interfaceID(),&pVoid );
  pool::IStorageExplorer* storageExplorer = (pool::IStorageExplorer*)pVoid;
  if ( !( sc.isSuccess() && storageExplorer ) ) {
    storageSvc->release();
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not retrieve a IStorageExplorer interface",
                                         "RelationalStorageService::RelationalPoolDatabaseUtilities::queryDatabase" );
  }  
  pool::Session* sessionHandle = 0;
  if ( ! ( storageSvc->startSession( pool::READ, pool::POOL_RDBMS_StorageType.type(), sessionHandle ).isSuccess() ) ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not start a session.",
                                         "RelationalStorageService::RelationalPoolDatabaseUtilities::queryDatabase" );
  }

  pool::FileDescriptor* fileDesc = new pool::FileDescriptor( m_connectionString, m_connectionString );
  if ( ! ( storageSvc->connect( sessionHandle, pool::READ, *fileDesc ).isSuccess() ) ) {
    storageSvc->release();
    return dbInfo; 
  }
  dbInfo.connectionString = m_connectionString;
  pool::DatabaseConnection* connection = fileDesc->dbc();
  pool::Transaction* transaction = 0;
  if ( ! ( storageSvc->startTransaction( connection, transaction ).isSuccess() ) ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not start a transaction.",
                                          "RelationalStorageService::RelationalPoolDatabaseUtilities::queryDatabase" );
  }

  std::map<std::string,ContainerInfo> containers;
  std::vector<const pool::Token*> containerTokens;
  storageExplorer->containers( *fileDesc, containerTokens );
  if(containerTokens.size()>0){
    for ( std::vector<const pool::Token*>::const_iterator iToken = containerTokens.begin();
                                               iToken != containerTokens.end(); ++iToken ) {
      pool::DbType tech((*iToken)->technology());
      // no selection on the tokens...
      pool::DbSelect* selection = new pool::DbSelect( "" );
      ContainerInfo cont;
      cont.name = (*iToken)->contID();
      cont.technology = pool::DbType((*iToken)->technology()).storageName();
      storageExplorer->select( *fileDesc, cont.name, *selection );
      pool::Token* objectToken = 0;
      unsigned int iObj = 0;
      while ( storageExplorer->next( *selection, objectToken ).isSuccess() ) {
         iObj++;
	       TypeH theType = pool::DbReflex::forGuid(objectToken->classID());
         if(theType){
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
	         std::string className = theType.Name(ROOT::Reflex::SCOPED);
#else
	         std::string className = theType.Name(Reflex::SCOPED);
#endif
           if(cont.containerClassName=="") cont.containerClassName = className;
	         std::map<std::string,unsigned int>::iterator iTable = cont.classes.find(className);
           if(iTable == cont.classes.end()){
             unsigned int st = 1;
             cont.classes.insert(std::make_pair(className,st));
           } else {
             iTable->second++;
           }
         }
         const_cast<Token*>(objectToken)->release();
      }
      cont.size = iObj;
      containers.insert(std::make_pair(cont.name,cont));      
      delete selection;  
    }
  } 

  if ( ! ( storageSvc->endTransaction( transaction, pool::Transaction::TRANSACT_COMMIT ).isSuccess() ) ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not end a transaction.",
                                          "RelationalStorageService::RelationalPoolDatabaseUtilities::queryDatabase" );
  }

  if ( ! ( storageSvc->disconnect( *fileDesc ).isSuccess() ) ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not disconnect.",
                                                                             "RelationalStorageService::RelationalPoolDatabaseUtilities::queryDatabase" );
  }
  
  storageSvc->release();

  // now add the remaining information
  // open a storage transaction
  std::auto_ptr<pool::Context> ctx(new pool::Context);
  DbSession sess;
  sess.open( ctx.get() );
  DbDomain dom(POOL_RDBMS_StorageType);
  // open database
  DbDatabase database(POOL_RDBMS_StorageType);
  dom.open(sess,POOL_RDBMS_StorageType,pool::READ);
  pool::Guid guid;
  pool::Guid::create( guid );
  if(!database.open(dom,m_connectionString,guid.toString(),pool::READ).isSuccess()){
    throw RelationalStorageService::RelationalStorageServiceException("Failure while opening database.","RelationalPoolDatabaseUtilities::queryDatabase");
  }
  DbTransaction trans;
  trans.set(Transaction::TRANSACT_START);
  database.transAct(trans);

  // open (->create) the specified container
  for(std::map<std::string,ContainerInfo>::iterator iC=containers.begin();
      iC!=containers.end();++iC){
    DbType tech = DbType::getType(iC->second.technology);
    DbTypeInfo* dbTypeInfo = DbTypeInfo::create(iC->second.containerClassName);
    DbContainer container(POOL_RDBMS_StorageType);
    if(!container.open(database,iC->second.name,dbTypeInfo,tech,pool::READ).isSuccess()){
      throw RelationalStorageService::RelationalStorageServiceException("Failure while opening container.","RelationalPoolDatabaseUtilities::queryDatabase");
    }
    // get the pure Relational hook for the forced update of the container
    RelationalStorageService::RelationalPoolContainer* contHandle = dynamic_cast<RelationalStorageService::RelationalPoolContainer*>(container.info());
    iC->second.mappingVersion=contHandle->mappingVersion();
    iC->second.className=contHandle->className();
    // close container
    container.close();
  }

  // close the transaction
  trans.set(Transaction::TRANSACT_COMMIT);
  database.transAct(trans);
  // close database, domain and session
  database.close();
  dom.close();
  sess.close();
  
  dbInfo.containers = containers;
  return dbInfo;
}

void pool::RelationalPoolDatabaseUtilities::updateDatabase(bool dry){
  coral::MessageStream log( pool::RelationalStorageService::serviceName() );
  DatabaseInfo dbInfo = queryDatabase();
  if(dbInfo.connectionString==""){
    throw pool::RelationalStorageService::RelationalStorageServiceException( "POOL Database not found on "+m_connectionString, "RelationalStorageService::RelationalPoolDatabaseUtilities::updateDatabase" );
  }
  if(dbInfo.containers.size()){

    std::auto_ptr<pool::Context> ctx(new pool::Context);
    DbSession sess;
    sess.open( ctx.get() );
    DbDomain dom(POOL_RDBMS_StorageType);
    // open database
    DbDatabase database(POOL_RDBMS_StorageType);
    dom.open(sess,POOL_RDBMS_StorageType,pool::UPDATE);
    pool::Guid guid;
    pool::Guid::create( guid );
    if(!database.open(dom,m_connectionString,guid.toString(),pool::UPDATE).isSuccess()){
      throw RelationalStorageService::RelationalStorageServiceException("Failure while opening database.","RelationalPoolDatabaseUtilities::updateDatabase");
    }
    log << "Opening POOL database: "<<m_connectionString<<coral::MessageStream::endmsg;
    std::map<std::string,ContainerInfo>::const_iterator iCont;
    for(iCont = dbInfo.containers.begin();iCont!=dbInfo.containers.end();iCont++){
      log << "    Container name=\"" <<iCont->second.name << "\" technology="<<iCont->second.technology<<" total entries="<< iCont->second.size << coral::MessageStream::endmsg;
      std::map<std::string, unsigned int>::const_iterator iClass;
      for(iClass=iCont->second.classes.begin();iClass!=iCont->second.classes.end();iClass++){
	      log << "        Class "<<iClass->first<<" entries="<<iClass->second<<coral::MessageStream::endmsg;
      }
      updateContainer(database,iCont->second,dry);
    }
    // close database, domain and session
    database.close();
    dom.close();
    sess.close();
  } else {
    log << coral::Info << "Database "<< dbInfo.connectionString <<": no containers found." << coral::MessageStream::endmsg;
    log << coral::Info << "Status=[UPTODATE]" << coral::MessageStream::endmsg;
  }
}

void pool::RelationalPoolDatabaseUtilities::updateContainer(const std::string& contName,bool dry){
  coral::MessageStream log( pool::RelationalStorageService::serviceName() );
  DatabaseInfo dbInfo = queryDatabase();
  if(dbInfo.connectionString==""){
    throw pool::RelationalStorageService::RelationalStorageServiceException( "POOL Database not found on "+m_connectionString, "RelationalStorageService::RelationalPoolDatabaseUtilities::updateContainer" );
  }
  if(!dbInfo.containers.size()){
    throw pool::RelationalStorageService::RelationalStorageServiceException( "No container found on "+m_connectionString, "RelationalStorageService::RelationalPoolDatabaseUtilities::updateContainer" );    
  }  
  
  std::map<std::string,ContainerInfo>::const_iterator iCont = dbInfo.containers.find(contName);
  if(iCont!=dbInfo.containers.end()){

    std::auto_ptr<pool::Context> ctx(new pool::Context);
    DbSession sess;
    sess.open( ctx.get() );
    DbDomain dom(POOL_RDBMS_StorageType);
    // open database
    DbDatabase database(POOL_RDBMS_StorageType);
    dom.open(sess,POOL_RDBMS_StorageType,pool::UPDATE);
    pool::Guid guid;
    pool::Guid::create( guid );
    if(!database.open(dom,m_connectionString,guid.toString(),pool::UPDATE).isSuccess()){
      throw RelationalStorageService::RelationalStorageServiceException("Failure while opening database.","RelationalPoolDatabaseUtilities::updateContainer");
    }

    log << "    Container name=\"" <<iCont->second.name << "\" technology="<<iCont->second.technology<<" - "<< iCont->second.size << " entries " << coral::MessageStream::endmsg;
    std::map<std::string, unsigned int>::const_iterator iClass;
    for(iClass=iCont->second.classes.begin();iClass!=iCont->second.classes.end();iClass++){
      log << "    Class "<<iClass->first<<" - "<<iClass->second<<" entries."<<coral::MessageStream::endmsg;
    }
    updateContainer(database,iCont->second,dry);

    // close database, domain and session
    database.close();
    dom.close();
    sess.close();
  } else {
    throw RelationalStorageService::RelationalStorageServiceException("Container not found.","RelationalPoolDatabaseUtilities::updateContainer");
  }
}

void pool::RelationalPoolDatabaseUtilities::dropContainer( const std::string& contName,bool full){
  init();
  
  // clean up containers
  std::auto_ptr<pool::Context> ctx(new pool::Context());
  RelationalStorageService::RelationalPoolDatabase::dropContainer( ctx.get(),m_connectionString, contName, full);
}

void pool::RelationalPoolDatabaseUtilities::setUpDatabase(const pool::ObjectRelationalDatabaseDescriptor& dbstruct){
  std::auto_ptr<pool::Context> ctx(new pool::Context);
  DbSession sess;
  sess.open(ctx.get());
  DbDomain dom(POOL_RDBMS_StorageType);
  // open database
  DbDatabase database(POOL_RDBMS_StorageType);
  dom.open(sess,POOL_RDBMS_StorageType,pool::UPDATE);
  pool::Guid guid;
  pool::Guid::create( guid );
  if(!database.open(dom,m_connectionString,guid.toString(),pool::UPDATE).isSuccess()){
    throw RelationalStorageService::RelationalStorageServiceException("Failure while opening database.","RelationalPoolDatabaseUtilities::setUpDatabase");
  }

  // loop over container to be created
  for(std::vector<ObjectRelationalContainerDescriptor*>::const_iterator iContainer=dbstruct.containers().begin();
      iContainer!=dbstruct.containers().end();iContainer++){

      bool recreatePolicy = (*iContainer)->recreatePolicy();
      DbType tech;
      std::string contName = (*iContainer)->name();
      std::string contClassName = (*iContainer)->classStores()[0]->className();

      std::vector<std::pair<std::string,std::string> > classes;
      std::string contType = (*iContainer)->type();
      coral::MessageStream log( pool::RelationalStorageService::serviceName());
      log << coral::Debug << "Creating container \""<< contName<<"\", type=\""<<contType<<"\"."<<coral::MessageStream::endmsg;
      if(contType == ObjectRelationalContainerDescriptor::homogeneousContainerTypeName()){
        tech = pool::POOL_RDBMS_HOMOGENEOUS_StorageType;
        if((*iContainer)->classStores().size()>1){
          log << coral::Warning << "Only one class is admitted for the homogeneous container \""<< contName<<"\", ignoring the other ("<<(*iContainer)->classStores().size()-1<<") classes specified." << coral::MessageStream::endmsg;
        }        
      } else {

        // for polymorhic container loop also over classes
        for(std::vector<ObjectRelationalClassDescriptor*>::const_iterator iClass=(*iContainer)->classStores().begin();iClass!=(*iContainer)->classStores().end();iClass++){
          std::string mappingVersion = (*iClass)->mappingVersion();
          classes.push_back(std::make_pair((*iClass)->className(),mappingVersion));
        }
        tech = pool::POOL_RDBMS_POLYMORPHIC_StorageType;
      }

      // open a storage transaction 
      DbTransaction trans;
      trans.set(Transaction::TRANSACT_START);
      database.transAct(trans);

      // open (->create) the specified container
      DbTypeInfo* dbTypeInfo = DbTypeInfo::create(contClassName);
      DbContainer newContainer(POOL_RDBMS_StorageType);
      if(!newContainer.open(database,contName,dbTypeInfo,tech,pool::UPDATE).isSuccess()){
        throw RelationalStorageService::RelationalStorageServiceException("Failure while opening container.","RelationalPoolDatabaseUtilities::setUpDatabase");
      }

      // Creating the persistent shapes & links.
      for(unsigned int ic=0;ic<classes.size();ic++){
        DbTypeInfo* shape = DbTypeInfo::create(classes[ic].first);
        if(database.addShape(shape).isSuccess()){
          Token* link = new Token(newContainer.token());
          link->setClassID(shape->shapeID());
          if(!database.makeLink(link, link->oid()).isSuccess()){
            throw RelationalStorageService::RelationalStorageServiceException("Failure while creating link.","RelationalPoolDatabaseUtilities::setUpDatabase");
          }
          link->release();
        } else {
            throw RelationalStorageService::RelationalStorageServiceException("Failure while registering shape.","RelationalPoolDatabaseUtilities::setUpDatabase");
        }
      }

      // get the pure Relational hook for the forced update of the container
      RelationalStorageService::RelationalPoolContainer* contHandle = dynamic_cast<RelationalStorageService::RelationalPoolContainer*>(newContainer.info());
      if(!contHandle->updateContent(classes,recreatePolicy).isSuccess()){
        throw RelationalStorageService::RelationalStorageServiceException("Failure in the container update procedure.","RelationalPoolDatabaseUtilities::setUpDatabase");
      }
      // close container
      newContainer.close();
      // close the transaction
      trans.set(Transaction::TRANSACT_COMMIT);
      database.transAct(trans);
  }
  // close database, domain and session
  database.close();
  dom.close();
  sess.close();
}

void pool::RelationalPoolDatabaseUtilities::updateContainer(pool::DbDatabase& db, const pool::RelationalPoolDatabaseUtilities::ContainerInfo& info, bool dry){
  // open a storage transaction 
  DbTransaction trans;
  trans.set(Transaction::TRANSACT_START);
  db.transAct(trans);

  DbType tech = DbType::getType(info.technology);
  // open (->create) the specified container
  DbTypeInfo* dbTypeInfo = DbTypeInfo::create(info.containerClassName);
  DbContainer container(POOL_RDBMS_StorageType);
  if(!container.open(db,info.name,dbTypeInfo,tech,pool::UPDATE).isSuccess()){
    throw RelationalStorageService::RelationalStorageServiceException("Failure while opening container.","RelationalPoolDatabaseUtilities::updateContainer");
  }
  std::vector<std::pair<std::string,std::string> > classes;
  for(std::map<std::string,unsigned int>::const_iterator iC=info.classes.begin();iC!=info.classes.end();iC++){
    classes.push_back(std::make_pair(iC->first,""));
  } 
  // get the pure Relational hook for the forced update of the container
  RelationalStorageService::RelationalPoolContainer* contHandle = dynamic_cast<RelationalStorageService::RelationalPoolContainer*>(container.info());
  if(!contHandle->updateContent(classes,false,dry).isSuccess()){
    throw RelationalStorageService::RelationalStorageServiceException("Failure in the container update procedure.","RelationalPoolDatabaseUtilities::updateContainer");
  }
  // close container
  container.close();
  // close the transaction
  trans.set(Transaction::TRANSACT_COMMIT);
  db.transAct(trans);
}

void pool::RelationalPoolDatabaseUtilities::init(){
  initialize();
  m_connectionString = getProperty(connectionStringProperty());
  if(m_connectionString==""){
    throw pool::RelationalStorageService::RelationalStorageServiceException("Connection string not provided.","RelationalPoolDatabaseUtilities::connectToDatabase");
  }
  return;
}



