//====================================================================
//
//        Package    : RootStorageSvc (The POOL project)
//
//        Author     : M.Frank
//====================================================================
#define POOL_ROOTDOMAIN_CPP 1

// Framework include files
#include "POOLCore/strcasecmp.h"
#include "StorageSvc/DbOption.h"
#include "RootDomain.h"
#include "RootDBase.h"

#include "TSystem.h"
#include "TError.h"
#include "TFile.h"
#include "TROOT.h"
#include "TTree.h"

using namespace pool;

/// Standard Constuctor
RootDomain::RootDomain(IOODatabase* idb)
: DbDomainImp(idb),
  m_defCompression(1),
  m_defSplitLevel(99),
  m_defAutoSave(16*1024*1024),
  m_defBufferSize(16*1024)
{
}

/// Standard destructor
RootDomain::~RootDomain()   {
}

/// Check for Database existence
bool RootDomain::existsDbase(const std::string& nam)  {
  return RootDBase::exists(nam);
}

/// Set domain specific options
DbStatus RootDomain::setOption(const DbOption& opt)  {
  const char* n = opt.name().c_str();
  switch( ::toupper(n[0]) )  {
    case 'A':
      if ( !strcasecmp(n, "ABORT_LEVEL") )  {
        int lvl = gErrorAbortLevel;
        DbStatus sc = opt._getValue(lvl);
        if ( sc.isSuccess() )  {
          gErrorAbortLevel = lvl;
        }
        return sc;
      }
      break;
    case 'D':
      if ( !strcasecmp(n, "DEFAULT_COMPRESSION") )  {
        return opt._getValue(m_defCompression);
      }
      else if ( !strcasecmp(n, "DEFAULT_SPLITLEVEL") )  {
        return opt._getValue(m_defSplitLevel);
      }
      else if ( !strcasecmp(n, "DEFAULT_AUTOSAVE") )  {
        return opt._getValue(m_defAutoSave);
      }
      else if ( !strcasecmp(n, "DEFAULT_BUFFERSIZE") )  {
        return opt._getValue(m_defBufferSize);
      }
      break;
    case 'E':
      if ( !strcasecmp(n, "ERROR_LEVEL") )  {
        int lvl = gErrorIgnoreLevel;
        DbStatus sc = opt._getValue(lvl);
        if ( sc.isSuccess() )  {
          gErrorIgnoreLevel = lvl;
        }
        return sc;
      }
      break;
    case 'F':
      if ( !strncasecmp(n+5, "READSTREAMERINFO",15) )  {
        int val = 1;
        DbStatus sc = opt._getValue(val);
        if ( sc.isSuccess() )  {
          TFile::SetReadStreamerInfo(val != 0 ? kTRUE : kFALSE);
        }
        return sc;
      }
      break;
    case 'T':
      if ( !strcasecmp(n+5,"MAX_SIZE") )  {
        long long int max_size = TTree::GetMaxTreeSize();
        DbStatus sc = opt._getValue(max_size);
        if ( sc.isSuccess() )  {
          TTree::SetMaxTreeSize(max_size);
        }
        return sc;
      }
    default:
      break;
  }
  return Error;  
}

/// Access domain specific options
DbStatus RootDomain::getOption(DbOption& opt) const   {
  const char* n = opt.name().c_str();
  switch( ::toupper(n[0]) )  {
    case 'A':
      if ( !strcasecmp(n, "ABORT_LEVEL") )  {
        return opt._setValue(int(gErrorAbortLevel));
      }
      break;
    case 'C':
      if ( !strcasecmp(n, "CLASS") )  {
        return opt._setValue((void*)ROOT::GetROOT()->GetClass(opt.option().c_str(), kTRUE));
      }
      break;
    case 'D':
      if ( !strcasecmp(n, "DEFAULT_COMPRESSION") )  {
        return opt._setValue(int(m_defCompression));
      }
      else if ( !strcasecmp(n, "DEFAULT_SPLITLEVEL") )  {
        return opt._setValue(int(m_defSplitLevel));
      }
      else if ( !strcasecmp(n, "DEFAULT_AUTOSAVE") )  {
        return opt._setValue(int(m_defAutoSave));
      }
      else if ( !strcasecmp(n, "DEFAULT_BUFFERSIZE") )  {
        return opt._setValue(int(m_defBufferSize));
      }
      break;
    case 'E':
      if ( !strcasecmp(n, "ERRNO") )  {
        return opt._setValue(int(gSystem->GetErrno()));
      }
      else if ( !strcasecmp(n, "ERROR_LEVEL") )  {
        return opt._setValue(int(gErrorIgnoreLevel));
      }
      break;
    case 'N':
      if ( !strcasecmp(n, "NUM_CLASSES") )  {
        return opt._setValue(int(ROOT::GetROOT()->GetNclasses()));
      }
      if ( !strcasecmp(n, "NUM_TYPES") )  {
        return opt._setValue(int(ROOT::GetROOT()->GetNtypes()));
      }
      break;
    case 'T':
      if ( !strcasecmp(n+5,"MAX_SIZE") )  {
        long long int val = TTree::GetMaxTreeSize();
        return opt._setValue(val);
      }
    default:
      break;
  }
  return Error;  
}
