//====================================================================
//        Root Database factories implementation
//--------------------------------------------------------------------
//
//        Package    : POOLCore/RootDb (The POOL project)
//
//        Author     : M.Frank
//====================================================================
#include "RootOODb.h"
#include "RootDBase.h"
#include "RootDomain.h"
#include "RootClassLoader.h"
#include "RootKeyContainer.h"
#include "RootTreeContainer.h"
#include "StorageSvc/DbInstanceCount.h"
#include "Reflex/Builder/ClassBuilder.h"
#include <memory>
using namespace pool;

static RootClassLoader* _classLoader();
typedef const std::vector<void*>& _Args;

#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
template <class T> static void* ctor(void* m, _Args, void*)  {  return ::new(m) T();      }
template <class T> static void* dtor(void* m, _Args, void*)  {  ((T*)m)->~T(); return 0;  }
static void* any_init(void* , _Args, void*) { static int r=1; _classLoader(); return &r; }

#else

template <class T> static void ctor(void* ret, void* m, _Args, void*)
{  if(ret)*(void**)ret = ::new(m) T(); else ::new(m) T();           }
template <class T> static void dtor(void*, void* m, _Args, void*)  
{  ((T*)m)->~T();                                                   }
static void any_init(void* ret, void*, _Args, void*) 
{ int r=1; _classLoader(); *(::new(ret) (int)) = r;                 }
#endif

struct RootAny {
  RootAny()          {    DbInstanceCount::increment(this);  }
  virtual ~RootAny() {    DbInstanceCount::decrement(this);  }
};

POOL_BEGIN_STATIC_BLOCK(RootAny)
  using namespace ROOT::Reflex;
  ClassBuilderT<RootAny>()
    .AddFunctionMember<void(void)>("RootAny",  ctor<RootAny>, 0, 0, PUBLIC | CONSTRUCTOR )
    .AddFunctionMember<void(void)>("~RootAny", dtor<RootAny>, 0, 0, PUBLIC | DESTRUCTOR | VIRTUAL )
    .AddFunctionMember<int(void)>("init", any_init, 0, 0, PUBLIC | STATIC );
POOL_END_STATIC_BLOCK(RootAny)

static RootClassLoader* _classLoader()  {
  static std::auto_ptr<RootClassLoader> ldr;
  if ( 0 == ldr.get() )  {
    ldr = std::auto_ptr<RootClassLoader>(new RootClassLoader);
  }
  return ldr.get();
}

namespace pool {  void* createLoader()  {    return _classLoader();  }    }
/// Standard Constructor
RootOODb::RootOODb(void* ctxt) : OODatabaseImp(ctxt, ROOT_StorageType)  {
  DbInstanceCount::increment(this);
  setClassLoader(_classLoader());
}

/// Standard Constructor
RootOODb::RootOODb(void* ctxt, DbType typ) : OODatabaseImp(ctxt, typ)   {
  DbInstanceCount::increment(this);
  setClassLoader(_classLoader());
}

/// Standard Destructor
RootOODb::~RootOODb()  {
  DbInstanceCount::decrement(this);
}

/// Create Root Domain object
IDbDomain* RootOODb::createDomain()  {
  return new RootDomain(this);        
}

/// Create Root Database object (TFile)
IDbDatabase* RootOODb::createDatabase()  {
  return new RootDBase(this);         
}

/// Create Root Container object
IDbContainer* RootOODb::createContainer(const DbType& typ) { 
  if ( typ.match(ROOTKEY_StorageType) )  {
    return new RootKeyContainer(this);
  }
  else if ( typ.match(ROOTTREE_StorageType) )    {
    return new RootTreeContainer(this);
  }
  else if ( typ.match(ROOT_StorageType) )    {
    return new RootTreeContainer(this);
  }
  return 0;
}
