//====================================================================
//    Root Database implementation
//--------------------------------------------------------------------
//
//    Package    : RootStorageSvc (The POOL project)
//
//    Author     : M.Frank
//====================================================================
#ifndef POOL_ROOTTREECONTAINER_H
#define POOL_ROOTTREECONTAINER_H 1

// Framework include files
#include "StorageSvc/DbDatabase.h"
#include "StorageSvc/DbContainerImp.h"
#include "StorageSvc/DbImplementation.h"

// Forward declarations
class TObject;
class TBranch;
class TBuffer;
class TTree;
class TLeaf;
class TClass;
class RootSelect;

/*
 * POOL namespace declaration
 */
namespace pool  { 

  // Forward declaration
  class DbColumn;
  class RootDBase;

  /** @class RootTreeContainer RootTreeContainer.h src/RootTreeContainer.h
    *
    * Description:
    * ROOT specific implementation of Database container.
    * Since objects in root are stored in "trees with branches",
    * this object corresponds to a tuple (Tree/Branch), where
    * each object type (determined by the location of the transient
    * object within the data store) is accessed by the "Event" number
    * inside its tree.
    *
    * @author  M.Frank
    * @date    1/8/2002
    * @version 1.0
    */
  class RootTreeContainer : public DbContainerImp   {
    /// Definiton of a branch descriptor
    struct BranchDesc {
    public:
      TClass*          clazz;
      TBranch*         branch;
      TLeaf*           leaf;
      void*            object;
      void*            buffer;
      const DbColumn*  column;
      IClassHandler*   handler;
      BranchDesc()
      : clazz(0), 
        branch(0),
        leaf(0),
        object(0),
        buffer(0),
        column(0),
        handler(0) {}
      BranchDesc( TClass* cl, 
                  TBranch* b,
                  TLeaf* l,
                  void* o,
                  const DbColumn* c,
                  IClassHandler* h)
      : clazz(cl), 
        branch(b),
        leaf(l),
        object(0),
        buffer(o),
        column(c),
        handler(h) {}
      BranchDesc(const BranchDesc& c)
      : clazz(c.clazz), 
        branch(c.branch),
        leaf(c.leaf),
        object(c.object),
        buffer(c.buffer),
        column(c.column),
        handler(c.handler) {}
    };
    /// Definition of the branch container
    typedef std::vector<BranchDesc> Branches;
    /// Reference to the root tree object
    TTree*             m_tree;
    /// reference to exact type description
    const DbTypeInfo*  m_type;
    /// Branch map
    Branches           m_branches;
    /// Parent Database handle
    DbDatabase         m_dbH;
    /// Root database file reference
    RootDBase*         m_rootDb;
    /// Branch name, if branch is specified by user.
    std::string        m_branchName;
    /// Number of bytes written/read during last operation. Set to -1 if it failed.
    int                m_ioBytes;

  private:

    /// Add item branch
    DbStatus addBranch( const DbColumn* col,
                        BranchDesc& dsc,
                        const std::string& desc);

    /// Add BLOB
    DbStatus addObject( const DbColumn* col, 
                        BranchDesc& dsc,
                        const std::string& desc,
                        int defSplitLevel, 
                        int defBufferSize);

    /// Find entry identified by his number (=primary key) in the Database
    DbStatus selectRow( const DataCallBack* call,
                        const DbLink&       linkH,
                        DbAccessMode        mode);

    // Routine needed for TRANSACT_FLUSH, if branch is specified by user.
    DbStatus finishTransAct();

    /// Access branch by name
    TBranch* branch(const std::string& nam)  const;

  public:

    /// Standard constructor
    RootTreeContainer(IOODatabase* idb);

    /// Standard destructor
    virtual ~RootTreeContainer();

    /// Close the container and deallocate resources
    virtual DbStatus close();

    /// Open the container for object access
    virtual DbStatus open(const DbDatabase& dbH, 
                          const std::string& nam, 
                          const DbTypeInfo* info,
                          DbAccessMode mod);

    /// Access options
    /** @param opt      [IN]  Reference to option object.
      *
      * @return DbStatus code indicating success or failure.  
      */
    virtual DbStatus getOption(DbOption& opt) const;

    /// Set options
    /** @param opt      [IN]  Reference to option object.
      *
      * @return DbStatus code indicating success or failure.  
      */
    virtual DbStatus setOption(const DbOption& opt);

    /// Ask if a given shape is supported
    virtual DbStatus isShapeSupported(const DbTypeInfo* typ) const;

    /// Define selection criteria
    virtual DbStatus select(DbSelect& criteria);

    /// Number of entries within the container
    virtual long long int size();

    /// Fetch next object address of the selection to set token
    virtual DbStatus fetch(DbSelect& sel);

    /// Fetch refined object address. Default implementation returns identity
    virtual DbStatus fetch(const DbLink& linkH, DbLink& stmt);

    /// Find object by object identifier and load it into memory
    /** @param  call      [IN]   Callback to load data
      * @param  oid      [OUT]   Object OID
      * @param  mode      [IN]   Object access mode
      *
      * @return Status code indicating success or failure.
      */
    virtual DbStatus loadObject(DataCallBack* call,
                                DbLink&       oid,
                                DbAccessMode  mode);

    /// Commit single entry to container
    virtual DbStatus writeObject(TransactionStack::value_type& entry);

    /// Execute end of object modification requests during a transaction
    /** @param refTr    [IN]     Transaction reference
      *
      * @return DbStatus code indicating success or failure.  
      */
    virtual DbStatus endTransaction(DbTransaction& refTr);
  };
}
#endif //POOL_ROOTTREECONTAINER_H
