//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id: RulesFCHostMatcher.cpp,v 1.4 2008/03/04 09:17:02 witoldp Exp $
//
// Description:
//	Class RulesFCHostMatcher
//
// Environment:
//	This software was developed for the ATLAS collaboration.
//
// Author List:
//      Andy Salnikov
//
//------------------------------------------------------------------------

//-----------------------
// This Class's Header --
//-----------------------
#include "RulesFCHostMatcher.h"

//-----------------
// C/C++ Headers --
//-----------------
#include <string>
#include <boost/regex.hpp>
#include <boost/tokenizer.hpp>

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "POOLCore/Exception.h"
//#include "CoralBase/MessageStream.h"

//-----------------------------------------------------------------------
// Local Macros, Typedefs, Structures, Unions and Forward Declarations --
//-----------------------------------------------------------------------

// include/declaration for gethostname()
#ifdef _WIN32
#include <Winsock2.h>
#else
#include <unistd.h>
#endif // _WIN32

namespace {

  // message stream
  //coral::MessageStream fclog("RulesFileCatalog" ) ;

  unsigned int str2num ( std::string str ) ;

  typedef std::pair<unsigned int,unsigned int> range ;
  std::list<range> getHostNumbers ( const std::string& spec ) ;

  // host name
  std::string localHostname () 
  {
    static std::string hostname ;
    // get the host name
    if ( hostname.empty() ) {
      char buf[256] ;
      if ( gethostname ( buf, 256 ) ) {
        hostname = "localhost" ;
        throw pool::Exception ( "failed to determine local host name", "RulesFCHostMatcher", "FileCatalog" ) ;
      } else {
        hostname = buf ;
      }
    }
    return hostname ;
  }
}

//		----------------------------------------
// 		-- Public Function Member Definitions --
//		----------------------------------------

// Constructor
pool::RulesFCHostMatcher::RulesFCHostMatcher(const std::string& hostname) 
  : m_hostname ( hostname )
{
}

// Destructor
pool::RulesFCHostMatcher::~RulesFCHostMatcher() 
{
}

// Match the given XML element agains the current host name
bool 
pool::RulesFCHostMatcher::match ( const std::string& hostRe, const std::string& hostNumbers ) const 
{
  // if only host-numbers exists then make regex that matches 
  // any hostname with numbers
  std::string host = hostRe ;
  if ( host.empty() && ! hostNumbers.empty() ) {
    host = ".*%N.*" ;
  }

  if ( host.empty() ) {
    // both are empty - matches every hostname
    return true ;
  }

  // replace %N with regexp for a number
  std::string::size_type numPos = host.find ( "%N" ) ;
  if ( numPos != std::string::npos ) {
    host.replace ( numPos, 2, "([0-9]+)" ) ;
  }

  // build regexp
  boost::regex re ;
  try {
    re = boost::regex( host ) ;
  } catch ( const std::exception& e ) {
    throw pool::Exception ( "error in regex \""+host+"\": "+e.what(), "RulesFCHostMatcher::match", "FileCatalog" ) ;
  }

  //fclog << coral::Debug << " host = " << host << coral::MessageStream::endmsg;

  // match regexp
  boost::smatch match ;
  std::string myhostname = m_hostname ;
  if ( myhostname.empty() ) myhostname = ::localHostname() ;
  bool r = boost::regex_match ( myhostname, match, re ) ;
  if ( ! r ) {
    // no match
    return false ;
  }

  // matching, check the host number
  std::list<range> numbers = getHostNumbers ( hostNumbers ) ;
  if ( numbers.empty() ) {
    // nothing to check, say we are OK
    return true ;
  }

  // OK, we got few numbers to check
  if ( numPos == std::string::npos ) {
    // host name did not have %N in it, get any number form the local host name
    re = boost::regex( ".*([0-9]+).*" ) ;
    r = boost::regex_match ( myhostname, match, re ) ;
    if ( ! r ) {
      // host name does not have any number in it?
      return false ;
    }
  }

  // get the number from the name
  //fclog << coral::Debug << " num-match = " << match.str(1) << coral::MessageStream::endmsg;
  unsigned int num = str2num( match.str(1) ) ;
  for ( std::list<range>::const_iterator it = numbers.begin() ; it != numbers.end() ; ++ it ) {
    if ( it->first <= num && num <= it->second ) return true ;
  }

  // not in any range
  return false ;
}



namespace {

  unsigned int str2num ( std::string str ) 
  {
    // strip whitespaces
    std::string::size_type x = str.find_first_not_of( ' ' ) ;
    if ( x != std::string::npos ) str.erase(0,x) ;
    x = str.find_last_not_of( ' ' ) ;
    if ( x != std::string::npos ) str.erase(x+1) ;

    // conversion
    char* endptr = 0 ;
    unsigned int res = strtoul ( str.c_str(), &endptr, 10 );
    if ( *endptr != '\0' ) {
      // part of string left unconverted which is an error
      throw pool::Exception ( "invalid number format: "+str, "RulesFCHostMatcher::match", "FileCatalog" ) ;
    }
    return res ;
  }

  std::list<range> 
  getHostNumbers ( const std::string& spec ) 
  {
    std::list<range> result ;

    // spec is a string denoting the list of the ranges of numbers or single numbers,
    // for example 1-5,8,24-125. Separator could be either coma or semi-colon.

    typedef boost::tokenizer<boost::char_separator<char> > tokenizer ;
    boost::char_separator<char> sep(",;");
    tokenizer tokens ( spec, sep ) ;
    for ( tokenizer::iterator tok_iter = tokens.begin() ; tok_iter != tokens.end() ; ++tok_iter ) {
      const std::string& rangeStr = *tok_iter ;
      std::string::size_type dash = rangeStr.find('-') ;
      if ( dash == std::string::npos ) {
        // maybe single number
        unsigned int r = ::str2num ( rangeStr ) ;
        result.push_back ( range ( r, r ) ) ;
        //fclog << coral::Debug << " single number = " << r << coral::MessageStream::endmsg;
      } else {
        // must be a pair of numbers
        unsigned int r0 = ::str2num ( std::string( rangeStr, 0, dash ) ) ;
        unsigned int r1 = ::str2num ( std::string( rangeStr, dash+1 ) ) ;
        result.push_back ( range ( r0, r1 ) ) ;
        //fclog << coral::Debug << " range = " << r0 << "-" << r1 << coral::MessageStream::endmsg;
      }
    }

    return result ;

  }

}
