//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id: RulesFileCatalog.cpp,v 1.6 2008/03/11 16:54:28 witoldp Exp $
//
// Description:
//	Class RulesFileCatalog
//
// Environment:
//	This software was developed for the ATLAS collaboration.
//
// Author List:
//      Andy Salnikov
//
//--------------------------------------------------------------------------

//-----------------------
// This Class's Header --
//-----------------------
#include "RulesFileCatalog.h"

//-----------------
// C/C++ Headers --
//-----------------
#include <list>
#include <string>

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "FileCatalog/FCException.h"
#include "CoralBase/MessageStream.h"
#include "Reflex/PluginService.h"
#include "xercesc/dom/DOMElement.hpp"
#include "xercesc/dom/DOMException.hpp"
#include "xercesc/dom/DOMNamedNodeMap.hpp"
#include "xercesc/dom/DOMNode.hpp"
#include "xercesc/dom/DOMNodeList.hpp"
#include "xercesc/parsers/XercesDOMParser.hpp"
#include "xercesc/sax/HandlerBase.hpp"
#include "xercesc/sax/SAXException.hpp"
#include "xercesc/util/PlatformUtils.hpp"
#include "xercesc/util/XMLException.hpp"
#include "xercesc/util/XMLString.hpp"
#include "xercesc/util/XMLURL.hpp"

#include "RulesDTDResolver.h"
#include "RulesFCHostMatcher.h"
#include "RulesPFNRule.h"

//-----------------------------------------------------------------------
// Local Macros, Typedefs, Structures, Unions and Forward Declarations --
//-----------------------------------------------------------------------

using namespace pool ;
using namespace xercesc ;

PLUGINSVC_FACTORY_WITH_ID(RulesFileCatalog, std::string("rulescatalog"), FCImpl*())

namespace {

  // message stream
  coral::MessageStream fclog("RulesFileCatalog") ;
  
  std::string xml2str ( const XMLCh* xmlstr ) 
  {
    char* buf = XMLString::transcode ( xmlstr ) ;
    std::string res ( buf ) ;
    XMLString::release ( &buf ) ;
    return res ;
  }

  // get the element attribute value as a string
  std::string getAttrValue ( xercesc::DOMElement* elem, const char* attr ) 
  {
    XMLCh* name = XMLString::transcode ( attr ) ;
    const XMLCh* attrVal = elem->getAttribute( name ) ;
    XMLString::release ( &name ) ;
    return xml2str ( attrVal ) ;
  }

  // get the element attribute value as a string, if first element attribute is empty then try second element
  std::string getAttrValue ( xercesc::DOMElement* elem1, xercesc::DOMElement* elem2, const char* attr ) 
  {
    std::string res ;
    XMLCh* name = XMLString::transcode ( attr ) ;
    const XMLCh* attrVal = elem1->getAttribute( name ) ;
    if ( XMLString::stringLen(attrVal) == 0 ) {
      attrVal = elem2->getAttribute( name ) ;
    }
    XMLString::release ( &name ) ;
    return xml2str ( attrVal ) ;
  }

  // add one rule
  pool::RulesPFNRule* getPFNRule ( xercesc::DOMElement* pfnRuleElem, 
                                               xercesc::DOMElement* rulesElem, 
                                               const pool::RulesFCHostMatcher& hostMatcher ) ;

}

//		----------------------------------------
// 		-- Public Function Member Definitions --
//		----------------------------------------

//----------------
// Constructors --
//----------------
pool::RulesFileCatalog::RulesFileCatalog()
  : FCImpl()
  , m_activeParent(0)
  , m_rules()
{
}

//--------------
// Destructor --
//--------------
pool::RulesFileCatalog::~RulesFileCatalog()
{
  // just in case client forgot to call disconnect
  disconnect() ;
}

//-------------
// Methods   --
//-------------
    
/** 
 *  RulesFileCatalog connection method,
 *  opens the rules file and reads it.
 *
 *  @param  url DB connection parameters in url format
 */
void 
pool::RulesFileCatalog::connect()
{

  // initialize xercesc
  static bool xercesInitialized = false ;
  if ( ! xercesInitialized ) {
    xercesInitialized = true ;
    XMLPlatformUtils::Initialize();
  }

  // setup XML parser
  XercesDOMParser* parser = new XercesDOMParser;
  parser->setValidationScheme( XercesDOMParser::Val_Auto );
  parser->setDoNamespaces( false );
  HandlerBase errHandler ;
  parser->setErrorHandler( &errHandler );
  RulesDTDResolver dtdresolver ;
  parser->setEntityResolver( &dtdresolver );

  // parse the thing
  try {
    parser->parse( m_url.c_str() );
  } catch ( const SAXException& e ) {
    throw FCconnectionException ( "RulesFileCatalog::connect", ::xml2str( e.getMessage() ) ) ;
  } catch ( const XMLException& e ) {
    throw FCconnectionException ( "RulesFileCatalog::connect", ::xml2str( e.getMessage() ) ) ;
  } catch ( const DOMException& e ) {
    throw FCconnectionException ( "RulesFileCatalog::connect", ::xml2str( e.getMessage() ) ) ;
  }

  // get top-level pointer
  DOMDocument* doc = parser->getDocument();
  if ( ! doc ) {
    throw FCconnectionException ( "RulesFileCatalog::connect", "parser returned empty document for URL "+m_url ) ;
  }
  DOMElement* root = doc->getDocumentElement();

  // collect all parent URLs
  std::list<std::string> parentURLs ;

  // get the parent catalog URLs
  XMLCh* tag1 = XMLString::transcode ( "data_sources" ) ;
  XMLCh* tag2 = XMLString::transcode ( "parent_catalog" ) ;
  DOMNodeList* nodes1 = root->getChildNodes() ;
  for ( unsigned int i = 0 ; i < nodes1->getLength() ; ++ i ) {
    if ( XMLString::compareString( tag1, nodes1->item(i)->getNodeName() ) == 0 ) {
      DOMNode* srcNode = nodes1->item(i) ;
      DOMNodeList* nodes2 = srcNode->getChildNodes() ;
      for ( unsigned int j = 0 ; j < nodes2->getLength() ; ++ j ) {
        if ( XMLString::compareString( tag2, nodes2->item(j)->getNodeName() ) == 0 ) {
          DOMElement* parentElem = dynamic_cast<DOMElement*> ( nodes2->item(j) ) ;
          const std::string& url = ::getAttrValue ( parentElem, "contact" ) ;
          parentURLs.push_back ( url ) ;
        }
      }
    }
  }
  XMLString::release ( &tag1 ) ;
  XMLString::release ( &tag2 ) ;

  if ( parentURLs.empty() ) {
    // need at least one catalog
    throw FCconnectionException ( "RulesFileCatalog::connect", "No parent catalogs found in "+m_url ) ;
  }


  // make an object which matches hosts in the rules with current host name
  RulesFCHostMatcher hostMatcher ;

  // get all the PFN rules
  tag1 = XMLString::transcode ( "rules" ) ;
  tag2 = XMLString::transcode ( "pfn_rule" ) ;
  nodes1 = root->getChildNodes() ;
  for ( unsigned int i = 0 ; i < nodes1->getLength() ; ++ i ) {
    if ( XMLString::compareString( tag1, nodes1->item(i)->getNodeName() ) == 0 ) {
      DOMElement* rulesElem = dynamic_cast<DOMElement*>( nodes1->item(i) ) ;
      DOMNodeList* nodes2 = rulesElem->getChildNodes() ;
      for ( unsigned int j = 0 ; j < nodes2->getLength() ; ++ j ) {
        if ( XMLString::compareString( tag2, nodes2->item(j)->getNodeName() ) == 0 ) {
          DOMElement* pfnElem = dynamic_cast<DOMElement*> ( nodes2->item(j) ) ;
          RulesPFNRule* rule = ::getPFNRule ( pfnElem, rulesElem, hostMatcher ) ;
          if ( rule ) m_rules.push_back ( rule ) ;
        }
      }
    }
  }
  XMLString::release ( &tag1 ) ;
  XMLString::release ( &tag2 ) ;

  fclog << coral::Debug << "found " << m_rules.size() << " rules" << coral::MessageStream::endmsg;

  // done with XML, can get rid of everything
  delete parser ;
  parser = 0 ;

  // Now try to connect to every parent, skip those that cannot be found
  for ( std::list<std::string>::const_iterator it = parentURLs.begin() ; it != parentURLs.end() ; ++ it ) {

    try {
      FCImpl* impl = FCImplBuilder::BuildFCImpl ( *it ) ;
      impl->connect() ;
      m_activeParent = impl ;
    } catch ( const std::exception& e ) {
      // probably means that contact string is not good, plugin not found, etc...
      // try to get as far as we can, discard this error
      fclog << coral::Warning << "failed to connect to parent catalog: " << e.what() << coral::MessageStream::endmsg;
    }

  }

  // must have at least one parent to be usable
  if ( ! m_activeParent ) {
    throw FCconnectionException ( "RulesFileCatalog::connect", "Failed to connect to parent catalogs" ) ;
  }

}

/**
 *  finish
 */
void 
pool::RulesFileCatalog::disconnect() const
{

  // disconnect from all parent catalogs, destroy them all
  if ( m_activeParent ) m_activeParent->disconnect() ;
  delete m_activeParent ;
  m_activeParent = 0 ;

  // clear all the rules too
  for ( RulesList::const_iterator it = m_rules.begin() ; it != m_rules.end() ; ++ it ) {
    delete *it ;
  }
  m_rules.clear() ;

}

///Start the catalog transaction
void 
pool::RulesFileCatalog::start()const
{
  if ( m_activeParent ) m_activeParent->start() ;
}

///Commit the catalog transaction
void 
pool::RulesFileCatalog::commit(pool::FileCatalog::CommitMode cm) const
{
  if ( m_activeParent ) m_activeParent->commit( cm ) ;
}

///Rollback the catalog transaction
void 
pool::RulesFileCatalog::rollback() const
{
  if ( m_activeParent ) m_activeParent->rollback() ;
}

/**
 *  Register a PFN when a new file is created, returns the corresponding FileID
 *
 *  @param pfn [IN] the Physical file name
 *  @param filetype [IN] file technology type
 *  @param fid [OUT] File ID of the file
 *  @param fsta [IN] file status
 */
void 
pool::RulesFileCatalog::registerPFN( const std::string&, const std::string&, pool::FileCatalog::FileID& ) const
{
  throw pool::Exception ( "operation is not supported", "RulesFileCatalog::registerPFN", "FileCatalog" ) ;
}

/**
 *  Register LFN
 *  @param pf [IN] PFN
 *  @param lf [IN] LFN
 */
void 
pool::RulesFileCatalog::registerLFN(const std::string&, const std::string&) const
{
  throw pool::Exception ( "operation is not supported", "RulesFileCatalog::registerLFN", "FileCatalog" ) ;
}

/**
 *  Add a replica file name in the catalog.
 *  @param pf [IN] the PFN of the master copy file.
 *  @param rpf [IN] the PFN of the replica to be added in the catalog.
 */
void 
pool::RulesFileCatalog::addReplicaPFN(const std::string&, const std::string&) const
{
  throw pool::Exception ( "operation is not supported", "RulesFileCatalog::addReplicaPFN", "FileCatalog" ) ;
}

void 
pool::RulesFileCatalog::addPFNtoGuid(const pool::FileCatalog::FileID&, const std::string&, const std::string&) const
{
  throw pool::Exception ( "operation is not supported", "RulesFileCatalog::addPFNtoGuid", "FileCatalog" ) ;
}

/**
 *  Rename a PFN in the catalog. The use case is the file is moved.
 *  @param pf [IN] the old PFN
 *  @param newpf [IN] the new PFN
 */
void 
pool::RulesFileCatalog::renamePFN(const std::string&, const std::string&) const
{
  throw pool::Exception ( "operation is not supported", "RulesFileCatalog::renamePFN", "FileCatalog" ) ;
}

void 
pool::RulesFileCatalog::lookupFileByPFN(const std::string& pf, pool::FileCatalog::FileID& fid, std::string& ftype) const
{
  FCImpl* parentCatalog = getParentCatalog () ;
  parentCatalog->lookupFileByPFN( pf, fid, ftype ) ;
}

void 
pool::RulesFileCatalog::lookupFileByLFN(const std::string& lf, pool::FileCatalog::FileID& fid) const
{
  FCImpl* parentCatalog = getParentCatalog () ;
  parentCatalog->lookupFileByLFN( lf, fid ) ;
}

void 
pool::RulesFileCatalog::lookupBestPFN(const pool::FileCatalog::FileID& fid, 
                                                   const pool::FileCatalog::FileOpenMode& omode, 
                                                   const pool::FileCatalog::FileAccessPattern& amode, 
                                                   std::string& pf, 
                                                   std::string& filetype) const
{
  FCImpl* parentCatalog = getParentCatalog () ;
  std::string pfn ;
  parentCatalog->lookupBestPFN( fid, omode, amode, pfn, filetype ) ;
  for ( RulesList::const_iterator it = m_rules.begin() ; it != m_rules.end() ; ++ it ) {
    fclog << coral::Debug << "RulesFileCatalog::lookupBestPFN -- before rule: " << pfn << coral::MessageStream::endmsg;
    pfn = (*it)->process ( pfn ) ;
    fclog << coral::Debug << "RulesFileCatalog::lookupBestPFN -- after rule: " << pfn << coral::MessageStream::endmsg;
  }
  pf = pfn ;
}

void 
pool::RulesFileCatalog::insertPFN(pool::PFNEntry&) const
{
  throw pool::Exception ( "operation is not supported", "RulesFileCatalog::insertPFN", "FileCatalog" ) ;
}

void 
pool::RulesFileCatalog::insertLFN(pool::LFNEntry&) const
{
  throw pool::Exception ( "operation is not supported", "RulesFileCatalog::insertLFN", "FileCatalog" ) ;
}

void 
pool::RulesFileCatalog::deletePFN(const std::string&) const
{
  throw pool::Exception ( "operation is not supported", "RulesFileCatalog::deletePFN", "FileCatalog" ) ;
}

void 
pool::RulesFileCatalog::deleteLFN(const std::string&) const
{
  throw pool::Exception ( "operation is not supported", "RulesFileCatalog::deleteLFN", "FileCatalog" ) ;
}

void 
pool::RulesFileCatalog::deleteEntry(const pool::FileCatalog::FileID&) const
{
  throw pool::Exception ( "operation is not supported", "RulesFileCatalog::deleteEntry", "FileCatalog" ) ;
}

bool 
pool::RulesFileCatalog::isReadOnly() const
{
  return true ;
}

bool 
pool::RulesFileCatalog::retrievePFN(const std::string& query, pool::FCBuf<pool::PFNEntry>& buf, const size_t& start )
{
  FCImpl* parentCatalog = getParentCatalog () ;
  bool res = parentCatalog->retrievePFN( query, buf, start ) ;

  std::vector<PFNEntry>& entries = buf.getData() ;
  fclog << coral::Debug << "RulesFileCatalog::retrievePFN -- " << entries.size() << " entries in the buffer" << coral::MessageStream::endmsg;
  for ( std::vector<PFNEntry>::iterator eit = entries.begin() ; eit != entries.end() ; ++ eit ) {

    std::string pfn = eit->pfname() ;
    for ( RulesList::const_iterator it = m_rules.begin() ; it != m_rules.end() ; ++ it ) {
      fclog << coral::Debug << "RulesFileCatalog::retrievePFN -- before rule: " << pfn << coral::MessageStream::endmsg;
      pfn = (*it)->process ( pfn ) ;
      fclog << coral::Debug << "RulesFileCatalog::retrievePFN -- after rule: " << pfn << coral::MessageStream::endmsg;
    }
  
    eit->setValue ( pfn, eit->guid(), eit->filetype() ) ;

  }

  return res ;
}

bool 
pool::RulesFileCatalog::retrieveLFN(const std::string& query, pool::FCBuf<pool::LFNEntry>& buf, const size_t& start )
{
  FCImpl* parentCatalog = getParentCatalog () ;
  return parentCatalog->retrieveLFN( query, buf, start ) ;
}

bool 
pool::RulesFileCatalog::retrieveGuid(const std::string& query, pool::FCBuf<pool::FileCatalog::FileID>& buf, const size_t& start )
{
  FCImpl* parentCatalog = getParentCatalog () ;
  return parentCatalog->retrieveGuid( query, buf, start ) ;
}

/**
 *  Get one of the catalogs to be used for redirection. Default algorithm always
 *  uses the first availbale, can be overriden in subcalss.
 */
pool::FCImpl* 
pool::RulesFileCatalog::getParentCatalog () const 
{
  if ( ! m_activeParent ) {
    throw FCconnectionException ( "RulesFileCatalog::getParentCatalog", "Parent catalogs is not set" ) ;
  }
  return m_activeParent ;
}

namespace {

  // add one rule
  pool::RulesPFNRule* getPFNRule ( xercesc::DOMElement* pfnRuleElem, 
                                               xercesc::DOMElement* rulesElem, 
                                               const pool::RulesFCHostMatcher& hostMatcher )
  {
    // attributes
    std::string host = ::getAttrValue ( pfnRuleElem, rulesElem, "host" ) ;
    std::string hostNumbers = ::getAttrValue ( pfnRuleElem, rulesElem, "host_numbers" ) ;

    fclog << coral::Debug << "found new rule: host='" 
      << host << "' host_numbers='" << hostNumbers << "'" << coral::MessageStream::endmsg;

    if ( ! hostMatcher.match ( host, hostNumbers ) ) {
      // host does not match
      fclog << coral::Debug << "host does not match" << coral::MessageStream::endmsg;
      return 0 ;
    }

    std::string pfnRe = ::getAttrValue ( pfnRuleElem, rulesElem, "pfn_regexp" ) ;
    std::string replace = ::getAttrValue ( pfnRuleElem, rulesElem, "replace" ) ;
    std::string xrootd = ::getAttrValue ( pfnRuleElem, rulesElem, "xrootd" ) ;

    fclog << coral::Debug << "making new rule: pfn_regexp='" 
      << pfnRe << "' replace='" << replace << "' xrootd='" << xrootd << "'" << coral::MessageStream::endmsg;

    // Build the rule object
    pool::RulesPFNRule* res = new pool::RulesPFNRule ( pfnRe, replace, xrootd ) ;

    return res ;

  }
}
