#ifndef POOL_RULESFILECATALOG_H
#define POOL_RULESFILECATALOG_H

//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id: RulesFileCatalog.h,v 1.2 2008/02/28 22:53:35 salnikov Exp $
//
// Description:
//	Class RulesFileCatalog.
//
// Environment:
//      This software was developed for the ATLAS collaboration.  If you
//      use all or part of it, please give an appropriate acknowledgement.
//
//------------------------------------------------------------------------

//-----------------
// C/C++ Headers --
//-----------------
#include <vector>

//----------------------
// Base Class Headers --
//----------------------
#include "FileCatalog/FCImpl.h"

//-------------------------------
// Collaborating Class Headers --
//-------------------------------

//------------------------------------
// Collaborating Class Declarations --
//------------------------------------

//		---------------------
// 		-- Class Interface --
//		---------------------


/**
 *  C++ source file code template. The first sentence is a brief summary of 
 *  what the class is for. It is followed by more detailed information
 *  about how to use the class. This doc comment must immediately preced the 
 *  class definition.
 *
 *  Additional paragraphs with more details may follow; separate paragraphs
 *  with a blank line. The last paragraph before the tags (preceded by @) 
 *  should be the identification and copyright, as below.
 *
 *  This software was developed for the ATLAS collaboration.  If you
 *  use all or part of it, please give an appropriate acknowledgement.
 *
 *  @see TemplateTemplate
 *
 *  @version $Id: RulesFileCatalog.h,v 1.2 2008/02/28 22:53:35 salnikov Exp $
 *
 *  @author Andy Salnikov
 */

namespace pool {

class RulesPFNRule ;

class RulesFileCatalog : public pool::FCImpl {
public:

  // Constructors
  RulesFileCatalog() ;

  // Destructor
  virtual ~RulesFileCatalog() ;

  /** 
   *  RulesFileCatalog connection method,
   *  opens the rules file and reads it.
   *
   *  @param  url DB connection parameters in url format
   */
  virtual void connect();

  /**
   *  finish
   */
  virtual void disconnect() const;

  ///Start the catalog transaction
  virtual void start()const;

  ///Commit the catalog transaction
  virtual void commit(pool::FileCatalog::CommitMode cm = pool::FileCatalog::REFRESH) const;

  ///Rollback the catalog transaction
  virtual void rollback() const;

  /**
   *  Register a PFN when a new file is created, returns the corresponding FileID
   *
   *  @param pfn [IN] the Physical file name
   *  @param filetype [IN] file technology type
   *  @param fid [OUT] File ID of the file
   *  @param fsta [IN] file status
   */
  virtual void registerPFN( const std::string& pfn, const std::string& filetype, pool::FileCatalog::FileID& fid ) const;

  /**
   *  Register LFN
   *  @param pf [IN] PFN
   *  @param lf [IN] LFN
   */
  virtual void registerLFN(const std::string& pf, const std::string& lf) const;

  /**
   *  Add a replica file name in the catalog.
   *  @param pf [IN] the PFN of the master copy file.
   *  @param rpf [IN] the PFN of the replica to be added in the catalog.
   */
  virtual void addReplicaPFN(const std::string& pf, const std::string& rpf) const;

  virtual void addPFNtoGuid(const pool::FileCatalog::FileID& guid, const std::string& pf, const std::string& filetype) const;

  /**
   *  Rename a PFN in the catalog. The use case is the file is moved.
   *  @param pf [IN] the old PFN
   *  @param newpf [IN] the new PFN
   */
  virtual void renamePFN(const std::string& pf, const std::string& newpf) const;

  virtual void lookupFileByPFN(const std::string& pf, pool::FileCatalog::FileID& fid, std::string& ftype) const;

  virtual void lookupFileByLFN(const std::string& lf, pool::FileCatalog::FileID& fid) const;

  virtual void lookupBestPFN(const pool::FileCatalog::FileID& fid, 
                             const pool::FileCatalog::FileOpenMode& omode, 
                             const pool::FileCatalog::FileAccessPattern& amode, 
                             std::string& pf, 
                             std::string& filetype) const;

  virtual void insertPFN(pool::PFNEntry& pentry) const;

  virtual void insertLFN(pool::LFNEntry& lentry) const;

  virtual void deletePFN(const std::string&  pf) const;

  virtual void deleteLFN(const std::string& lf) const;

  virtual void deleteEntry(const pool::FileCatalog::FileID& guid) const;

  virtual bool isReadOnly() const;

  virtual bool retrievePFN(const std::string& query, pool::FCBuf<pool::PFNEntry>& buf, const size_t& start );

  virtual bool retrieveLFN(const std::string& query, pool::FCBuf<pool::LFNEntry>& buf, const size_t& start );

  virtual bool retrieveGuid(const std::string& query, pool::FCBuf<pool::FileCatalog::FileID>& buf, const size_t& start );

protected:

  // Helper functions

  /**
   *  Get one of the catalogs to be used for redirection. Default algorithm always
   *  uses the first availbale, can be overriden in subcalss.
   */
  virtual pool::FCImpl* getParentCatalog () const ;

private:

  // Types
  typedef std::vector<RulesPFNRule*> RulesList ;

  // Data members
  mutable pool::FCImpl* m_activeParent ;
  mutable RulesList m_rules ;

  // Note: if your class needs a copy constructor or an assignment operator, 
  //  make one of the following public and implement it.
  RulesFileCatalog( const RulesFileCatalog& );                // Copy Constructor
  RulesFileCatalog& operator= ( const RulesFileCatalog& );    // Assignment op

};

} // namespace pool

#endif // POOL_RULESFILECATALOG_H
