//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id: RulesPFNRule.cpp,v 1.1 2008/02/27 17:25:03 salnikov Exp $
//
// Description:
//	Class RulesPFNRule
//
// Environment:
//	This software was developed for the ATLAS collaboration.
//
// Author List:
//      Andy Salnikov
//
//------------------------------------------------------------------------

//-----------------------
// This Class's Header --
//-----------------------
#include "RulesPFNRule.h"

//-----------------
// C/C++ Headers --
//-----------------

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "POOLCore/Exception.h"

//-----------------------------------------------------------------------
// Local Macros, Typedefs, Structures, Unions and Forward Declarations --
//-----------------------------------------------------------------------

//		----------------------------------------
// 		-- Public Function Member Definitions --
//		----------------------------------------

/**
 *  Constructor
 *
 *  @param pfnRe    regular expression for PFN matching, can be empty
 *  @param replace  replacement expression, can be empty
 *  @param xrootd   host name of xrootd server, can be empty
 */
pool::RulesPFNRule::RulesPFNRule( const std::string& pfnRe, 
                                  const std::string& replace, 
                                  const std::string& xrootd )
  : m_re()
  , m_replacement(replace)
  , m_xrootd(xrootd)
{
  // check arg consistency
  if ( ! replace.empty() && pfnRe.empty() ) {
      throw pool::Exception ( "have replacement string but no regexp", "RulesPFNRule", "FileCatalog" ) ;
  }

  // compile regexp
  if ( ! pfnRe.empty() ) {
    try {
      m_re = boost::regex( pfnRe ) ;
    } catch ( const std::exception& e ) {
      throw pool::Exception ( "error in regex \""+pfnRe+"\": "+e.what(), "RulesPFNRule", "FileCatalog" ) ;
    }
  }
}

// Destructor
pool::RulesPFNRule::~RulesPFNRule()
{
}

// Process PFN and return modified (or unchanged) PFN
std::string 
pool::RulesPFNRule::process ( const std::string& aPfn ) const 
{
  std::string pfn ( aPfn ) ;

  // if match string is there then match PFN against it, if it does
  // not then retuyrtn PFN unchanged
  if ( ! m_re.empty() ) {
    if ( ! boost::regex_match ( pfn, m_re ) ) {
      return pfn ;
    }
  }

  // do the replacements if replacement string is not empty
  if ( ! m_replacement.empty() ) {
    try {
      pfn = boost::regex_replace ( pfn, m_re, m_replacement ) ;
    } catch ( const std::exception& e ) {
      throw pool::Exception ( std::string("error in regex_replace: ")+e.what(), "RulesPFNRule::process", "FileCatalog" ) ;
    }
  }

  // if xroot string was given then redirect everything to that xrootd server
  if ( ! m_xrootd.empty() ) {
    // just append root protocol/host name to the PFN, if PFN already
    // has a protocol or hosthane in it this will break
    pfn = "root://"+m_xrootd+"/"+pfn ;
  }

  return pfn ;
}
