// $Id: IClassLoader.h,v 1.19 2008/05/13 14:39:31 witoldp Exp $
//
//  Package    : pool/StorageSvc (The pool framework)
//
//  @author      M.Frank
//
//====================================================================
#ifndef POOL_ICLASSLOADER_H
#define POOL_ICLASSLOADER_H
// $Header: /cvs/PF/pool/StorageSvc/StorageSvc/IClassLoader.h,v 1.19 2008/05/13 14:39:31 witoldp Exp $

// Framework include files
#include "StorageSvc/pool.h"
#include "Reflex/Type.h"
#include "RVersion.h"

// C++ include files
#include <vector>
#include <string>

/*
 *   POOL namespace declaration
 */
namespace pool    {

  // Forward declarations
  class IClassHandler;

  /** @class IClassLoader IClassLoader.h StorageSvc/IClassLoader.h
    *
    * Definition of the IClassLoader class.
    * 
    * Description:
    *  Object definition or a class, which handles the dynamic loading
    *  of reflection information for accessing persistent storage.
    *
    * @author  Markus Frank                                               
    * @version 1                                                         
    */
  class IClassLoader  {
  protected:
    /// Destructor (called only by sub-classes)
    virtual ~IClassLoader()   {     }

  public:
    /// Reflection class declaration
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
    typedef ROOT::Reflex::Type TypeH;
#else
    typedef Reflex::Type TypeH;
#endif

    /// Definition of class descriptor
    typedef std::pair<TypeH, void* > Descriptor;

    /// Release resource
    virtual void release() = 0;

    /// Access reflection information of a given class
    /** @param name    [IN]  Class name as it is known to the LCG dictionary.
      * @param force   [IN]  Flag to force the class loading.
      * @return Pair with the pointers to the LCG and native class description.
      */
    virtual Descriptor getClass(    const std::string& name, 
                                    bool force=true) = 0;

    /// Access reflection information of a given class
    /** @param name    [IN]  Class name as it is known to the native dictionary.
      * @param force   [IN]  Flag to force the class loading.
      * @return Pair with the pointers to the LCG and native class description.
      */
    virtual Descriptor getNativeClass(const std::string& name, 
                                      bool force=true) = 0;
    /// Load the class unconditionally
    /** @param name    [IN]  Class name as it is known to the native dictionary.
      * @return  DbStatus code indicating success or failure.
      */
    virtual DbStatus loadClass(       const std::string& name) = 0;

    /// Load the native class unconditionally.
    /** Note, that the name of the native class may not correspond
      * to the class name of the LCG reflection interface
      * 
      * @param   name   [IN]   Name of the native class
      * @return  DbStatus code indicating success or failure.
      */
    virtual DbStatus loadNativeClass( const std::string& name) = 0;

    /// Load the native class unconditionally.
    /** Note, that the name of the native classes may not correspond
      * to the class name of the LCG reflection interface
      * 
      * @param   names  [IN]   Vector with the names of the native classes
      *                        to be loaded.
      * @return  DbStatus code indicating success or failure.
      */
    virtual DbStatus loadNativeClasses(const std::vector<std::string>& names) = 0;

    /// Unload the class
    /** @param name    [IN]  Class name as it is known to the LCG dictionary.
      * @return Status code indicating success or failure.
      */
    virtual DbStatus unloadClass(     const std::string& name) = 0;

    /// Access to the class handler by it's LCG class name
    /** @param name    [IN]  Class name as it is known to the LCG dictionary.
      * @param force   [IN]  Flag to force the class loading.
      * @return Pair with the pointers to the LCG and native class description.
      */
    virtual IClassHandler* handler( const std::string& name,
                                    bool force_load=true) = 0;

    /// Access to the class handler by it's native class name
    /** @param name    [IN]  Class name as it is known to the native dictionary.
      * @param force   [IN]  Flag to force the class loading.
      * @return Pair with the pointers to the LCG and native class description.
      */
    virtual IClassHandler* nativeHandler( const std::string& name,
                                          bool force_load=true) = 0;
    /// Set a user handler in place to deal with special class treatment
    /** @param lcg_name [IN]  Class name as it is known to the LCG dictionary.
      * @param nat_name [IN]  Class name as it is known to the native dictionary.
      * @param hndl     [IN]  Pointer to class handler
      * @return Status code indicating success or failure.
      */
    virtual DbStatus setHandler(      const std::string& lcg_name, 
                                      const std::string& nat_name, 
                                      IClassHandler* hndl) = 0;

    /// Remove a user handler
    /** @param name    [IN]  Class name as it is known to the LCG dictionary.
      * @return Status code indicating success or failure.
      */
    virtual DbStatus removeHandler(   const std::string& cl_name) = 0;
  };
}      // End namespace pool
#endif // POOL_ICLASSLOADER_H
