#include "SealBase/SharedLibrary.h"
#include "FileCatalog/URIParser.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/FCEntry.h"
#include "FileCatalog/IFCAction.h"
#include "DataSvc/DataSvcFactory.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/Placement.h"
#include "StorageSvc/DbType.h"
#include "PersistencySvc/ITransaction.h"
#include "DataSvc/Ref.h"
#include "DataSvc/IDataSvc.h"
#include "AttributeList/AttributeList.h"

#include <iostream> 
#include <vector>
#include <string>

#include "Event.h"

#include "Collection/Collection.h"
#include "Collection/CollectionProxy.h"
#include "Collection/CollectionSvc.h"



using namespace std;
using namespace pool;
using namespace seal;



const AttributeList& getEventMetaData(){
  
  static int cnt = 0;
  static AttributeList* al = 0;
  static AttributeListSpecification als;
  
  if(!al){
    als.push_back("NMuon", "int");
    als.push_back("NZ0","int");
    als.push_back("NHiggs","int");
    al = new AttributeList(als);
  }
  
  ++cnt;
    
  (*al)["NMuon"].setValue(cnt);
  (*al)["NZ0"].setValue(cnt/2);
  (*al)["NHiggs"].setValue(cnt/4);
  
  return *al; 
}



MetaDataEntry* getRunMetaData(){
  
  static int cnt = 0;
  static MetaDataEntry* al = 0;
  static const string crew[] = {"Wallace, Gromit",
				"Wallace, Gromit, Feathers"};
  
  if(!al){
    al = new MetaDataEntry();
    al->addAttributeSpec("Lumi","double");
    al->addAttributeSpec("ShiftCrew","string");
  }
  
  ++cnt;
  
  al->setAttributeValue("Lumi",cnt*23/42.0);
  al->setAttributeValue("ShiftCrew",crew[cnt%2]);
  
  return al; 
}



void printAttributes( const AttributeList& attrList ){
  const AttributeListSpecification& attrListSpec = attrList.attributeListSpecification();
  for( unsigned int i = 0; i < attrListSpec.size(); ++i ){
    const AttributeSpecification& attrSpec = attrListSpec[i];
    string attrName = attrSpec.name();
    string attrValue = attrList[attrName].getValueAsString();
    cout << attrName << "=" << attrValue << "\t";
  }
  cout << endl;
}




int main(){
  
  const int nRuns = 5;
  const int nEvents = 3;
  
  const string collType = "RootCollection";
  const string collConnection = "./";
  const string collNamePrefix = "EvtColl_";
  const string collQuery = "ShiftCrew like '%Feathers%";
  const string objQuery = "NHiggs>0 && NZ0%2==0";
  const string logicalName = "";
  
  
  
  
  try{
    
    const string plaformDependentName = seal::SharedLibrary::libname("MultiCollection_BasicFunctionalityDict");
    seal::SharedLibrary::load(plaformDependentName);
    
    pool::URIParser p("file:PoolFileCatalog.xml");
    pool::IFileCatalog* fileCatalog = new pool::IFileCatalog();
    p.parse();
    fileCatalog->addReadCatalog( p.contactstring() );
    fileCatalog->connect();
    
    IDataSvc* dataSvc = pool::DataSvcFactory::instance( fileCatalog );
    
    {  
      fileCatalog->start();
      
      pool::DatabaseConnectionPolicy policy;
      policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
      policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
      dataSvc->session().setDefaultConnectionPolicy( policy );
      
      pool::Placement placement;
      placement.setDatabase( "testFile.pool", pool::DatabaseSpecification::PFN );
      placement.setContainerName( "TestContainer" );
      placement.setTechnology( pool::ROOTTREE_StorageType.type() );    
      
      

      
      // fill a vector with references to the events we want to store in the collections 
      vector< Ref<Event> > EventRefVec;
      
      dataSvc->transaction().start( pool::ITransaction::UPDATE );
      
      for(int i=0; i<nEvents*nRuns; ++i){
	Ref<Event> ref(dataSvc, new Event(true));
	ref.markWrite(placement);
	EventRefVec.push_back(ref);
      }
      
      dataSvc->transaction().commit();
      fileCatalog->commit();
      
      fileCatalog->start();
      dataSvc->transaction().start( pool::ITransaction::READ );
      
      // setup an iterator for the Ref<Event>s
      vector< Ref<Event> >::const_iterator EventRefIter = EventRefVec.begin();
      
      // setup of the collection catalog
      IFileCatalog* collCatalog = new IFileCatalog; 
      collCatalog->setWriteCatalog("xmlcatalog_file:CollectionCatalog.xml");
      collCatalog->connect();
      FCAdmin act;
      collCatalog->setAction(act);
      collCatalog->start();
      act.createMetaDataSpec(*getRunMetaData());
      collCatalog->commit();
      
      CollectionSvc::setDefaultCatalog(collCatalog);
      
      
      // create and register the collections
      cout << "Create and register " << nRuns << " collections" << endl;
      for(int run=1; run<=nRuns; ++run){
	ostringstream oss;
	oss << collNamePrefix << run;
	string collName = oss.str(); 
	
	Collection<Event> evtColl = 
	  CollectionSvc::createAndRegister<Event>( dataSvc, 
						   collType, 
						   collConnection, 
						   collName,
						   logicalName,
						   getRunMetaData());
	for (int i=0; i<nEvents; ++i)
	  evtColl.add(*EventRefIter++, getEventMetaData());
	evtColl.commit();
      }
      
      // test query and iteration
      cout << "Query and iteration test:" << endl;
      Collection<Event>::Iterator iter = 
	CollectionSvc::select<Event>( dataSvc, objQuery, collQuery);
      
      while( iter.next() ){
        cout << "[OVAL] "; 
	iter->print();
	printAttributes(iter.attributeList());
      }
    }
    
  }catch( std::exception& exception ) {
    cout << "Exception: " << exception.what() << endl;
  }
}
