#include "Collection/Collection.h"
#include "AttributeList/AttributeList.h"
#include "AttributeList/AttributeListSpecification.h"
#include "FileCatalog/URIParser.h"
#include "FileCatalog/IFileCatalog.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/Placement.h"
#include "StorageSvc/DbType.h"

#include "CoralBase/MessageStream.h"
#include "POOLCore/Exception.h"
#include "TestDictionary/SimpleTestClass.h"
#include "PoolBench/ProcessTime.h"
#include "util.h"

#include "StorageSvc/DbStatus.h"
#include "SealBase/SharedLibrary.h"

#include <fstream>
#include <iomanip>
#include <math.h>
#include <stdexcept>

const std::string TEST_LIBRARY = "TestDictionaryDict";
const std::string FILE_CATALOG_TYPE = "xmlcatalog";
const std::string FILE_CATALOG_NAME = "PoolFileCatalog.xml";
const pool::DbType DB_TYPE = pool::ROOTTREE_StorageType;

int main(int argc, const char *argv[])
{
  try
  {
    pool::DbStatus sc = pool::DbStatus::Success;
    bool isInt = true;

    
    // pool::POOLContext::setMessageVerbosityLevel( coral::Info );

    coral::MessageStream log( "Collection_ExplicitWritePerformance", 
      coral::Info );
    int msgLevel = coral::Info - 1;
    char* msgLevelGetEnv = ::getenv( "POOL_OUTMSG_LEVEL" );
    char msgLevelSetEnv[20] = "";
    if( msgLevelGetEnv )
    {
      msgLevel = atoi( msgLevelGetEnv );
      if( ( msgLevel < (seal::Msg::Nil-1) ) || 
        ( msgLevel > (coral::Fatal-1) ) )
      {
	log << coral::Fatal << "Environment variable POOL_OUTMSG_LEVEL "
          << "is set to " << msgLevelGetEnv << " which is out of required "
            << " range of [" << (seal::Msg::Nil-1) << "," 
              << (coral::Fatal-1) << "]" << coral::MessageStream::endmsg;
	return 1;
      }
    }
    else
    {
      log << coral::Warning << "Environment variable POOL_OUTMSG_LEVEL "
        << "not set. Setting it to (coral::Info-1)" << coral::MessageStream::endmsg;
      std::sprintf( msgLevelSetEnv, "%i", 2 );
      ::setenv( "POOL_OUTMSG_LEVEL", msgLevelSetEnv, 1 );
    }

    log << coral::Info << "Starting main()" << coral::MessageStream::endmsg;

    if( argc < 5 )
    {
      log << coral::Fatal 
	  << "Usage: test_Collection_ExplicitWritePerformance "
	  << "<collection name prefix> "
          << "<# of events (at creation)> "
	  << "<# of attributes> "
          << "<collection type (Root, MySQL, MySQLlt, RALMySQL or RALOracle, )> "
	  << "Optional arguments (use - char to mask out those not needed): "
	  << "[# of events to process before next commit DEFAULT=1] "
          << "[# of events to append (for update) DEFAULT = 0] "
          << "[# of attributes of type Token DEFAULT = 0] "
	  << "[# of events to process before next print to screen "
          << "DEFAULT = 1000] "
          << "[name of performance data output file DEFAULT = Test_Write.log] "
	  << "[printout units (seconds, millisec, microsec or nanosec) "
	  << "DEFAULT = millisec] "
	  << "[random seed int for attribute generation DEFAULT=7232003] "
	  << coral::MessageStream::endmsg;
      return 1;
    }

    std::string basename = std::string(argv[1])+"_"+argv[2]+"_"+argv[3]+"_"+
      argv[4]+"_";

    std::string dbFile = basename;
    dbFile += "File.pool";
    std::string containerName = basename;
    containerName += "Container";
    std::string collectionName = basename; 
    collectionName += "Collection";
    std::string collectionType = argv[4]; 
    collectionType += "Collection";

    int numEventsCreate = 0;
    isInt = Util::to_value(argv[2], numEventsCreate);
    if( !isInt || (numEventsCreate<0) )
    {
      log << coral::Fatal << "Number of events must be a positive "
        << "integer. Got: " << argv[2] << coral::MessageStream::endmsg;
      return 1;
    }

    int numAttributes = 0;
    isInt = Util::to_value(argv[3], numAttributes);
    if( !isInt || (numAttributes<0) )
    {
      log << coral::Fatal << "Number of attributes must be a positive "
        << "integer. Got: " << argv[3] << coral::MessageStream::endmsg;
      return 1;
    }

    std::string collectionDb = "";
    char* dbConnectGetEnv;
    if( !strcmp( argv[4], "Root" ) )
    { 
      collectionDb = "";
    }
    else if( !strcmp( argv[4], "MySQL" ) || 
             !strcmp( argv[4], "MySQLlt" ) )
    {   
      dbConnectGetEnv = ::getenv( "POOL_MYSQL_CONNECT" );
      if( dbConnectGetEnv )
      {
	collectionDb = dbConnectGetEnv;
      }
      else
      {
	log << coral::Warning << "Environment variable POOL_MYSQL_CONNECT "
	  << "not set. Setting MySQL database connection to `" 
	    << "mysql://regrTester:rT2003@lxshare070d.cern.ch/regression'"
	      << coral::MessageStream::endmsg;
	collectionDb = 
          "mysql://regrTester:rT2003@lxshare070d.cern.ch/regression"; 
      }
    }
    else if( !strcmp( argv[4], "RALMySQL" ) )
    {   
      dbConnectGetEnv = ::getenv( "POOL_RAL_MYSQL_CONNECT" );
      if( dbConnectGetEnv )
      {
	collectionDb = dbConnectGetEnv;
      }
      else
      {
	log << coral::Warning << "Environment variable POOL_RAL_MYSQL_CONNECT "
	  << "not set. Setting RAL MySQL database connection to `" 
	    << "mysql://pcitdb59.cern.ch/ralcolls'"
	      << coral::MessageStream::endmsg;
	collectionDb = 
          "mysql://pcitdb59.cern.ch/ralcolls"; 
      }
    }
    else if( !strcmp( argv[4], "RALOracle" ) )
    {   
      dbConnectGetEnv = ::getenv( "POOL_RAL_ORACLE_CONNECT" );
      if( dbConnectGetEnv )
      {
	collectionDb = dbConnectGetEnv;
      }
      else
      {
	log << coral::Warning << "Environment variable POOL_RAL_ORACLE_CONNECT "
	  << "not set. Setting RAL Oracle database connection to `" 
	    << "oracle://devdb10.cern.ch/ral_writer'"
	      << coral::MessageStream::endmsg;
	collectionDb = 
          "oracle://devdb10.cern.ch/ral_writer"; 
      }
    }
    else
    {
      log << coral::Fatal << "Command line argument " << argv[4]
	<< " is not an expected database type." << coral::MessageStream::endmsg;
      return 1;
    }

    int skipCommitInterval = 1;
    if( argc > 5 )
    {
      if( strcmp(argv[5], "-") ) 
      {
        isInt = Util::to_value(argv[5], skipCommitInterval);
        if( !isInt || (skipCommitInterval<1) )
        {
          log << coral::Fatal << "Number of events to process before a "
            << "collection commit must be an integer > 0. Got: " << argv[5]
              << coral::MessageStream::endmsg;
	  return 1;
        }
      }
    }

    int numEventsAppend = 0;
    int numEvents = numEventsCreate;
    pool::DatabaseConnectionPolicy::Mode existingFileOpenMode
      = pool::DatabaseConnectionPolicy::OVERWRITE;
    pool::ICollection::OpenMode collectionOpenMode
      = pool::ICollection::CREATE_AND_OVERWRITE;
    if( argc > 6 )
    {
      if( strcmp(argv[6], "-") )
      {
        isInt = Util::to_value(argv[6], numEventsAppend);
        if( !isInt || (numEventsAppend<0) )
        {
          log << coral::Fatal << "Number of events to append must be a "
            << "positive integer. Got: " << argv[6] << coral::MessageStream::endmsg;
          return 1;
        }
        if( numEventsAppend > 0 )
        {
          numEvents = numEventsAppend;
          existingFileOpenMode = pool::DatabaseConnectionPolicy::UPDATE;
          collectionOpenMode = pool::ICollection::UPDATE;
        }
      }
    }

    int numTokenAttributes = 0;
    if( argc > 7 )
    {
      if( strcmp(argv[7], "-") )
      {
        isInt = Util::to_value(argv[7], numTokenAttributes);
        if( !isInt || (numTokenAttributes<0) || 
          numTokenAttributes > (numAttributes-2) )
        {
          log << coral::Fatal << "Number of attributes of type Token must "
            << "be a positive integer <= (total number of attributes - 2). "
              << "Got: " << argv[7] << coral::MessageStream::endmsg;
          return 1;
        }
      }
    }

    int skipPrintInterval = 1000;
    if( argc > 8 )
    {
      if( strcmp(argv[8], "-") )
      {
        isInt = Util::to_value(argv[8], skipPrintInterval);
        if( !isInt || (skipPrintInterval<1) )
        {
          log << coral::Fatal << "Number of events to process before "
            << "print to screen must be an integer > 0. Got: " << argv[8] 
            << coral::MessageStream::endmsg;
          return 1;
        }
      }
    }

    std::string logFileName = "Test_Write.log";
    if( argc > 9 )
      if( strcmp(argv[9], "-") )
        logFileName = argv[9];
    
    std::ofstream logFileStrm;
    logFileStrm.open( logFileName.c_str(), std::ios::out | std::ios::app );
    if( logFileStrm.bad() )
    {
      log << coral::Fatal << "Could not open file: " 
	<< logFileName << std::endl;
    }
    logFileStrm.setf(std::ios::fixed);

    double scale = pool::ProcessTime::MILLISEC;;
    std::string units = " ms.";
    int prec = 6;
    if( argc > 10 )
    {
      if( strcmp(argv[10], "millisec") && strcmp(argv[10], "-") )
      {
        if( !strcmp( argv[10], "seconds" ) )
        {
	  scale  = pool::ProcessTime::SECONDS;
	  units = " s.";
	  prec = 9;
        }
        else if( !strcmp( argv[10], "microsec" ) )
        {
	  scale  = pool::ProcessTime::MICROSEC;
	  units = " us.";
	  prec = 3;
        }
        else if( !strcmp( argv[10], "nanosec" ) )
        {
	  scale  = pool::ProcessTime::NANOSEC;
          units = " ns.";
	  prec = 0;
        }
        else
        {
	  log << coral::Fatal << "Argument " << argv[10]
	    << " is not a valid unit type." << coral::MessageStream::endmsg;
	  return 1;
        }
      }
    }

    int randSeed = 7232003;
    if( argc > 11 )
    {
      if( strcmp(argv[11], "-") )
      {
        if( !Util::to_value(argv[11], randSeed) )
        {
          log << coral::Fatal << "Random seed must be an integer. Got: " 
            << argv[11] << coral::MessageStream::endmsg;
          return 1;
        }
      }
    }

    try
    {
      const std::string platformDependentName =
        seal::SharedLibrary::libname( TEST_LIBRARY );
      seal::SharedLibrary::load( platformDependentName );
    }
    catch ( seal::SharedLibraryError* error )
    {
      throw std::runtime_error( error->explainSelf().c_str() );
    }

    pool::IFileCatalog* fileCatalog = new pool::IFileCatalog;
    pool::URIParser p( FILE_CATALOG_TYPE + "_file:" + FILE_CATALOG_NAME );
    p.parse();
    fileCatalog->setWriteCatalog( p.contactstring() );
    fileCatalog->connect();
    fileCatalog->start();

    pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fileCatalog );

    pool::DatabaseConnectionPolicy policy;
    policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE);
    policy.setWriteModeForExisting( existingFileOpenMode );
    policy.setReadMode( pool::DatabaseConnectionPolicy::UPDATE );
    dataSvc->session().setDefaultConnectionPolicy( policy );

    pool::Placement placement;
    placement.setDatabase( dbFile, pool::DatabaseSpecification::PFN );
    placement.setContainerName( containerName );
    placement.setTechnology( DB_TYPE.type() );

    std::vector<std::string> refAttrName(numTokenAttributes);
    std::vector<pool::Placement> placeRefAttr(numTokenAttributes);
    if( numTokenAttributes > 0 )
    {
      char attrNum[128];
      for(int i=0; i<numTokenAttributes; i++)
      {
        std::sprintf(attrNum, "%i", i+1);
        refAttrName[i] = "ref_";
        refAttrName[i] += attrNum;
      }

     log << coral::Info << "[OVAL] Collection_Update: Creating Placement "
        << " objects to store the Ref attributes." << coral::MessageStream::endmsg;
      std::string contName;
      for(int i=0; i<numTokenAttributes; i++)
      {
        std::sprintf(attrNum, "%i", i+1);
        placeRefAttr[i].setDatabase(dbFile, pool::DatabaseSpecification::PFN);
        contName = containerName;
        contName += "Ref";
        contName += attrNum;
        placeRefAttr[i].setContainerName( contName );
        placeRefAttr[i].setTechnology( DB_TYPE.type() );
      }
    }

    log << coral::Info << "Creating a collection object:" 
      << coral::MessageStream::endmsg;
    if( msgLevel < coral::Debug )
    {  
      std::cout << "[OVAL] Collection type is " << collectionType << std::endl;
      std::cout << "[OVAL] Collection location is " << collectionDb 
        << std::endl;
      std::cout << "[OVAL] Collection name is " << collectionName << std::endl;
      std::cout << "[OVAL] Collection open mode is " << collectionOpenMode 
	<< std::endl;
    }
    pool::ProcessTime t1;
    pool::Collection< SimpleTestClass > collection( dataSvc, collectionType,
      collectionDb, collectionName, collectionOpenMode );
    pool::ProcessTime t2;
    pool::ProcessTime::Clock_t openCollectionRealTime = 0.0;
    pool::ProcessTime::Clock_t openCollectionUserTime = 0.0;
    pool::ProcessTime::Clock_t openCollectionSystemTime = 0.0;
    openCollectionRealTime = (t2.getRealTime() - t1.getRealTime())*scale;
    openCollectionUserTime = (t2.getUserTime() - t1.getUserTime())*scale;
    openCollectionSystemTime = (t2.getSystemTime() - t1.getSystemTime())*scale;
    logFileStrm << "Total time to open the collection: " << std::endl;
    logFileStrm << std::setprecision( prec ) << "   ";
    logFileStrm << "Real: " << std::setw(14) << openCollectionRealTime 
      << units << "   User: " << std::setw(14) << openCollectionUserTime 
        << units << "   System: " << std::setw(14) << openCollectionSystemTime 
	  << units << std::endl;
    logFileStrm << std::endl;

    log << coral::Debug << "Getting the attribute list specification." 
      << coral::MessageStream::endmsg;
    pool::AttributeListSpecification& attribSpec = 
      collection.attributeListSpecification();

    char attribName[20];
    if( collectionOpenMode == pool::ICollection::CREATE_AND_OVERWRITE )
    {
      if( numAttributes > 0 ) attribSpec.push_back( "run", "int" );
      if( numAttributes > 1 ) attribSpec.push_back( "event", "int" );
      for( int i=1; i<=(numAttributes-numTokenAttributes-2); i++ )
      {
	std::sprintf( attribName, "attribute%i", i+2 );
        if( i % 2)
          attribSpec.push_back( attribName, "int" );
        else
	  attribSpec.push_back( attribName, "float" );
      }
      for( int i=0; i<numTokenAttributes; i++ )
      {
        attribSpec.push_back( refAttrName[i], "string" );
      }
    }

    if( msgLevel < coral::Debug )
    {
      std::cout << "[OVAL] ";
      attribSpec.print( std::cout );
      std::cout << std::endl;
    }

    log << coral::Info << "Starting event loop to add objects and "
      "meta data to the collection." << coral::MessageStream::endmsg;
    int attribIntVal = 0;
    float attribFloatVal = 0.0;
    pool::ProcessTime::Clock_t markWriteRealTime = 0.0;
    pool::ProcessTime::Clock_t markWriteUserTime = 0.0;
    pool::ProcessTime::Clock_t markWriteSystemTime = 0.0;
    pool::ProcessTime::Clock_t fillMetaDataRealTime = 0.0;
    pool::ProcessTime::Clock_t fillMetaDataUserTime = 0.0;
    pool::ProcessTime::Clock_t fillMetaDataSystemTime = 0.0;
    pool::ProcessTime::Clock_t addObjectRealTime = 0.0;
    pool::ProcessTime::Clock_t addObjectUserTime = 0.0;
    pool::ProcessTime::Clock_t addObjectSystemTime = 0.0;
    pool::ProcessTime::Clock_t commitCollectRealTime = 0.0;
    pool::ProcessTime::Clock_t commitCollectUserTime = 0.0;
    pool::ProcessTime::Clock_t commitCollectSystemTime = 0.0;
    int eventCounter = 0;
    for ( int run = 1; run <= 1; run++)
    {
      for ( int evt=1; evt <= numEvents; evt++ )
      {
	eventCounter++;

	sc = dataSvc->transaction().start( pool::ITransaction::UPDATE );
	if( !sc.isSuccess() )
	{
	log << coral::Fatal << "Error when starting database transaction."
	  << coral::MessageStream::endmsg;
	}

	log << coral::Debug << "Creating a SimpleTestClass object." 
	  << coral::MessageStream::endmsg;
	pool::Ref<SimpleTestClass> 
	  simpleTestClass( dataSvc, new SimpleTestClass() );

	log << coral::Debug << "Marking the object for writing."
	  << coral::MessageStream::endmsg;
	pool::ProcessTime t3;
	simpleTestClass.markWrite( placement );
	pool::ProcessTime t4;
	markWriteRealTime += t4.getRealTime() - t3.getRealTime();
	markWriteUserTime += t4.getUserTime() - t3.getUserTime();
	markWriteSystemTime += t4.getSystemTime() - t3.getSystemTime();

	log << coral::Debug << "Setting the object's data." 
          << coral::MessageStream::endmsg;
	simpleTestClass->setNonZero();

	if( msgLevel < coral::Debug )
	{
	  std::cout << "[OVAL] The data is: "; 
	  simpleTestClass->streamOut( std::cout );
	  std::cout << std::endl;
	}

	log << coral::Debug << "Setting the object's attributes." 
	  << coral::MessageStream::endmsg;
	pool::AttributeList attributeList( attribSpec );
	pool::ProcessTime t5;
        if ( numAttributes > 0 ) attributeList[ "run" ].setValue( run );
	if ( numAttributes > 1 ) attributeList[ "event" ].setValue( evt );
	srand( (int) (randSeed*sin(evt)) );
	for( int i=1; i<=(numAttributes-numTokenAttributes-2); i++ )
  	{
	  std::sprintf( attribName, "attribute%i", i+2);
	  if( i % 2)
	  {
	    attribIntVal = 1 + (int) (10.0*rand()/(RAND_MAX+1.0));
	    attributeList[ attribName ].setValue( attribIntVal );
	  }
	  else
	  {
	    attribFloatVal = rand()/(RAND_MAX+1.0);
	    attributeList[ attribName ].setValue( attribFloatVal);
	  }
        }
	for( int i=0; i<numTokenAttributes; i++ )
        {
          pool::Ref<SimpleTestClass>
            refAttr( dataSvc, new SimpleTestClass() );
          refAttr.markWrite( placeRefAttr[i] );
          refAttr->setNonZero();
          if( msgLevel < coral::Debug )
          {
            std::cout << "[OVAL] Ref attribute " << refAttrName[i] 
              << " has it's data set to ";
            refAttr->streamOut(std::cout);
            std::cout << std::endl;
          }
          attributeList[refAttrName[i]].setValue( refAttr.toString() );
        }

	pool::ProcessTime t6;
	fillMetaDataRealTime += t6.getRealTime() - t5.getRealTime();
	fillMetaDataUserTime += t6.getUserTime() - t5.getUserTime();
	fillMetaDataSystemTime += t6.getSystemTime() - t5.getSystemTime();

	if( msgLevel < coral::Debug )
	{
	  std::cout << "[OVAL] ";
	  attributeList.print( std::cout );
	  std::cout << std::endl;
	}

	if( !(eventCounter % numEvents) )
	{
	  logFileStrm << "Meta data of last event processed is:" << std::endl;
	  logFileStrm << "   ";
	  attributeList.print( logFileStrm );
	  logFileStrm << std::endl;
	  logFileStrm << std::endl;
	}

	log << coral::Debug << "Adding object and attributes to "
	  << "collection." << coral::MessageStream::endmsg;
	pool::ProcessTime t7; 
	collection.add( simpleTestClass, attributeList );
	pool::ProcessTime t8;
	addObjectRealTime += t8.getRealTime() - t7.getRealTime();
	addObjectUserTime += t8.getUserTime() - t7.getUserTime();
	addObjectSystemTime += t8.getSystemTime() - t7.getSystemTime();

	if( msgLevel <= coral::Debug )
	  log << coral::Debug << coral::MessageStream::endmsg;

	if( !(eventCounter % skipCommitInterval) )
	{
	  log << coral::Debug << "Committing last changes made to "
	    << "the collection." << coral::MessageStream::endmsg;
	  pool::ProcessTime t9;
	  collection.commit();
	  pool::ProcessTime t10;
	  commitCollectRealTime += (t10.getRealTime() - t9.getRealTime());
	  commitCollectUserTime += (t10.getUserTime() - t9.getUserTime());
	  commitCollectSystemTime += (t10.getSystemTime() - 
	    t9.getSystemTime());
	}

	sc = dataSvc->transaction().commit();
	if( !sc.isSuccess() )
	{
	  log << coral::Fatal << "Error when comitting database "
            << "transaction." << coral::MessageStream::endmsg;
	}

	if( !(eventCounter % skipPrintInterval) )
	  log << coral::Info << "Processed " << eventCounter 
	    << " events." << coral::MessageStream::endmsg;
      }
    }

    log << coral::Debug << "Committing any changes made to "
      << "the collection since last commit in loop." << coral::MessageStream::endmsg;
    pool::ProcessTime t11;
    collection.commit();
    pool::ProcessTime t12;
    commitCollectRealTime += (t12.getRealTime() - t11.getRealTime());
    commitCollectUserTime += (t12.getUserTime() - t11.getUserTime());
    commitCollectSystemTime += (t12.getSystemTime() - t11.getSystemTime());

    double markWriteAveRealTime = (markWriteRealTime/numEvents)*scale;
    double markWriteAveUserTime = (markWriteUserTime/numEvents)*scale;
    double markWriteAveSystemTime = (markWriteSystemTime/numEvents)*scale;
    double fillMetaDataAveRealTime = (fillMetaDataRealTime/numEvents)*scale;
    double fillMetaDataAveUserTime = (fillMetaDataUserTime/numEvents)*scale;
    double fillMetaDataAveSystemTime = 
      (fillMetaDataSystemTime/numEvents)*scale;

    double addObjectAveRealTime = (addObjectRealTime/numEvents)*scale;
    double addObjectAveUserTime = (addObjectUserTime/numEvents)*scale;
    double addObjectAveSystemTime = (addObjectSystemTime/numEvents)*scale;
    double commitCollectAveRealTime = (commitCollectRealTime/numEvents)*scale;
    double commitCollectAveUserTime = (commitCollectUserTime/numEvents)*scale;
    double commitCollectAveSystemTime = 
      (commitCollectSystemTime/numEvents)*scale;

    logFileStrm << "Average time to mark an object for writing: " << std::endl;
    logFileStrm << std::setprecision( prec ) << "   ";
    logFileStrm << "Real: " << std::setw(14) << markWriteAveRealTime << units 
      << "   User: " << std::setw(14) << markWriteAveUserTime << units 
	<< "   System: " << std::setw(14) << markWriteAveSystemTime 
	    << units << std::endl;
    logFileStrm << std::endl;

    logFileStrm << "Average time to fill an object's meta data: " << std::endl;
    logFileStrm << std::setprecision( prec ) << "   ";
    logFileStrm << "Real: " << std::setw(14) << fillMetaDataAveRealTime 
      << units << "   User: " << std::setw(14) << fillMetaDataAveUserTime 
        << units << "   System: " << std::setw(14) 
          << fillMetaDataAveSystemTime << units << std::endl;
    logFileStrm << std::endl;

    logFileStrm << "Average time to add an object to the collection: " 
      << std::endl;
    logFileStrm << std::setprecision( prec ) << "   ";
    logFileStrm << "Real: " << std::setw(14) << addObjectAveRealTime << units
      << "   User: " << std::setw(14) << addObjectAveUserTime << units
	<< "   System: " << std::setw(14) << addObjectAveSystemTime << units 
	  << std::endl;
    logFileStrm << std::endl; 

    logFileStrm << "Average time to commit changes made to the collection: " 
      << std::endl;
    logFileStrm << std::setprecision( prec ) << "   ";
    logFileStrm << "Real: " << std::setw(14) << commitCollectAveRealTime 
      << units << "   User: " << std::setw(14) << commitCollectAveUserTime 
      << units << "   System: " << std::setw(14) 
      << commitCollectAveSystemTime << units << std::endl;
    logFileStrm << std::endl;

    log << coral::Info << "Closing the collection." << coral::MessageStream::endmsg;
    pool::ProcessTime t13;
    collection.close();
    pool::ProcessTime t14;
    pool::ProcessTime::Clock_t closeCollectionRealTime = 0.0;
    pool::ProcessTime::Clock_t closeCollectionUserTime = 0.0;
    pool::ProcessTime::Clock_t closeCollectionSystemTime = 0.0;
    closeCollectionRealTime = (t14.getRealTime() - t13.getRealTime())*scale;
    closeCollectionUserTime = (t14.getUserTime() - t13.getUserTime())*scale;
    closeCollectionSystemTime = (t14.getSystemTime() 
      - t13.getSystemTime())*scale;
    logFileStrm << "Total time to close the collection: " << std::endl;
    logFileStrm << std::setprecision( prec ) << "   ";
    logFileStrm << "Real: " << std::setw(14) << closeCollectionRealTime 
      << units << "   User: " << std::setw(14) << closeCollectionUserTime 
        << units << "   System: " << std::setw(14) 
          << closeCollectionSystemTime << units << std::endl;
    logFileStrm << std::endl;
    logFileStrm << std::endl;
    logFileStrm << std::endl;

    std::cout << "Total number of events written: " << numEvents << std::endl;
    std::cout << "Number of attributes per event: " << numAttributes 
      << std::endl;

    std::cout << "Time to open collection: ";
    (t2-t1).printPrettyOn( std::cout );
    std::cout << std::endl;
    std::cout << "Time to loop through events: ";
    (t13-t2).printPrettyOn( std::cout );
    std::cout << std::endl;
    std::cout << "Time to close collection: ";
    (t14-t13).printPrettyOn( std::cout );
    std::cout << std::endl;
    std::cout << "Total time: ";
    (t14-t1).printPrettyOn( std::cout );
    std::cout << std::endl;

    fileCatalog->commit();

    delete dataSvc;

    fileCatalog->disconnect();
    delete fileCatalog;

    logFileStrm.close();

    log << coral::Info << "Returning from main()" << coral::MessageStream::endmsg;

    return 0;
  }
  catch( pool::Exception& poolException )
  {
    std::cerr << "pool::Exception: " << poolException.what() << std::endl;
    return 1;
  }
  catch( std::exception& exception )
  {
    std::cerr << "std::Exception: " << exception.what() << std::endl;
    return 1;
  }
  catch( ... )
  {
    std::cerr << "Unknown exception. " << std::endl;
    return 1;
  }
}
