#include "Collection/Collection.h"
#include "AttributeList/AttributeList.h"
#include "AttributeList/AttributeListSpecification.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"
#include "FileCatalog/IFCAction.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/Placement.h"
#include "StorageSvc/DbType.h"

#include "POOLCore/Exception.h"
#include "TestDictionary/SimpleTestClass.h"

#include "StorageSvc/DbStatus.h"
#include "SealBase/SharedLibrary.h"

#include <stdexcept>

const int NUM_FILES = 3;
const int NUM_RUNS_PER_FILE = 2;
const std::string TEST_LIBRARY = "TestDictionaryDict";
const std::string FILE_CATALOG_TYPE = "xmlcatalog";
const std::string FILE_CATALOG_NAME = "PoolFileCatalog.xml";
const std::string DB_FILE_PREFIX = "TestDbFile";
const std::string DB_FILE_SUFFIX = "pool";
const pool::DbType DB_TYPE = pool::ROOTTREE_StorageType;
const std::string CONTAINER_NAME_PREFIX = "TestContainer";
const std::string COLLECTION_NAME = "TestCollection";

int main(int argc, const char *argv[])
{
  try
  { 
    pool::DbStatus sc = pool::DbStatus::Success;
    char fileNum[2];
    int runCounter = 0;
    std::string collectionDb = "";

    if( argc < 2 )
    {
      std::cerr << "Collection_MultiFileUpdate: Incorrect usage. Must specify "
	<< " collection type as command line argument." << std::endl;
      return 1;
    }

    std::string collectionType = std::string( argv[1] );
    if( collectionType == "RootCollection" )
    {
      collectionDb = "";
    }
    else if( collectionType == "MySQLCollection" || collectionType ==
      "MySQLltCollection" )
    {
      collectionDb = 
        "mysql://regrTester:rT2003@lxshare070d.cern.ch/regression";
    }
    else if ( collectionType == "RelationalCollectionMySQL" )
    {
      collectionDb = "mysql://pcitdb59.cern.ch/ralcolls";
    }
    else if ( collectionType == "RelationalCollectionOracle" )
    {
      collectionDb = "oracle://devdb10.cern.ch/ral_writer";
    }
    else
    {
      std::cerr << "Collection_MultiFileUpdate: Collection type " 
        << collectionType << " is not presently handled by this program." 
          << std::endl;
      return 1;
    }

    std::cout << "[OVAL] Collection_MulitFileUpdate: Initialising the "
      << "SEAL services." << std::endl;
    
    // pool::POOLContext::setMessageVerbosityLevel( coral::Info );

    std::cout << "[OVAL] Collection_MultiFileUpdate: Loading the shared "
      << "libraries." << std::endl;
    try
    {
      const std::string platformDependentName =
        seal::SharedLibrary::libname( TEST_LIBRARY );
      seal::SharedLibrary::load( platformDependentName );
    }
    catch( seal::SharedLibraryError* error)
    {
      throw std::runtime_error( error->explainSelf().c_str() );
    }

    std::cout << "[OVAL] Collection_MultiFileUpdate: Creating a file catalog." 
      << std::endl;
    pool::IFileCatalog* fileCatalog = new pool::IFileCatalog;
    pool::URIParser p( FILE_CATALOG_TYPE + "_file:" + FILE_CATALOG_NAME );
    p.parse();
    fileCatalog->setWriteCatalog( p.contactstring() );
    fileCatalog->addReadCatalog( p.contactstring() );

    std::cout << "[OVAL] Collection_MultiFileUpdate: Establishing connection "
      << "to the file catalog." << std::endl;
    fileCatalog->connect();

    std::cout << "[OVAL] Collection_MultiFileUpdate: Starting the file "
      << "catalog transaction." << std::endl;
    fileCatalog->start();

    std::cout << "[OVAL] Collection_MultiFileUpdate: Creating a data service "
      << "instance." << std::endl;
    pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fileCatalog );
    if ( dataSvc ) { 
      std::cout << "[OVAL] Collection_MultiFileUpdate: Setting the database "
        << "connection policy for this session." << std::endl;
      pool::DatabaseConnectionPolicy policy;
      policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::UPDATE );
      policy.setReadMode( pool::DatabaseConnectionPolicy::UPDATE );
      dataSvc->session().setDefaultConnectionPolicy( policy );
      std::cout << "[OVAL] Collection_MultiFileUpdate: Write mode for an "
        << "existing database is set to: " 
          << policy.writeModeForExisting() << std::endl;
      std::cout << "[OVAL] Collection_MultiFileUpdate: Read mode for a "
        << "database is set to " << policy.readMode() << std::endl;

      std::vector<std::string> dbFile(NUM_FILES);
      std::cout << "[OVAL] Collection_MultiFileUpdate: Declaring " << NUM_FILES
        << " database files." << std::endl;
      for( int i=0; i<NUM_FILES; i++ ) {
	std::sprintf(fileNum, "%i", i+1);
	dbFile[i] = DB_FILE_PREFIX;
	dbFile[i] += fileNum;
	dbFile[i] += ".";
	dbFile[i] += DB_FILE_SUFFIX;
      }

      std::vector<pool::Placement> placement(NUM_FILES);
      std::string contName;
      for( int i=0; i<NUM_FILES; i++ )
      {
	std::cout << "[OVAL] Collection_MultiFileUpdate: Creating a "
          << "Placement object for file " << dbFile[i] << std::endl;
	placement[i].setDatabase( dbFile[i], pool::DatabaseSpecification::PFN);
	std::sprintf(fileNum, "%i", i+1);
	contName = CONTAINER_NAME_PREFIX;
	contName += fileNum;
	placement[i].setContainerName( contName );
	placement[i].setTechnology( DB_TYPE.type() );
      }

      std::cout << "[OVAL] Collection_MultiFileUpdate: Starting a database "
	<< "transaction for UPDATE." << std::endl;
      sc = dataSvc->transaction().start( pool::ITransaction::UPDATE );
      if( !sc.isSuccess() )
      {
	std::cout << "Collection_Read: Error when starting database "
          << "transaction." << std::endl;
      }

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_MultiFileUpdate: Creating a collection "
	<< "object for UPDATE." << std::endl;
      std::cout << "[OVAL] Collection type is " << collectionType
        << std::endl;
      std::cout << "[OVAL] Collection location is " << collectionDb 
        << std::endl;
      std::cout << "[OVAL] Collection name is " << COLLECTION_NAME
	<< std::endl; 
      pool::Collection< SimpleTestClass > collection( dataSvc, collectionType,
	collectionDb, COLLECTION_NAME, pool::ICollection::UPDATE ); 
      std::cout << std::endl;

      std::cout << "[OVAL] Collection_MultiFileUpdate: Retrieving the "
        << "attribute list specification." << std::endl;
      const pool::AttributeListSpecification& attribSpec = 
	collection.attributeListSpecification();
      std::cout << "[OVAL] ";
      attribSpec.print(std::cout);
      std::cout << std::endl;

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_MultiFileUpdate: Creating a collection " 
        << "iterator." << std::endl;
      pool::Collection< SimpleTestClass >::Iterator collIter1 = 
        collection.select();

      std::cout << "[OVAL] Collection_MultiFileUpdate: Looping over all "
        << "existing collection elements." << std::endl;
      std::cout << std::endl;

      while ( collIter1.next() ) 
      {
	std::cout << "[OVAL] New element found." << std::endl;

	std::cout << "[OVAL] The data is: ";
	collIter1->streamOut( std::cout );
	std::cout << std::endl;

	pool::AttributeList attributeList = collIter1.attributeList();
	std::cout << "[OVAL] ";
	attributeList.print(std::cout);
	std::cout << std::endl;

	attributeList["run"].getValue(runCounter);
      }

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_MultiFileUpdate: Adding objects and "
	<< "attributes to the collection." << std::endl;
      std::cout << std::endl;
      runCounter++;
      for( int i = 0; i<NUM_FILES; i++ )
      {
	std::cout << "Writing to file " << dbFile[i] << std::endl;
	std::cout << std::endl;
	for (int run = runCounter; run < runCounter+NUM_RUNS_PER_FILE; run++)
	{
	  for (int evt = 1; evt <= 5; evt++)
          {
            std::cout << "[OVAL] Creating a SimpleTestClass object." 
              << std::endl;
            pool::Ref<SimpleTestClass> 
	    simpleTestClass( dataSvc, new SimpleTestClass() );

            std::cout << "[OVAL] Marking the object for writing." << std::endl;
	    simpleTestClass.markWrite( placement[i] );

            std::cout << "[OVAL] Setting the object's data." << std::endl;
            simpleTestClass->setNonZero();

            std::cout << "[OVAL] The data is: ";
	    simpleTestClass->streamOut(std::cout);
	    std::cout << std::endl;

	    std::cout << "[OVAL] Setting the object's attributes." 
              << std::endl;
            pool::AttributeList attributeList(attribSpec);
	    attributeList[ "run" ].setValue( run );
            attributeList[ "event" ].setValue( evt );

	    std::cout << "[OVAL] ";
	    attributeList.print(std::cout);
	    std::cout << std::endl;

            std::cout << "[OVAL] Adding object and attributes to collection." 
              << std::endl; 
	    collection.add( simpleTestClass, attributeList );

	    std::cout << std::endl;
          }
	}
	runCounter += 2;
      }

      std::cout << "[OVAL] Collection_MultiFileUpdate: Rolling back last "
        << "changes made to the collection." << std::endl;
      collection.rollback();

      std::cout << "[OVAL] Collection_MultiFileUpdate: Committing all changes "
	<< "made to the collection." << std::endl;
      collection.commit();

      std::cout << "[OVAL] Collection_MultiFileUpdate: Testing temporary "
	<< "closing of collection." << std::endl;
      collection.close();
      std::cout << "[OVAL] Collection_MultiFileUpdate: Collection is open? " 
	<< collection.isOpen() << std::endl;

      std::cout << "[OVAL] Collection_MultiFileUpdate: Testing reopening of "
	<< "collection." << std::endl;
      collection.open();
      std::cout << "[OVAL] Collection_MultiFileUpdate: Collection is open? "
	<< collection.isOpen() << std::endl;

      std::cout << "[OVAL] Collection_MultiFileUpdate: Creating another "
	<< "collection iterator." << std::endl;
      pool::Collection< SimpleTestClass >::Iterator collIter2 = 
	collection.select();
      std::cout << "[OVAL] Collection_MultiFileUpdate: Looping over all "
	<< "collection elements to verify that the new elements were erased:" 
	  << std::endl;
      std::cout << std::endl;

      while ( collIter2.next() ) 
      {
	std::cout << "[OVAL] New element found." << std::endl;

	std::cout << "[OVAL] The data is: ";
	collIter2->streamOut( std::cout );
	std::cout << std::endl;

	pool::AttributeList attributeList = collIter2.attributeList();
	std::cout << "[OVAL] ";
	attributeList.print(std::cout);
	std::cout << std::endl;

	attributeList["run"].getValue(runCounter);
      }

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_MultiFileUpdate: Adding more objects "
	<< "and attributes to the collection:" << std::endl;
      std::cout << std::endl;
      runCounter++;
      for( int i = 0; i<NUM_FILES; i++ )
      {
	std::cout << "Writing to file " << dbFile[i] << std::endl;
	std::cout << std::endl;
	for (int run = runCounter; run < runCounter+NUM_RUNS_PER_FILE; run++)
	{
	  for (int evt = 1; evt <= 5; evt++)
          {
	    std::cout << "[OVAL] Creating a SimpleTestClass object." 
              << std::endl;
	    pool::Ref<SimpleTestClass> 
            simpleTestClass( dataSvc, new SimpleTestClass() );

	    std::cout << "[OVAL] Marking the object for writing." << std::endl;
	    simpleTestClass.markWrite( placement[i] );
	    if( evt == 1 )
	    {
	      fileCatalog->commit();
              fileCatalog->start();
	    }

	    std::cout << "[OVAL] Setting the object's data." << std::endl;
	    simpleTestClass->setNonZero();

	    std::cout << "[OVAL] The data is: ";
            simpleTestClass->streamOut(std::cout);
            std::cout << std::endl;

	    std::cout << "[OVAL] Setting the object's attributes."
	      << std::endl;
            pool::AttributeList attributeList(attribSpec);
	    attributeList[ "run" ].setValue( run );
	    attributeList[ "event" ].setValue( evt );

	    std::cout << "[OVAL] ";
	    attributeList.print(std::cout);
	    std::cout << std::endl;

	    std::cout << "[OVAL] Adding object and attributes to collection." 
	      << std::endl; 
            collection.add(simpleTestClass, attributeList);

	    std::cout << std::endl;
          }
	}
	runCounter += 2;
      }

      std::cout << "[OVAL] Collection_MultiFileUpdate: Committing all changes "
	<< "made to the collection." << std::endl;
      collection.commit();

      std::cout << "[OVAL] Collection_MultiFileUpdate: Trying to roll back "
        << "last changes made to collection after they were commited." 
          << std::endl;
      collection.rollback();

      std::cout << "[OVAL] Collection_MultiFileUpdate: Creating another "
	<< "collection iterator." << std::endl;
      pool::Collection< SimpleTestClass >::Iterator collIter3 =  
        collection.select();
      std::cout << "[OVAL] Collection_MultiFileUpdate: Looping over all "
        << "collection elements to verify that the new elements were not "
          << "rolled back:" << std::endl;
      std::cout << std::endl;

      while ( collIter3.next() ) 
      {
	std::cout << "[OVAL] New element found." << std::endl;

	std::cout << "[OVAL] The data is: ";
	collIter3->streamOut( std::cout );
	std::cout << std::endl;

	pool::AttributeList attributeList = collIter3.attributeList();
	std::cout << "[OVAL] ";
	attributeList.print(std::cout);
	std::cout << std::endl;
      }

      std::cout << "[OVAL] Collection_MultiFileUpdate: Closing the collection."
	<< std::endl;
      collection.close();

      std::cout << "[OVAL] Collection_MultiFileUpdate: Ending the database "
	<< "transaction." << std::endl;
      sc = dataSvc->transaction().commit();
      if( !sc.isSuccess() )
      {
	std::cout << "[OVAL] Collection_MultiFileUpdate: Error when ending "
          << "database transaction." << std::endl;
      }

      std::cout << "[OVAL] Collection_MultiFileUpdate: Committing all file "
	<< "catalog changes." << std::endl;
      fileCatalog->commit();

      std::cout << "[OVAL] Collection_MultiFileUpdate: Disconnecting from the "
	<< "database." << std::endl;
    }
    delete dataSvc;

    std::cout << "[OVAL] Collection_MultiFileUpdate: Disconnecting from the "
      << "file catalog." << std::endl;
    fileCatalog->disconnect();
    delete fileCatalog;

    std::cout << "[OVAL] Collection_MultiFileUpdate: Returning from main()" 
      << std::endl;
    return 0;
  }
  catch( pool::Exception& poolException )
  {
    std::cerr << "pool::Exception: " << poolException.what() << std::endl;
    return 1;
  }
  catch( std::exception& exception )
  {
    std::cerr << "std::Exception: " << exception.what() << std::endl;
    return 1;
  }
  catch( ... )
  {
    std::cerr << "Unknown exception. " << std::endl;
    return 1;
  }
}
