#include "Collection/Collection.h"
#include "Collection/CollectionProxy.h"
#include "AttributeList/AttributeList.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"
#include "FileCatalog/IFCAction.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"

#include "POOLCore/Exception.h"
#include "TestDictionary/SimpleTestClass.h"

#include "StorageSvc/DbStatus.h"
#include "SealBase/SharedLibrary.h"

#include <stdexcept>

namespace pool  {
   int numTokenInstances();
}

const std::string TEST_LIBRARY = "test_TestDictionaryDict";
const std::string FILE_CATALOG_TYPE = "xmlcatalog";
const std::string FILE_CATALOG_NAME = "PoolFileCatalog.xml";
const std::string DB_FILE = "TestDbFile.pool";
const std::string CONTAINER_NAME = "TestContainer";

int main(int argc, const char *argv[])
{
  try
  {
    pool::DbStatus sc = pool::DbStatus::Success;
    std::string collectionDb = "";
    std::string collection1Name = "TestCollection1";
    std::string collection2Name = "TestCollection2";
    unsigned int evtCounter = 0;

    if( argc < 2 )
    {
      std::cerr << "Collection_Read: Incorrect usage. Must specify "
	<< "collection type as command line argument." << std::endl;
      return 1;
    }

     const char* CoralEVar =  "CORAL_AUTH_PATH";
     ::setenv(CoralEVar, "/afs/cern.ch/sw/lcg/app/pool/db", 0 );
     char *authPath = ::getenv( CoralEVar );
     std::cout << "Using Coral Auth path: " << authPath << std::endl;

    std::string collectionType = std::string( argv[1] );
    if( collectionType == "RootCollection" )
    {
      collectionDb = "";
    }
    else if( collectionType == "MySQLCollection" || collectionType ==
      "MySQLltCollection" )
    {
      collectionDb = 
        "mysql://regrTester:rT2003@lxshare070d.cern.ch/regression";
    }
    else if ( collectionType == "RelationalCollectionMySQL" )
    {
       //collectionDb = "mysql://pcitdb59.cern.ch/ralcolls";
      collectionDb = "mysql://itrac424.cern.ch/RAL_WRITER";
      collectionType = "RelationalCollection";
    }
    else if ( collectionType == "RelationalCollectionOracle" )
    {
      collectionDb = "oracle://devdb10/ral_writer";
      collectionType = "RelationalCollection";
    }
    else if( collectionType == "ImplicitCollection" )
    {
      collectionDb = "PFN:"+DB_FILE;
      collection1Name = CONTAINER_NAME;
      collection2Name = CONTAINER_NAME;
    }
    else
    {
      std::cerr << "Collection_Read: Collection type " << collectionType
	<< " is not presently handled by this program." << std::endl;
      return 1;
    }

    std::cout << "[OVAL] Collection_Read: Initialising the SEAL services."
      << std::endl;
    
    // pool::POOLContext::setMessageVerbosityLevel( coral::Info );

    std::cout << "[OVAL] Collection_Read: Loading the shared libraries." 
      << std::endl;
    try
    {
      const std::string platformDependentName =
        seal::SharedLibrary::libname( TEST_LIBRARY );
      seal::SharedLibrary::load( platformDependentName );
    }
    catch ( seal::SharedLibraryError* error)
    {
       std::cerr<< "Failed shared library loading for " << TEST_LIBRARY << std::endl;
       throw std::runtime_error( error->explainSelf().c_str() );
    }

    std::cout << "[OVAL] Collection_Read: Creating a file catalog." 
      << std::endl;
    pool::IFileCatalog* fileCatalog = new pool::IFileCatalog;
    pool::URIParser p( FILE_CATALOG_TYPE + "_file:" + FILE_CATALOG_NAME );
    p.parse();
    fileCatalog->addReadCatalog( p.contactstring() );

    std::cout << "[OVAL] Collection_Read: Establishing a connection to the "
      << "file catalog." << std::endl;
    fileCatalog->connect();

    std::cout << "[OVAL] Collection_Read: Starting the file catalog "
      << "transaction." << std::endl;
    fileCatalog->start();

    std::cout << "[OVAL] Collection_Read: Creating a data service instance." 
      << std::endl;
    pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fileCatalog );
    if ( dataSvc ) 
    {
      std::cout << "[OVAL] Collection_Read: Setting the database connection "
        << "policy for this session." << std::endl;
      pool::DatabaseConnectionPolicy policy;
      policy.setReadMode( pool::DatabaseConnectionPolicy::READ );
      dataSvc->session().setDefaultConnectionPolicy( policy );
      std::cout << "[OVAL] Collection_Read: Read mode for a database is "
        << "set to: " << policy.readMode() << std::endl;

      std::cout << "[OVAL] Collection_Read: Starting a database transaction "
        << "for " << "READ." << std::endl;
      sc = dataSvc->transaction().start( pool::ITransaction::READ );
      if( !sc.isSuccess() )
      {
        std::cout << "[OVAL] Collection_Read: Error when starting database " 
		  << "transaction." << std::endl;
      }
      
      if(1) {
	 std::cout << std::endl;
	 std::cout << "[OVAL] Collection_Read: Opening a collection object for "
		   << "READ: " << std::endl;
	 std::cout << "[OVAL] Collection type is " << collectionType
		   << std::endl;
	 std::cout << "[OVAL] Collection location is " << collectionDb 
		   << std::endl;
	 std::cout << "[OVAL] Collection name is " << collection1Name
		   << std::endl;
	 pool::Collection< SimpleTestClass > collection1( dataSvc, collectionType,
							  collectionDb, collection1Name, pool::ICollection::READ );
	 std::cout << std::endl;

	 if(0) {
	    std::cout << std::endl;
	    std::cout << "[OVAL] Collection_Read: Creating a collection iterator "
		      << "for fetches without caching." << std::endl;
	    pool::Collection< SimpleTestClass >::Iterator coll1Iter1 = 
	       collection1.select( "", "", "FetchOne" );
	    std::cout << "[OVAL] Collection_Read: Looping over all collection "
		      << "elements:" << std::endl;

	    evtCounter = 0;
	    while ( coll1Iter1.next() ) 
	    {
	       std::cout << "[OVAL] New element found." << std::endl;

	       std::cout << "[OVAL] The data is: ";
	       coll1Iter1->streamOut( std::cout );
	       std::cout << std::endl;

	       if( collectionType == "ImplicitCollection" )	    {
		  evtCounter++;
		  std::cout << "[OVAL] Events counted: " << evtCounter << std::endl;
	       }
	       else	    {
		  pool::AttributeList attributeList = coll1Iter1.attributeList();
		  std::cout << "[OVAL] ";
		  attributeList.print( std::cout );
		  std::cout << std::endl;
	       }

	       std::string dbFileId = coll1Iter1.ref().token()->dbID();
	       std::string bestPFN;
	       std::string fileType;
	       pool::IFCAction action;
	       fileCatalog->setAction( action );
	       action.lookupBestPFN( dbFileId, pool::FileCatalog::READ, 
				     pool::FileCatalog::SEQUENTIAL, bestPFN, fileType );
	       std::cout << "[OVAL] Physical file name is " << bestPFN
			 << std::endl;
	       std::cout << "[OVAL] File type is " << fileType << std::endl;
	    }
	 }
	 if(1) {
	    std::cout << std::endl;
	    std::cout << "[OVAL] Collection_Read: Creating a collection iterator " 
		      << "for fetches with caching." << std::endl;
	    pool::Collection< SimpleTestClass >::Iterator coll1Iter2 = 
	       collection1.select( "", "", "FetchAll" );
	    std::cout << "[OVAL] Collection_Read: Looping over all collection "
		      << "elements:" << std::endl;

	    evtCounter = 0;
	    while ( coll1Iter2.next() ) 
	    {
	       std::cout << "[OVAL] New element found." << std::endl;

	       std::cout << "[OVAL] The data is: ";
	       coll1Iter2->streamOut( std::cout );
	       std::cout << std::endl;

	       if( collectionType == "ImplicitCollection" )	    {
		  evtCounter++;
		  std::cout << "[OVAL] Events counted: " << evtCounter << std::endl;
	       }
	       else	    {
		  pool::AttributeList attributeList = coll1Iter2.attributeList();
		  std::cout << "[OVAL] ";
		  attributeList.print( std::cout );
		  std::cout << std::endl;
	       }

	       std::string dbFileId = coll1Iter2.ref().token()->dbID();
	       std::string bestPFN;
	       std::string fileType;
	       pool::IFCAction action;
	       fileCatalog->setAction( action );  
	       action.lookupBestPFN( dbFileId, pool::FileCatalog::READ, 
				     pool::FileCatalog::SEQUENTIAL, bestPFN, fileType);
	       std::cout << "[OVAL] Physical file name is " << bestPFN
			 << std::endl;
	       std::cout << "[OVAL] File type is " << fileType << std::endl;
	    }
	 }

	 std::cout << "[OVAL] Collection_Read: Closing the collection." 
		   << std::endl;
	 collection1.close();

	 std::cout << "[OVAL] Collection_Read: Testing; Collection is open? " 
		   << collection1.isOpen() << std::endl;
      }

      

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_Read: Opening a collection proxy object "
		<< "for READ: " << std::endl;
      std::cout << "[OVAL] Collection type is " << collectionType
		<< std::endl;
      std::cout << "[OVAL] Collection location is " << collectionDb 
		<< std::endl;
      std::cout << "[OVAL] Collection name is " << collection1Name 
		<< std::endl;
      pool::CollectionProxy collectionProxy1( dataSvc, collectionType, 
					      collectionDb, collection1Name, pool::ICollection::READ );
      std::cout << std::endl;

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_Read: Creating a collection proxy "
        << "iterator for fetches without caching." << std::endl;
      pool::CollectionProxy::Iterator collProxy1Iter1 = 
	collectionProxy1.select( "", "", "FetchOne" );
      std::cout << "[OVAL] Collection_Read: Looping over all collection "
        << "proxy elements:" << std::endl;

      evtCounter = 0;
      while ( collProxy1Iter1.next() )
      {
	std::cout << "[OVAL] New element found." << std::endl;

	std::cout << "[OVAL] Is element of type SimpleTestClass? "
	  << collProxy1Iter1.isCurrentObjectOfType<SimpleTestClass>() 
	    << std::endl;

	std::cout << "[OVAL] The data is: ";
	collProxy1Iter1.ref<SimpleTestClass>()->streamOut( std::cout );
        std::cout << std::endl;

        if( collectionType == "ImplicitCollection" )
	{
          evtCounter++;
          std::cout << "[OVAL] Events counted: " << evtCounter << std::endl;
        }
        else
        {
	  pool::AttributeList attributeList = collProxy1Iter1.attributeList();
	  std::cout << "[OVAL] ";
	  attributeList.print( std::cout );
	  std::cout << std::endl;
        }

        std::string dbFileId = 
          collProxy1Iter1.ref<SimpleTestClass>().token()->dbID();
	std::string bestPFN;
	std::string fileType;
	pool::IFCAction action;
	fileCatalog->setAction( action );
	action.lookupBestPFN( dbFileId, pool::FileCatalog::READ, 
	  pool::FileCatalog::SEQUENTIAL, bestPFN, fileType );
	std::cout << "[OVAL] Physical file name is " << bestPFN
	  << std::endl;
        std::cout << "[OVAL] File type is " << fileType << std::endl;
      }

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_Read: Creating a collection proxy " 
	<< "iterator for fetches with caching." << std::endl;
      pool::CollectionProxy::Iterator collProxy1Iter2 = 
	collectionProxy1.select( "", "", "FetchAll" );
      std::cout << "[OVAL] Collection_Read: Looping over all collection "
        << "proxy elements:" << std::endl;

      evtCounter = 0;
      while ( collProxy1Iter2.next() ) 
      {
	std::cout << "[OVAL] New element found." << std::endl;

	std::cout << "[OVAL] Is element of type SimpleTestClass? "
	  << collProxy1Iter2.isCurrentObjectOfType<SimpleTestClass>() 
	    << std::endl;

	std::cout << "[OVAL] The data is: ";
	collProxy1Iter2.ref<SimpleTestClass>()->streamOut( std::cout );
	std::cout << std::endl;

        if( collectionType == "ImplicitCollection" )
	{
          evtCounter++;
          std::cout << "[OVAL] Events counted: " << evtCounter << std::endl;
        }
        else
        {
	  pool::AttributeList attributeList = collProxy1Iter2.attributeList();
          std::cout << "[OVAL] ";
          attributeList.print( std::cout );
          std::cout << std::endl;
        }

	std::string dbFileId = 
          collProxy1Iter2.ref<SimpleTestClass>().token()->dbID();
	std::string bestPFN;
	std::string fileType;
	pool::IFCAction action;
	fileCatalog->setAction( action );
	action.lookupBestPFN( dbFileId, pool::FileCatalog::READ, 
	  pool::FileCatalog::SEQUENTIAL, bestPFN, fileType );
	std::cout << "[OVAL] Physical file name is " << bestPFN
	  << std::endl;
	std::cout << "[OVAL] File type is " << fileType << std::endl;
      }

      std::cout << "[OVAL] Collection_Read: Closing the collection proxy." 
        << std::endl;
      collectionProxy1.close();

      std::cout << "[OVAL] Collection_Read: Testing; Collection proxy is "             << "open? " << collectionProxy1.isOpen() << std::endl;

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_Read: Opening a collection object for "
        << "READ of collection written via a collection proxy: " << std::endl;
      std::cout << "[OVAL] Collection type is " << collectionType
        << std::endl;
      std::cout << "[OVAL] Collection location is " << collectionDb 
        << std::endl;
      std::cout << "[OVAL] Collection name is " << collection2Name 
        << std::endl;
      pool::Collection< SimpleTestClass > collection2( dataSvc, collectionType,
	collectionDb, collection2Name, pool::ICollection::READ );
      std::cout << std::endl;

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_Read: Creating a collection iterator "
        << "for fetches without caching." << std::endl;
      pool::Collection< SimpleTestClass >::Iterator coll2Iter = 
	collection2.select( "", "", "FetchOne" );
      std::cout << "[OVAL] Collection_Read: Looping over all collection "
        << "elements:" << std::endl;

      evtCounter = 0;
      while ( coll2Iter.next() ) 
      {
	std::cout << "[OVAL] New element found." << std::endl;

	std::cout << "[OVAL] The data is: ";
	coll2Iter->streamOut( std::cout );
        std::cout << std::endl;

        if( collectionType == "ImplicitCollection" )
	{
          evtCounter++;
          std::cout << "[OVAL] Events counted: " << evtCounter << std::endl;
        }
        else
        {
	  pool::AttributeList attributeList = coll2Iter.attributeList();
	  std::cout << "[OVAL] ";
	  attributeList.print( std::cout );
	  std::cout << std::endl;
        }

        std::string dbFileId = coll2Iter.ref().token()->dbID();
	std::string bestPFN;
	std::string fileType;
	pool::IFCAction action;
        fileCatalog->setAction( action );
        action.lookupBestPFN( dbFileId, pool::FileCatalog::READ,
	  pool::FileCatalog::SEQUENTIAL, bestPFN, fileType );
	std::cout << "[OVAL] Physical file name is " << bestPFN
	  << std::endl;
        std::cout << "[OVAL] File type is " << fileType << std::endl;
      }

      std::cout << "[OVAL] Collection_Read: Closing the collection." 
        << std::endl;
      collection2.close();

      std::cout << "[OVAL] Collection_Read: Testing; Collection is open? " 
	<< collection2.isOpen() << std::endl;

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_Read: Opening a collection proxy "
        << "object for READ of collection written via a collection proxy: " 
          << std::endl;
      std::cout << "[OVAL] Collection type is " << collectionType
        << std::endl;
      std::cout << "[OVAL] Collection location is " << collectionDb 
        << std::endl;
      std::cout << "[OVAL] Collection name is " << collection2Name 
        << std::endl;
      pool::CollectionProxy collectionProxy2( dataSvc, collectionType,
	collectionDb, collection2Name, pool::ICollection::READ );
      std::cout << std::endl;

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_Read: Creating a collection proxy "
        << "iterator for fetches without caching." << std::endl;
      pool::CollectionProxy::Iterator collProxy2Iter = 
	collectionProxy2.select( "", "", "FetchOne" );
      std::cout << "[OVAL] Collection_Read: Looping over all collection "
        << "elements:" << std::endl;

      evtCounter = 0;
      while ( collProxy2Iter.next() ) 
      {
	std::cout << "[OVAL] New element found." << std::endl;

	std::cout << "[OVAL] Is element of type SimpleTestClass? "
	  << collProxy2Iter.isCurrentObjectOfType<SimpleTestClass>() 
	    << std::endl;

	std::cout << "[OVAL] The data is: ";
	collProxy2Iter.ref<SimpleTestClass>()->streamOut( std::cout );
        std::cout << std::endl;

        if( collectionType == "ImplicitCollection" )
	{
          evtCounter++;
          std::cout << "[OVAL] Events counted: " << evtCounter << std::endl;
        }
        else
        {
          pool::AttributeList attributeList = collProxy2Iter.attributeList();
	  std::cout << "[OVAL] ";
	  attributeList.print( std::cout );
	  std::cout << std::endl;
        }

        std::string dbFileId = 
          collProxy2Iter.ref<SimpleTestClass>().token()->dbID();
	std::string bestPFN;
	std::string fileType;
        pool::IFCAction action;
        fileCatalog->setAction( action );
        action.lookupBestPFN( dbFileId, pool::FileCatalog::READ,
          pool::FileCatalog::SEQUENTIAL, bestPFN, fileType );
	std::cout << "[OVAL] Physical file name is " << bestPFN
	  << std::endl;
        std::cout << "[OVAL] File type is " << fileType << std::endl;
      }

      std::cout << "[OVAL] Collection_Read: Closing the collection." 
        << std::endl;
      collectionProxy2.close();

      std::cout << "[OVAL] Collection_Read: Testing; Collection is open? " 
	<< collectionProxy2.isOpen() << std::endl;

      std::cout << "[OVAL] Collection_Read: Ending the datbase transaction." 
	<< std::endl;
      sc = dataSvc->transaction().commit();
      if( !sc.isSuccess() )
      {
	std::cout << "[OVAL] Collection_Read: Error when ending database "
          << "transaction." << std::endl;
      }

      std::cout << "[OVAL] Collection_Read: Disconnecting from the database." 
        << std::endl;
    
    }
    delete dataSvc;

    std::cout << "[OVAL] Collection_Read: Disconnecting from the file "
      "catalog." << std::endl;
    fileCatalog->disconnect();
    delete fileCatalog;


    std::cout << "[OVAL] Collection_Read: Tokens unreleased:" << pool::numTokenInstances() << std::endl;
    std::cout << "[OVAL] Collection_Read: Returning from main()" << std::endl;
    
    return 0;
  }
  catch( pool::Exception& poolException )
  {
    std::cerr << "pool::Exception: " << poolException.what() << std::endl;
    return 1;
  }
  catch( std::exception& exception )
  {
    std::cerr << "std::exception: " << exception.what() << std::endl;
    return 1;
  }
  catch( ... )
  {
    std::cerr << "Unknown exception. " << std::endl;
    return 1;
  }
}
