#include "POOLCore/Token.h"
#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/ISession.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "TNamed.h"
#include "TestRefClasses/TestClass.h"
#include "DataSvc_AppendWrite.h"
#include "Collection/Collection.h"

#ifndef WIN32
# include <unistd.h>
#endif

pool::AppendWriteTest::AppendWriteTest(
                       std::string& dbName1, const DbType& storage_type1,
                       std::string& dbName2, const DbType& storage_type2,
                       std::string& dbName3, const DbType& storage_type3) 
            :ObjectStorageTest(dbName1,storage_type1,dbName2,storage_type2),
             m_dbName3(dbName3),m_type3(storage_type3){
  m_sequence_type = 0;
}

void pool::AppendWriteTest::reWrite(unsigned int NC){

  std::cout << "starting reWrite procedure..."<<std::endl;
  Placement place1, place2;
  place1.setDatabase(m_dbName3,pool::DatabaseSpecification::PFN );
  place1.setContainerName("test3_cont_1");
  place1.setTechnology(m_type3.type());
  place2.setDatabase(m_dbName2,pool::DatabaseSpecification::PFN );
  place2.setContainerName("test2_cont_1");
  place2.setTechnology(m_type2.type());

  // Start an update transaction
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
    printError("test ERROR: Could not start an UPDATE transaction");
  }  

  std::vector<pool::Ref<TestClass> > refTst(NC);
  unsigned int i;
  for (i = 0; i < NC; ++i )   {
    char txt1[256], txt2[256];
    ::sprintf(txt1,"Name1 %d", i);
    ::sprintf(txt2,"Title %d", i);

    pool::Ref<TNamed> refNamed(m_dataSvc,new TNamed(txt1, txt2));

    OId id(i);
    pool::Ref<TestClass> reftmp(m_dataSvc,new TestClass(i,id));        
        
    refTst[i] = reftmp;
                   

    try{
      refNamed.markWrite(place1);
    } catch (const RefException& e){
      printException(e);        
    }
    try{
      refTst[i].markWrite(place2);
      m_tokens.push_back(refTst[i].toString());
    } catch (const RefException& e){
      printException(e);
    }
    refTst[i]->setTNamed(refNamed);  
    refTst[i]->setLabel("WRITE");
  }
  printM("Commit transaction...");
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the WRITE transaction." );
  }

  printM("Acid read-back test:");
  if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
    printError("test ERROR: Could not start a READ transaction." );
  }
  for(i=0;i<NC;i++){
    if(!compareExpected("WRITE",i,*refTst[i])) break;
  }
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the READ transaction." );
  }

  // disconnect 
  m_dataSvc->session().disconnectAll();

  printM("Clearing cache");
  m_dataSvc->clearCache();

  return;
}

void pool::AppendWriteTest::readTokens(unsigned int expected){
  std::cout << "starting READ using available tokens..."<<std::endl;
  if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
    printError("test ERROR: Could not start a READ transaction." );
  }
  unsigned int i = 0;
  for(i=0;i<m_tokens.size();i++){
    Ref<TestClass> refT(m_dataSvc,m_tokens[i]);
    if(!compareExpected("WRITE",refT->id(),*refT)) break;
  }

  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the READ transaction." );
  }
  std::cout << OVALTAG << "Read back correctly "<<i<<" objects. Expected="<<expected<<std::endl;
  if(i!=expected){
    printError("test ERROR: number of object retrieved different from expected");
  }
}

void pool::AppendWriteTest::selectTestSequence(SequenceType type){
  m_sequence_type = type;
}

void pool::AppendWriteTest::readDbContent(unsigned int n_expected){

  printM("Starting read procedure of full db contents"); 
  // Starting a read transaction
  if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
    printError("test ERROR: Could not start a READ transaction." );
  }

  // Open an implicit collection
  // Reading back the object(s) in a container.
  pool::Collection<TestClass> collection( m_dataSvc, "ImplicitCollection",
					     "PFN:test_AppendWrite2.root", 
                                              "test2_cont_1",
					     pool::ICollection::READ );
  pool::Collection<TestClass>::Iterator itObj = collection.select();
  
  unsigned int i=0;
  itObj.next();
  while ( itObj.isValid() ){
    i++;
    pool::Ref<TestClass> refObj = itObj.ref();

    unsigned int j = refObj->id();
    if(!compareExpected("WRITE",j,*refObj)) break;

    refObj = 0;
    itObj.next();
  }

  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the READ transaction." );
  }

  std::cout << OVALTAG <<  
    " Read back " << i << " Objects. Expected="<<n_expected << std::endl;
  if(i!=n_expected)
    printError("test ERROR: number of objects retrieved different from expected");

  // disconnect 
  m_dataSvc->session().disconnectAll();

  printM("Clear cache...");
  m_dataSvc->clearCache();
  printM("Read procedure completed");
  return;  
}

void pool::AppendWriteTest::execute(){
  switch(m_sequence_type){
    case WRITE: {
      std::cout << "EXECUTING WRITE TEST "<<std::endl;
      write(500);
      read(500,"WRITE");
      break;
    }
    case APPEND: {
      std::cout << "EXECUTING APPEND TEST "<<std::endl;
      reWrite(500);
      readTokens(1000);
      break;
    }
    case READ: {
      std::cout << "EXECUTING READ TEST "<<std::endl;
      readDbContent(1000);
      break;
    }
    case FULL: {
      std::cout << "EXECUTING FULL TEST "<<std::endl;
      write(500);
      read(500,"WRITE");
      reWrite(500);
      readTokens(1000);
      break;
    }
  }
}

int main(int argc, char** argv)     {
  typedef pool::AppendWriteTest Test;
  std::string f1 = "test_AppendWrite1.root";
  std::string f2 = "test_AppendWrite2.root";
  std::string f3 = "test_AppendWrite3.root";
  Test theTest(f1, pool::ROOTKEY_StorageType,
               f2, pool::ROOTKEY_StorageType,
               f3, pool::ROOTKEY_StorageType);
  if(argc > 1){
    if(argv[1][0]=='F'){
      theTest.selectTestSequence(Test::FULL);
    } else if(argv[1][0]=='W'){
      theTest.selectTestSequence(Test::WRITE);
    } else if(argv[1][0]=='U'){
      theTest.selectTestSequence(Test::APPEND);
    } else if(argv[1][0]=='R'){
      theTest.selectTestSequence(Test::READ);
    }
  } else {
    theTest.selectTestSequence(Test::FULL);
  }
  theTest.run();
  return 0;
}
