#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "POOLCore/Token.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "TestBase/TestBase.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "ooContObj.h"
#include "ooScopeObj.h"
#include "RTLoader.h"
#ifdef _WIN32
#include <process.h>
#else
#include <unistd.h>
#endif
#include <stdexcept>

namespace pool  {

  class CastDynamicTest : public TestBase {
    private:
    const std::string m_dbName1;
    const std::string m_dbName2;
    const DbType m_type1;
    const DbType m_type2;
    std::string m_fid1;
    std::string m_fid2;
    const Token* rtlTok;
    const Token* contTok;

    public: 
    CastDynamicTest(std::string& dbName1, const DbType& storage_type1,
                        std::string& dbName2, const DbType& storage_type2)
      :TestBase(),m_dbName1(dbName1),m_dbName2(dbName2),
       m_type1(storage_type1),m_type2(storage_type2), rtlTok(0),contTok(0){
      std::vector< std::string > libraries;
      libraries.push_back( "test_DataSvc_CastDynamicDict" );
      init( libraries );       
      m_fid1 = fileId(m_dbName1,m_type1);
      m_fid2 = fileId(m_dbName2,m_type2);
      if(m_fid1 == "" || m_fid2 == ""){
        printError("test ERROR: cannot retrieve file id");
        return;
      }
    }

    void insertScope(const Ref<ooScopeObj> contScope, const Ref<ooScopeObj> rtlScope ){
	contScope->myMap.insert(std::make_pair("RTloadername", rtlScope));    
    }

    Ref<ooScopeObj> extractScope(const Ref<ooScopeObj> contScope){
       Ref<ooScopeObj> rtlScope = contScope->myMap["RTloadername"];
       return rtlScope;
    }   

    /// write objects to the persistent store
    void write(long l) {
      // Start an update transaction
      if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
        printError("test ERROR: Could not start an UPDATE transaction");
      }  

        Placement place1, place2;
        place1.setDatabase(m_dbName1,pool::DatabaseSpecification::PFN );
        place1.setContainerName("test_cont_1");
        place1.setTechnology(m_type1.type());
        place2.setDatabase(m_dbName2,pool::DatabaseSpecification::PFN );
        place2.setContainerName("test_cont_2");
        place2.setTechnology(m_type2.type());

	Ref<ooContObj> cont(m_dataSvc, new ooContObj);
        cont.markWrite(place1);
	Ref<RTLoader> rtl(m_dataSvc, new RTLoader(l));
        rtl.markWrite(place2);

        contTok = cont.token();
        (const_cast<Token*>(contTok))->addRef();

        insertScope(cont, rtl);

      printM("Commit transaction...");
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the WRITE transaction." );
      }
      printM("transaction WRITE committed");
    }

    /// read objects from the persistent store
    void read(long l)        {
      // Starting a read transaction
      if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
        printError("test ERROR: Could not start a READ transaction." );
      }

      // Read first object
      pool::Ref<ooContObj> cont(m_dataSvc,*contTok);
      (const_cast<Token*>(contTok))->release();


      Ref<ooScopeObj> rtlScope = extractScope(cont);

      Ref<RTLoader> rtl;
      rtl.castDynamic(rtlScope); 

      if(!rtl) {
        printError("test ERROR: rtl null found!");
      } else {
        long v = rtl->value();
        if(v!=l) {
          printError("Test ERROR: wrong value received");
        } else {
          printM("OK, the value is correct.");
        }
      }
       
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the READ transaction." );
      }
      printM("transaction READ committed");
  
      printM("Clear cache...");
      m_dataSvc->clearCache();
      return;
    }

    void execute(){ 
      write(999);
      read(999);
    }
  };

}

int main(int, char**)     {
  typedef pool::CastDynamicTest Test;
  std::string f1 = "ref1.root";
  std::string f2 = "ref2.root";
  Test theTest(f1, pool::ROOTKEY_StorageType,f2, pool::ROOTKEY_StorageType );
  theTest.run();
  return 0;
}
