#include "RefTestClass.h"

#include "StressTestBase/StressTestApp.h"
#include "StressTestBase/SimpleAsciiCollection.h"

#include "POOLCore/Token.h"
#include "FileCatalog/URIParser.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "DataSvc/Ref.h"
#include "DataSvc/DataSvcFactory.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"

#include <iostream>
#include <fstream>
#include <string>

#include <unistd.h>


using std::cout;
using std::endl;
using std::string;
using std::ifstream;
using std::ios;

using pool::Ref;
using pool::Token;
using pool::URIParser;
using pool::IDataSvc;


class RefTestApp : public pool::StressTestApp {
public:
  RefTestApp(int argc, const char* argv[])
      : StressTestApp(argc,argv)
    { }

  int userMain();
  
protected:
  void CreateTests();
  void IsNull();
  void FromToken();

  void CopyTests();
  void CopyConstructor();
  void AssignmentOperator();

  void OperatorTests();
  void Ptr();
  void Not();
  void Star();

  void EqualityTests();
  void Equality();
  void Inequality();

  void MiscTests();
  void ContextReset();
  void CacheLocation();
private:
};






int RefTestApp::userMain()
{
  CreateTests();
  CopyTests();
  OperatorTests();
  EqualityTests();
  MiscTests();

  cout << "All done" << endl;
  
  return 0;
}






void RefTestApp::CreateTests()
{
  int firstInt = 1;
  int secondInt = 2;

  string step = "========== Ref Creation Tests =========="; ovalmsg(step);
  
  RefTestClass *pTestObj1 = new RefTestClass(firstInt);
  const RefTestClass *pConstTestObj1 = new RefTestClass(secondInt);

  step = "Check init of non const ref";
  if (pTestObj1->data() == firstInt) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);

  step = "Check init of const ref";
  if (pConstTestObj1->data() == secondInt) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);

  IsNull();
  FromToken();
}

void RefTestApp::IsNull()
{
  string step = "Check is_null method";

  RefTestClass *c = new RefTestClass(1);
  
  Ref<RefTestClass> r3;
  Ref<RefTestClass> r(context(),c);
  if ( r3.isNull() ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  if ( !(r.isNull()) ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);
}

void RefTestApp::FromToken()
{
  catalog().start();

  Ref<RefTestClass> r1(context(),new RefTestClass(1));

  string dbFile = "GenRefTest.root";
  //ifstream dbcheck (dbFile.c_str(),ios::in);
  //::unlink(dbFile.c_str());

  pool::Placement place1;
  place1.setDatabase(dbFile,pool::DatabaseSpecification::PFN);
  place1.setContainerName("cont_1");
  place1.setTechnology(pool::ROOTTREE_StorageType.type());
  
  string collFile = "GenRefTest.coll";
  ::unlink(collFile.c_str());
  pool::SimpleAsciiCollection coll(collFile);
  
  context()->transaction().start( pool::ITransaction::UPDATE );
  r1.markWrite(place1);
  coll.add(r1.token()->toString());
  coll.write();
  context()->transaction().commit();
  
  catalog().commit();



  string step = "Create a Ref from the externalised token string";
  Ref<RefTestClass> r4(context(),coll.collection()[0]);
  if ( r4 == r1 ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);

  step = "Check token-as-string returned from Ref";
  if ( r4.toString() == coll.collection()[0] ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
    cout << "From Ref " << r4.toString() << endl << "Externalised " << coll.collection()[0] << endl;
  }
  ovalmsg(step);

  step = "Check token ptr returned from Ref";
  const Token *tok = r4.token();
  if ( tok->toString() == coll.collection()[0] ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
    cout << "From Token " << tok->toString() << endl << "Externalised " << coll.collection()[0] << endl;
  }
  ovalmsg(step);

  step = "Check ptr to context returned from Ref";
  //if ( r4.dataSvc() == context() ) {
  //  step="PASSED "+step;
  //} else {
  //  step="FAILED "+step;
  //}
  step += " (This check commented out- no longer valid?)";
  ovalmsg(step);
}
  
void RefTestApp::CopyTests()
{
  string step = "========== Ref Copy Tests =========="; ovalmsg(step);

  CopyConstructor();
  AssignmentOperator();
}

void RefTestApp::CopyConstructor()
{
  string step = "Creating new Ref via Copy Constructor";

  RefTestClass *c = new RefTestClass(1);
  
  Ref<RefTestClass> r(context(),c);
  Ref<RefTestClass> s(r);
  
  if (r->data() == s->data()) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);
}

void RefTestApp::AssignmentOperator()
{
  string step = "Creating new Ref via Assignment Operator";

  RefTestClass *c = new RefTestClass(1);
  
  Ref<RefTestClass> r(context(),c);
  Ref<RefTestClass> s,t;
  t = s = r;

  if (r->data() == t->data()) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);
}

void RefTestApp::OperatorTests()
{
  string step = "========== Misc Operator Tests =========="; ovalmsg(step);

  Ptr();
  Not();
  Star();
}

void RefTestApp::Ptr()
{
  string step = "Check returned real pointer";

  RefTestClass *c = new RefTestClass(1);
  RefTestClass *pFoilObj = new RefTestClass(10);

  Ref<RefTestClass> r1(context(),c);
  if ( (r1.ptr()) == c ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  if (! ((r1.ptr()) == pFoilObj) ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);

  delete pFoilObj;
}

void RefTestApp::Not()
{
  string step = "Check ! operator";

  RefTestClass *c = new RefTestClass(1);
  
  Ref<RefTestClass> r3;
  Ref<RefTestClass> r(context(),c);
  if ( !r3 ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  if ( !(!r) ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);
}

void RefTestApp::Star()
{
  string step = "Check star operator";

  RefTestClass *c = new RefTestClass(1);
  
  Ref<RefTestClass> r(context(),c);
  if (*r == *c) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);
}

void RefTestApp::EqualityTests()
{
  string step = "========== Misc Equality Tests =========="; ovalmsg(step);

  Equality();
  Inequality();
}

void RefTestApp::Equality()
{
  string step = "Check Ref == Ref equality (checks the cache location?)";

  Ref<RefTestClass> r1(context(),new RefTestClass(1));
  Ref<RefTestClass> r2(context(),new RefTestClass(1));
  Ref<RefTestClass> r3(context(),new RefTestClass(2));
  Ref<int> r4(context(), new int(10));
  
  if ( r2 == r1 ) {
    step="FAILED "+step;
  } else {
    step="PASSED "+step;
  }
  if ( r3 == r1 ) {
    step="FAILED "+step;
  } else {
    step="PASSED "+step;
  }
  ovalmsg(step);
}

void RefTestApp::Inequality()
{
  Ref<RefTestClass> r1(context(),new RefTestClass(1));
  Ref<RefTestClass> r3(context(),new RefTestClass(2));

  RefTestClass *pTestObj = new RefTestClass(5);
  
  string step = "Check Ref != Ref equality";
  
  if ( r3 != r1 ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  if ( r3 != r3 ) {
    step="FAILED "+step;
  } else {
    step="PASSED "+step;
  }
  ovalmsg(step);
  
  step = "Check Ref != Obj equality ";
  
  if ( r3 != pTestObj ){
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);
  
  step = "Check Obj != Ref equality";
  
  if ( pTestObj != r3 ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);

  
  delete pTestObj;
}

void RefTestApp::MiscTests()
{
  string step = "========== Misc Tests =========="; ovalmsg(step);

  ContextReset();
  CacheLocation();
}

void RefTestApp::ContextReset()
{
  string step = "Check resetting of Ref context- no longer available";

//  Ref<RefTestClass> r4(context(),new RefTestClass(1));
//
//  IDataSvc* altContext = pool::DataSvcFactory::instance(catalog());
//  r4.setDataSvc(altContext);
//
//  if ( r4.dataSvc() == context() ) {
//    step="FAILED "+step;
//  } else {
//    step="PASSED "+step;
//  }
//  ovalmsg(step);
}

void RefTestApp::CacheLocation()
{
  string step = "TO DO? Check Cache location query (effectively checked by operator==)";

//  Ref<RefTestClass> r4(context(),new RefTestClass(1));
//  Ref<RefTestClass> r5(context(),new RefTestClass(1));
//
//  pool::CacheLocation *loc4 = r4.cacheLocation();
//  pool::CacheLocation *loc5 = r5.cacheLocation();
//
//  if ( *loc4 = *loc5 ) {
//    step="FAILED "+step;
//  } else {
//    step="PASSED "+step;
//  }
  
  ovalmsg(step);
}

  
// low level main
int main(int argc, const char *argv[])
{
  std::string lib("DataSvc_GeneralRefDict");
  
  std::cout << "Creating app" << std::endl;
  RefTestApp app(argc,argv);
  std::cout << "Loading libraries" << std::endl;
  app.loadLib(lib);
  std::cout << "Running app" << std::endl;
  try {
    app.run();
  } catch (...) {
    std::cout << "An exception was thrown" << std::endl;
  }
  return 0;
}
