#include "POOLCore/Token.h"
#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "TestBase/TestBase.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "TreeClass.h"
#include "TNamed.h"
#include <stdexcept>

#ifndef WIN32
# include <unistd.h>
#endif

namespace pool  {

  class ObjectTreeTest : public TestBase {
  private:
    const std::string m_dbName1;
    const std::string m_dbName2;
    const DbType m_type1;
    const DbType m_type2;
    std::string m_fid1;
    std::string m_fid2;
    pool::Guid m_treeClassGuid;
  public: 
    ObjectTreeTest(std::string& dbName1, const DbType& storage_type1,
      std::string& dbName2, const DbType& storage_type2)
      :TestBase(),m_dbName1(dbName1),m_dbName2(dbName2),
      m_type1(storage_type1),m_type2(storage_type2),m_treeClassGuid(){
        m_fid1 = fileId(m_dbName1,m_type1);
        m_fid2 = fileId(m_dbName2,m_type2);
        if(m_fid1 == "" || m_fid2 == ""){
          printError("test ERROR: cannot retrieve file id");
          return;
        }
      };

    bool compareExpected(unsigned int i,const TreeClass& obj){
      TreeClass t(i);   
      char txt1[256], txt2[256], txt3[256];
      ::sprintf(txt1,"Name1 %d", i);
      ::sprintf(txt2,"Title %d", i);
      ::sprintf(txt3,"%d", i);
      TNamed* nam = new TNamed(txt1,txt2);
      t.setTNamed(nam);
      std::string str = "TreeClass_";
      str +=txt3;
      t.setName(str);              
      bool comp = (t== obj);
      if(!comp){
        printError("test ERROR: object retrieved different from the expected one");
        printM("Expected:-->");
        t.dump();
        printM("Retrieved:-->");
        obj.dump();
      }
      return comp;
    }

    /// write objects to the persistent store
    void write(unsigned int NC){


      Placement place1, place2;  
      place1.setDatabase(m_dbName1,pool::DatabaseSpecification::PFN );
      place1.setContainerName("test1_cont_1");
      place1.setTechnology(m_type1.type());
      place2.setDatabase(m_dbName2,pool::DatabaseSpecification::PFN );
      place2.setContainerName("test2_cont_1");
      place2.setTechnology(m_type2.type());

      // Start an update transaction
      if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
        printError("test ERROR: Could not start an UPDATE transaction");
      }  

      std::vector<TNamed*> pNamed(NC);
      pool::Ref<TreeClass> refTree(m_dataSvc);
      std::vector<pool::Ref<TreeClass> > refs(NC);
      unsigned int i;
      for (i = 0; i < NC; ++i )   {
        char txt1[256], txt2[256], txt3[256];
        ::sprintf(txt1,"Name1 %d", i);
        ::sprintf(txt2,"Title %d", i);
        ::sprintf(txt3,"%d", i);

        pool::Ref<TNamed> refNamed(m_dataSvc,pNamed[i]=
          new TNamed(txt1, txt2));
        TreeClass* nextTree =  new TreeClass(i);
        if(refTree) refTree->setNext(nextTree);         
        refTree = nextTree;
        std::string str = "TreeClass_";
        str +=txt3;
        refTree->setName(str);

        try{
          refNamed.markWrite(place1);
        } catch (const RefException& e){
          printException(e);        
        }
        refs[i] = refTree;
        try{
          refs[i].markWrite(place2);
          m_treeClassGuid = refs[i].token()->classID();
        } catch (const RefException& e){
          printException(e);
        }
        refTree->setTNamed(refNamed); 
      }
      printM("Commit transactions...");
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the WRITE transaction." );
      }

      std::cout << "Written "<< i << " objects in the database\n";

      printM("Acid read-back test:");
      if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
        printError("test ERROR: Could not start a READ transaction." );
      }
      Ref<const TreeClass> tree;
      tree = refs[0];
      for(i=0;i<NC;i++){
        if(!tree){
          printError("test ERROR: object tree is null");
          break;
        }
        bool comp = compareExpected(i,*tree);
        if(!comp) break;
        tree = tree->nextTree();
      }
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the READ transaction." );
      }

      printM("Clearing cache");
      m_dataSvc->clearCache();
      for(i=0;i<NC;i++){
        try{
          refs[i]->named()->GetName();
          printError("test ERROR: Exception was not throw after cache clearing\n");
        }
        catch (const RefException& /* e */){  // expected behavor
        }
      }

      return;
    }

    /// read objects from the persistent store
    void read(unsigned int n_expected)        {
      if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
        printError("test ERROR: Could not start a READ transaction." );
      }

      Token *tok = new Token();
      tok->setDb(m_fid2);
      tok->setCont("test2_cont_1");
      tok->setTechnology(m_type2.type());
      tok->setClassID(m_treeClassGuid);
      tok->oid().first  = 1;
      tok->oid().second = 0;
      // Read first object
      pool::Ref<TreeClass> refObj(m_dataSvc,*tok);
      tok->release();
      pool::Ref<const TreeClass> tree;
      tree = refObj;
      unsigned int i;
      for ( i = 0; ; ++i )   {
        if(!tree) break;
        bool comp = compareExpected(i,*tree);
        if(!comp) break;
        tree = tree->nextTree();
      }

      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the READ transaction." );
      }

      std::cout << "Object in tree: "<< i << std::endl;

      if(i!=n_expected) {
        printError("test ERROR: number of object retrieved different from expected");
      }
      printM("Clear cache...");
      m_dataSvc->clearCache();
      return;
    }

    void execute(){
      write(1000);
      read(1000);
    }
  };

}

int main(int, char**)     {
  typedef pool::ObjectTreeTest Test;
  std::string f1 = "test1.root";
  std::string f2 = "test2.root";
  std::vector< std::string > libraries;
  libraries.push_back( "test_DataSvc_ObjectTreeDict" );
  Test::init( libraries );
  Test theTest(f1, pool::ROOTTREE_StorageType, f2, pool::ROOTTREE_StorageType);
  theTest.run();
  return 0;
}
