#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "POOLCore/Token.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "TestBase/TestBase.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "TestRefClasses/TestClass.h"
#include "TNamed.h"
#include <stdexcept>

#ifndef WIN32
# include <unistd.h>
#endif

namespace pool  {

  class PtrOwnershipTest : public TestBase {
    private:
    const std::string m_dbName1;
    const std::string m_dbName2;
    const DbType m_type1;
    const DbType m_type2;
    std::string m_fid1;
    std::string m_fid2;
    pool::Guid m_classGuid;

    bool compareExpected(unsigned int i,const TestClass& obj){
        OId oid(i);
        TestClass t(i,oid);   
	char txt1[256], txt2[256];
        ::sprintf(txt1,"Name1 %d", i);
        ::sprintf(txt2,"Title %d", i);
        TNamed* nam = new TNamed(txt1,txt2);
        t.setTNamed(nam);        
        bool comp = (t== obj);
        if(!comp){
          printError("test ERROR: object retrieved different from the expected one");
          printM("Expected:-->");
          t.dump();
          printM("Retrieved:-->");
          obj.dump();
        }
        return comp;
    }
    public: 
    PtrOwnershipTest(std::string& dbName1, const DbType& storage_type1,
                     std::string& dbName2, const DbType& storage_type2)
      :TestBase(),m_dbName1(dbName1),m_dbName2(dbName2),
       m_type1(storage_type1),m_type2(storage_type2),m_classGuid(){
      std::vector< std::string > libraries;
      libraries.push_back( "test_TestRefClassesDict" );
      init( libraries );       
      m_fid1 = fileId(m_dbName1,m_type1);
      m_fid2 = fileId(m_dbName2,m_type2);
      if(m_fid1 == "" || m_fid2 == ""){
        printError("test ERROR: cannot retrieve file id");
        return;
      }
    };

    /// write objects to the persistent store
    void write(unsigned int NC)       {
 
      Placement place1, place2;
      place1.setDatabase(m_dbName1,pool::DatabaseSpecification::PFN );
      place1.setContainerName("test1_cont_1");
      place1.setTechnology(m_type1.type());
      place2.setDatabase(m_dbName2,pool::DatabaseSpecification::PFN );
      place2.setContainerName("test2_cont_1");
      place2.setTechnology(m_type2.type());

      // Start an update transaction
      if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
        printError("test ERROR: Could not start an UPDATE transaction");
      }  

      std::vector<pool::Ref<TNamed> > refN(NC);
      std::vector<pool::Ref<TestClass> > refTst(NC);
      unsigned int i;
      for (i = 0; i < NC; ++i )   {
        char txt1[256], txt2[256];
        ::sprintf(txt1,"Name1 %d", i);
        ::sprintf(txt2,"Title %d", i);

        pool::Ref<TNamed> refNamed1(m_dataSvc,new TNamed(txt1, txt2));
        OId id(i);
        pool::Ref<TestClass> reftmp(m_dataSvc,new TestClass(i,id));        
        
      	refN[i] = refNamed1.ptr();        

        try{
          refNamed1.markWrite(place1);
        } catch (const RefException& e){
          printException(e);        
        }

        refTst[i] = reftmp;
        Ref<TestClass> r2(m_dataSvc,reftmp);
        try{
          reftmp.markWrite(place2);
          r2.markWrite(place2);
          m_classGuid = r2.token()->classID();
        } catch (const RefException& e){
          printException(e);
        }
        refTst[i]->setTNamed(refN[i]);   // Assign object pointer to reference
        if(reftmp.token() != 
           refTst[i].token()){
          printError("test ERROR: the cache location is not shared");
        }
     }
      printM("Commit transaction...");
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the WRITE transaction." );
      }

      printM("Read after commit test:");
      if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
        printError("test ERROR: Could not start a READ transaction." );
      }
      for(i=0;i<NC;i++){
 	char txt[256];
        ::sprintf(txt,"Name1 %d", i);
        bool comp = ::strcmp(txt,refN[i]->GetName())==0;
        if(!comp){
          printError("test ERROR: object retrieved different from the expected one");
	  std::cout << "Expected: "<< txt << std::endl;
	  std::cout << "Retrieved: "<< refN[i]->GetName() << std::endl;
          break;
        }
      }
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the READ transaction." );
      }
      printM("Clearing cache");
      m_dataSvc->clearCache();
      //dynamic_cast<CacheSvc&>(m_dataSvc->cacheSvc()).dumpCacheContent();
      printM("Read after cache clear test:");
      for(i=0;i<NC;i++){
 	char txt[256];
        ::sprintf(txt,"Name1 %d", i);
        bool comp = ::strcmp(txt,refN[i]->GetName())==0;
        if(!comp){
          printError("test ERROR: object retrieved different from the expected one");
	  std::cout << "Expected: "<< txt << std::endl;
	  std::cout << "Retrieved: "<< refN[i]->GetName() << std::endl;
          break;
        }
      }
      return;
    }

    /// read objects from the persistent store
    void read(unsigned int n_expected) {
      // Starting a read transaction
      if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
        printError("test ERROR: Could not start a READ transaction." );
      }

      unsigned int i;
      for ( i = 0;  ; ++i )   {

        Token *tok = new Token();
        // Since this example does not use collections,
        // we do it here by faking the tokens to be 
        // returned by the collections
        // 
        tok->setDb(m_fid2);
        tok->setCont("test2_cont_1");
        tok->setTechnology(m_type2.type());
        tok->setClassID(m_classGuid);
        tok->oid().first  = 1;
        tok->oid().second = i;

        // Read first object
      	pool::Ref<TestClass> refObj(m_dataSvc,*tok);
        tok->release();

        if( !refObj ){
          std::cout << OVALTAG 
            << " End of reading after " << i << " Objects." << std::endl;
          break;
        }

        if(!compareExpected(i,*refObj)) break;
        
      	pool::Ref<TestClass> refTmp;
        refTmp = refObj.ptr();
        refObj.reset();
        try{
          refTmp->named();
          printError("test ERROR: Exception was not throw after ref reset\n");
        } catch (const RefException& /* e */){  // expected behavor
        }
        if(!compareExpected(i,*refObj)) break;
           
        // testing Ref consistency
	pool::Ref<TestClass> reftmp(m_dataSvc);
        reftmp = refObj.ptr();
        if(reftmp.token()!=refObj.token()){
          printError("test ERROR: the token is not shared");
        }
        Ref<TestClass> r1;
        r1 = refObj.ptr();
        if(r1!=reftmp) printError("test ERROR: operator equals failed");       
        if(r1==0) printError("test ERROR: (1) operator != failed");
        if(!(r1!=0)) printError("test ERROR: (1) operator == failed");
        reftmp = r1;
        r1 = refObj;
        if(r1!=reftmp) printError("test ERROR: operator equals failed");       
        r1.clear();
        r1 = refObj.ptr(); 
        if(r1.token()!=refObj.token()){
          printError("test ERROR: the token is not shared (2)");
        }
        Ref<TestClass> r2 = r1;
        if(r2!=reftmp) printError("test ERROR: operator equals failed");       
        if(r2==0) printError("test ERROR: (2) operator != failed");
        if(!(r2!=0)) printError("test ERROR: (2) operator == failed");
        Ref<TestClass> r3;
        if(r3!=0) printError("test ERROR: (3) operator != failed");
        if(!(r3==0)) printError("test ERROR: (3) operator == failed");
        OId idobj(9999);
        Ref<TestClass> r4(m_dataSvc,new TestClass(9999,idobj));
        if(r4==0) printError("test ERROR: (4) operator != failed");
        if(!(r4!=0)) printError("test ERROR: (4) operator == failed");
        Ref<TestClass> r5(m_dataSvc,refObj.ptr());
        if(r5==0) printError("test ERROR: (5) operator != failed");
        if(!(r5!=0)) printError("test ERROR: (5) operator == failed");
              
 
        // Implicit object read and pointer conversion 
        const TNamed* pNam = refObj->named();   
 	char txt[256];
        ::sprintf(txt,"Name1 %d", i);
        bool comp = ::strcmp(txt,pNam->GetName())==0;
        if(!comp){
          printError("test ERROR: object retrieved different from the expected one");
	  std::cout << "Expected: "<< txt << std::endl;
	  std::cout << "Retrieved: "<< pNam->GetName() << std::endl;
          break;
        }

        Ref<TestClass> nRef;
        OId oid(i);
        nRef = new TestClass(i,oid);
        nRef->setTNamed(pNam);
        refObj = nRef;
        if(!compareExpected(i,*refObj)) break;

      }

      printM("First read cycle completed.");

      unsigned int nr = i;
      std::vector<Ref<TestClass> > refs(nr); 
      std::vector<Ref<TestClass> > reft(nr); 
      for ( i = 0;  ; ++i )   {
        Token *tok = new Token();
        // Since this example does not use collections,
        // we do it here by faking the tokens to be 
        // returned by the collections
        // 
        tok->setDb(m_fid2);
        tok->setCont("test2_cont_1");
        tok->setTechnology(m_type2.type());
        tok->setClassID(m_classGuid);
        tok->oid().first  = 1;
        tok->oid().second = i;

        // Read first object
      	pool::Ref<TestClass> refObj(m_dataSvc,*tok);
        tok->release();
        if( !refObj ){
          break;
        }
        refs[i] = refObj;
        reft[i] = refObj.ptr();
        if(!compareExpected(i,*refObj)) break;
      }

      m_dataSvc->resetCache();

      for ( i = 0; i<reft.size() ; ++i )   {
        try{
          reft[i]->named();
          printError("test ERROR: Exception was not throw after ref reset\n");
        } catch (const RefException& /* e */){  // expected behavor
        }      
      }

      for ( i = 0; i<refs.size() ; ++i )   {
        if(!compareExpected(i,*refs[i])) break;
      }

      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the READ transaction." );
      }

      if(i!=n_expected)
        printError("test ERROR: number of objects retrieved different from expected"); 

      std::cout << "Clear cache..." << std::endl;
      m_dataSvc->clearCache();
      return;
    }

    void execute(){
      write(1000);
      read(1000);
    }
  };

}

int main(int, char**)     {
  typedef pool::PtrOwnershipTest Test;
  std::string f1 = "test1.root";
  std::string f2 = "test2.root";
  Test theTest(f1, pool::ROOT_StorageType,f2,pool::ROOT_StorageType);
  theTest.run();
  return 0;
}
