#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "POOLCore/Token.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/ICacheSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "TestBase/TestBase.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "TestRefClasses/TestClass.h"
#include "TNamed.h"
#include <stdexcept>

#ifndef WIN32
# include <unistd.h>
#endif

namespace pool  {

  class SimpleMultiWriteTest : public TestBase {
    private:
    const DbType m_type;
    Token* tok1;
    Token* tok2;
    Token* tok3;
    bool compareExpected(std::string label, unsigned int i, const TestClass& obj){
      OId oid(i);
      TestClass t(i,oid);   
      char txt1[256], txt2[256];
      ::sprintf(txt1,"Name1 %d", i);
      ::sprintf(txt2,"Title %d", i);
      t.setLabel(label);
      bool comp = (t== obj);
      if(!comp){
        printError("test ERROR: object retrieved different from the expected one");
        printM("Expected:-->");
        t.dump();
        printM("Retrieved:-->");
        obj.dump();
      }
      return comp;
    }

    public: 
    SimpleMultiWriteTest(const DbType& storage_type)
                   :TestBase(),m_type(storage_type),
                    tok1(0),tok2(0),tok3(0){
      std::vector< std::string > libraries;
      libraries.push_back( "test_TestRefClassesDict" );
      init( libraries );       
    };

    void write1(){
      printM("Executing two mark_write operations");
      // Start an update transaction
      if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
        printError("test ERROR: Could not start an UPDATE transaction");
      }  
      OId id(99);
      pool::Ref<TestClass> a1(m_dataSvc, new TestClass(99,id));
      {
        Placement place;
        place.setDatabase("MultiWrite1.root",
                          pool::DatabaseSpecification::PFN );
        place.setContainerName("test1_cont1");
        place.setTechnology(m_type.type());
        a1.markWrite(place);
        tok1 = const_cast<Token*>(a1.token());
        tok1->addRef();
      }
      pool::Ref<TestClass> a2(m_dataSvc, a1.ptr());
      {
        Placement place;
        place.setDatabase("MultiWrite1.root",
                          pool::DatabaseSpecification::PFN );
        place.setContainerName("test1_cont2");
        place.setTechnology(m_type.type()); 
        a2.markWrite(place);
        tok2 = const_cast<Token*>(a2.token());
        if(!(*tok2==*tok1)){
          printError("test ERROR: token expected equals are different");
        }
        tok2->addRef();
      }
      printM("Commit transactions...");
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the WRITE transaction." );
      }
    }

    void write2(){
      printM("Executing mark_multiwrite operations");
      // Start an update transaction
      if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
        printError("test ERROR: Could not start an UPDATE transaction");
      }  
      OId id(99);
      pool::Ref<TestClass> a1(m_dataSvc, new TestClass(99,id));
      {
        Placement place;
        place.setDatabase("MultiWrite2.root",
                          pool::DatabaseSpecification::PFN );
        place.setContainerName("test2_cont1");
        place.setTechnology(m_type.type());
        a1.markWrite(place);
        tok1 = const_cast<Token*>(a1.token());
        tok1->addRef();
      }
      pool::Ref<TestClass> a2(m_dataSvc, a1.ptr());
      {
        Placement place;
        place.setDatabase("MultiWrite2.root",
                          pool::DatabaseSpecification::PFN );
        place.setContainerName("test2_cont2");
        place.setTechnology(m_type.type()); 
        a2.markMultiwrite(place);
        tok2 = const_cast<Token*>(a2.token());
        tok2->addRef();
        if(*tok2==*tok1){
          printError("test ERROR: token expected different are equals");
        }
      }
      pool::Ref<TestClass> a3(m_dataSvc, a2.ptr());
      {
        Placement place;
        place.setDatabase("MultiWrite2.root",
                          pool::DatabaseSpecification::PFN );
        place.setContainerName("test2_cont3");
        place.setTechnology(m_type.type()); 
        a3.markMultiwrite(place);
        tok3 = const_cast<Token*>(a3.token());
        tok3->addRef();
        if(*tok3==*tok1){
          printError("test ERROR: token expected different are equals");
        }
      }
      printM("Commit transactions...");
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the WRITE transaction." );
      }
    }

   /// read objects from the persistent store
   void read(){

     printM("Starting read procedure"); 
     // Starting a read transaction
     if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
       printError("test ERROR: Could not start a READ transaction." );
     }

     // Read objects
     if(tok1){
       pool::Ref<TestClass> refObj1(m_dataSvc,*tok1);
       tok1->release();
       if(!refObj1) {
         printError("test ERROR: tok1 not found in the DB");
       } else {
         if(!compareExpected("TestClass",99,*refObj1)) {
           printError("test ERROR: object different from expected");
         }
       }
     } else {
       printError("test ERROR: token tok1 is null");
     }

     if(tok2){
       pool::Ref<TestClass> refObj2(m_dataSvc,*tok2);
       tok2->release();
       if(!refObj2) {
         printError("test ERROR: tok1 not found in the DB");
       } else {
         if(!compareExpected("TestClass",99,*refObj2)) {
           printError("test ERROR: object different from expected");
         }
       }
     } else {
       printError("test ERROR: token tok2 is null");
     }

     if(tok3){
       pool::Ref<TestClass> refObj3(m_dataSvc,*tok3);
       tok3->release();
       if(!refObj3) {
         printError("test ERROR: tok1 not found in the DB");
       } else {
         if(!compareExpected("TestClass",99,*refObj3)) {
           printError("test ERROR: object different from expected");
         }
       }
     }

     if ( ! m_dataSvc->transaction().commit() ) {
       printError("test ERROR: Could not commit the READ transaction." );
     }

     printM("Clear cache...");
     m_dataSvc->clearCache();
     printM("Read procedure completed");
     return;
   }

   void writeTwoTransactions(){
      printM("Executing multiWrite in two transactions");
      //Token *t1, *t2, *t3, *t4;

      // Start an update transaction
      if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
        printError("test ERROR: Could not start an UPDATE transaction");
      }  
      pool::Ref<TNamed> refNamed(m_dataSvc,new TNamed("Name1", "Title1"));
      {
        Placement place;
        place.setDatabase("MultiWrite3.root",
                          pool::DatabaseSpecification::PFN );
        place.setContainerName("test3_cont0");
        place.setTechnology(m_type.type());
        refNamed.markMultiwrite(place);
      }
      OId id1(111);
      pool::Ref<TestClass> a1(m_dataSvc, new TestClass(111,id1));
      a1->setTNamed(refNamed);
      {
        Placement place;
        place.setDatabase("MultiWrite3.root",
                          pool::DatabaseSpecification::PFN );
        place.setContainerName("test3_cont1");
        place.setTechnology(m_type.type());
        a1.markMultiwrite(place);
        //t1 = const_cast<Token*>(a1.token());
        //t1->addRef();
      }
      OId id2(222);
      pool::Ref<TestClass> a2(m_dataSvc, new TestClass(222,id2));
      {
        Placement place;
        place.setDatabase("MultiWrite3.root",
                          pool::DatabaseSpecification::PFN );
        place.setContainerName("test3_cont2");
        place.setTechnology(m_type.type()); 
        a2.markMultiwrite(place);
        //t2 = const_cast<Token*>(a2.token());
        //t2->addRef();
      }
      OId id3(333);
      pool::Ref<TestClass> a3(m_dataSvc, new TestClass(333,id3));
      {
        Placement place;
        place.setDatabase("MultiWrite3.root",
                          pool::DatabaseSpecification::PFN );
        place.setContainerName("test3_cont3");
        place.setTechnology(m_type.type()); 
        a3.markMultiwrite(place);
        //t3 = const_cast<Token*>(a3.token());
        //t3->addRef();
      }
      printM("Commit transactions...");
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the WRITE transaction." );
      }
      // Start an update transaction
      if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
        printError("test ERROR: Could not start an UPDATE transaction");
      }  
      {
        Placement place;
        place.setDatabase("MultiWrite4.root",
                          pool::DatabaseSpecification::PFN );
        place.setContainerName("test4_cont0");
        place.setTechnology(m_type.type());
        refNamed.markMultiwrite(place);
      }
      {
        Placement place;
        place.setDatabase("MultiWrite4.root",
                          pool::DatabaseSpecification::PFN );
        place.setContainerName("test4_cont1");
        place.setTechnology(m_type.type());
        a1.markMultiwrite(place);
        //t4 = const_cast<Token*>(a1.token());
        //t4->addRef();
      }
      printM("Commit transactions...");
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the WRITE transaction." );
      }

      //printM("Clear cache");
      //m_dataSvc->clearCache();      
      //std::cout << OVALTAG << "Cache entries="<< 
      //  m_dataSvc->cacheSvc().cacheSize() << std::endl;
      //dynamic_cast<CacheSvc&>(m_dataSvc->cacheSvc()).dumpCacheContent();

      //if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::READ ) ) ) {
      //  printError("test ERROR: Could not start an UPDATE transaction");
      //}  
      //pool::Ref<TestClass> r1(m_dataSvc,*t1);
      //if(r1->id() != 111) printError("test ERROR: value retrieved different form expected");      
      //t1->release(); 
      //pool::Ref<TestClass> r2(m_dataSvc,*t2);
      //if(r2->id() != 222) printError("test ERROR: value retrieved different form expected");       
      //t2->release(); 
      //pool::Ref<TestClass> r3(m_dataSvc,*t3);
      //if(r3->id() != 333) printError("test ERROR: value retrieved different form expected");       
      //t3->release(); 
      //pool::Ref<TestClass> r4(m_dataSvc,*t4);
      //if(r4->id() != 111) printError("test ERROR: value retrieved different form expected");       
      //t4->release(); 
      //if ( ! m_dataSvc->transaction().commit() ) {
      //  printError("test ERROR: Could not commit the WRITE transaction." );
      //}

   }


    void execute(){
      write1();
      read();
      write2();
      read();
      writeTwoTransactions();
    }
  };

}

int main(int, char**)     {
  typedef pool::SimpleMultiWriteTest Test;
  //pool::ObjectDeletePolicy::GLOBAL_DEFAULT = 
  //      pool::ObjectDeletePolicy::DONOT_DELETE;
  Test theTest(pool::ROOTTREE_StorageType);
  theTest.run();
  return 0;
}
