#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/ISession.h"
#include "POOLCore/Token.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "TNamed.h"
#include "TestRefClasses/TestClass.h"
#include "DataSvc_WriteTwice.h"

#ifndef WIN32
# include <unistd.h>
#endif

pool::WriteTwiceTest::WriteTwiceTest( std::string& dbName1, const DbType& storage_type1,
                                      std::string& dbName2, const DbType& storage_type2 )
  :ObjectStorageTest(dbName1,storage_type1, dbName2,storage_type2),m_sequence_type(0),m_classGuid(){}

bool pool::WriteTwiceTest::compareExpected(unsigned int i,const TestClass& obj){
  OId oid(i);
  TestClass t(i,oid);
  int tmp = i;   
  if(obj.id()!=(long)i){
    tmp=obj.id();
    i = obj.id();
  }
  if(tmp!=-1){
    char txt1[256], txt2[256];
    ::sprintf(txt1,"Name1 %d", i);
    ::sprintf(txt2,"Title %d", i);
    TNamed* nam = new TNamed(txt1,txt2);
    t.setTNamed(nam);        
    t.setLabel("WRITE");        
  }
  bool comp = (t== obj);
  if(!comp){
    printError("test ERROR: object retrieved different from the expected one");
    printM("Expected:-->");
    t.dump();
    printM("Retrieved:-->");
    obj.dump();
  }
  return comp;
}

bool pool::WriteTwiceTest::compareExpected(std::string,unsigned int i,const TestClass& obj)
{
  return compareExpected(i,obj);
}

bool pool::WriteTwiceTest::compareExpected(unsigned int i,const TNamed& obj){
  char txt1[256], txt2[256];
  ::sprintf(txt1,"Name1 %d", i);
  ::sprintf(txt2,"Title %d", i);
  TNamed nam(txt1,txt2);
  bool comp = TestClass::equals(nam,obj);
  if(!comp){
    printError("test ERROR: object retrieved different from the expected one");
    printM("Expected:-->");
    std::cout << "Name:  "<< nam.GetName() << std::endl;
    std::cout << "Title: "<< nam.GetTitle()<< std::endl;
    printM("Retrieved:-->");
    std::cout << "Name:  "<< obj.GetName() << std::endl;
    std::cout << "Title: "<< obj.GetTitle()<< std::endl;
  }
  return comp;
}

void pool::WriteTwiceTest::reWrite(unsigned int NC, unsigned int offset){

  // Start an update transaction
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
    printError("test ERROR: Could not start an UPDATE transaction");
  }  

  std::vector<pool::Ref<const TNamed> > refTst(NC);
  unsigned int i;

  std::string fid2= fileId(m_dbName2,m_type2);
  if(fid2 == ""){
    printError("test ERROR: cannot retrieve file id");
    return;
  }
  
  if(NC>tokens().size()){
    printError("test ERROR: too many object requested...");
    NC = tokens().size();
  }

  for (i = 0; i < NC; ++i )   {

    // Read main object
    pool::Ref<TestClass> refObjRead(m_dataSvc,tokens()[i]);

    pool::Ref<const TNamed> refNam(m_dataSvc);
    if(refObjRead){
      refNam = refObjRead->named();

      refNam.token();
      Placement place = refNam.placement();

      char txt1[256], txt2[256];
      ::sprintf(txt1,"Name1 %d", i+offset);
      ::sprintf(txt2,"Title %d", i+offset);
      refNam = new TNamed(txt1,txt2);

      try{
        refNam.markWrite(place);
        m_classGuid = refNam.token()->classID();
      } catch (const RefException& e){
        printException(e);
      }

      refTst[i] = refNam;

    }

  }
  printM("Commit transaction...");
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the WRITE transaction." );
  }

  printM("Acid read-back test:");
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::READ ) ) ) {
    printError("test ERROR: Could not start an READ transaction");
  }  
  for(i=0;i<NC;i++){
    if(!compareExpected(i+offset,*refTst[i])) break;
  }
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the READ transaction." );
  }

  // disconnect 
  m_dataSvc->session().disconnectAll();

  printM("Clearing cache");
  m_dataSvc->clearCache();
  return;
}

/// read objects from the persistent store
void pool::WriteTwiceTest::readNamed(unsigned int n_expected){

  if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
    printError("test ERROR: Could not start a READ transaction." );
  }

  std::string fid1= fileId(m_dbName1,m_type1);
  if(fid1 == ""){
    printError("test ERROR: cannot retrieve file id");
    return;
  }

  unsigned int i;
  for ( i = 0; ; ++i )   {
    Token *tok = new Token();
    // Since this example does not use collections,
    // we do it here by faking the tokens to be 
    // returned by the collections
    // 
    tok->setDb(fid1);
    tok->setCont("test1_cont_1");
    tok->setTechnology(m_type1.type());
    tok->setClassID(m_classGuid);
    tok->oid().first  = 1;
    tok->oid().second = i;

    pool::Ref<TNamed> refNam(m_dataSvc,*tok);
    tok->release();

    if( !refNam ){
      std::cout << OVALTAG 
      << " End of reading after " << i << " Objects." << std::endl;
      break;
    }

    if(!compareExpected(i,*refNam)) break;

    refNam = 0;
  }

  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the READ transaction." );
  }

  if(i!=n_expected)
    printError("test ERROR: number of object retrieved different than expected");

  // disconnect 
  m_dataSvc->session().disconnectAll();

  printM("Clear cache...");
  m_dataSvc->clearCache();
  return;
}


void pool::WriteTwiceTest::selectTestSequence(SequenceType type){
  m_sequence_type = type;
}

void pool::WriteTwiceTest::execute(){
  switch(m_sequence_type){
    case WRITE: {
      write(500);
      read(500);
      break;
    }
    case APPEND: {
      reWrite(500,500);
      readNamed(1000);
      break;
    }
    case READ: {
      readNamed(1000);
      break;
    }
    case FULL: {
      write(500);
      read(500);
      reWrite(500,500);
      readNamed(1000);
      break;
    }
  }
}

int main(int argc, char** argv)     {
  typedef pool::WriteTwiceTest Test;
  std::string f1 = "test_WriteTwice1.root";
  std::string f2 = "test_WriteTwice2.root";
  Test theTest(f1, pool::ROOTKEY_StorageType,
               f2, pool::ROOTKEY_StorageType);
  if(argc > 1){
    if(argv[1][0]=='F'){
      theTest.selectTestSequence(Test::FULL);
    } else if(argv[1][0]=='W'){
      theTest.selectTestSequence(Test::WRITE);
    } else if(argv[1][0]=='U'){
      theTest.selectTestSequence(Test::APPEND);
    } else if(argv[1][0]=='R'){
      theTest.selectTestSequence(Test::READ);
    }
  } else {
    theTest.selectTestSequence(Test::FULL);
  }
  theTest.run();
  return 0;
}
