#include "DataRegression_CollectionBase/CollectionRegressionTestDriver.h"

#include "CollectionBase/ICollection.h"
#include "CollectionBase/ICollectionIterator.h"
#include "CollectionBase/ICollectionFactory.h"

#include "AttributeList/AttributeList.h"
#include "AttributeList/AttributeListSpecification.h"

#include "POOLCore/Token.h"
#include "POOLCore/Guid.h"


#include <iomanip>
#include <limits>
#include <stdexcept>

using namespace std;

namespace pool{
  
  CollectionRegressionTestDriver::CollectionRegressionTestDriver():
    TestKernel(),
    m_token(0),
    m_attributeList(0)
  {
    try{
      m_token = new Token;
      AttributeListSpecification* als = new AttributeListSpecification(); 
      als->push_back("bool","bool");
      als->push_back("short","short");
      als->push_back("unsigned short","unsigned short");
      als->push_back("int","int");
      als->push_back("unsigned int","unsigned int");
      als->push_back("long","long");
      als->push_back("unsigned long","unsigned long");
      als->push_back("double","double");
      als->push_back("float","float");
      als->push_back("string","string");
      
      m_attributeList = new AttributeList(boost::shared_ptr<const AttributeListSpecification>(als));
      
      POOLContext::loadComponent( "SEAL/Services/Message Service" );
      POOLContext::setMessageVerbosityLevel( coral::Info );
      
    } catch (std::exception& e){
      std::cerr << "test ERROR:" << e.what() << "\n";
      addError();
    } 
  }
  
  
  
  CollectionRegressionTestDriver::~CollectionRegressionTestDriver(){
    m_token->release();
    delete m_attributeList;    
  }
  
  
  
  void CollectionRegressionTestDriver::readObjects(const std::string& name,
						   const std::string& connection,
						   const std::string& type){
    try{
      ICollection* collection = createCollection(name, connection, type, ICollection::READ);
      ICollectionIterator* iter = collection->select();
      while(iter->next()){
	cout << OVALTAG 
	     << " Token: "
	     << iter->token()->toString()
	     << endl;
	printAttributeList( iter->attributeList(), cout);
	cout << endl;
      }
      delete iter;
      delete collection;
    }catch (std::exception& e){
      std::cerr << "test ERROR:" << e.what() << "\n";
      addError();
    } 
  }
  
  
  
  void CollectionRegressionTestDriver::readCollections(const std::string& name,
						       const std::string& connection,
						       const std::string& type){
    
    cout << name << connection << type << endl;
    
  }
  


  void CollectionRegressionTestDriver::writeFlat( const std::string& name,
						  const std::string& connection,
						  const std::string& type,
						  const int& howmany ){
    try{
      ICollection* collection = createCollection(name, connection, type, ICollection::CREATE_AND_OVERWRITE);
      for(int i=0; i<howmany; ++i){
	fakeTokenAndAttributeList();
	cout << OVALTAG
	     << " Token: "
	     << m_token->toString()
	     << endl;
	printAttributeList( *m_attributeList, cout);
	cout << endl;
	collection->add( *m_token, *m_attributeList);
      }
      collection->commit();
      delete collection;
    }catch (std::exception& e){
      std::cerr << "test ERROR:" << e.what() << "\n";
      addError();
    } 
  }
  

  
  void CollectionRegressionTestDriver::writeMulti( const std::string& name,
						   const std::string& connection,
						   const std::string& type,
						   const int& howmany ){
    
    cout << name << connection << type << howmany << endl;
    
  }
  

  
  void CollectionRegressionTestDriver::fakeTokenAndAttributeList(){
    
    static unsigned long counter = 0;
    
    ++counter;
    
    Guid guid;
    //Guid::create(guid);
    guid.fromString("08610A10-0001-0000-38C2-FFBF24BE3241DBID");
    m_token->setClassID(guid);
    m_token->setDb(guid.toString()+"DBID");
    m_token->setCont(guid.toString()+"CONTID");
    m_token->setTechnology(numeric_limits<long>::max()-counter);
    m_token->oid().first = numeric_limits<unsigned long>::max()-counter;
    m_token->oid().first = numeric_limits<unsigned long>::min()+counter;
    
    bool b = counter % 2;
    (*m_attributeList)["bool"].setValue(b);
    short s = counter%2 ? numeric_limits<short>::min()+counter : numeric_limits<short>::max()-counter;
    (*m_attributeList)["short"].setValue(s);
    unsigned short us = counter%2 ? numeric_limits<unsigned short>::min()+counter : numeric_limits<unsigned short>::max()-counter;
    (*m_attributeList)["unsigned short"].setValue(us);
    int i = counter%2 ? numeric_limits<int>::min()+counter : numeric_limits<int>::max()-counter;
    (*m_attributeList)["int"].setValue(i);
    unsigned int ui = counter%2 ? numeric_limits<unsigned int>::min()+counter : numeric_limits<unsigned int>::max()-counter;
    (*m_attributeList)["unsigned int"].setValue(ui);
    long l = counter%2 ? numeric_limits<long>::min()+counter : numeric_limits<long>::max()-counter;
    (*m_attributeList)["long"].setValue(l);
    unsigned long ul = counter%2 ? numeric_limits<unsigned long>::min()+counter : numeric_limits<unsigned long>::max()-counter;
    (*m_attributeList)["unsigned long"].setValue(ul);
    double d = (counter%2 ? -1.0 : 1.0)*counter*1.00000000000001;
    (*m_attributeList)["double"].setValue(d);
    float f = (counter%2 ? -1.0 : 1.0)*counter*1.00001;
    (*m_attributeList)["float"].setValue(f);
    (*m_attributeList)["string"].setValue(m_token->toString());
  }
  
  
  
  void CollectionRegressionTestDriver::printAttributeList(const AttributeList& attributeList, std::ostream& os){
    
    bool b;
    attributeList["bool"].getValue(b);
    os << OVALTAG << " bool: " << b << endl; 
    short s;
    attributeList["short"].getValue(s);
    os << OVALTAG << " short: " << s << endl; 
    unsigned short us;
    attributeList["unsigned short"].getValue(us);
    os << OVALTAG << " unsigned short: " << us << endl; 
    int i;
    attributeList["int"].getValue(i);
    os << OVALTAG << " int: " << i << endl; 
    unsigned int ui;
    attributeList["unsigned int"].getValue(ui);
    os << OVALTAG << " unsigned int: " << ui << endl; 
    long l;
    attributeList["long"].getValue(l);
    os << OVALTAG << " long: " << l << endl; 
    unsigned long ul;
    attributeList["unsigned long"].getValue(ul);
    os << OVALTAG << " unsigned long: " << ul << endl; 
    double d;
    attributeList["double"].getValue(d);
    os << setprecision(15);
    os << OVALTAG << " double: " << d << endl; 
    float f;
    attributeList["float"].getValue(f);
    os << setprecision(6);
    os << OVALTAG << " float: " << f << endl; 
    os << OVALTAG << " string: " << attributeList["string"].getValueAsString() << endl;
    
  }
  
  
  
  ICollection* CollectionRegressionTestDriver::createCollection( const std::string& name,
								 const std::string& connection,
								 const std::string& type,
								 const ICollection::OpenMode& mode){
    try{
      ICollectionFactory* factory = ICollectionFactory::get();
      if (!factory) 
	throw std::runtime_error( "Could not retrieve a collection factory" );
      ICollection* coll = factory->create( type, 0, connection, name, mode );
      if (!coll) 
	throw std::runtime_error( "Could not create a collection" );
      return coll;
    } catch (std::exception& e){
      std::cerr << "test ERROR:" << e.what() << "\n";
      addError();
    } 
    return 0;
  }
}
