//#include "Reflex/Reflex.h"
#include "POOLCore/Token.h"
#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/ISession.h"
#include "DataSvc/Ref.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/ICacheSvc.h"
#include "DataSvc/RefException.h"
#include "TNamed.h"
#include "TestRefClasses/TestClass.h"
#include "ObjectStorageTest/ObjectStorageTest.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"

#ifndef WIN32
# include <unistd.h>
#endif

namespace pool_test {
  struct MyClass {
    int myData;
  };
}

bool pool::ObjectStorageTest::compareExpected(std::string label,unsigned int i,const TestClass& obj){
  OId oid(i);
  TestClass t(i,oid);   
  char txt1[256], txt2[256];
  ::sprintf(txt1,"Name1 %d", i);
  ::sprintf(txt2,"Title %d", i);
  TNamed* nam = new TNamed(txt1,txt2);
  t.setTNamed(nam);        
  t.setLabel(label);
  bool comp = (t== obj);
  if(!comp){
    printError("test ERROR: object retrieved different from the expected one");
    printM("Expected:-->");
    t.dump();
    obj.dump();
  }
  return comp;
}

pool::ObjectStorageTest::ObjectStorageTest(std::string& dbName1, const DbType& storage_type1,
                                           std::string& dbName2, const DbType& storage_type2)
                                           :TestBase(),m_dbName1(dbName1),m_dbName2(dbName2),
                                           m_type1(storage_type1),m_type2(storage_type2),m_tokens(){
  std::vector< std::string > libraries;
  libraries.push_back( "test_TestRefClassesDict" );
  std::cout << "init libraries " << std::endl;
  TestBase::init( libraries );
  std::cout << "libreries loaded." <<std::endl;
}

pool::ObjectStorageTest::~ObjectStorageTest(){}

/// write objects to the persistent store
void pool::ObjectStorageTest::write(unsigned int NC){
 
  printM("Starting write procedure"); 
  // construct placement objects
  Placement place1, place2;
  place1.setDatabase(m_dbName1,pool::DatabaseSpecification::PFN );
  place1.setContainerName("test1_cont_1");
  place1.setTechnology(m_type1.type());
  place2.setDatabase(m_dbName2,pool::DatabaseSpecification::PFN );
  place2.setContainerName("test2_cont_1");
  place2.setTechnology(m_type2.type());

  // Start an update transaction
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
    printError("test ERROR: Could not start an UPDATE transaction");
  }  

  // start a cycle of NC object created and stored in root files
  unsigned int i;
  for (i = 0; i < NC; ++i )   {
    // set up a text specifying the object index
    char txt1[256], txt2[256];
    ::sprintf(txt1,"Name1 %d", i);
    ::sprintf(txt2,"Title %d", i);

    // construct an object with a root class - Tnamed is used as an example,
    // to prove that root classes are supported like any other object type
    // A pool::Ref<T> object is costructed with the current dataSvc pointer (created in TestBase) 
    // and with the TNamed
    pool::Ref<TNamed> refNamed(m_dataSvc,new TNamed(txt1, txt2));
    //Construct Oid object - is a custom object, to give to TestClass some internal structure!
    OId id(i);
    // same as before: a Ref is costructed with the dataSvc and the newly constructed TestClass
    pool::Ref<TestClass> reftmp(m_dataSvc,new TestClass(i,id));           
 
    // Now make persistent the TNamed and the TestClass
    try{
      // register for writing in the dedicate place
      refNamed.markWrite(place1);
    } catch (const RefException& e){
      printException(e);        
    }

    try{
      // register for writing in the dedicated place
      reftmp.markWrite(place2);
      m_tokens.push_back(reftmp.toString());
    } catch (const RefException& e){
      printException(e);
    }

    // any operation on the pointers can be done even after the registration  
    reftmp->setLabel("WRITE");
    // set association relationship: the TNamed Ref<T> object is a data member of TestClass 
    reftmp->setTNamed(refNamed); // Assign object pointer to reference
  
  }

  // an additional operation outside the loop, to test an error condition
  // here we want to persistify an instance of the internal class MyClass 
  pool::Ref<pool_test::MyClass> refMyClass(m_dataSvc,new pool_test::MyClass);
  try{
    refMyClass.markWrite(place2);
    // MyClass is not in the dictionary: an exception must be thrown!
    printError("test ERROR: expected exception has been not thrown.");
  } catch (const pool::Exception& e){
    std::cout << OVALTAG << "Expected exception thrown."<< std::endl;
  }

  // check the number of marked objects: should be NC (TestClass) + NC (TNAmed)
  if(m_dataSvc->cacheSvc().numberOfMarkedObjects()!=NC*2){
    printError("test ERROR: wrong number of marked objects");
    std::cout << "Number of Object marked = "<< m_dataSvc->cacheSvc().numberOfMarkedObjects() << std::endl;
  }

  // commit the pool global transaction
  printM("Commit transactions...");
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the WRITE transaction." );
  }

  // the number of object register for write should be now 0!
  if(m_dataSvc->cacheSvc().numberOfMarkedObjects()!=0){
    printError("test ERROR: wrong number of marked objects");
    std::cout << "Number of Object marked after commit = "<< m_dataSvc->cacheSvc().numberOfMarkedObjects() << std::endl;
  }

  // disconnect
  m_dataSvc->session().disconnectAll();

  std::cout << OVALTAG << "Written on disk "<< NC << " objects\n";

  printM("Write procedure completed");
  return;
}


/// update objects on the persistent store
void pool::ObjectStorageTest::update(unsigned int n_expected,std::string expectedLabel){

  pool::DatabaseConnectionPolicy policy;
  m_dataSvc->session().setDefaultConnectionPolicy( policy );

  printM("Starting update procedure"); 
  // Starting a read transaction
  if ( ! m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) {
    printError("test ERROR: Could not start a UPDATE transaction." );
  }

  std::string fid2= fileId(m_dbName2,m_type2);
  if(fid2 == ""){
    printError("test ERROR: cannot retrieve file id");
    return;
  }

  unsigned int i=0;
  for ( i = 0; i<m_tokens.size() ; ++i )   {
    std::string tokString = m_tokens[i];

    // Read main object
    pool::Ref<TestClass> refObj(m_dataSvc,tokString);   

    if(!refObj ){
      printM("reference to requested object is empty.");
    }

    if(!compareExpected(expectedLabel,i,*refObj)) break;

    refObj.markUpdate();
    refObj->setLabel("UPDATE");

  }

  std::cout << OVALTAG 
      << " End of reading after " << i << " Objects." << std::endl;

  if(i!=n_expected)
    printError("test ERROR: number of objects retrieved different from expected");

  pool::Ref<pool_test::MyClass> refMyClass(m_dataSvc,new pool_test::MyClass);
  if(refMyClass.markUpdate()){
    printError("test ERROR: wrong expected exception return value.");
  } 

  if(m_dataSvc->cacheSvc().numberOfMarkedObjects()!=i){
    printError("test ERROR: wrong number of marked objects");
    std::cout << "Number of Object marked = "<< m_dataSvc->cacheSvc().numberOfMarkedObjects() << std::endl;
  }

  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the UPDATE transaction." );
  }

  if(m_dataSvc->cacheSvc().numberOfMarkedObjects()!=0){
    printError("test ERROR: wrong number of marked objects");
    std::cout << "Number of Object marked = "<< m_dataSvc->cacheSvc().numberOfMarkedObjects() << std::endl;
  }

  // disconnect 
  m_dataSvc->session().disconnectAll();

  printM("Clear cache...");
  m_dataSvc->clearCache();
  printM("Update procedure completed");
  return;
}

/// update objects on the persistent store
void pool::ObjectStorageTest::erase(unsigned int n_expected,std::string expectedLabel){

  pool::DatabaseConnectionPolicy policy;
  m_dataSvc->session().setDefaultConnectionPolicy( policy );

  printM("Starting delete procedure"); 
  // Starting an update transaction
  if ( ! m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) {
    printError("test ERROR: Could not start a UPDATE transaction." );
  }

  std::string fid2= fileId(m_dbName2,m_type2);
  if(fid2 == ""){
    printError("test ERROR: cannot retrieve file id");
    return;
  }

  unsigned int i=0;
  for ( i = 0; i<m_tokens.size() ; ++i )   {
    std::string tokString = m_tokens[i];

    // Read main object
    pool::Ref<TestClass> refObj(m_dataSvc,tokString);   

    if(!refObj ){
      printM("reference to requested obejct is empty.");
    }

    if(!compareExpected(expectedLabel,i,*refObj)) break;

    refObj.markDelete();
    refObj->setLabel("DELETE");

  }

  std::cout << OVALTAG << " End of reading after " << i << " Objects." << std::endl;

  if(i!=n_expected)
    printError("test ERROR: number of valid objects to delete different from expected");


  pool::Ref<pool_test::MyClass> refMyClass(m_dataSvc,new pool_test::MyClass);
  if(refMyClass.markDelete()){
    printError("test ERROR: wrong expected return value");
  } 

  if(m_dataSvc->cacheSvc().numberOfMarkedObjects()!=i){
    printError("test ERROR: wrong number of marked objects");
    std::cout << "Number of Object marked = "<< m_dataSvc->cacheSvc().numberOfMarkedObjects() << std::endl;
  }

  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the UPDATE transaction." );
  }

  if(m_dataSvc->cacheSvc().numberOfMarkedObjects()!=0){
    printError("test ERROR: wrong number of marked objects");
    std::cout << "Number of Object marked for delete after commit= "<< m_dataSvc->cacheSvc().numberOfMarkedObjects() << std::endl;
  }

  if(i!=n_expected){
    printError("test ERROR: number of objects retrieved different from expected");
    std::cout << "Expected:" << n_expected << "; retrieved:"<< i <<";"<<std::endl;
  }

  // disconnect 
  m_dataSvc->session().disconnectAll();

  printM("Clear cache...");
  m_dataSvc->clearCache();
  printM("Erase procedure completed");
  return;
}

/// read objects from the persistent store
void pool::ObjectStorageTest::read(unsigned int n_expected,std::string label){

  printM("Starting read procedure"); 
  // Starting a read transaction
  if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
    printError("test ERROR: Could not start a READ transaction." );
  }

  unsigned int i=0;
  for ( i = 0; i<m_tokens.size();)   {

    std::string tokString = m_tokens[i];

    // Read main object using the token constructed 'by hand'
    pool::Ref<TestClass> refObj(m_dataSvc,tokString);

    // check if something is referenced inside the db
    if(!refObj ){
      printM("ref read is empty.");
      break;
    }
    
    // check the object retrieved
    if(!compareExpected(label,i,*refObj)) break;

    bool okRef = false;

    // verify consistency of associated object (a few of cross checks)
    Ref<const TNamed> refNam;
    refNam = refObj->named();
    if ( refNam )   {
      if ( ! (! refNam) )   {
        if ( ! refNam.isNull() )  {
          pool::Ref<const TNamed> other(m_dataSvc,refObj->named());
          if ( refObj->named() == other )   {
            if ( !(refObj->named() != other) )   {
              const TNamed* t = other.ptr();
              if ( refObj->named() == t )   {
                t = refNam.ptr();
                if ( t == other )   {
                  okRef = true;
                }
              }
            }
          }
        }
      }
    }
    if(!okRef)
      printError("test ERROR: Something went awfully wrong with the refs...");
    refObj = 0;
    i++;
  }

  std::cout << OVALTAG 
      << " End of reading after " << i << " Objects." << std::endl;

  if(n_expected!=i){
    printError("test ERROR: number of object different from expected");
  }

  // no object marked are expected in a read transaction
  if(m_dataSvc->cacheSvc().numberOfMarkedObjects()!=0){
    printError("test ERROR: wrong number of marked objects");
    std::cout << "Number of Object marked = "<< m_dataSvc->cacheSvc().numberOfMarkedObjects() << std::endl;
  }

  // the transaction has to be committed anyway
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the READ transaction." );
  }

  // no object marked are expected in a read transaction
  if(m_dataSvc->cacheSvc().numberOfMarkedObjects()!=0){
    printError("test ERROR: wrong number of marked objects");
    std::cout << "Number of Object marked = "<< m_dataSvc->cacheSvc().numberOfMarkedObjects() << std::endl;
  }

  // check the number of not empty object retrieved 
  if(i!=n_expected)
    printError("test ERROR: number of objects retrieved different from expected");

  // disconnect 
  m_dataSvc->session().disconnectAll();

  printM("Clear cache...");
  m_dataSvc->clearCache();
  printM("Read procedure completed");
  return;
}

std::vector<std::string>& pool::ObjectStorageTest::tokens(){
  return m_tokens;
}
