#include <stdexcept>
#include <sstream>

#include "PerformanceStudy/PerformanceStudy.h"

#include "Reflex/SharedLibrary.h"
#include "SealKernel/Exception.h"

#include "CoralBase/MessageStream.h"

#include "FileCatalog/URIParser.h"
#include "FileCatalog/FCSystemTools.h"
#include "FileCatalog/IFileCatalog.h"

#include "DataSvc/IDataSvc.h"
#include "DataSvc/DataSvcFactory.h"

#include "PoolBench/ProcessTime.h"
#include "PoolBench/ProcessInfo.h"

PerformanceStudy::PerformanceStudy():
  m_fileCatalog( 0 ),
  m_dataSvc( 0 ),
  m_libs(),
  m_ellapsedTime( 0 ),
  m_initialMemory( 0 ),
  m_finalMemory( 0 )
{
  // Loads the seal message stream
  pool::POOLContext::loadComponent( "SEAL/Services/Message Service" );
  // Set the verbosity threshold to warnings
  // pool::POOLContext::setMessageVerbosityLevel( coral::Info );
}

PerformanceStudy::~PerformanceStudy()
{
  if ( m_dataSvc ) delete m_dataSvc;
  if ( m_fileCatalog ) delete m_fileCatalog;
}


void
PerformanceStudy::addRunTimeLibrary( const std::string& libName )
{
  m_libs.insert( libName );
}


void
PerformanceStudy::prepare()
{
  pool::URIParser p;
  p.parse();
//   m_fileCatalog = pool::FCSystemTools::createCatalog( p.prefix() );
//   if ( ! m_fileCatalog )
//     throw std::runtime_error( "Could not create a file catalog" );
//   m_fileCatalog->connect( p.url() );

  m_fileCatalog = new pool::IFileCatalog();
  m_fileCatalog->addReadCatalog( p.contactstring() );
  m_fileCatalog->connect();
  
  m_dataSvc = pool::DataSvcFactory::instance( m_fileCatalog );
  if ( ! m_dataSvc ) {
    throw std::runtime_error( "Could not create a data svc" );
  }
}

pool::IDataSvc&
PerformanceStudy::dataSvc()
{
  if ( ! m_dataSvc ) {
    throw std::runtime_error( "No data service installed" );
  }
  return *m_dataSvc;
}

pool::IFileCatalog&
PerformanceStudy::fileCatalog()
{
  if ( ! m_fileCatalog ) {
    throw std::runtime_error( "No file catalog is used" );
  }
  return *m_fileCatalog;
}

int
PerformanceStudy::run()
{
  coral::MessageStream msg( "PerformanceStudy::run", coral::Info );
  try {
    this->loadLibraries();
    msg << coral::Info << "Preparing the job." << coral::MessageStream::endmsg;
    this->prepare();
    m_initialMemory = this->currentMemoryUsage();
    msg << coral::Info << "Running the main method of the job." << coral::MessageStream::endmsg;
    pool::ProcessTime startTime;
    this->theMainMethod();
    pool::ProcessTime endTime;
    m_ellapsedTime = (endTime - startTime).getUserTime() * pool::ProcessTime::SECONDS;
    m_finalMemory = this->currentMemoryUsage();
    msg << coral::Info << "Finalizing." << coral::MessageStream::endmsg;
    this->finalize();
    return 0;
  }
  catch ( seal::Exception& sealError ) {
    msg << coral::Fatal << "SEAL Exception : " << sealError.what() << coral::MessageStream::endmsg;
    return 1;
  }
  catch ( std::exception& error ) {
    msg << coral::Fatal << "std exception : " << error.what() << coral::MessageStream::endmsg;
    return 1;
  }
  catch (...) {
    msg << coral::Fatal << "Funny error ..." << coral::MessageStream::endmsg;
    return 1;
  }
}

unsigned int
PerformanceStudy::currentMemoryUsage() const
{
  pool::ProcessInfo::update();
  return pool::ProcessInfo::memRSS();
}

void
PerformanceStudy::loadLibraries()
{
  coral::MessageStream msg( "PerformanceStudy::loadLibraries", coral::Info );
  for ( std::set< std::string >::const_iterator iLib = m_libs.begin();
        iLib != m_libs.end(); ++iLib ) {
    msg << coral::Info << "Loading library " << *iLib << coral::MessageStream::endmsg;
    
    Reflex::SharedLibrary libloader( "lib" + *iLib + ".so" );
    
    if(!libloader.Load()){
      msg << coral::Error << libloader.Error() << coral::MessageStream::endmsg;
    }
  }
}


double
PerformanceStudy::ellapsedTime() const
{
  return m_ellapsedTime;
}

unsigned int
PerformanceStudy::initialMemory() const
{
  return m_initialMemory;
}

unsigned int
PerformanceStudy::finalMemory() const
{
  return m_finalMemory;
}

void
PerformanceStudy::theMainMethod()
{}

void
PerformanceStudy::finalize()
{
  coral::MessageStream msg( "PerformanceStudy::finalize", coral::Info );
  msg << coral::Info << "CPU time (user): " << m_ellapsedTime << " s" << coral::MessageStream::endmsg;
  msg << coral::Info << "Initial memory : " << m_initialMemory / ( 1024*1024 ) << " MB" << coral::MessageStream::endmsg;
  msg << coral::Info << "Final   memory : " << m_finalMemory / ( 1024*1024 ) << " MB" << coral::MessageStream::endmsg;
}
