// $Id: ProcessInfo.cpp,v 1.1 2003/07/24 16:30:11 rado Exp $

#include "SealBase/ProcessInfo.h"
#include "PoolBench/ProcessInfo.h"

#include <sstream>

#include <errno.h>

namespace pool {
  
  bool ProcessInfo::s_autoUpdate = false;

  /**
   * Provides information about memory status in pages.
   * The columns are:
   * size       total program size
   * resident   resident set size
   * share      shared pages
   * trs        text (code)
   * drs        data/stack
   * lrs        library
   * dt         dirty pages
   */
  static size_t s_size   = 0;
  static size_t s_rss    = 0;
  static size_t s_share  = 0;
  static size_t s_text   = 0;
  static size_t s_data   = 0;
  static size_t s_libs   = 0;
  static size_t s_dirty  = 0;

  int readProcStatm()
  {
#if defined(_WIN32)
#  error Not implemented yet    
#elif defined(__linux)
    //unsigned int dummy = 0;
    
    pid_t mypid = seal::ProcessInfo::pid();
    
    std::stringstream proc_pid_statm;
    //std::stringstream proc_pid_status;

    proc_pid_statm  << "/proc/" << mypid << "/statm" << std::ends;
    //proc_pid_status << "/proc/" << mypid << "/status" << std::ends;
    
    FILE* proc_self_statm  = fopen( proc_pid_statm.str().c_str() , "r" );
    //FILE* proc_self_status = fopen( proc_pid_status.str().c_str(), "r" );
    
    if( proc_self_statm != NULL ) {
      fscanf( proc_self_statm, "%u %u %u %u %u %u %u",
              &s_size, &s_rss,  &s_share,
              &s_text, &s_data, &s_libs, &s_dirty );
    } else {
      perror( "pool::ProcessInfo::init :" );
    }
    
    return errno;
  }
#else
#  error Unsupported platform
#endif
  
  /**
   *  /proc/PID/stat
   *  Status information about the process.  This is used by ps(1).  It is defined in /usr/src/linux/fs/proc/array.c.
   *  The fields, in order, with their proper scanf(3) format specifiers, are:
   */
  static int           s_pid;     ///< %d  The process id.
  static std::string   s_comm;    ///< %s  The filename of the executable, in parentheses.
                                  ///<     This is visible whether or not the executable is swapped out.
  static char          s_state;   ///< %c  One  character  from  the string "RSDZTW" where:
                                  ///<     R is running, S is sleeping in an interruptible wait,
                                  ///<     D is waiting in uninterruptible disk sleep, Z is zombie,
                                  ///<     T is traced or stopped (on a signal), and W is paging.
  static int           s_ppid;    ///< %d  The PID of the parent.
  static int           s_pgrp;    ///< %d  The process group ID of the process.
  static int           s_session; ///< %d  The session ID of the process.
  static int           s_tty_nr;  ///< %d  The tty the process uses.
  static int           s_tpgid;   ///< %d  The process group ID of the process which currently owns the tty that the
                                  ///<     process is connected to.
  static unsigned long s_flags;   ///< %lu The flags of the process.
                                  ///<     The math bit is decimal 4, and the traced bit is decimal 10.
  static unsigned long s_minflt;  ///< %lu The number of minor faults the process has made which have not required
                                  ///<     loading a memory page from disk.

  /*
     cminflt %lu         The number of minor faults that the process and its children have made.
     majflt %lu          The number of major faults the process has made which have required loading a memory page from disk.
     cmajflt %lu         The number of major faults that the process and its children have made.
     utime %lu           The number of jiffies that this process has been scheduled in user mode.
     stime %lu           The number of jiffies that this process has been scheduled in kernel mode.
     cutime %ld          The number of jiffies that this process and its children have been scheduled in user mode.
     cstime %ld          The number of jiffies that this process and its children have been scheduled in kernel mode.
     priority %ld        The standard nice value, plus fifteen.  The value is never negative in the kernel.
     nice %ld            The nice value ranges from 19 (nicest) to -19 (not nice to others).
     0 %ld               This value is hard coded to 0 as a placeholder for a removed field.
     itrealvalue %ld     The time in jiffies before the next SIGALRM is sent to the process due to an interval timer.
     starttime %lu       The time in jiffies the process started after system boot.
  */
  static unsigned long s_vsize; ///< %lu  Virtual memory size in bytes.
  static          long s_vrss;  ///< %ld  Resident Set Size: number of pages the process has in real memory,
                                ///<      minus 3 for administrative purposes.
                                ///<      This is  just the  pages  which  count towards text, data, or stack space.
                                ///<      This does not include pages which have not been demand-loaded in,
                                ///<      or which are swapped out.
  /*
     rlim %lu            Current limit in bytes on the rss of the process (usually 4,294,967,295).
     startcode %lu       The address above which program text can run.
     endcode %lu         The address below which program text can run.
     startstack %lu      The address of the start of the stack.
     kstkesp %lu         The current value of esp (stack pointer), as found in the kernel stack page for the process.
     kstkeip %lu         The current EIP (instruction pointer).
     signal %lu          The bitmap of pending signals (usually 0).
     blocked %lu         The bitmap of blocked signals (usually 0, 2 for shells).
     sigignore %lu       The bitmap of ignored signals.
     sigcatch %lu        The bitmap of catched signals.
     wchan %lu           This is the "channel" in which the process is waiting.
                         It is the address of a system call, and can be looked up in a namelist  if  you  need  a textual name.
                         (If you have an up-to-date /etc/psdatabase, then try ps -l to see the WCHAN field in action.)
     nswap %lu           Number of pages swapped - not maintained.
     cnswap %lu          Cumulative nswap for child processes.
     exit_signal %d      Signal to be sent to parent when we die.
     processor %d        Processor number last executed on.    
  */
  
  static int           dummint;
  static long          dummlong;
  static unsigned long dummulong;
  static char          dummchar;
  static char          dummstr[8192];

#define PROC_STAT_FORMAT "%d %s %c \
%d %d %d %d %d \
%lu %lu %lu %lu %lu %lu %lu \
%ld %ld %ld %ld %ld %ld %lu \
%lu %ld \
%lu %lu %lu %lu %lu \
%lu %lu %lu %lu %lu \
%lu %lu %lu \
%d %d"  

  bool readProcStat()
  {
#if defined(_WIN32)
#  error Not implemented yet    
#elif defined(__linux)
    pid_t myPid = seal::ProcessInfo::pid();
    
    std::stringstream ps;
    ps << "/proc/" << myPid << "/stat" << std::ends;
    
    FILE* proc_pid_stat = fopen( ps.str().c_str(), "r" );
    
    int filled;
    
    if( proc_pid_stat != NULL )
    {
      filled = fscanf( proc_pid_stat,
                       PROC_STAT_FORMAT,
                       &dummint,   dummstr,    &dummchar,
                       &dummint,   &dummint,   &dummint,   &dummint,   &dummint,
                       &dummulong, &dummulong, &dummulong, &dummulong, &dummulong, &dummulong, &dummulong,
                       &dummlong,  &dummlong,  &dummlong,  &dummlong,  &dummlong,  &dummlong,  &dummulong,
                       &s_vsize,   &s_vrss,
                       &dummulong, &dummulong, &dummulong, &dummulong, &dummulong,
                       &dummulong, &dummulong, &dummulong, &dummulong, &dummulong,
                       &dummulong, &dummulong, &dummulong,
                       &dummint,   &dummint
        );

      if( filled == EOF ) {
        perror( "readProcStat: " );
        return false;
      }
    }
    
    return true;
#else
#  error Unsupported platform
#endif    
  }
    
  /// Return the total program size of the current process (bytes)
  size_t ProcessInfo::memSize()
  {
    updateBeforeRead();
    return s_vsize;
  }
  
  /// Return the program resident set size (roughly, bytes)
  size_t ProcessInfo::memRSS()
  {
    updateBeforeRead();
    // FIXME: do a proper platform dependent stuff
    // This hardwired Linux kernel memory page size
    return 4096*s_vrss;
  }
  
  /// Return the program's shared sgement size
  size_t ProcessInfo::memShared()
  {
    updateBeforeRead();
    return s_share;
  }
  
  /// Return the size program's text(code) segment
  size_t ProcessInfo::memText()
  {
    updateBeforeRead();
    return s_text;
  }
  
  /// Return the size of program's data segment
  size_t ProcessInfo::memData()
  {
    updateBeforeRead();
    return s_data;
  }
  
  /// Return the size of the segment of loaded libraries
  size_t ProcessInfo::memLibs()
  {
    updateBeforeRead();
    return s_libs;
  }  
  
  /// Return the size of dirty pages
  size_t ProcessInfo::memDirty()
  {
    updateBeforeRead();
    return s_dirty;
  }  
  
  int ProcessInfo::update()
  {
#if defined(_WIN32)
#  error Not implemented yet    
#elif defined(__linux)
    readProcStatm();
    readProcStat();
#else
#  error Unsupported platform
#endif    
  }
  
}

