#ifndef StressTestApp_h
#define StressTestApp_h

#include <vector>
#include <string>
#include <iostream>

#include "DataSvc/IDataSvc.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"
#include "FileCatalog/FCSystemTools.h"
#include "StorageSvc/DbType.h"


namespace pool {
  class IDataSvc;
  class IFileCatalog;
  class FCSystemTools;
}

namespace pool {
 /** @class StressTestApp StressTestApp.h Tests/Libraries/StressTestBase/StressTestApp.h

    StressTestApp is designed as a base class that initialises the
    environment for POOL test code. The user should override the function
    userMain to create their test.

    @author Tim Barrass
    @date 8/5/2003
*/
class StressTestApp {
public:
  StressTestApp(int argc, const char *argv[]);
  virtual ~StressTestApp();

  /** Low level main- call this to run the test.
    @return int 0 success, -1 fail
  */
  int  run();

  /** Dynamically load a single shared library, or a vector of libraries.
      Libraries filenames are "lib" + libName + ".so"
    @param libName The library name
  */
  virtual void loadLib(const std::string &libName);

  /** Dynamically load a vector of libraries.
      Libraries filenames are "lib" + libName + ".so"
    @param libraries A vector of libName strings (see other loadLib)
  */
  virtual void loadLib(const std::vector< std::string >& libraries);

protected:
  /** virtual placeholder for user's test code
  */
  virtual int userMain() = 0;
  
  /** Get a handle to the context (the DataSvc cache)
    @return pool::IDataSvc* Handle to the context
  */
  pool::IDataSvc     *context() const { return m_context; }

  /** Get catalog handle
    @return pool::FileCatalog* Handle to the catalog
  */
  pool::IFileCatalog& catalog() const { return *m_catalog; }

  /** Reset the current catalog.
    @param nuCatalog  Handle to the new catalog
  */
  void catalog( pool::IFileCatalog& nuCatalog ) { *m_catalog = nuCatalog; }
  
  /** Get create a catalog instance based on a URIParser string (e.g.
    p = "xmlcatalog_file:StressTest.xml"). The user is responsible for
    connecting starting and disconnecting the catalog.
    @return pool::FileCatalog* Handle to the catalog
    @param p The URIParser to init the catalog
  */
  pool::IFileCatalog& createCatalog(pool::URIParser& p);

  /** UNFINISHED? Only one catalog may be open at a time in a StressTestApp.
    This method handles the swap to a new catalog, including clearing
    the cache and initialising a new one in its place. The user is
    responsibe for updating all existing Refs to use the new cache (context).
    The new catalog is created with a URIParser object then connected and started.
    @return int 0 is success
    @param p The URIParser to init the catalog
  */
  int swapCatalog( pool::URIParser& p );
  
  /** Get vector of command line arguments. The first argument in the vector is
    app name.
    @return std::vector Vector of command line arguments, parsed as strings.
  */
  const std::vector<std::string> &cmdLine() const { return m_cmdLineArgs; }
      
  /** Register a filename with the catalog, checking first to see if it
    exists already.
    @return std::string The UID that uniquely determines the file in the
    catalog.
    @param nuDBFilename The file pfn string
    @param storageType The method by which objects are stored in the file.
  */
  virtual std::string registerFileWithCatalog(const std::string &nuDBFileName,
                                              const pool::DbType &storageType=pool::ROOTTREE_StorageType.majorType());
  
  /** Simple output of OVAL messages to stdout.
    @param msg The message string
  */
  virtual void ovalmsg(const std::string& msg) { std::cout << "[OVAL] " << msg << std::endl; }
  virtual void ovalmsg(const std::string& msg) const { std::cout << "[OVAL] " << msg << std::endl; }

private:
  pool::IFileCatalog      *m_catalog;
  pool::IDataSvc          *m_context;
  std::vector<std::string> m_cmdLineArgs;
};

} // ns pool

#endif // StressTestApp_h

