#include "StressTestBase/StressTestApp.h"

#include "DataSvc/DataSvcFactory.h"
#include "FileCatalog/URIParser.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/IFCAction.h"
#include "FileCatalog/FCSystemTools.h"
#include "StorageSvc/DbType.h"

#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/ISession.h"

#include "Reflex/SharedLibrary.h"
#include "libname.h"
#include "CoralBase/MessageStream.h"

#include <stdexcept>
#include <string>

#include "RVersion.h"

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
using namespace ROOT;
#endif
namespace {
  coral::MessageStream appLog("StressTestApp");
} 

pool::StressTestApp::StressTestApp(int argc, const char *argv[]) : m_catalog(0), m_context(0)
{
    for (int i=0; i<argc; i++) {
      m_cmdLineArgs.push_back(argv[i]);
    }
}



pool::StressTestApp::~StressTestApp()
{
  if (m_context != 0) {
    m_context->session().disconnectAll();
    delete m_context;
  }
  if (m_catalog != 0) {
    m_catalog->disconnect();
    delete m_catalog;
  }
}



int pool::StressTestApp::run()
{
  appLog << coral::Info << "Create pluginmanager" << coral::MessageStream::endmsg;
  //  seal::PluginManager* pm = seal::PluginManager::get();
  
  

  try{
    //    if ( ! pm ) {
    //      throw std::runtime_error( "Could not get the plugin manager instance" );
    //    }
    appLog << coral::Info << "Init pluginmanager" << coral::MessageStream::endmsg;
    //    pm->initialise();
    
    appLog << coral::Info << "Init message service" << coral::MessageStream::endmsg;
    //    pool::POOLContext::loadComponent( "SEAL/Services/Message Service" );
    // pool::POOLContext::setMessageVerbosityLevel( coral::Info );

    appLog << coral::Info << "Start catalog and context" << coral::MessageStream::endmsg;
    URIParser p("xmlcatalog_file:StressTestCatalog.xml");
    p.parse();
    
//     m_catalog   = createCatalog(p);
//     m_catalog->connect(p.url());
    m_catalog = new pool::IFileCatalog();
//     m_catalog->addReadCatalog( p.contactstring() );
    m_catalog->setWriteCatalog( p.contactstring() );
    m_catalog->connect();
    
    m_context   = DataSvcFactory::instance(m_catalog);

    appLog << coral::Info << "Create db access policy" << coral::MessageStream::endmsg;
    pool::DatabaseConnectionPolicy policy;
    policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
    policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
    m_context->session().setDefaultConnectionPolicy( policy );

    userMain();
  }
  catch (std::exception &e)
  {
    appLog << coral::Error << "Exception caught: " << e.what() << coral::MessageStream::endmsg;
    return -1;
  }
  return 0;
}


/** registerFileWithCatalog
  registers a filename with the catalog and returns the fid for use. Could wrap this internally
  in a catalog transaction- but what if user wants to register a large number of files? Easier
  for user to wrap in their own catalog transaction.  Not strictly necessary, as files are now
  registered automatically.
*/
std::string pool::StressTestApp::registerFileWithCatalog(const std::string &nuDBFileName, const pool::DbType &storageType)
{
  std::string pfn = nuDBFileName;
  std::string fid = "";
  std::string type = "";

  IFCAction action;
  m_catalog->setAction( action );
  action.lookupFileByPFN(pfn,fid,type);
  
  std::cout << "FID IS " << fid << std::endl;
  if ( fid.empty() ) {
    std::cout << "Need to register file" << std::endl;
    pool::DbType majorType(storageType);
    action.registerPFN(pfn,majorType.storageName(),fid);
    if ( fid.empty() )
      throw std::runtime_error("failed to register file `" + pfn +"'");
  }

  return fid;
}

pool::IFileCatalog& pool::StressTestApp::createCatalog(pool::URIParser& p)
// potential problem: if this method is used to create a catalog,
// who deals with shutting it down again?
{
    p.parse();

//     IFileCatalog* cat=0;
//     cat = FCSystemTools::createCatalog(p.prefix());
    IFileCatalog* cat = new pool::IFileCatalog();
    cat->addReadCatalog( p.contactstring() );
    
    if ( ! cat ) {
      throw std::runtime_error( "Could not retrieve a file catalog" );
    }

    return *cat;

}


int pool::StressTestApp::swapCatalog(URIParser& p)
{
  p.parse();
  
  IFileCatalog* cat = 0;
  
  try{
   context()->clearCache();

   catalog().commit();
   catalog().disconnect();
   
   *cat = createCatalog( p );
   catalog(*cat);
   catalog().connect();
   catalog().start();

   m_context = DataSvcFactory::instance( &catalog() );
   
  } catch (...) {
    if (cat)
      delete cat;                                
    throw std::runtime_error("Problem swapping catalog");  
  }

  return 0;
}



void pool::StressTestApp::loadLib(const std::string &libName)
{
  std::cout << "Loading library " << libName << std::endl;
  
  Reflex::SharedLibrary libloader( libname(libName) );
  
  if(!libloader.Load()){
    std::cout << libloader.Error() << std::endl;
  }
}



void pool::StressTestApp::loadLib( const std::vector< std::string >& libraries )
{  
  for ( std::vector<std::string>::const_iterator iLibrary = libraries.begin();
        iLibrary != libraries.end(); ++iLibrary ) {
    const std::string& library = *iLibrary;
    std::cout << "Loading library " << library << std::endl;

    Reflex::SharedLibrary libloader( libname(library) );
    
    if(!libloader.Load()){
      std::cout << libloader.Error() << std::endl;
    }
  }
}

