#include "PrimitivesContainerStreamers.h"
#include "Reflex/Member.h"
#include "Reflex/Object.h"
#include <stdexcept>
#include <cstring>

pool::BlobWriter::BlobWriter( const Reflex::Type& type ):
  m_type( type ),
  m_blob()
{}

pool::BlobWriter::~BlobWriter()
{}

const coral::Blob&
pool::BlobWriter::write( const void* addressOfInputData )
{
  // The actual object
  Reflex::Object theContainer( m_type, const_cast<void*>( addressOfInputData ) );

  // Retrieve the size of the container
  Reflex::Member sizeMethod = m_type.MemberByName( "size" );
  if ( ! sizeMethod )
    throw std::runtime_error( "No size method is defined for the container" );
  Reflex::Object ret = sizeMethod.Invoke(theContainer);
  if( !ret.TypeOf() || !ret.Address() )
    throw std::runtime_error( "Could not invoke the size method on the container" );
  size_t contrainerSize = *(static_cast<size_t*>(ret.Address()));

  // Retrieve the element size
  Reflex::Member beginMethod = m_type.MemberByName( "begin" );
  if ( ! beginMethod )
    throw std::runtime_error( "No begin method is defined for the container" );
  Reflex::Type iteratorType = beginMethod.TypeOf().ReturnType();
  Reflex::Member dereferenceMethod = iteratorType.MemberByName( "operator*" );
  if ( ! dereferenceMethod )
    throw std::runtime_error( "Could not retrieve the dereference method of the container's iterator" );
  size_t elementSize = dereferenceMethod.TypeOf().ReturnType().SizeOf();

  // allocate the blob
  m_blob.resize( contrainerSize * elementSize );
  void* startingAddress = m_blob.startingAddress();

  // Create an iterator
  Reflex::Object iteratorObject = beginMethod.Invoke( Reflex::Object( m_type, const_cast< void * > ( addressOfInputData ) ) );

  // Loop over the elements of the container
  Reflex::Member incrementMethod = iteratorObject.TypeOf().MemberByName( "operator++" );
  if ( ! incrementMethod )
    throw std::runtime_error( "Could not retrieve the increment method of the container's iterator" );

  for ( size_t i = 0; i < contrainerSize; ++i ) {
    void* elementAddress = dereferenceMethod.Invoke( iteratorObject ).Address();
    ::memcpy( startingAddress, elementAddress, elementSize );
    char* cstartingAddress = static_cast<char*>( startingAddress );
    cstartingAddress += elementSize;
    startingAddress = cstartingAddress;
    incrementMethod.Invoke( iteratorObject );
  }


  // Destroy the iterator
  iteratorObject.Destruct();

  return m_blob;
}



pool::BlobReader::BlobReader( const Reflex::Type& type ):
  m_type( type )
{}


pool::BlobReader::~BlobReader()
{}


void
pool::BlobReader::read( const coral::Blob& blobData,
                        void* containerAddress ) const
{
  // Retrieve the element size
  Reflex::Member beginMethod = m_type.MemberByName( "begin" );
  if ( ! beginMethod )
    throw std::runtime_error( "No begin method is defined for the container" );
  Reflex::Type iteratorType = beginMethod.TypeOf().ReturnType();
  Reflex::Member dereferenceMethod = iteratorType.MemberByName( "operator*" );
  if ( ! dereferenceMethod )
    throw std::runtime_error( "Could not retrieve the dereference method of the container's iterator" );

  size_t elementSize = dereferenceMethod.TypeOf().ReturnType().SizeOf();

  // Retrieve the container size
  size_t contrainerSize = blobData.size() / elementSize;

  // Retrieve the end method
  Reflex::Member endMethod = m_type.MemberByName( "end" );
  if ( ! endMethod )
   throw std::runtime_error( "Could not retrieve the end method of the container" );

  // Retrieve the insert method
  Reflex::Member insertMethod;
  for( unsigned int i = 0; i < m_type.FunctionMemberSize();i++){
    Reflex::Member im = m_type.FunctionMemberAt(i);
    if( im.Name() != std::string( "insert" ) ) continue;
    if( im.TypeOf().FunctionParameterSize() != 2) continue;
    insertMethod = im;
    break;
  }

  // Retrieve the clear method
  Reflex::Member clearMethod = m_type.MemberByName( "clear" );
  if ( ! clearMethod )
   throw std::runtime_error( "Could not retrieve the clear method of the container" );

  // Clear the container
  Reflex::Object containerObject( m_type, containerAddress );
  clearMethod.Invoke( containerObject );

  // Fill-in the elements
  const void* startingAddress = blobData.startingAddress();
  for ( size_t i = 0; i < contrainerSize; ++i ) {
    std::vector< void* > args( 2 );

    // call container.end()
    args[0] = endMethod.Invoke( containerObject ).Address();

    // call container.insert( container.end(), data )
    args[1] = const_cast<void*>( startingAddress );
    Reflex::Object pos = insertMethod.Invoke( containerObject, args );

    // this is required! (as it was in Reflect). 
    iteratorType.Destruct( args[0] );

    // this is required ! (as it was in Reflect). 
    iteratorType.Destruct( pos.Address() );

    const char* cstartingAddress = static_cast<const char*>( startingAddress );
    cstartingAddress += elementSize;
    startingAddress = cstartingAddress;
  }

}

