#include "TestDriver.h"
#include <stdexcept>
#include <iostream>
#include <sstream>
#include <memory>

#include "Reflex/SharedLibrary.h"
#include "Reflex/Type.h"
#include "libname.h"

#include "POOLCore/Token.h"


#include "StorageSvc/DbType.h"

#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"

#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"

#include "PersistencySvc/ITokenValidator.h"
#include "PersistencySvc/ITokenValidatorFactory.h"

#include "PersistencySvc/IDataTransform.h"
#include "PersistencySvc/IDataTransformFactory.h"

#include "PersistencySvc/IPersistencySvc.h"
#include "PersistencySvc/IPersistencySvcFactory.h"

#include "MyTestClasses.h"
#include "MyDataCallBack.h"

namespace pool  {
  int numTokenInstances();
}

pool::TestDriver::TestDriver():
  m_fileCatalog( 0 ),
  m_fileName( "pool.root" ),
  m_events( 100 )
{
  
  // pool::POOLContext::setMessageVerbosityLevel( coral::Info );
  
  std::cout << "[OVAL] Creating a file catalog" << std::endl;
  m_fileCatalog = new pool::IFileCatalog;
  pool::URIParser p;
  p.parse();
  m_fileCatalog->setWriteCatalog( p.contactstring() );
  m_fileCatalog->connect();
}

pool::TestDriver::~TestDriver()
{
  if ( m_fileCatalog ) delete m_fileCatalog;
  std::cout << "[OVAL] Number of floating tokens : " << pool::numTokenInstances() << std::endl;
}

void
pool::TestDriver::clearCache()
{
  for ( std::vector< pool::Token* >::iterator iToken = m_tokens.begin();
	iToken != m_tokens.end(); ++iToken ) {
    (*iToken)->release();
  }
  m_tokens.clear();
}

void
pool::TestDriver::loadLibraries( const std::vector<std::string>& libraries )
{
  for ( std::vector<std::string>::const_iterator iLibrary = libraries.begin();
        iLibrary != libraries.end(); ++iLibrary ) {
    const std::string& library = *iLibrary;
    std::cout << "Loading library " << library << std::endl;
    
    Reflex::SharedLibrary libloader( libname(library) );
    
    if(!libloader.Load()){
      std::cout << libloader.Error() << std::endl;
    }
  }
}

void
pool::TestDriver::write()
{
  pool::IFileCatalog& catalog = *m_fileCatalog;
  catalog.start();

  std::cout << "Creating the persistency service" << std::endl;
  pool::IPersistencySvcFactory* psfactory = pool::IPersistencySvcFactory::get();
  if ( ! psfactory ) {
    throw std::runtime_error( "Could not retrieve an IPersistencySvc factory" );
  }
  std::auto_ptr< pool::IPersistencySvc > persistencySvc( psfactory->create( "PersistencySvc", catalog ) );
  if ( ! persistencySvc.get() ) {
    throw std::runtime_error( "Could not create a PersistencySvc" );
  }

  // Set up the policy.
  pool::DatabaseConnectionPolicy policy;
  policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
  policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
  persistencySvc->session().setDefaultConnectionPolicy( policy );

  // Retrieving the class descriptions
  Reflex::Type class_MyTransientClass = Reflex::Type::ByName( "MyTransientClass" );
  Reflex::Type class_MyPersistentClass = Reflex::Type::ByName( "MyPersistentClass" );

  // Resistering the transformation
  MyDataCallBack callBackObject;
  persistencySvc->dataTransform().registerTransformation( class_MyTransientClass,
							  class_MyPersistentClass,
							  callBackObject );

  // Start an update transaction
  if ( ! ( persistencySvc->session().transaction().start( pool::ITransaction::UPDATE ) ) ) {
    throw std::runtime_error( "Could not start an update transaction" );
  }

  // Defining the placement objects
  pool::Placement placementHint_MyTransientClass;
  placementHint_MyTransientClass.setDatabase( m_fileName, pool::DatabaseSpecification::PFN );
  placementHint_MyTransientClass.setContainerName( "MyContainer" );
  placementHint_MyTransientClass.setTechnology( pool::ROOTTREE_StorageType.type() );
  placementHint_MyTransientClass.setShapeID( class_MyPersistentClass  );

  std::vector< MyTransientClass* > v_myTransientClass;

  for ( int i = 0; i < m_events; ++i ) {

    MyTransientClass* object_MyTransientClass = new MyTransientClass();
    v_myTransientClass.push_back( object_MyTransientClass );
    object_MyTransientClass->a = i;
    object_MyTransientClass->b = float(2.*i + 1.5);
    object_MyTransientClass->c = 3.*i + 1.6;
    object_MyTransientClass->d = i%12;

    pool::Token* token_MyTransientClass = persistencySvc->registerForWrite( placementHint_MyTransientClass,
									    object_MyTransientClass,
									    class_MyTransientClass );
    if ( ! token_MyTransientClass ) {
      throw std::runtime_error( "Could not write an object" );
    }
    m_tokens.push_back( token_MyTransientClass );
  }

  std::cout << "Committing the transaction." << std::endl;
  if ( ! persistencySvc->session().transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction." );
  }

  // Removing the "cache"
  for ( std::vector< MyTransientClass* >::iterator iObject = v_myTransientClass.begin();
	iObject != v_myTransientClass.end(); ++iObject ) {
    delete *iObject;
  }
  v_myTransientClass.clear();

  catalog.commit();
}


void
pool::TestDriver::read()
{
  pool::IFileCatalog& catalog = *m_fileCatalog;
  catalog.start();

  std::cout << "Creating the persistency service" << std::endl;
  pool::IPersistencySvcFactory* psfactory = pool::IPersistencySvcFactory::get();
  if ( ! psfactory ) {
    throw std::runtime_error( "Could not retrieve an IPersistencySvc factory" );
  }
  std::auto_ptr< pool::IPersistencySvc > persistencySvc( psfactory->create( "PersistencySvc", catalog ) );
  if ( ! persistencySvc.get() ) {
    throw std::runtime_error( "Could not create a PersistencySvc" );
  }

  // Retrieving the class descriptions
  Reflex::Type class_MyTransientClass = Reflex::Type::ByName( "MyTransientClass" );
  Reflex::Type class_MyPersistentClass = Reflex::Type::ByName( "MyPersistentClass" );

  // Resistering the transformation
  MyDataCallBack callBackObject;
  persistencySvc->dataTransform().registerTransformation( class_MyTransientClass,
							  class_MyPersistentClass,
							  callBackObject );

  // Starting a read transaction
  if ( ! persistencySvc->session().transaction().start( pool::ITransaction::READ ) ) {
    throw std::runtime_error( "Could not start a read transaction." );
  }

  std::cout << "Reading back the objects." << std::endl;
  for ( int i = 0; i < m_events; ++i ) {

    std::pair< void*, Reflex::Type > objectAndClass_myTransientClass = persistencySvc->readObject( class_MyTransientClass, *( m_tokens[ i ] ) );
    if ( objectAndClass_myTransientClass.first == 0 ) {
      throw std::runtime_error( "Could not read the stored data" );
    }
    if ( !objectAndClass_myTransientClass.second ) {
      throw std::runtime_error( "Could not retrieve the stored class" );
    }
    if ( objectAndClass_myTransientClass.second.Name() != "MyTransientClass" ) {
      throw std::runtime_error( "read wrong class type." );
    }
    MyTransientClass* object_myTransientClass = reinterpret_cast< MyTransientClass* >( objectAndClass_myTransientClass.first );
    bool OK = true;
    if ( object_myTransientClass->a != i ) {
      std::cerr << "i : " << i << std::endl
		<< "a : " << object_myTransientClass->a << std::endl;
      OK = false;
    }
    float b_expected = float(2*i + 1.5);
    if ( object_myTransientClass->b != b_expected ) {
      std::cerr << "b expected : " << b_expected << std::endl
		<< "b : " << object_myTransientClass->b << std::endl;
      OK = false;
    }
    double c_expected = 123; 
    if ( object_myTransientClass->c != c_expected ) {
      std::cerr << "i : " << i << std::endl
		<< "c expected : " << c_expected << std::endl
		<< "c : " << object_myTransientClass->c << std::endl;
      OK = false;
    }
    short d_expected = i%12;
    if ( object_myTransientClass->d != d_expected ) {
      std::cerr << "d expected : " << d_expected << std::endl
		<< "d : " << object_myTransientClass->d << std::endl;
      OK = false;
    }
    if ( ! OK )
      throw std::runtime_error( "MyTransientClass object written is different from object read" );
    delete object_myTransientClass;
  }

  std::cout << "Committing the transaction." << std::endl;
  if ( ! persistencySvc->session().transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction." );
  }

  catalog.commit();
}
