#include "PrimitivesContainerStreamers.h"
#include "Reflex/Member.h"
#include "Reflex/Object.h"
#include <stdexcept>
#include <cstring>

pool::BlobWriter::BlobWriter( const Reflex::Type& type ):
  m_type( type ),
  m_blob()
{}

pool::BlobWriter::~BlobWriter()
{}

const coral::Blob&
pool::BlobWriter::write( const void* addressOfInputData )
{
  // The actual object
  Reflex::Object theContainer( m_type, const_cast<void*>( addressOfInputData ) );

  // Retrieve the size of the container
  Reflex::Member sizeMethod = m_type.MemberByName( "size" );
  if ( ! sizeMethod )
    throw std::runtime_error( "No size method is defined for the container" );
#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
  Reflex::Object ret = sizeMethod.Invoke(theContainer);

  if( !ret.TypeOf() || !ret.Address() )
    throw std::runtime_error( "Could not invoke the size method on the container" );
  size_t containerSize = *(static_cast<size_t*>(ret.Address()));

#else
  /*
  Reflex::Type retType =  sizeMethod.TypeOf().ReturnType();  
  char* retbuf = ::new char[retType.SizeOf()];
  Reflex::Object ret(retType, retbuf);
  sizeMethod.Invoke(theContainer, &ret);
  */
  size_t containerSize = 0;
  sizeMethod.Invoke(theContainer, containerSize);
  
  // if( !ret.TypeOf() || !ret.Address() )
  //  throw std::runtime_error( "Could not invoke the size method on the container" );

  // size_t containerSize = *(static_cast<size_t*>(ret.Address()));
  
  // delete [] retbuf;
      
#endif
  
  // Retrieve the element size
  Reflex::Member beginMethod = m_type.MemberByName( "begin" );
  if ( ! beginMethod )
    throw std::runtime_error( "No begin method is defined for the container" );
  Reflex::Type iteratorType = beginMethod.TypeOf().ReturnType();
  Reflex::Member dereferenceMethod = iteratorType.MemberByName( "operator*" );
  if ( ! dereferenceMethod )
    throw std::runtime_error( "Could not retrieve the dereference method of the container's iterator" );
  size_t elementSize = dereferenceMethod.TypeOf().ReturnType().SizeOf();

  // allocate the blob
  m_blob.resize( containerSize * elementSize );
  void* startingAddress = m_blob.startingAddress();

  // Create an iterator
#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
  Reflex::Object iteratorObject = beginMethod.Invoke( Reflex::Object( m_type, const_cast< void * > ( addressOfInputData ) ) );
#else
  Reflex::Type retType2 =  beginMethod.TypeOf().ReturnType();
  char* retbuf2 = ::new char[retType2.SizeOf()];
  Reflex::Object iteratorObject(retType2, retbuf2);
  beginMethod.Invoke( Reflex::Object( m_type, const_cast< void * > ( addressOfInputData ) ), &iteratorObject );
#endif

  // Loop over the elements of the container
  Reflex::Member incrementMethod = iteratorObject.TypeOf().MemberByName( "operator++" );
  if ( ! incrementMethod )
    throw std::runtime_error( "Could not retrieve the increment method of the container's iterator" );

  /*
#if ROOT_VERSION_CODE >= ROOT_VERSION(5,21,3)
    Reflex::Type retType3 =  dereferenceMethod.TypeOf().ReturnType();
    char* retbuf3 = ::new char[retType3.SizeOf()];
#endif
  */

  for ( size_t i = 0; i < containerSize; ++i ) {

#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
    void* elementAddress = dereferenceMethod.Invoke( iteratorObject ).Address();
#else
    /*Reflex::Object temp(retType3, retbuf3);

    dereferenceMethod.Invoke( iteratorObject, &temp);
    void* elementAddress = temp.Address();
    */
    void* elementAddress = 0;
    dereferenceMethod.Invoke( iteratorObject, elementAddress);
#endif
    ::memcpy( startingAddress, elementAddress, elementSize );
    char* cstartingAddress = static_cast<char*>( startingAddress );
    cstartingAddress += elementSize;
    startingAddress = cstartingAddress;

#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
    incrementMethod.Invoke( iteratorObject);
#else
    incrementMethod.Invoke( iteratorObject, 0);
#endif
  }

  /*
#if ROOT_VERSION_CODE >= ROOT_VERSION(5,21,3)
  delete [] retbuf3;
#endif
  */

  // Destroy the iterator
  iteratorObject.Destruct();
  return m_blob;
}



pool::BlobReader::BlobReader( const Reflex::Type& type ):
  m_type( type )
{}


pool::BlobReader::~BlobReader()
{}


void
pool::BlobReader::read( const coral::Blob& blobData,
                        void* containerAddress ) const
{
  // Retrieve the element size
  Reflex::Member beginMethod = m_type.MemberByName( "begin" );
  if ( ! beginMethod )
    throw std::runtime_error( "No begin method is defined for the container" );
  Reflex::Type iteratorType = beginMethod.TypeOf().ReturnType();
  Reflex::Member dereferenceMethod = iteratorType.MemberByName( "operator*" );
  if ( ! dereferenceMethod )
    throw std::runtime_error( "Could not retrieve the dereference method of the container's iterator" );

  size_t elementSize = dereferenceMethod.TypeOf().ReturnType().SizeOf();

  // Retrieve the container size
  size_t contrainerSize = blobData.size() / elementSize;

  // Retrieve the end method
  Reflex::Member endMethod = m_type.MemberByName( "end" );
  if ( ! endMethod )
   throw std::runtime_error( "Could not retrieve the end method of the container" );

  // Retrieve the insert method
  Reflex::Member insertMethod;
  for( unsigned int i = 0; i < m_type.FunctionMemberSize();i++){
    Reflex::Member im = m_type.FunctionMemberAt(i);
    if( im.Name() != std::string( "insert" ) ) continue;
    if( im.TypeOf().FunctionParameterSize() != 2) continue;
    insertMethod = im;
    break;
  }

  // Retrieve the clear method
  Reflex::Member clearMethod = m_type.MemberByName( "clear" );
  if ( ! clearMethod )
   throw std::runtime_error( "Could not retrieve the clear method of the container" );

  // Clear the container
  Reflex::Object containerObject( m_type, containerAddress );

#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
  clearMethod.Invoke( containerObject);
#else
  clearMethod.Invoke( containerObject ,0 );
#endif

  // Fill-in the elements
  const void* startingAddress = blobData.startingAddress();
  for ( size_t i = 0; i < contrainerSize; ++i ) {
    std::vector< void* > args( 2 );

    // call container.end()
#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
    args[0] = endMethod.Invoke( containerObject ).Address();
#else
    Reflex::Type retType =  endMethod.TypeOf().ReturnType();
    char* retbuf = ::new char[retType.SizeOf()];
    Reflex::Object ret(retType, retbuf);
    endMethod.Invoke(containerObject, &ret);
    args[0] = ret.Address();
    //    delete [] retbuf;
#endif
    
    // call container.insert( container.end(), data )
    args[1] = const_cast<void*>( startingAddress );
#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
    Reflex::Object pos = insertMethod.Invoke( containerObject, args );
#else
    insertMethod.Invoke( containerObject, 0, args );
#endif

    // this is required! (as it was in Reflect). 
    iteratorType.Destruct( args[0] );

    // this is required ! (as it was in Reflect). 
#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
    iteratorType.Destruct( pos.Address() );
#endif

    const char* cstartingAddress = static_cast<const char*>( startingAddress );
    cstartingAddress += elementSize;
    startingAddress = cstartingAddress;
#if ROOT_VERSION_CODE >= ROOT_VERSION(5,21,3)
    //    delete [] retbuf1;
#endif
  }

}

