#include "TestDriver.h"
#include "TestClasses.h"
#include "RelationalAccess/ConnectionService.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/ITableDescription.h"
#include "RelationalAccess/ITransaction.h"
#include "Collection/Collection.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/ISession.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/IFCAction.h"
#include "FileCatalog/IFCContainer.h"
#include "FileCatalog/FCEntry.h"
#include "FileCatalog/FCException.h"
#include <cstdlib>
#include <stdexcept>
#include "ObjectRelationalAccess/ObjectRelationalMappingUtilities.h"
static std::string testClassLibrary("test_RelationalStorageService_ComplexDataDict");
static std::string inputFile("input.xml");
static std::string outputFile("output.xml");

pool::TestDriver::TestDriver( const std::string& conn):TestBase(),
                                                        m_connection(conn){
  setMessageVerbosityLevel( coral::Debug );  
  std::vector<std::string> libs;
  libs.push_back(testClassLibrary);
  init(libs);
}

pool::TestDriver::~TestDriver(){
}

void pool::TestDriver::createDatabase(unsigned int nobjects){
  std::cout << "##### creating and filling POOL Database " << std::endl;  
  
  pool::Placement place1, place2;
  place1.setDatabase(m_connection, pool::DatabaseSpecification::PFN );
  place1.setContainerName("Cont1");
  place1.setTechnology(pool::POOL_RDBMS_HOMOGENEOUS_StorageType.type());
  place2.setDatabase(m_connection, pool::DatabaseSpecification::PFN );
  place2.setContainerName("Cont2");
  place2.setTechnology(pool::POOL_RDBMS_POLYMORPHIC_StorageType .type());
  pool::Placement place3;
  place3.setDatabase(m_connection, pool::DatabaseSpecification::PFN );
  place3.setContainerName("Cont3");
  place3.setTechnology(pool::POOL_RDBMS_HOMOGENEOUS_StorageType.type());
  pool::Placement place4;
  place4.setDatabase(m_connection, pool::DatabaseSpecification::PFN );
  place4.setContainerName("BlobComplex_basic");
  place4.setTechnology(pool::POOL_RDBMS_HOMOGENEOUS_StorageType.type());

  std::cout << "Starting UPDATE transaction..." << std::endl;
  // Start an update transaction
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
    printError("test ERROR: Could not start an UPDATE transaction");
  }  

  unsigned int i;
  for (i = 0; i < nobjects; ++i )   {
    pool::Ref<BlobComplex> blobComplex(m_dataSvc,new BlobComplex);
    unsigned int loc_i = i;
    blobComplex->fill(loc_i);
    try
    {
      blobComplex.markWrite(place4);
    } catch (const pool::RefException& e){
      printException(e);
    }      
  }
      
  std::cout << "Commit transaction..." << std::endl;
  if ( ! m_dataSvc->transaction().commit() ) {
     printError("test ERROR: Could not commit the WRITE transaction." );
  }
  std::cout << OVALTAG << "Written on disk "<< i << " objects\n";

  m_dataSvc->session().disconnectAll();

}

void pool::TestDriver::cleanUp(){
  std::cout << "##### cleaning up existing schema " << std::endl;

  coral::ConnectionService connServ;
  coral::ISessionProxy* session = connServ.connect( m_connection, coral::Update );
  coral::ISchema& schema = session->nominalSchema();
  session->transaction().start();
  schema.dropIfExistsTable( "POOL_OR_MAPPING_COLUMNS" );
  schema.dropIfExistsTable( "POOL_OR_MAPPING_ELEMENTS" );
  schema.dropIfExistsTable( "POOL_OR_CLASS_VERSIONS" );
  schema.dropIfExistsTable( "POOL_OR_MAPPING_VERSIONS" );
  schema.dropIfExistsTable( "POOL_CONT_OTHER" );
  schema.dropIfExistsTable( "POOL_CONT___LINKS" );
  schema.dropIfExistsTable( "POOL_CONT___PARAMS" );
  schema.dropIfExistsTable( "POOL_CONT___SHAPES" );
  schema.dropIfExistsTable( "POOL_RSS_CONTAINERS" );
  schema.dropIfExistsTable( "POOL_RSS_DB" );
  schema.dropIfExistsTable( "POOL_RSS_SEQ" );
  schema.dropIfExistsTable( "BLOBCOMPLEX_OBJECT_2" );
  schema.dropIfExistsTable( "BLOBCOMPLEX_OBJECT_3" );
  schema.dropIfExistsTable( "BLOBCOMPLEX_OBJECT_4" );
  schema.dropIfExistsTable( "BLOBCOMPLEX_OBJECT_5" );
  schema.dropIfExistsTable( "BLOBCOMPLEX_OBJECT_1" );
  schema.dropIfExistsTable( "BLOBCOMPLEX_OBJECTS" );
  schema.dropIfExistsTable( "BLOBCOMPLEX" );
  session->transaction().commit();
  delete session;
  std::cout << OVALTAG <<  "All POOL Database Tables removed." <<std::endl; 
}

void pool::TestDriver::readBackData(unsigned int expected){

  std::cout << "##### reading back data from POOL database " << std::endl;

  // Start a read transaction
  m_dataSvc->transaction().start( pool::ITransaction::READ );
  unsigned int numberOfObj = 0;
  pool::Collection< BlobComplex > blobCollection( m_dataSvc,
                                                   "ImplicitCollection",
                                                   "PFN:" + m_connection,
                                                   "BlobComplex_basic",
                                                   pool::ICollection::READ );
  pool::Collection< BlobComplex >::Iterator blobObj = blobCollection.select();

  numberOfObj = 0;
  while ( blobObj.next() ) {
    unsigned int seed = blobObj.ref()->index;
    BlobComplex r;
    r.fill(seed);
    if(blobObj.ref()->operator!=(r)){
      printError("TEST ERROR: BlobComplex object read back different from expected.");
      break;
    }
    numberOfObj++;
  }
  std::cout << OVALTAG << "Read back " <<numberOfObj<< " object of type BlobComplex" <<std::endl;
                   
  if(numberOfObj!=expected){
    printError("TEST ERROR: number of BlobComplex objects read back different from expected.");
  }

  // commit the transaction
  m_dataSvc->transaction().commit();
  m_dataSvc->session().disconnectAll();

}
#ifdef _WIN32
#include <windows.h>
#endif

void pool::TestDriver::execute(){ 
  cleanUp();
  //runCustomMapping();
  createDatabase(20);
#ifdef _WIN32
  ::Sleep(1000);
#else
  sleep(1);
#endif
  readBackData(20);
  cleanUp();
}
