#include "TestDriver.h"
#include "TestClasses.h"
#include "RelationalAccess/IRelationalService.h"
#include "RelationalAccess/IRelationalDomain.h"
#include "RelationalAccess/ConnectionService.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/ITableDescription.h"
#include "RelationalAccess/ITransaction.h"
#include "Collection/Collection.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/ISession.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/IFCAction.h"
#include "FileCatalog/IFCContainer.h"
#include "FileCatalog/FCEntry.h"
#include "FileCatalog/FCException.h"
#include <cstdlib>
#include <stdexcept>
#include "CoralCommon/Sleep.h"

static std::string testClassLibrary("test_RelationalStorageService_PoolPtrIODict");

pool::TestDriver::TestDriver( const std::string& connectionString):TestBase(),m_connectionString(connectionString){
  std::vector<std::string> libs;
  libs.push_back(testClassLibrary);
  init(libs);
  setMessageVerbosityLevel( coral::Debug );
}

pool::TestDriver::~TestDriver(){
}

void pool::TestDriver::createDatabase(unsigned int nobjects){
  std::cout << "##### creating and filling POOL Database " << std::endl;  
  
  pool::Placement place1, place2;
  place1.setDatabase(m_connectionString, pool::DatabaseSpecification::PFN );
  place1.setContainerName("Cont1");
  place1.setTechnology(pool::POOL_RDBMS_HOMOGENEOUS_StorageType.type());
  place2.setDatabase(m_connectionString, pool::DatabaseSpecification::PFN );
  place2.setContainerName("Cont2");
  place2.setTechnology(pool::POOL_RDBMS_POLYMORPHIC_StorageType .type());

  std::cout << "Starting UPDATE transaction..." << std::endl;
  // Start an update transaction
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
    printError("test ERROR: Could not start an UPDATE transaction");
  }

  unsigned int i;
  for (i = 0; i < nobjects; ++i )   {
    pool::Ref<SimplePtrClass> simple(m_dataSvc,new SimplePtrClass(i));

    try{
      simple.markWrite(place1);
    } catch (const pool::RefException& e){
      printException(e);        
    }
  }
      
  std::cout << "Commit transaction..." << std::endl;
  if ( ! m_dataSvc->transaction().commit() ) {
     printError("test ERROR: Could not commit the WRITE transaction." );
  }
  std::cout << OVALTAG << "Written on disk "<< i << " objects\n";

  m_dataSvc->session().disconnectAll();

}

void pool::TestDriver::cleanUp(){
  std::cout << "##### cleaning up existing schema " << std::endl;

  coral::ConnectionService  connServ;
  coral::ISessionProxy* session = connServ.connect(m_connectionString );
  coral::ISchema& schema = session->nominalSchema();
  session->transaction().start();
  schema.dropIfExistsTable( "POOL_OR_MAPPING_COLUMNS" );
  schema.dropIfExistsTable( "POOL_OR_MAPPING_ELEMENTS" );
  schema.dropIfExistsTable( "POOL_OR_CLASS_VERSIONS" );
  schema.dropIfExistsTable( "POOL_OR_MAPPING_VERSIONS" );
  schema.dropIfExistsTable( "POOL_CONT___LINKS" );
  schema.dropIfExistsTable( "POOL_CONT___PARAMS" );
  schema.dropIfExistsTable( "POOL_CONT___SHAPES" );
  schema.dropIfExistsTable( "POOL_RSS_CONTAINERS" );
  schema.dropIfExistsTable( "POOL_RSS_DB" );
  schema.dropIfExistsTable( "POOL_RSS_SEQ" );
  schema.dropIfExistsTable( "SIMPLEPTRCLASS_ARR" );
  schema.dropIfExistsTable( "SIMPLEPTRCLASS_MEM" );
  schema.dropIfExistsTable( "SIMPLEPTRCLASS" );
  session->transaction().commit();
  delete session;
  std::cout << OVALTAG <<  "All POOL Database Tables removed." <<std::endl; 
}

void pool::TestDriver::readBackData(unsigned int expected){

  std::cout << "##### reading back data from POOL database " << std::endl;

  // Start read transaction I
  m_dataSvc->transaction().start( pool::ITransaction::READ );
  pool::Collection< SimplePtrClass > simpleCollection0( m_dataSvc,
                                                   "ImplicitCollection",
                                                   "PFN:" + m_connectionString,
                                                   "Cont1",
                                                   pool::ICollection::READ );
  pool::Collection< SimplePtrClass >::Iterator simpleObj0 = simpleCollection0.select();
  unsigned int numberOfObj = 0;
  while ( simpleObj0.next() ) {
    // here loading all data including the ptr
    simpleObj0.ref()->mid;
    int seed = simpleObj0.ref()->mid;
    // now accessing the ptr
    std::cout << "Loader="<<simpleObj0.ref()->mem.loader()<<std::endl;
  try {
    simpleObj0.ref()->mem->m_id;
  } catch (const pool::Exception& e){
    std::cout << "ACCESSING PTR"<<e.what()<<std::endl;
  }
  
    SimplePtrClass r(seed);
    if(simpleObj0.ref()->operator!=(r)){
      printError("TEST ERROR: SimpleClass object read back different from expected.");
      break;
    }
    numberOfObj++;
  }
  std::cout << OVALTAG << "Read back " <<numberOfObj<< " object of type SimplePtrClass" <<std::endl;
                   
  if(numberOfObj!=expected){
    printError("TEST ERROR: number of SimplePtrClass objects read back different from expected.");
  }
  // commit the transaction
  m_dataSvc->transaction().commit();
  
  std::cout << "##### reading again... " << std::endl;
  pool::Ref< SimplePtrClass > lastRef;
  // Start read transaction I
  m_dataSvc->transaction().start( pool::ITransaction::READ );
  pool::Collection< SimplePtrClass > simpleCollection1( m_dataSvc,
                                                   "ImplicitCollection",
                                                   "PFN:" + m_connectionString,
                                                   "Cont1",
                                                   pool::ICollection::READ );
  pool::Collection< SimplePtrClass >::Iterator simpleObj1 = simpleCollection1.select();
  numberOfObj = 0;
  while ( simpleObj1.next() ) {
    // here loading all data but the ptr...
    simpleObj1.ref()->mid;
    numberOfObj++;
    lastRef = simpleObj1.ref();
  }
  std::cout << OVALTAG << "Read back " <<numberOfObj<< " object of type SimplePtrClass" <<std::endl;
                   
  if(numberOfObj!=expected){
    printError("TEST ERROR: number of SimplePtrClass objects read back different from expected.");
  }
  
  // commit the transaction
  m_dataSvc->transaction().commit();
  try {
    // now accessing back the ptr...=";
    lastRef->mem->m_id;    
  } catch (const pool::Exception& e){
    std::cout << OVALTAG << "OK, got the exception, underlying pointer is null."<<std::endl;
    //std::cout << e.what() << std::endl;
  }
  
  
  m_dataSvc->session().disconnectAll();

}

void pool::TestDriver::execute(){
  DClass* d = new DClass;
  BClass* b = d;
  if( b ){}
  pool::Ptr<BClass> p0(d);
  pool::Ptr<DClass> p1;
  p1.cast(p0);
  pool::Ptr<DClass> p2(d);
  BClass& br = *p0;
  std::cout << OVALTAG << "BR mi="<<br.mi<<std::endl;
  DClass* dp = p1.get();
  std::cout << OVALTAG<< "DP (0) mi="<<dp->mi<<" ml="<<dp->ml<<std::endl;
  dp = p1.get();
  std::cout << OVALTAG<< "DP (1) mi="<<dp->mi<<" ml="<<dp->ml<<std::endl;
  pool::Ptr<BClass> p3(p1);
  pool::Ptr<BClass> p4;
  p4 = p1;
  if(p1){
    if(p1==p0){
      std::cout << OVALTAG << "P1 and P0 not equals!"<<std::endl;
    }
    if( p0==p1){
      std::cout << OVALTAG << "P0 and P1 equals!"<<std::endl;
    }
    if(p1==p2){
      std::cout << OVALTAG << "P1 and P2 equals!"<<std::endl;
    }
    if(p1.get()==d){
      std::cout << OVALTAG << "P1 and d equals!"<<std::endl;
    }
    if(d==p0.get()){
      std::cout << OVALTAG << "d and P0 equals!"<<std::endl;
    }
    if(p1!=p0){
    }
    if(!p1){}
    if(p1==p3){
      std::cout << OVALTAG << "P1 and P3 equals!"<<std::endl;
    }
    if(p1==p4){
      std::cout << OVALTAG << "P1 and P4 equals!"<<std::endl;
    }
  }
  if(!p1.isLoaded()){
      std::cout << OVALTAG << "P1 is not loaded"<<std::endl;    
  }
  

  cleanUp();
  createDatabase(10);
  coral::sys::sleep(1);
  readBackData(10);
  cleanUp();
}

