#include "TestDriver.h"
#include "TestClasses.h"
#include "RelationalAccess/IRelationalService.h"
#include "RelationalAccess/IRelationalDomain.h"
#include "RelationalAccess/ConnectionService.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/ITableDescription.h"
#include "RelationalAccess/ITransaction.h"
#include "Collection/Collection.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/ISession.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/IFCAction.h"
#include "FileCatalog/IFCContainer.h"
#include "FileCatalog/FCEntry.h"
#include "FileCatalog/FCException.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingUtilities.h"
#include <cstdlib>
#include <stdexcept>

static std::string testClassLibrary("test_RelationalStorageService_SimpleWriteReadDict");
static std::string inputFile("input.xml");
static std::string outputFile("output.xml");

pool::TestDriver::TestDriver( const std::string& connectionString):TestBase(),m_connectionString(connectionString){
  std::vector<std::string> libs;
  libs.push_back(testClassLibrary);
  init(libs);
  setMessageVerbosityLevel( coral::Debug );
}

pool::TestDriver::~TestDriver(){
}

void pool::TestDriver::createDatabase(unsigned int nobjects){
  std::cout << "##### creating and filling POOL Database " << std::endl;  
  
  pool::Placement place1, place2;
  place1.setDatabase(m_connectionString, pool::DatabaseSpecification::PFN );
  place1.setContainerName("Cont1");
  place1.setTechnology(pool::POOL_RDBMS_HOMOGENEOUS_StorageType.type());
  place2.setDatabase(m_connectionString, pool::DatabaseSpecification::PFN );
  place2.setContainerName("Cont2");
  place2.setTechnology(pool::POOL_RDBMS_POLYMORPHIC_StorageType .type());
  

  std::cout << "Starting UPDATE transaction..." << std::endl;
  // Start an update transaction
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
    printError("test ERROR: Could not start an UPDATE transaction");
  }  

  unsigned int i;
  for (i = 0; i < nobjects; ++i )   {
    pool::Ref<SimpleClass> simple(m_dataSvc,new SimpleClass(i));
    pool::Ref<mynamespace::OtherClass> other(m_dataSvc,new mynamespace::OtherClass(i));
    pool::Ref<DataClass> data(m_dataSvc,new DataClass(i));

    try{
      simple.markWrite(place1);
    } catch (const pool::RefException& e){
      printException(e);        
    }
    try{
      other.markWrite(place2);
    } catch (const pool::RefException& e){
      printException(e);
    }
    try{
      data.markWrite(place2);
    } catch (const pool::RefException& e){
      printException(e);
    }
  }
      
  std::cout << "Commit transaction..." << std::endl;
  if ( ! m_dataSvc->transaction().commit() ) {
     printError("test ERROR: Could not commit the WRITE transaction." );
  }
  std::cout << OVALTAG << "Written on disk "<< i << " objects\n";

  m_dataSvc->session().disconnectAll();

}

void pool::TestDriver::cleanUp(){
  std::cout << "##### cleaning up existing schema " << std::endl;

  coral::ConnectionService  connServ;
  coral::ISessionProxy* session = connServ.connect(m_connectionString );
  coral::ISchema& schema = session->nominalSchema();
  session->transaction().start();
  schema.dropIfExistsTable( "POOL_OR_CLASS_VERSIONS" );
  schema.dropIfExistsTable( "POOL_OR_MAPPING_COLUMNS" );
  schema.dropIfExistsTable( "POOL_OR_MAPPING_ELEMENTS" );
  schema.dropIfExistsTable( "POOL_OR_MAPPING_VERSIONS" );
  schema.dropIfExistsTable( "POOL_CONT_OTHER" );
  schema.dropIfExistsTable( "POOL_CONT___LINKS" );
  schema.dropIfExistsTable( "POOL_CONT___PARAMS" );
  schema.dropIfExistsTable( "POOL_CONT___SHAPES" );
  schema.dropIfExistsTable( "POOL_RSS_CONTAINERS" );
  schema.dropIfExistsTable( "POOL_RSS_DB" );
  schema.dropIfExistsTable( "POOL_RSS_SEQ" );
  schema.dropIfExistsTable( "SIMPLEC_M_ARRAYDATA" );
  schema.dropIfExistsTable( "SIMPLEC_M_MAP" );
  schema.dropIfExistsTable( "SIMPLECLASS" );
  schema.dropIfExistsTable( "OTHERCL_M_ARRAYDATA" );
  schema.dropIfExistsTable( "OTHERCLASS" );
  schema.dropIfExistsTable( "DATACLASS" );
  schema.dropIfExistsTable( "POOL_CONT_CONT2" );
  session->transaction().commit();
  delete session;
  std::cout << OVALTAG <<  "All POOL Database Tables removed." <<std::endl; 
}

void pool::TestDriver::readBackData(unsigned int expected){

  std::cout << "##### reading back data from POOL database " << std::endl;

  // Start a read transaction
  m_dataSvc->transaction().start( pool::ITransaction::READ );
  pool::Collection< SimpleClass > simpleCollection( m_dataSvc,
                                                   "ImplicitCollection",
                                                   "PFN:" + m_connectionString,
                                                   "Cont1",
                                                   pool::ICollection::READ );
  pool::Collection< SimpleClass >::Iterator simpleObj = simpleCollection.select();

  unsigned int numberOfObj = 0;
  while ( simpleObj.next() ) {
    int seed = simpleObj.ref()->m_intData;
    SimpleClass r(seed);
    if(simpleObj.ref()->operator!=(r)){
      printError("TEST ERROR: SimpleClass object read back different from expected.");
      break;
    }
    numberOfObj++;
  }
  std::cout << OVALTAG << "Read back " <<numberOfObj<< " object of type SimpleClass" <<std::endl;
                   
  if(numberOfObj!=expected){
    printError("TEST ERROR: number of SimpleClass objects read back different from expected.");
  }
  
  pool::Collection< mynamespace::OtherClass > otherCollection( m_dataSvc,
                                                   "ImplicitCollection",
                                                   "PFN:" + m_connectionString,
                                                   "Cont2",
                                                   pool::ICollection::READ );
  pool::Collection< mynamespace::OtherClass >::Iterator otherObj = otherCollection.select();

  numberOfObj = 0;
  while ( otherObj.next() ) {
    int seed = otherObj.ref()->m_intData;
    mynamespace::OtherClass r(seed);
    if(otherObj.ref()->operator!=(r)){
      printError("TEST ERROR: OtherClass object read back different from expected.");
      break;
    }
    numberOfObj++;
  }
  std::cout << OVALTAG << "Read back " <<numberOfObj<< " object of type OtherClass" <<std::endl;
                                
  if(numberOfObj!=expected){
    printError("TEST ERROR: number of OtherClass objects read back different from expected.");
  }

  pool::Collection< DataClass > dataCollection( m_dataSvc,
                                                   "ImplicitCollection",
                                                   "PFN:" + m_connectionString,
                                                   "Cont2",
                                                   pool::ICollection::READ );
  pool::Collection< DataClass >::Iterator dataObj = dataCollection.select();

  numberOfObj = 0;
  while ( dataObj.next() ) {
    int seed = dataObj.ref()->m_i;
    DataClass r(seed);
    if(dataObj.ref()->operator!=(r)){
      printError("TEST ERROR: DataClass object read back different from expected.");
      break;
    }
    numberOfObj++;
  }
  std::cout << OVALTAG << "Read back " <<numberOfObj<< " object of type DataClass" <<std::endl;
                   
  if(numberOfObj!=expected){
    printError("TEST ERROR: number of DataClass objects read back different from expected.");
  }
  
  // commit the transaction
  m_dataSvc->transaction().commit();
  m_dataSvc->session().disconnectAll();

}

#ifdef _WIN32
#include <windows.h>
#endif

void pool::TestDriver::execute(){ 
  cleanUp();
  createDatabase(100);
#ifdef _WIN32
  :: Sleep(1000);
#else
  sleep(1);
#endif
  readBackData(100);
  cleanUp();
}
