#include "TestDriver.h"
#include "TestDictionary/SimpleTestClass.h"
#include "TestDictionary/TestClassPrimitives.h"

#include "Reflex/Type.h"
#include "Reflex/SharedLibrary.h"
#include "libname.h"

#include "POOLCore/Guid.h"
#include "POOLCore/Token.h"


#include "StorageSvc/Shape.h"
#include "StorageSvc/poolDb.h"
#include "StorageSvc/IStorageSvc.h"
#include "StorageSvc/IStorageExplorer.h"
#include "StorageSvc/DbSelect.h"
#include "StorageSvc/DbReflex.h"
#include "StorageSvc/DatabaseConnection.h"
#include "StorageSvc/FileDescriptor.h"
#include "StorageSvc/DataCallBack.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/DbObjectCallBack.h"

#include <stdexcept>
#include <iostream>
#include <sstream>
#include <memory>

#include "RVersion.h"

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
using namespace ROOT;
#endif

namespace pool { int numTokenInstances(); }

static const std::string file = "pool_test.root";
static const std::string container = "container";
static const int nObjects = 100;


TestDriver::TestDriver()
{
  
  // pool::POOLContext::setMessageVerbosityLevel( coral::Info );
}

TestDriver::~TestDriver()
{
  std::cout << "[OVAL] Number of floating tokens : " << pool::numTokenInstances() << std::endl;
}

void
TestDriver::loadLibraries( const std::vector<std::string>& libraries )
{
  for ( std::vector<std::string>::const_iterator iLibrary = libraries.begin();
	iLibrary != libraries.end(); ++iLibrary ) {
    const std::string& library = *iLibrary;
    std::cout << "Loading library " << library << std::endl;

    Reflex::SharedLibrary libloader( libname(library) );
    
    if(!libloader.Load()){
      std::cout << libloader.Error() << std::endl;
    }
  }
}

void
TestDriver::testWriting()
{
  pool::IStorageSvc* storSvc = pool::createStorageSvc( 0, "StorageSvc");
  if ( ! storSvc ) {
    throw std::runtime_error( "Could not create a StorageSvc object" );
  }
  storSvc->addRef();
  pool::Session* sessionHandle = 0;
  if ( ! ( storSvc->startSession( poolDb::CREATE, pool::ROOT_StorageType.type(), sessionHandle ).isSuccess() ) ) {
    throw std::runtime_error( "Could not start a session." );
  }

  pool::FileDescriptor* fd = new pool::FileDescriptor( file, file );
  if ( ! ( storSvc->connect( sessionHandle, poolDb::CREATE, *fd ).isSuccess() ) ) {
    throw std::runtime_error( "Could not start a connection." );
  }
  pool::DatabaseConnection* connection = fd->dbc();
  pool::Transaction* transaction = 0;
  if ( ! ( storSvc->startTransaction( connection, transaction ).isSuccess() ) ) {
    throw std::runtime_error( "Could not start a transaction." );
  }

  // Retrieve the dictionary
  const Reflex::Type class_SimpleTestClass = Reflex::Type::ByName( "SimpleTestClass" );
  if ( ! class_SimpleTestClass ) {
    throw std::runtime_error( "Could not retrieve the dictionary for class SimpleTestClass" );
  }
  //std::string guid_nam_SimpleTestClass = class_SimpleTestClass->propertyList()->getProperty( "ClassID" );
  //if ( guid_nam_SimpleTestClass.empty() ) {
  //  std::ostringstream error;
  //  error << "There is no ClassID property for class \"SimpleTestClass\"" << std::ends;
  //  throw std::runtime_error( error.str() );
  //}
  //pool::Guid guid_SimpleTestClass(guid_nam_SimpleTestClass);
  const Reflex::Type class_TestClassPrimitives = Reflex::Type::ByName( "TestClassPrimitives" );
  if ( ! class_TestClassPrimitives ) {
    throw std::runtime_error( "Could not retrieve the dictionary for class TestClassPrimitives" );
  }
  //std::string guid_nam_TestClassPrimitives = class_TestClassPrimitives->propertyList()->getProperty( "ClassID" );
  //if ( guid_nam_TestClassPrimitives.empty() ) {
  //  std::ostringstream error;
  //  error << "There is no ClassID property for class \"TestClassPrimitives\"" << std::ends;
  //  throw std::runtime_error( error.str() );
  //}
  //pool::Guid guid_TestClassPrimitives(guid_nam_TestClassPrimitives);

  std::vector< SimpleTestClass* > myObjects_SimpleTestClass;
  std::vector< TestClassPrimitives* > myObjects_TestClassPrimitives;
  for ( int i = 0; i < nObjects; ++i ) {
    myObjects_SimpleTestClass.push_back( new SimpleTestClass() );
    SimpleTestClass* myObject_SimpleTestClass = myObjects_SimpleTestClass.back();
    myObject_SimpleTestClass->data = i;

    // Creating a callback object
    std::auto_ptr<pool::DataCallBack> cb_SimpleTestClass( new pool::DbObjectCallBack(class_SimpleTestClass ) );
    cb_SimpleTestClass->setObject( myObject_SimpleTestClass );

    // Creating the persistent shape.
    const pool::Shape* shape_SimpleTestClass = 0;
    if ( storSvc->getShape( *fd, class_SimpleTestClass, shape_SimpleTestClass ) == pool::IStorageSvc::SHAPE_NOT_AVAILIBLE ) {
      storSvc->createShape( *fd, container, class_SimpleTestClass, shape_SimpleTestClass );
    }
    if ( ! shape_SimpleTestClass ) {
      throw std::runtime_error( "Could not create a persistent shape." );
    }
  
    // Writing the object.
    pool::Token* token_SimpleTestClass;
    if ( ! ( storSvc->allocate( transaction, *fd,
				container, pool::ROOTKEY_StorageType.type(),
				shape_SimpleTestClass, cb_SimpleTestClass.get(), token_SimpleTestClass ).isSuccess() ) ) {
      throw std::runtime_error( "Could not write an object" );
    }
    //token_SimpleTestClass->setClassID( guid_SimpleTestClass );
    cb_SimpleTestClass.release();
    delete token_SimpleTestClass;


    // The second class
    myObjects_TestClassPrimitives.push_back( new TestClassPrimitives() );
    TestClassPrimitives* myObject_TestClassPrimitives = myObjects_TestClassPrimitives.back();
    myObject_TestClassPrimitives->setNonZero();

    // Creating a callback object
    std::auto_ptr<pool::DataCallBack> cb_TestClassPrimitives( new pool::DbObjectCallBack( class_TestClassPrimitives ) );
    cb_TestClassPrimitives->setObject( myObject_TestClassPrimitives );

    // Creating the persistent shape.
    const pool::Shape* shape_TestClassPrimitives = 0;
    if ( storSvc->getShape( *fd, class_TestClassPrimitives, shape_TestClassPrimitives ) == pool::IStorageSvc::SHAPE_NOT_AVAILIBLE ) {
      storSvc->createShape( *fd, container, class_TestClassPrimitives, shape_TestClassPrimitives );
    }
    if ( ! shape_TestClassPrimitives ) {
      throw std::runtime_error( "Could not create a persistent shape." );
    }
  
    // Writing the object.
    pool::Token* token_TestClassPrimitives;
    if ( ! ( storSvc->allocate( transaction, *fd,
				container, pool::ROOTKEY_StorageType.type(),
				shape_TestClassPrimitives, cb_TestClassPrimitives.get(), token_TestClassPrimitives ).isSuccess() ) ) {
      throw std::runtime_error( "Could not write an object" );
    }
    //token_TestClassPrimitives->setClassID( guid_TestClassPrimitives );
    cb_TestClassPrimitives.release();
    delete token_TestClassPrimitives;

  }

  // Closing the transaction.
  if ( ! ( storSvc->endTransaction( transaction, pool::Transaction::TRANSACT_COMMIT ).isSuccess() ) ) {
    throw std::runtime_error( "Could not end a transaction." );
  }

  // Clearing the cache
  for ( std::vector< SimpleTestClass* >::iterator iObject_SimpleTestClass = myObjects_SimpleTestClass.begin();
	iObject_SimpleTestClass != myObjects_SimpleTestClass.end(); ++iObject_SimpleTestClass ) delete *iObject_SimpleTestClass;
  myObjects_SimpleTestClass.clear();

  for ( std::vector< TestClassPrimitives* >::iterator iObject_TestClassPrimitives = myObjects_TestClassPrimitives.begin();
	iObject_TestClassPrimitives != myObjects_TestClassPrimitives.end(); ++iObject_TestClassPrimitives ) delete *iObject_TestClassPrimitives;
  myObjects_TestClassPrimitives.clear();

  // Disconnecting
  if ( ! ( storSvc->disconnect( *fd ).isSuccess() ) ) {
    throw std::runtime_error( "Could not disconnect." );
  }
  delete fd;
  if ( ! ( storSvc->endSession( sessionHandle ).isSuccess() ) ) {
    throw std::runtime_error( "Could not end correctly the session." );
  }
  storSvc->release();
}


void
TestDriver::testReadingParallelSameContainer()
{
  pool::IStorageSvc* storSvc = pool::createStorageSvc( 0, "StorageSvc");
  if ( ! storSvc ) {
    throw std::runtime_error( "Could not create a StorageSvc object" );
  }
  storSvc->addRef();
  void* pVoid = 0;
  pool::DbStatus sc = storSvc->queryInterface( pool::IStorageExplorer::interfaceID(), &pVoid );
  pool::IStorageExplorer* storageExplorer = (pool::IStorageExplorer*)pVoid;
  if ( !( sc == pool::DbStatus::Success && storageExplorer ) ) {
    storSvc->release();
    throw std::runtime_error( "Could not retrieve a IStorageExplorer interface" );
  }

  pool::Session* sessionHandle = 0;
  if ( ! ( storSvc->startSession( poolDb::READ, pool::ROOT_StorageType.type(), sessionHandle ).isSuccess() ) ) {
    throw std::runtime_error( "Could not start a session." );
  }

  pool::FileDescriptor* fd = new pool::FileDescriptor( file, file );
  sc = storSvc->connect( sessionHandle, poolDb::READ, *fd );
  if ( sc != pool::DbStatus::Success ) {
    throw std::runtime_error( "Could not start a connection." );
  }

  pool::DatabaseConnection* connection = fd->dbc();
  pool::Transaction* transaction = 0;
  if ( ! ( storSvc->startTransaction( connection, transaction ).isSuccess() ) ) {
    throw std::runtime_error( "Could not start a transaction." );
  }

  // Fetch the containers
  std::vector<const pool::Token*> containerTokens;
  storageExplorer->containers( *fd, containerTokens );
  if ( containerTokens.size() != 1 ) {
    throw std::runtime_error( "Unexpected number of containers" );
  }
  const pool::Token* containerToken = containerTokens.front();
  const std::string containerName = containerToken->contID();
  if ( containerName != container ) {
    throw std::runtime_error( "Container name read is different from the container name written" );
  }

  // Fetch the objects in the container (Initialize the iterators)
  pool::DbSelect selectionObject_SimpleTestClass("");
  sc = storageExplorer->select( *fd, containerToken->contID(), selectionObject_SimpleTestClass );
  if ( ! sc.isSuccess() ) {
    throw std::runtime_error( "Could not start an implicit collection iteration" );
  }
  pool::DbSelect selectionObject_TestClassPrimitives("");
  sc = storageExplorer->select( *fd, containerToken->contID(), selectionObject_TestClassPrimitives );
  if ( ! sc.isSuccess() ) {
    throw std::runtime_error( "Could not start an implicit collection iteration" );
  }

  // Start the parallel iteration
  int iObject = 0;
  while ( iObject < nObjects ) {

    // First iterator
    pool::Token* objectToken_SimpleTestClass = 0;
    if ( ! ( storageExplorer->next( selectionObject_SimpleTestClass, objectToken_SimpleTestClass ).isSuccess() ) ) {
      throw std::runtime_error( "Could not retrieve the object token" );
    }

    const pool::Shape* shape_SimpleTestClass = 0;
    if ( storSvc->getShape( *fd, objectToken_SimpleTestClass->classID(), shape_SimpleTestClass ) != pool::IStorageSvc::IS_PERSISTENT_SHAPE ) {
      throw std::runtime_error( "Could not fetch the persistent shape for SimpleTestClass" );
    }
    pool::DataCallBack* cb_SimpleTestClass = new pool::DbObjectCallBack( pool::DbReflex::forGuid( objectToken_SimpleTestClass->classID() ) );
    cb_SimpleTestClass->setHandler( 0 );
    Reflex::Type classH_SimpleTestClass;
    if ( ! ( storSvc->read( *fd, *objectToken_SimpleTestClass, cb_SimpleTestClass, shape_SimpleTestClass, classH_SimpleTestClass ) ).isSuccess() ) {
      throw std::runtime_error( "failed to read a SimpleTestClass object back from the persistency" );
    }
    
    if ( classH_SimpleTestClass.Name() != "SimpleTestClass" ) {
      throw std::runtime_error( "read wrong class type." );
    }

    SimpleTestClass* object_SimpleTestClass = reinterpret_cast< SimpleTestClass* >( const_cast<void*>( cb_SimpleTestClass->object() ) );
    if ( object_SimpleTestClass->data != iObject ) {
      throw std::runtime_error( "Object read different from object written" );
    }
    delete object_SimpleTestClass;
    delete cb_SimpleTestClass;
    objectToken_SimpleTestClass->release();

    // Skip the next object for this iterator
    objectToken_SimpleTestClass = 0;
    if ( ! ( storageExplorer->next( selectionObject_SimpleTestClass, objectToken_SimpleTestClass ).isSuccess() ) ) {
      throw std::runtime_error( "Could not retrieve the object token" );
    }
    objectToken_SimpleTestClass->release();


    // Second iterator
    // Skip the next object for this iterator
    pool::Token* objectToken_TestClassPrimitives = 0;
    if ( ! ( storageExplorer->next( selectionObject_TestClassPrimitives, objectToken_TestClassPrimitives ).isSuccess() ) ) {
      throw std::runtime_error( "Could not retrieve the object token" );
    }
    objectToken_TestClassPrimitives->release();

    // Read the next object for this iterator
    objectToken_TestClassPrimitives = 0;
    if ( ! ( storageExplorer->next( selectionObject_TestClassPrimitives, objectToken_TestClassPrimitives ).isSuccess() ) ) {
      throw std::runtime_error( "Could not retrieve the object token" );
    }

    const pool::Shape* shape_TestClassPrimitives = 0;
    if ( storSvc->getShape( *fd, objectToken_TestClassPrimitives->classID(), shape_TestClassPrimitives ) != pool::IStorageSvc::IS_PERSISTENT_SHAPE ) {
      throw std::runtime_error( "Could not fetch the persistent shape for TestClassPrimitives" );
    }
    pool::DataCallBack* cb_TestClassPrimitives = new pool::DbObjectCallBack( pool::DbReflex::forGuid( objectToken_TestClassPrimitives->classID() ) );
    cb_TestClassPrimitives->setHandler( 0 );
    Reflex::Type classH_TestClassPrimitives;
    if ( ! ( storSvc->read( *fd, *objectToken_TestClassPrimitives, cb_TestClassPrimitives, shape_TestClassPrimitives, classH_TestClassPrimitives ) ).isSuccess() ) {
      throw std::runtime_error( "failed to read a TestClassPrimitives object back from the persistency" );
    }
    
    if ( classH_TestClassPrimitives.Name() != "TestClassPrimitives" ) {
      throw std::runtime_error( "read wrong class type." );
    }

    TestClassPrimitives* object_TestClassPrimitives = reinterpret_cast< TestClassPrimitives* >( const_cast<void*>( cb_TestClassPrimitives->object() ) );
    TestClassPrimitives referenceObject;
    referenceObject.setNonZero();
    if ( *object_TestClassPrimitives != referenceObject ) {
      throw std::runtime_error( "Object read different from object written" );
    }
    delete object_TestClassPrimitives;
    delete cb_TestClassPrimitives;
    objectToken_TestClassPrimitives->release();

    // Continue the iteration
    ++iObject;
  }


  if ( iObject != nObjects ) {
    throw std::runtime_error( "Objects read different from objects written" );
  }

  // Closing the transaction.
  if ( ! ( storSvc->endTransaction( transaction, pool::Transaction::TRANSACT_COMMIT ).isSuccess() ) ) {
    throw std::runtime_error( "Could not end a transaction." );
  }

  if ( ! ( storSvc->disconnect( *fd ).isSuccess() ) ) {
    throw std::runtime_error( "Could not disconnect." );
  }
  delete fd;

  std::cout << "Closing the session" << std::endl;
  if ( ! ( storSvc->endSession( sessionHandle ).isSuccess() ) ) {
    throw std::runtime_error( "Could not end correctly the session." );
  }
  storageExplorer->release();
  storSvc->release();
}
