/**
 * @file CollCopy.cpp
 * @brief Utility to copy POOL collections
 * @author K. Karr <Kristo.Karr@cern.ch>
 * $Id: CollCopy.cpp,v 1.28 2008/03/27 08:39:08 witoldp Exp $ 
 */

#include "CollectionBase/ICollection.h"
#include "CollectionBase/ICollectionFactory.h"
#include "CollectionBase/ICollectionIterator.h"
#include "AttributeList/AttributeListSpecification.h"
#include "AttributeList/AttributeList.h"
#include "FileCatalog/IFileCatalog.h"
#include "PersistencySvc/ISession.h"
#include "CoralBase/MessageStream.h"
#include "POOLCore/Token.h"
#include "POOLCore/Exception.h"
#include "CmdLineArgs.h"

#include "StorageSvc/DbStatus.h"

#include <iostream>
#include <string>
#include <map>


const std::string COLLECTION_CATALOG_CONNECTION =
  "xmlcatalog_file:CollectionCatalog.xml";


int main(int argc, const char *argv[])
{
  try
  { 
    pool::DbStatus sc = pool::DbStatus::Success;
    std::vector<std::string> srcCollNameVec;
    std::vector<std::string> srcCollTypeVec;
    std::vector<std::string> srcCollConnectVec;
    std::vector<std::string> dstCollNameVec;
    std::vector<std::string> dstCollTypeVec;
    std::vector<std::string> dstCollConnectVec;
    std::vector<std::string> collCatalogReadVec;

    CmdLineArgs cmdLineArgs(argc, argv);

    if( cmdLineArgs.hasQual("-help") ||
        !cmdLineArgs.hasQual("-src") ||
        !cmdLineArgs.hasQual("-dst") )
    {
      std::cerr << "Usage: CollCopy "
                << "<-src [+]<<input collection name> <input collection type>> > "
                << "(List of name-type pairs is space separated; "
                << "Choices for collection types are RootCollection, "
                << "MySQLCollection, MySQLltCollection, RelationalCollection, "
                << "PHYSICAL_NAME, LOGICAL_NAME or GUID) "
                << "<-dst [+]<<output collection name> <output collection type>> > "
                << "(WARNING: output collections will overwrite existing "
                << "ones of same names) "
                << "Optional arguments: "
                << "[-srcconnect [Input database connection string]] "
                << "(Note: The default value is \"\" and this argument MUST "
                << "be specified for collections of type MySQLCollection, "
                << "MySQLltCollection and RelationalCollection. Also, for "
                << "collections of type RelationalCollection the environment " 
                << "variables POOL_AUTH_USER and POOL_AUTH_PASSWORD "
                << "must be set) "
                << "[-dstconnect [Output database connection string]] "
                << "(Note: The default value is \"\" and for collections of "
                << "type MySQLCollection, MySQLltCollection and "
                << "RelationalCollection this will be set to the value of "
                << "the input database connection string if no value is "
                << "specified. In that case, the input database connection "
                << "string MUST be specified and satisfy the conditions "
                << "described above) "
                << "[-query [predicate string (DEFAULT = \"\")]] "
                << "[-ccread [+][read collection catalog connection string "
                << "(DEFAULT = \"\")]] "
		<< "[-queryopt [Choices are space separated list of the "
                << "following: \"FetchOne\" (no caching) or \"FetchAll\" "
                << "(caching) and/or \"SELECT <comma separated attribute list>\" "
                << "(DEFAULT = \"FetchOne SELECT *\")]] "
                << "[-ccwrite [write collection catalog connection string "
                << "(DEFAULT = \"\")]] "
                << "[-noattrib] (no attributes copied, only event tokens) "
                << "[-tokattrib <token attribute name>] (token attributes "
                << "of given name copied in place of event tokens) "
                << "[-nevtpercommit [Max. number of events to process between "
                << "collection commits (DEFAULT = 10000)]] "
                << "[-nevtperprint [Number of events to process before "
                << "each print of status to the screen. \"all\" => "
                << "skip printing for all events (DEFAULT = \"all\")]]"
                << std::endl;
      return 1;
    }

    std::string inputConnection = cmdLineArgs.getArg( "-srcconnect" );

    std::string srcCollName = "";
    srcCollName = cmdLineArgs.getArg( "-src" );
    if( (srcCollName == "") || 
        (srcCollName == "-src") ||
        (srcCollName == "-dst") ||
        (srcCollName == "-query") ||
        (srcCollName == "-queryopt") ||
        (srcCollName == "-srcconnect") ||
        (srcCollName == "-dstconnect") ||
        (srcCollName == "-ccread") ||
        (srcCollName == "-ccwrite") ||
        (srcCollName == "-noattrib") ||
        (srcCollName == "-tokattrib") ||
        (srcCollName == "-nevtpercommit") ||
        (srcCollName == "-nevtperprint") )
    {
        std::cerr << "CollCopy: ERROR No input collection specified on " 
                  << "command line." << std::endl;
        return 1;
    }
    std::string srcCollType = "";
    std::string srcCollConnect = "";
    while( (srcCollName != "") && 
           (srcCollName != "-src") &&
           (srcCollName != "-dst") &&
           (srcCollName != "-query") &&
           (srcCollName != "-queryopt") &&
           (srcCollName != "-srcconnect") && 
           (srcCollName != "-dstconnect") && 
           (srcCollName != "-ccread") && 
           (srcCollName != "-ccwrite") && 
           (srcCollName != "-noattrib") &&
           (srcCollName != "-tokattrib") &&
           (srcCollName != "-nevtpercommit") &&
           (srcCollName != "-nevtperprint") )
    {
      srcCollType = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      if( srcCollType == "MySQLCollection" || 
          srcCollType == "MySQLltCollection" ||
          srcCollType == "RelationalCollection" )
      {
        if ( inputConnection != "" )
        {
          srcCollConnect = inputConnection;
        }
        else
        {
          std::cerr << "CollCopy: ERROR Must specify an input database "
                    << "connection string for collections of type " 
                    << "MySQL or Oracle." << std::endl;
          return 1;
        }
      }
      else if ( srcCollType == "RootCollection" || 
                srcCollType == "PHYSICAL_NAME" ||
                srcCollType == "LOGICAL_NAME" || 
                srcCollType == "GUID" )
      {
        srcCollConnect = "";
      }
      else
      {
        std::cerr << "CollCopy: ERROR Collection type " << srcCollType 
                  << " is not presently handled by this program." 
                  << std::endl;
        return 1;
      }
      srcCollNameVec.push_back( srcCollName );
      srcCollTypeVec.push_back( srcCollType );
      srcCollConnectVec.push_back( srcCollConnect );
      srcCollName = cmdLineArgs.getArg( cmdLineArgs.getPos() );
    }

    std::string outputConnection = cmdLineArgs.getArg( "-dstconnect" );

    std::string dstCollName = "";
    dstCollName = cmdLineArgs.getArg( "-dst" );
    if( (dstCollName == "") || 
        (dstCollName == "-src") ||
        (dstCollName == "-dst") ||
        (dstCollName == "-query") ||
        (dstCollName == "-queryopt") ||
        (dstCollName == "-srcconnect") ||
        (dstCollName == "-dstconnect") ||
        (dstCollName == "-ccread") ||
        (dstCollName == "-ccwrite") ||
        (dstCollName == "-noattrib") ||
        (dstCollName == "-tokattrib") ||
        (dstCollName == "-nevtpercommit") ||
        (dstCollName == "-nevtperprint") )
    {
        std::cerr << "CollCopy: ERROR No output collection specified on " 
                  << "command line." << std::endl;
        return 1;
    }
    std::string dstCollType = "";
    std::string dstCollConnect = "";
    while( (dstCollName != "") && 
           (dstCollName != "-src") &&
           (dstCollName != "-dst") &&
           (dstCollName != "-query") &&
           (dstCollName != "-queryopt") &&
           (dstCollName != "-srcconnect") &&
           (dstCollName != "-dstconnect") &&
           (dstCollName != "-ccread") && 
           (dstCollName != "-ccwrite") && 
           (dstCollName != "-noattrib") &&
           (dstCollName != "-tokattrib") &&
           (dstCollName != "-nevtpercommit") &&
           (dstCollName != "-nevtperprint") )
    {
      dstCollType = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      if( dstCollType == "MySQLCollection" || 
          dstCollType == "MySQLltCollection" ||
          dstCollType == "RelationalCollection" )
      {
        if ( outputConnection != "" )
        {
          dstCollConnect = outputConnection;
        }
        else
        {
          if ( inputConnection != "" )
	  {
            dstCollConnect = inputConnection;
          }
          else
          {
            std::cerr << "CollCopy: ERROR Must specify an output database "
                      << "connection string for collections of type " 
                      << "MySQL or Oracle unless an input database " 
                      << "connection string was already specified." 
                      << std::endl;
            return 1;
          }
        }
      }
      else if ( dstCollType == "RootCollection" || 
                dstCollType == "PHYSICAL_NAME" ||
                dstCollType == "LOGICAL_NAME" || 
                dstCollType == "GUID" )
      {
        dstCollConnect = "";
      }
      else
      {
        std::cerr << "CollCopy: ERROR Collection type " << dstCollType 
                  << " is not presently handled by this program." 
                  << std::endl;
        return 1;
      }
      dstCollNameVec.push_back( dstCollName );
      dstCollTypeVec.push_back( dstCollType );
      dstCollConnectVec.push_back( dstCollConnect );
      dstCollName = cmdLineArgs.getArg( cmdLineArgs.getPos() );
    }

    std::string query = cmdLineArgs.getArg( "-query" );

    std::string queryOptions = "";
    if ( cmdLineArgs.hasQual( "-queryopt" ) )
    {
      queryOptions = cmdLineArgs.getArg("-queryopt");
      if( (queryOptions == "-src") ||
          (queryOptions == "-dst") ||
          (queryOptions == "-srcconnect") ||
          (queryOptions == "-dstconnect") ||
          (queryOptions == "-query") ||
          (queryOptions == "-queryopt") ||
          (queryOptions == "-ccread") ||
          (queryOptions == "-ccwrite") ||
          (queryOptions == "-noattrib") ||
          (queryOptions == "-tokattrib") ||
          (queryOptions == "-nevtpercommit") ||
          (queryOptions == "-nevtperprint") )
      {
        std::string queryOptions = "";
      }
    }

    std::string collCatalogRead = cmdLineArgs.getArg( "-ccread" );
    if( (collCatalogRead == "") ||
        (collCatalogRead == "-src") ||
        (collCatalogRead == "-dst") ||
        (collCatalogRead == "-srcconnect") ||
        (collCatalogRead == "-dstconnect") ||
        (collCatalogRead == "-query") ||
        (collCatalogRead == "-queryopt") ||
        (collCatalogRead == "-ccread") ||
        (collCatalogRead == "-ccwrite") ||
        (collCatalogRead == "-noattrib") ||
        (collCatalogRead == "-tokattrib") ||
        (collCatalogRead == "-nevtpercommit") ||
        (collCatalogRead == "-nevtperprint") )
    {
      collCatalogRead = COLLECTION_CATALOG_CONNECTION;
    }
    while( (collCatalogRead != "") &&
           (collCatalogRead != "-src") &&
           (collCatalogRead != "-dst") &&
           (collCatalogRead != "-srcconnect") &&
           (collCatalogRead != "-dstconnect") &&
           (collCatalogRead != "-query") &&
           (collCatalogRead != "-queryopt") &&
           (collCatalogRead != "-ccread") &&
           (collCatalogRead != "-ccwrite") &&
           (collCatalogRead != "-noattrib") &&
           (collCatalogRead != "-tokattrib") &&
           (collCatalogRead != "-nevtpercommit") &&
           (collCatalogRead != "-nevtperprint") )
    {
      collCatalogReadVec.push_back( collCatalogRead );
      collCatalogRead = cmdLineArgs.getArg( cmdLineArgs.getPos() );
    }

    std::string collCatalogWrite = cmdLineArgs.getArg( "-ccwrite" );
    if ( collCatalogWrite == "" )
    {
      collCatalogWrite = COLLECTION_CATALOG_CONNECTION;
    }

    bool noAttrib = false;
    if ( cmdLineArgs.hasQual("-noattrib") )
    {
      noAttrib = true;
    }

    std::string tokAttribName = cmdLineArgs.getArg( "-tokattrib" );

    unsigned int numEventsPerCommit = 10000;
    if ( cmdLineArgs.hasQual( "-nevtpercommit" ) )
    {
      std::string numEventsPerCommitStr = cmdLineArgs.getArg( "-nevtpercommit");
      if ( (numEventsPerCommitStr != "") &&
           (numEventsPerCommitStr != "-src") &&
           (numEventsPerCommitStr != "-dst") &&
           (numEventsPerCommitStr != "-srcconnect") &&
           (numEventsPerCommitStr != "-dstconnect") &&
           (numEventsPerCommitStr != "-query") &&
           (numEventsPerCommitStr != "-queryopt") &&
           (numEventsPerCommitStr != "-ccread") &&
           (numEventsPerCommitStr != "-noattrib") &&
           (numEventsPerCommitStr != "-nevtpercommit") &&
           (numEventsPerCommitStr != "-nevtperprint") )
      {
        numEventsPerCommit = atoi( numEventsPerCommitStr.c_str() );
      }
    }

    int numEventsPerPrint = -1;
    if ( cmdLineArgs.hasQual( "-nevtperprint" ) )
    {
      std::string numEventsPerPrintStr = cmdLineArgs.getArg("-nevtperprint");
      if ( numEventsPerPrintStr == "all" )
      {
        numEventsPerPrint = -1;
      }
      else if ( (numEventsPerPrintStr != "") &&
                (numEventsPerPrintStr != "-src") &&
                (numEventsPerPrintStr != "-dst") &&
                (numEventsPerPrintStr != "-srcconnect") &&
                (numEventsPerPrintStr != "-dstconnect") &&
                (numEventsPerPrintStr != "-query") &&
                (numEventsPerPrintStr != "-queryopt") &&
                (numEventsPerPrintStr != "-ccread") &&
                (numEventsPerPrintStr != "-noattrib") &&
                (numEventsPerPrintStr != "-nevtpercommit") &&
                (numEventsPerPrintStr != "-nevtperprint") )
      {
        numEventsPerPrint = atoi( numEventsPerPrintStr.c_str() );
      }
    }

    if(!::getenv( "POOL_OUTMSG_LEVEL" ))
    {
      //      // pool::POOLContext::setMessageVerbosityLevel(coral::Warning);
    }
    else
    {
      coral::MessageStream pms("get threshold");
      //      // pool::POOLContext::setMessageVerbosityLevel( pms.threshold() );
    }

    pool::IFileCatalog* collCatalog = new pool::IFileCatalog();
    for ( unsigned int i=0; i<collCatalogReadVec.size(); i++ )
    {
      collCatalog->addReadCatalog( collCatalogReadVec[i] );
    }
    collCatalog->setWriteCatalog( collCatalogWrite );
    collCatalog->connect();

    pool::ICollectionFactory* collFactory = pool::ICollectionFactory::get();
    collFactory->setDefaultCatalog( collCatalog );

    pool::ISession* session = 0;

    std::map<unsigned int, pool::ICollection*> dstCollMap;
    for ( unsigned int i=0; i<dstCollNameVec.size(); i++ )
    {
      pool::ICollection* dstCollection = collFactory->create(
          dstCollTypeVec[i],
          session,
          dstCollConnectVec[i], 
          dstCollNameVec[i], 
          pool::ICollection::CREATE_AND_OVERWRITE );

      dstCollMap[i] = dstCollection;
    }

    unsigned int evtCounterTotal = 0;
    std::map<unsigned int, pool::ICollection*> srcCollMap;
    for ( unsigned int i=0; i<srcCollNameVec.size(); i++ )
    {
      pool::ICollection* srcCollection = collFactory->create( 
          srcCollTypeVec[i],
          session,
          srcCollConnectVec[i], 
          srcCollNameVec[i], 
          pool::ICollection::READ );

      srcCollMap[i] = srcCollection;

      pool::ICollectionIterator* srcCollIter = 
        srcCollection->select( query, "", queryOptions );

      const pool::AttributeListSpecification& firstAttribListSpec =
	srcCollMap[0]->attributeListSpecification();

      const pool::AttributeListSpecification& nextAttribListSpec =
	srcCollection->attributeListSpecification();

      if ( !noAttrib && !(nextAttribListSpec.equals(firstAttribListSpec)) )
      {
	std::cout << "CollCopy: WARNING Schemas of two or more collections "
                  << "are different and attributes have not been requested to "
                  << "be excluded from copy. By definition, Token attribute "
                  << "schema of equivalent collections of type RootCollection, "
                  << "MySQLCollection, MySQLltCollection are expected to differ "
                  << "slightly and will be handled properly here. Other types of "
                  << "cases could lead to unexpected errors." << std::endl;
      }
      else if ( ((query != "") || (queryOptions != "")) && 
                !(nextAttribListSpec.equals(firstAttribListSpec)) )
      {
        std::cout << "CollCopy: WARNING The schemas of one or more input "
                  << "collections are different and a query has been "
                  << "requested. This may lead to unexpected errors."
                  << std::endl;
      }

      unsigned int evtCounter = 0;
      while ( srcCollIter->next() ) 
      {
        evtCounter++;
        evtCounterTotal++;

        pool::Token* token = new pool::Token( srcCollIter->token() );

        pool::AttributeList attribList = srcCollIter->attributeList();

        if ( tokAttribName != "" )
	{
	  std::string type = attribList[ tokAttribName ].spec().type_name();
          if ( type == "Token" )
	  {
            attribList[ tokAttribName ].getValue( *token );
          }
          else if ( type == "string" )
          {
	    std::string tokenStr = "";
            attribList[ tokAttribName ].getValue( tokenStr );
            token->fromString( tokenStr );
          }
          else
          {
            std::cerr << "CollCopy: ERROR Unexpected type used to store Token."
                      << std::endl;
            return 1;
          }
        }

        for ( unsigned int i=0; i<dstCollNameVec.size(); i++ )
        {
          if ( !noAttrib )
          {
            dstCollMap[i]->add( *token, attribList );
          }
          else
          {
            const pool::AttributeListSpecification emptyAttribListSpec;
            pool::AttributeList emptyAttribList( emptyAttribListSpec );
            dstCollMap[i]->add( *token, emptyAttribList );
          }

          if( !(evtCounterTotal % numEventsPerCommit) )
          {
            dstCollMap[i]->commit();
          }

          if ( (numEventsPerPrint >= 0) &&
               !(evtCounterTotal % numEventsPerPrint) )
          {
            std::cout << "CollCopy: " << evtCounterTotal << " events copied" << std::endl;
          }
        }

        token->release();
      }

      srcCollection->close();
    }

    for ( unsigned int i=0; i<dstCollNameVec.size(); i++ )
    {
      dstCollMap[i]->commit();
      dstCollMap[i]->close();
    }

    collCatalog->disconnect();
    delete collCatalog;

    std::cout << "CollCopy: Finished copying input collection(s) ";
    for( unsigned int i=0; i<srcCollNameVec.size(); i++ )
    {
      std::cout << "`" << srcCollNameVec[i] << ":" 
                << srcCollTypeVec[i] << "' ";
    }
    std::cout << "to output collection(s) ";
    for( unsigned int i=0; i<dstCollNameVec.size(); i++ )
    {
      std::cout << "`" << dstCollNameVec[i] << ":" 
                << dstCollTypeVec[i] << "' ";
    }
    std::cout << std::endl;

    return 0;
  }
  catch( pool::Exception& poolException )
  {
    std::cerr << "pool::Exception: " << poolException.what() << std::endl;;
    return 1;
  }
  catch( std::exception& exception )
  {
    std::cerr << "std::exception: " << exception.what() << std::endl;;
    return 1;
  }
  catch( ... )
  {
    std::cerr << "Unknown exception caught in main()." << std::endl;;
    return 1;
  }
}
