/**
 * @file CollListPFN.cpp
 * @brief Utility to list the physical file names used by a POOL collection
 * @author K. Karr <Kristo.Karr@cern.ch>
 * $Id: CollListPFN.cpp,v 1.25 2008/03/27 08:39:13 witoldp Exp $
 */
#include "CollectionBase/TokenList.h"
#include "CollectionBase/ICollectionDescription.h"
#include "CollectionBase/ICollectionColumn.h"
#include "CollectionBase/ICollectionQuery.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/CollectionRowBuffer.h"
#include "CollectionBase/CollectionService.h"
#include "CollectionBase/ICollection.h"

#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/IFCAction.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/IDataSvc.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "CoralBase/MessageStream.h"
#include "POOLCore/Exception.h"
#include "POOLCore/Token.h"
#include "CmdLineArgs.h"

#include "StorageSvc/DbStatus.h"

#include <iostream>


int main(int argc, const char *argv[])
{
  try
  {
    pool::DbStatus sc = pool::DbStatus::Success;
    std::vector<std::string> collNameVec;
    std::vector<std::string> collTypeVec;
    std::vector<std::string> collConnectVec;
    std::vector<std::string> fileCatalogReadVec;
    std::vector<std::string> collCatalogReadVec;

    CmdLineArgs cmdLineArgs(argc, argv);

    if( cmdLineArgs.hasQual("-help") ||
        !cmdLineArgs.hasQual("-src") )
    {
      std::cerr << "Usage: CollListPFN "
                << "<-src [+]<<input collection name> <input collection type>> > "
                << "(List of name-type pairs is space separated; "
                << "Choices for collection types are RootCollection, "
                << "MySQLCollection, MySQLltCollection, RelationalCollection, "
                << "PHYSICAL_NAME, LOGICAL_NAME or GUID) "
                << "Optional arguments: "
                << "[-srcconnect [Input database connection string]] "
                << "(Note: The default is \"\" and this argument MUST be "
                << "specified for collections of type MySQLCollection, "
                << "MySQLltCollection and RelationalCollection. Also, for "
                << "collections of type RelationalCollection the environment " 
                << "variables POOL_AUTH_USER and POOL_AUTH_PASSWORD "
                << "must be set) "
                << "[-query [predicate string (DEFAULT = \"\")]] "
		<< "[-queryopt [Choices are space separated list of the "
                << "following: \"FetchOne\" (no caching) or \"FetchAll\" "
                << "(caching) and/or \"SELECT <comma separated attribute list>\" "
                << "(DEFAULT = \"FetchOne SELECT *\")]] "
                << "[-fcread [+][read file catalog connection string "
                << "(DEFAULT = \"\")]] "
                << "[-ccread [+][read collection catalog connection string "
                << "(DEFAULT = \"\")]]"
                << std::endl;
      return 1;
    }

    std::string connection = cmdLineArgs.getArg( "-srcconnect" );

    std::string collName = "";
    collName = cmdLineArgs.getArg( "-src" );
    if( (collName == "") ||
        (collName == "-src") ||
        (collName == "-query") ||
        (collName == "-queryopt") ||
        (collName == "-srcconnect") ||
        (collName == "-fcread") ||
        (collName == "-ccread") )
    {
        std::cerr << "CollListPFN: ERROR No input collection specified on " 
                  << "command line." << std::endl;
        return 1;
    }

    std::string collType = "";
    std::string collConnect = "";
    while( (collName != "") &&
           (collName != "-src") &&
           (collName != "-query") &&
           (collName != "-queryopt") &&
           (collName != "-srcconnect") &&
           (collName != "-fcread") &&
           (collName != "-ccread") )
    {
      collType = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      if( collType == "MySQLCollection" || 
          collType == "MySQLltCollection" ||
          collType == "RelationalCollection" )
      {
	 if( connection.empty() ) {
	    std::cerr << "CollListPFN: ERROR Must specify a database "
		      << "connection string for collections of type " 
		      << "MySQL or Oracle." << std::endl;
	    return 1;
	 }
	 collConnect = connection;
      }
      else if( collType == "RootCollection" ||
	       collType == "PHYSICAL_NAME" ||
	       collType == "LOGICAL_NAME" ||
	       collType == "GUID" )
      {
        collConnect = "";
      }
      else {
        std::cerr << "CollListPFN: ERROR Collection type " << collType 
                  << " is not presently handled by this program." 
                  << std::endl;
        return 1;
      }
      collNameVec.push_back( collName );
      collTypeVec.push_back( collType );
      collConnectVec.push_back( collConnect );
      collName = cmdLineArgs.getArg( cmdLineArgs.getPos() );
    }

    std::string query = cmdLineArgs.getArg( "-query" );

    std::string queryOptions = "";
    if ( cmdLineArgs.hasQual( "-queryopt" ) )
    {
      queryOptions = cmdLineArgs.getArg("-queryopt");
      if( (queryOptions == "-src") ||
          (queryOptions == "-query") ||
          (queryOptions == "-queryopt") ||
          (queryOptions == "-srcconnect") ||
          (queryOptions == "-fcread") ||
          (queryOptions == "-ccread") )
      {
        std::string queryOptions = "";
      }
    }


    std::string fileCatalogRead = cmdLineArgs.getArg( "-fcread" );
    while( fileCatalogRead.size() && fileCatalogRead[0] != '-' )  {
       fileCatalogReadVec.push_back( fileCatalogRead );
       fileCatalogRead = cmdLineArgs.getArg( cmdLineArgs.getPos() );
    }

    std::string collCatalogRead = cmdLineArgs.getArg( "-ccread" );
    while( collCatalogRead.size() && collCatalogRead[0] != '-' )  {
       collCatalogReadVec.push_back( collCatalogRead );
       collCatalogRead = cmdLineArgs.getArg( cmdLineArgs.getPos() );
    }


    // MN: this line prevents hang on exit when only RootCollections are used
    // no idea why
    //    pool::POOLContext::loadComponent( "CORAL/Services/ConnectionService" );

    if(!::getenv( "POOL_OUTMSG_LEVEL" ))
    {
      //      // pool::POOLContext::setMessageVerbosityLevel(coral::Warning);
    }
    else
    {
      coral::MessageStream pms("get threshold");
      //      // pool::POOLContext::setMessageVerbosityLevel( pms.threshold() );
    }

    coral::MessageStream log( "CollListPFN" );
    log << coral::Info << "creating CollectionService" << coral::MessageStream::endmsg;
    pool::ICollectionService *collectionService( new pool::CollectionService() );

    // FIX - does the next line connect to the catalog by itself?
    for( unsigned int i=0; i<collCatalogReadVec.size(); i++ )      {
       collectionService->addReadCollectionCatalog( collCatalogReadVec[i] );
    }
    // collCatalog->connect();
    for( unsigned int i=0; i<fileCatalogReadVec.size(); i++ )      {
       collectionService->addReadFileCatalog( fileCatalogReadVec[i] );
    }
    collectionService->getFileCatalog()->start(); 
    
    pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( collectionService->getFileCatalog() );
    pool::DatabaseConnectionPolicy policy;
    policy.setReadMode( pool::DatabaseConnectionPolicy::READ );
    dataSvc->session().setDefaultConnectionPolicy( policy );
    sc = dataSvc->transaction().start( pool::ITransaction::READ );
    if( !sc.isSuccess() )
    {
      std::cerr << "CollListPFN: ERROR Cannot start database transaction." 
                << std::endl;
      return 1;
    }

    std::map<int, pool::ICollection*> collMap;
    std::map< std::string, int > pfnMap;
    int fileCounter = 0;
    for( unsigned int i=0; i<collNameVec.size(); i++ )
    {
       bool readOnly( true );
       pool::ICollection* collection = collectionService->handle( collNameVec[i], collTypeVec[i], collConnectVec[i], readOnly, &dataSvc->session() );
       collMap[i] = collection;

       log << coral::Debug
	   << " Creating query for the collection" << coral::MessageStream::endmsg;
       pool::ICollectionQuery *collQuery = collection->newQuery();
       collQuery->setCondition( query );
       if( queryOptions.size() ) {
	  collQuery->addToOutputList( queryOptions );
       } else {
	  collQuery->selectAllTokens();
       }
       log << coral::Debug
	   << " Executing the query " << coral::MessageStream::endmsg;
       pool::ICollectionCursor& cursor = collQuery->execute();

       while( cursor.next() ) {
	  std::string dbFileId = cursor.eventRef().dbID();
	  std::string bestPFN;
	  std::string fileType;
	  pool::IFCAction action;
	  collectionService->getFileCatalog()->setAction( action );
	  action.lookupBestPFN( dbFileId, 
				pool::FileCatalog::READ, 
				pool::FileCatalog::SEQUENTIAL, 
				bestPFN, 
				fileType );

	  std::map< std::string, int >::iterator mapIter;
	  mapIter = pfnMap.find( bestPFN );
	  if( mapIter == pfnMap.end() ) {
	     pfnMap[ bestPFN ] = fileCounter++;
	     std::cout << bestPFN << std::endl;
	  }
       }
    }

    for( unsigned int i=0; i<collNameVec.size(); i++ ) {      
      collMap[i]->close();
    }

    sc = dataSvc->transaction().commit();
    if( !sc.isSuccess() )
    {
      std::cerr << "CollListPFN: ERROR Cannot commit database transaction."
                << std::endl;
      return 1;
    }
    delete dataSvc;
    return 0;
  }
  catch( pool::Exception& poolException )
  {
    std::cerr << "pool::Exception: " << poolException.what() << std::endl;
    return 1;
  }
  catch( std::exception& exception )
  {
    std::cerr << "std::Exception: " << exception.what() << std::endl;
    return 1;
  }
  catch( ... )
  {
    std::cerr << "Unknown exception. " << std::endl;
    return 1;
  }
}
