#include "TestApp.h"

#include "FileCatalog/IFileCatalogImpl.h"
#include "FileCatalog/IFileCatalogEditorImpl.h"
#include "FileCatalog/IMetaDataCatalogImpl.h"
#include "FileCatalog/IMetaDataCatalogEditorImpl.h"
#include "FileCatalog/FCException.h"
#include "FileCatalog/MetaDataProxy.h"
#include "FileCatalog/ILFNIterator.h"

#include "POOLCore/Guid.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeSpecification.h"

#include "SealBase/TimeInfo.h"

#include <stdexcept>
#include <iostream>

#include "../../src/XPFN.h"
#include "../../src/XLFN.h"

TestApp::TestApp():
  TestBase()
{}


TestApp::~TestApp()
{}

void
TestApp::run()
{
  this->write();
  seal::TimeInfo::sleep(1);
  this->read();

  this->fill();
  seal::TimeInfo::sleep(1);
  this->query();
}



void
TestApp::write()
{
  const std::string dbService = "file:testmeta.xml";
  this->resetCatalog( dbService );
  pool::IFileCatalogImpl* catalog = this->retrieveCatalog( dbService );
  catalog->connect( pool::FC::Update );
  catalog->start();

  pool::IMetaDataCatalogImpl& metaDataCatalog = catalog->metadataCatalog();
  pool::IMetaDataCatalogEditorImpl& metaDataCatalogEditor = metaDataCatalog.editor();

  coral::AttributeList metaData;
  metaData.extend< unsigned short >( "Year" );
  metaData.extend< unsigned int >( "Run_Number" );
  metaData.extend< std::string >( "Description" );
  metaDataCatalogEditor.createSchema( metaData );

  catalog->commit();
  catalog->disconnect();
  delete catalog;
}



void
TestApp::read()
{
  const std::string dbService = "file:testmeta.xml";
  pool::IFileCatalogImpl* catalog = this->retrieveCatalog( dbService );
  catalog->connect( pool::FC::ReadOnly );
  catalog->start();

  pool::IMetaDataCatalogImpl& metaDataCatalog = catalog->metadataCatalog();  

  std::cout << "Retrieving the meta data desription" << std::endl;

  const coral::AttributeList& metaData = metaDataCatalog.describeMetaData();

  for ( coral::AttributeList::const_iterator iMetaData = metaData.begin();
        iMetaData != metaData.end(); ++iMetaData ) {
    std::cout << iMetaData->specification().name() << " : " <<  iMetaData->specification().typeName() << std::endl;
  }

  catalog->commit();
  catalog->disconnect();
  delete catalog;
}



void
TestApp::fill()
{
  const std::string dbService = "file:testmeta.xml";
  pool::IFileCatalogImpl* catalog = this->retrieveCatalog( dbService );
  catalog->connect( pool::FC::Update );
  catalog->start();

  pool::IMetaDataCatalogImpl& metaDataCatalog = catalog->metadataCatalog();
  pool::IMetaDataCatalogEditorImpl& metaDataCatalogEditor = metaDataCatalog.editor();
  pool::MetaDataProxy metaData;
  metaData.value() = metaDataCatalog.describeMetaData();

  pool::xmlfc::XPFN pfn;

  std::string pf = "pfn1";
  std::string type = "ROOT_All";
  pool::Guid guid;
  pool::Guid::create(guid);

  pfn.setValues( pf, guid.toString(), type );
  metaData.setguid( guid.toString() );
  metaData.value()["Year"].data< unsigned short >() = 2006;
  metaData.value()["Run_Number"].data< unsigned int >() = 1;
  metaData.value()["Description"].data< std::string >() = "Data";

  catalog->editor().insertPFN( pfn );
  metaDataCatalogEditor.insertMetaData( metaData );

  pf = "pfn1_2";
  pfn.setValues( pf, guid.toString(), type );
  catalog->editor().insertPFN( pfn );




  pf = "pfn2";
  pool::Guid::create(guid);
  pfn.setValues( pf, guid.toString(), type );

  metaData.setguid( guid.toString() );
  metaData.value()["Year"].data< unsigned short >() = 2006;
  metaData.value()["Run_Number"].data< unsigned int >() = 2;
  metaData.value()["Description"].data< std::string >() = "Data";

  catalog->editor().insertPFN( pfn );
  metaDataCatalogEditor.insertMetaData( metaData );


  pool::xmlfc::XLFN lfn;
  std::string lf = "lf1";
  lfn.setValues( lf, guid.toString() );
  catalog->editor().insertLFN( lfn );
  lf = "lf2";
  lfn.setValues( lf, guid.toString() );
  catalog->editor().insertLFN( lfn );


  catalog->commit();
  catalog->disconnect();
  delete catalog;
}




void
TestApp::query()
{
  const std::string dbService = "file:testmeta.xml";
  pool::IFileCatalogImpl* catalog = this->retrieveCatalog( dbService );
  catalog->connect( pool::FC::ReadOnly );
  catalog->start();

  pool::IMetaDataCatalogImpl& metaDataCatalog = catalog->metadataCatalog();
  pool::ILFNIterator* lfnIterator = metaDataCatalog.newLFNIterator( "Run_Number > 1" );

  while ( lfnIterator->hasNext() ) {
    const pool::LFN& lfn = lfnIterator->next();
    std::cout << "lfn : " << lfn.name() << std::endl;
  }

  delete lfnIterator;


  lfnIterator = metaDataCatalog.newLFNIterator( "Run_Number < 2" );

  bool result = lfnIterator->hasNext();
  delete lfnIterator;

  catalog->commit();
  catalog->disconnect();
  delete catalog;

  if ( result )
    throw std::runtime_error( "unexpected result from the query" );
}
